import argparse
import sys
from typing import (
    Any,
    AnyStr,
    Callable,
    ClassVar,
    Dict,
    IO,
    Iterator,
    List,
    Optional as Opt,
    Tuple,
    Type,
    TypeVar,
    Union,
)


class AnalysisUnitKind(object):
    """
    Specify a kind of analysis unit. Specification units provide an interface
    to the outer world while body units provide an implementation for the
    corresponding interface.
    """

    unit_specification: str
    unit_body: str


class LookupKind(object):
    """"""

    recursive: str
    flat: str
    minimal: str


class FindAllMode(object):
    """"""

    references: str
    derivedtypes: str


class RefResultKind(object):
    """
    Kind for the result of a cross reference operation.

    - ``NoRef`` is for no reference, it is the null value for this enum.

    - ``Precise`` is when the reference result is precise.

    - ``Imprecise`` is when there was an error computing the precise result,
    and a result was gotten in an imprecise fashion.

    - ``Error`` is for unrecoverable errors (either there is no imprecise path
    for the request you made, or the imprecise path errored out too.
    """

    noref: str
    precise: str
    imprecise: str
    error: str


class GrammarRule(object):
    """
    Gramar rule to use for parsing.
    """

    parent_list_rule: str
    protected_type_decl_rule: str
    protected_op_rule: str
    protected_el_rule: str
    protected_def_rule: str
    protected_decl_rule: str
    task_item_rule: str
    task_def_rule: str
    task_type_decl_rule: str
    subtype_decl_rule: str
    interface_type_def_rule: str
    unconstrained_index_rule: str
    array_type_def_rule: str
    discrete_subtype_definition_rule: str
    constraint_list_rule: str
    signed_int_type_def_rule: str
    mod_int_type_def_rule: str
    derived_type_def_rule: str
    discriminant_assoc_rule: str
    discriminant_constraint_rule: str
    index_constraint_rule: str
    digits_constraint_rule: str
    delta_constraint_rule: str
    range_constraint_rule: str
    constraint_rule: str
    discriminant_spec_rule: str
    discr_spec_list_rule: str
    discriminant_part_rule: str
    enum_literal_decl_rule: str
    formal_discrete_type_def_rule: str
    record_def_rule: str
    range_spec_rule: str
    real_type_def_rule: str
    sexpr_or_box_rule: str
    ordinary_fixed_point_def_rule: str
    decimal_fixed_point_def_rule: str
    floating_point_def_rule: str
    record_type_def_rule: str
    access_def_rule: str
    enum_type_def_rule: str
    type_def_rule: str
    variant_rule: str
    anonymous_type_decl_rule: str
    type_decl_rule: str
    variant_part_rule: str
    component_def_rule: str
    component_item_rule: str
    component_decl_rule: str
    component_list_rule: str
    generic_decl_rule: str
    generic_formal_part_rule: str
    generic_formal_decl_rule: str
    formal_subp_decl_rule: str
    renaming_clause_rule: str
    generic_renaming_decl_rule: str
    generic_instantiation_rule: str
    exception_decl_rule: str
    basic_decls_rule: str
    package_renaming_decl_rule: str
    package_decl_rule: str
    basic_decl_rule: str
    object_decl_rule: str
    sub_object_decl_rule: str
    ext_ret_stmt_object_decl_rule: str
    defining_id_list_rule: str
    number_decl_rule: str
    contract_case_assoc_rule: str
    contract_cases_expr_rule: str
    aspect_assoc_rule: str
    aspect_spec_rule: str
    single_task_decl_rule: str
    overriding_indicator_rule: str
    entry_decl_rule: str
    component_clause_rule: str
    aspect_clause_rule: str
    param_spec_rule: str
    param_specs_rule: str
    subp_spec_rule: str
    expr_fn_rule: str
    null_subp_decl_rule: str
    abstract_subp_decl_rule: str
    subp_renaming_decl_rule: str
    simple_subp_decl_rule: str
    subp_decl_rule: str
    with_clause_rule: str
    context_item_rule: str
    use_clause_rule: str
    use_package_clause_rule: str
    use_type_clause_rule: str
    subtype_indication_rule: str
    discrete_subtype_indication_rule: str
    constrained_subtype_indication_rule: str
    type_expr_rule: str
    anonymous_type_rule: str
    mode_rule: str
    pragma_argument_rule: str
    pragma_rule: str
    subunit_rule: str
    library_unit_body_rule: str
    library_unit_renaming_decl_rule: str
    library_item_rule: str
    compilation_unit_rule: str
    compilation_rule: str
    decl_part_rule: str
    entry_body_rule: str
    protected_body_rule: str
    protected_body_stub_rule: str
    task_body_rule: str
    task_body_stub_rule: str
    package_body_stub_rule: str
    package_body_rule: str
    terminate_alternative_rule: str
    select_stmt_rule: str
    accept_stmt_rule: str
    case_alt_rule: str
    case_stmt_rule: str
    ext_return_stmt_rule: str
    iblock_stmt_rule: str
    block_stmt_rule: str
    iloop_stmt_rule: str
    loop_stmt_rule: str
    compound_stmt_rule: str
    if_stmt_rule: str
    raise_stmt_rule: str
    delay_stmt_rule: str
    abort_stmt_rule: str
    body_rule: str
    body_stub_rule: str
    subp_body_stub_rule: str
    recov_decl_part_rule: str
    subp_body_rule: str
    handled_stmts_rule: str
    exception_handler_rule: str
    stmts_rule: str
    label_rule: str
    stmt_rule: str
    call_stmt_rule: str
    simple_stmt_rule: str
    null_stmt_rule: str
    assignment_stmt_rule: str
    goto_stmt_rule: str
    exit_stmt_rule: str
    return_stmt_rule: str
    requeue_stmt_rule: str
    identifier_rule: str
    char_literal_rule: str
    string_literal_rule: str
    defining_id_rule: str
    dec_literal_rule: str
    int_literal_rule: str
    num_literal_rule: str
    null_literal_rule: str
    allocator_rule: str
    for_loop_param_spec_rule: str
    quantified_expr_rule: str
    case_expr_rule: str
    case_expr_alt_rule: str
    raise_expr_rule: str
    if_expr_rule: str
    conditional_expr_rule: str
    box_expr_rule: str
    others_designator_rule: str
    aggregate_assoc_rule: str
    regular_aggregate_rule: str
    bracket_aggregate_rule: str
    aggregate_rule: str
    direct_name_rule: str
    param_assoc_rule: str
    call_suffix_rule: str
    qualified_name_rule: str
    qual_name_internal_rule: str
    name_rule: str
    defining_name_rule: str
    direct_name_or_target_name_rule: str
    target_name_rule: str
    update_attr_aggregate_rule: str
    update_attr_content_rule: str
    multidim_array_assoc_rule: str
    subtype_name_rule: str
    static_name_rule: str
    primary_rule: str
    paren_expr_rule: str
    declare_expr_rule: str
    factor_rule: str
    term_rule: str
    unop_term_rule: str
    simple_expr_rule: str
    boolean_op_rule: str
    discrete_range_rule: str
    choice_rule: str
    choice_list_rule: str
    rel_op_rule: str
    membership_choice_rule: str
    membership_choice_list_rule: str
    relation_rule: str
    expr_rule: str


default_grammar_rule: str


class InvalidField(Exception):
    pass


class NodeDataEvaluationError(Exception):
    pass


class InvalidInput(Exception):
    pass


class InvalidSymbolError(Exception):
    pass


class InvalidUnitNameError(Exception):
    pass


class NativeException(Exception):
    pass


class PreconditionFailure(Exception):
    pass


class PropertyError(Exception):
    pass


class TemplateArgsError(Exception):
    pass


class TemplateFormatError(Exception):
    pass


class TemplateInstantiationError(Exception):
    pass


class StaleReferenceError(Exception):
    pass


class UnknownCharset(Exception):
    pass


class AnalysisContext(object):
    """
    This type represents a context for all source analysis. This is the first
    type you need to create to use Libadalang. It will contain the results of
    all analysis, and is the main holder for all the data.

    You can create several analysis contexts if you need to, which enables you,
    for example to:

    * analyze several different projects at the same time;

    * analyze different parts of the same projects in parallel.

    In the current design, contexts always keep all of their analysis units
    allocated. If you need to get this memory released, the only option at your
    disposal is to destroy your analysis context instance.
    """

    def __init__(
        self,
        charset: Opt[str] = None,
        unit_provider: Opt[UnitProvider] = None,
        with_trivia: bool = True,
        tab_stop: int = 8,
        *,
        _c_value: Any = None
    ) -> None:
        """
        Create a new analysis context.

        ``Charset`` will be used as a default charset to decode input sources
        in analysis units. Please see ``GNATCOLL.Iconv`` for several supported
        charsets. Be careful: passing an unsupported charset is not guaranteed
        to raise an error here. If no charset is provided, ``"iso-8859-1"`` is
        the default.

        .. todo:: Passing an unsupported charset here is not guaranteed to
        raise an error right here, but this would be really helpful for users.

        When ``With_Trivia`` is true, the parsed analysis units will contain
        trivias.

        If provided, ``Unit_Provider`` will be used to query the file name that
        corresponds to a unit reference during semantic analysis. If it is
        ``None``, the default one is used instead.

        ``Tab_Stop`` is a positive number to describe the effect of tabulation
        characters on the column number in source files.
        """

    def __eq__(self, other: Any) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    def get_from_file(
        self,
        filename: AnyStr,
        charset: Opt[str] = None,
        reparse: bool = False,
        rule: str = default_grammar_rule,
    ) -> AnalysisUnit:
        """
        Create a new analysis unit for ``Filename`` or return the existing one
        if any. If ``Reparse`` is true and the analysis unit already exists,
        reparse it from ``Filename``.

        ``Rule`` controls which grammar rule is used to parse the unit.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If any failure occurs, such as file opening, decoding, lexing or
        parsing failure, return an analysis unit anyway: errors are described
        as diagnostics of the returned analysis unit.
        """

    def get_from_buffer(
        self,
        filename: AnyStr,
        buffer: AnyStr,
        charset: Opt[str] = None,
        reparse: bool = False,
        rule: str = default_grammar_rule,
    ) -> AnalysisUnit:
        """
        Create a new analysis unit for ``Filename`` or return the existing one
        if any. Whether the analysis unit already exists or not, (re)parse it
        from the source code in ``Buffer``.

        ``Rule`` controls which grammar rule is used to parse the unit.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If any failure occurs, such as file opening, decoding, lexing or
        parsing failure, return an analysis unit anyway: errors are described
        as diagnostics of the returned analysis unit.
        """

    def get_from_provider(
        self, name: AnyStr, kind: str, charset: Opt[str] = None, reparse: bool = False
    ) -> AnalysisUnit:
        """
        Create a new analysis unit for ``Name``/``Kind`` or return the existing
        one if any. If ``Reparse`` is true and the analysis unit already
        exists, reparse it from ``Filename``.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If the unit name cannot be tuned into a file name, raise an
        ``InvalidUnitNameError`` exception. If any other failure occurs, such
        as file opening, decoding, lexing or parsing failure, return an
        analysis unit anyway: errors are described as diagnostics of the
        returned analysis unit.
        """

    def discard_errors_in_populate_lexical_env(self, discard: bool) -> None:
        """
        Debug helper. Set whether ``Property_Error`` exceptions raised in
        ``Populate_Lexical_Env`` should be discarded. They are by default.
        """


class AnalysisUnit(object):
    """
    This type represents the analysis of a single file.
    """

    class TokenIterator(object):
        """
        Iterator over the tokens in an analysis unit.
        """

        def __init__(self, first: Token) -> None:
            ...

        def __iter__(self) -> AnalysisUnit.TokenIterator:
            ...

        def __next__(self) -> Token:
            ...

        def next(self) -> Token:
            ...

    def __init__(self, context: AnalysisContext, c_value: Any) -> None:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    @property
    def context(self) -> AnalysisContext:
        """
        Return the context that owns this unit.
        """

    def reparse(self, buffer: Opt[AnyStr] = None, charset: Opt[str] = None) -> None:
        """
        Reparse an analysis unit from a buffer, if provided, or from the
        original file otherwise. If ``Charset`` is empty or ``None``, use the
        last charset successfuly used for this unit, otherwise use it to decode
        the content of the source file.

        If any failure occurs, such as decoding, lexing or parsing failure,
        diagnostic are emitted to explain what happened.
        """

    def populate_lexical_env(self) -> None:
        """
        Create lexical environments for this analysis unit, according to the
        specifications given in the language spec.

        If not done before, it will be automatically called during semantic
        analysis. Calling it before enables one to control where the latency
        occurs.

        Depending on whether errors are discarded (see
        ``Discard_Errors_In_Populate_Lexical_Env``), raise a ``Property_Error``
        on failure.
        """

    @property
    def root(self) -> AdaNode:
        """
        Return the root node for this unit, or ``None`` if there is none.

        :rtype: AdaNode
        """

    @property
    def first_token(self) -> Token:
        """
        Return a reference to the first token scanned in this unit.
        """

    @property
    def last_token(self) -> Token:
        """
        Return a reference to the last token scanned in this unit.
        """

    @property
    def text(self) -> str:
        """
        Return the source buffer associated to this unit.
        """

    @property
    def token_count(self) -> int:
        """
        Return the number of tokens in this unit.
        """

    @property
    def trivia_count(self) -> int:
        """
        Return the number of trivias in this unit. This is 0 for units that
        were parsed with trivia analysis disabled.
        """

    def lookup_token(self, sloc: Sloc) -> Token:
        """
        Look for a token in this unit that contains the given source location.
        If this falls before the first token, return the first token. If this
        falls between two tokens, return the token that appears before. If this
        falls after the last token, return the last token. If there is no token
        in this unit, return no token.
        """

    def iter_tokens(self) -> AnalysisUnit.TokenIterator:
        """
        Iterator over the tokens in an analysis unit.
        """

    @property
    def filename(self) -> str:
        """
        Return the filename this unit is associated to.
        """

    @property
    def diagnostics(self) -> List[Diagnostic]:
        """
        Diagnostics for this unit.
        """

    def __repr__(self) -> str:
        ...


class Sloc(object):
    """
    Location in a source file. Line and column numbers are one-based.
    """

    line: int
    column: int

    def __init__(self, line: int, column: int) -> None:
        ...

    def __bool__(self) -> bool:
        ...

    def __nonzero__(self) -> bool:
        ...

    def __lt__(self, other: Sloc) -> bool:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...


class SlocRange(object):
    """
    Location of a span of text in a source file.
    """

    start: Sloc
    end: Sloc

    def __init__(self, start: Sloc, end: Sloc) -> None:
        ...

    def __bool__(self) -> bool:
        ...

    def __nonzero__(self) -> bool:
        ...

    def __lt__(self, other: SlocRange) -> bool:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...


class Diagnostic(object):
    """
    Diagnostic for an analysis unit: cannot open the source file, parsing
    error, ...
    """

    sloc_range: SlocRange
    message: str

    def __init__(self, sloc_range: SlocRange, message: str) -> None:
        ...

    @property
    def as_text(self) -> str:
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...


class Token(object):
    """
    Reference to a token in an analysis unit.
    """

    @property
    def next(self) -> Opt[Token]:
        """
        Return a reference to the next token in the corresponding analysis
        unit.
        """

    @property
    def previous(self) -> Opt[Token]:
        """
        Return a reference to the previous token in the corresponding analysis
        unit.
        """

    def range_until(self, other: Token) -> Iterator[Token]:
        """
        Return an iterator on the list of tokens that spans between `self` and
        `other` (included). This returns an empty list if the first token
        appears after the other one in the source code. Raise a ``ValueError``
        if both tokens come from different analysis units.
        """

    def is_equivalent(self, other: Token) -> bool:
        """
        Return whether ``L`` and ``R`` are structurally equivalent tokens. This
        means that their position in the stream won't be taken into account,
        only the kind and text of the token.
        """

    @property
    def kind(self) -> str:
        """
        Kind for this token.
        """

    @property
    def is_trivia(self) -> bool:
        """
        Return whether this token is a trivia. If it's not, it's a regular
        token.
        """

    @property
    def index(self) -> int:
        """
        Zero-based index for this token/trivia. Tokens and trivias get their
        own index space.
        """

    @property
    def text(self) -> str:
        """
        Return the text of the given token.
        """

    @classmethod
    def text_range(cls, first: Token, last: Token) -> str:
        """
        Compute the source buffer slice corresponding to the text that spans
        between the ``First`` and ``Last`` tokens (both included). This yields
        an empty slice if ``Last`` actually appears before ``First``.

        This raises a ``ValueError`` if ``First`` and ``Last`` don't belong to
        the same analysis unit.
        """

    @property
    def sloc_range(self) -> SlocRange:
        """
        Return the source location range of the given token.
        """

    def __eq__(self, other: Any) -> bool:
        """
        Return whether the two tokens refer to the same token in the same unit.

        Note that this does not actually compares the token data.
        """

    def __hash__(self) -> int:
        ...

    def __repr__(self) -> str:
        ...

    def __lt__(self, other: Opt[Token]) -> bool:
        """
        Consider that None comes before all tokens. Then, sort by unit, token
        index, and trivia index.
        """

    def __le__(self, other: Opt[Token]) -> bool:
        ...

    def __gt__(self, other: Opt[Token]) -> bool:
        ...

    def __ge__(self, other: Opt[Token]) -> bool:
        ...

    def to_data(self) -> dict:
        """
        Return a dict representation of this Token.
        """


class UnitProvider(object):
    """
    Interface to fetch analysis units from a name and a unit kind.

    The unit provider mechanism provides an abstraction which assumes that to
    any couple (unit name, unit kind) we can associate at most one source file.
    This means that several couples can be associated to the same source file,
    but on the other hand, only one one source file can be associated to a
    couple.

    This is used to make the semantic analysis able to switch from one analysis
    units to another.
    """

    def __init__(self, c_value: Any) -> None:
        """
        This constructor is an implementation detail, and is not meant to be
        used directly.
        """

    @classmethod
    def for_project(
        cls,
        project_file: AnyStr,
        project: Opt[AnyStr] = None,
        scenario_vars: Opt[Dict[str, str]] = None,
        target: Opt[AnyStr] = None,
        runtime: Opt[AnyStr] = None,
    ) -> UnitProvider:
        ...

    @classmethod
    def auto(
        cls, input_files: Iterator[AnyStr], charset: Opt[AnyStr] = None
    ) -> UnitProvider:
        ...


class AdaNode(object):
    """
    Root node class for the Ada syntax tree.
    """

    is_list_type: ClassVar[bool]

    @property
    def p_declarative_scope(self) -> DeclarativePart:
        """
        Return the scope of definition of this basic declaration.
        """

    @property
    def p_complete(self) -> List[CompletionItem]:
        """
        Return possible completions at this point in the file.
        """

    @property
    def p_valid_keywords(self) -> List[str]:
        """
        Return the list of keywords that are valid at this point in the file.

        .. note:: This is work in progress. It will return all keywords for
        now, without looking at the context.
        """

    @property
    def p_generic_instantiations(self) -> List[GenericInstantiation]:
        """
        Return the potentially empty list of generic package/subprogram
        instantiations that led to the creation of this entity. Outer-most
        instantiations appear last.
        """

    @property
    def p_semantic_parent(self) -> AdaNode:
        """
        Return the semantic parent for this node, if applicable, null
        otherwise.
        """

    @property
    def p_parent_basic_decl(self) -> BasicDecl:
        """
        Return the parent basic decl for this node, if applicable, null
        otherwise.

        .. note:: If the parent BasicDecl of the given node is a generic
        declaration, this call will return the instantiation from which the
        node was retrieved instead, if any.
        """

    def p_filter_is_imported_by(
        self, units: List[AnalysisUnit], transitive: bool
    ) -> List[AnalysisUnit]:
        """
        Filters out among the list of given units those that cannot refer to
        the unit in which this node lies. If transitive is True, the whole
        transitive closure of imports will be used to find a reference to the
        unit of this node.
        """

    @property
    def p_xref_entry_point(self) -> bool:
        """
        Designates entities that are entry point for the xref solving
        infrastructure. If this returns true, then resolve_names can be called
        on it.

        .. note:: For convenience, and unlike what is defined in the ARM wrt.
        complete contexts for name resolution, ``xref_entry_points`` can be
        nested.
        """

    @property
    def p_resolve_names(self) -> bool:
        """
        This will resolve names for this node. If the operation is successful,
        then type_var and ref_var will be bound on appropriate subnodes of the
        statement.
        """

    @property
    def p_standard_unit(self) -> AnalysisUnit:
        """
        Static method. Return the analysis unit corresponding to the Standard
        package.
        """

    def p_std_entity(self, sym: str) -> AdaNode:
        """
        Static property. Return an entity from the standard package with name
        `sym`.
        """

    @property
    def p_bool_type(self) -> AdaNode:
        """
        Static method. Return the standard Boolean type.
        """

    @property
    def p_int_type(self) -> AdaNode:
        """
        Static method. Return the standard Integer type.
        """

    @property
    def p_universal_int_type(self) -> AdaNode:
        """
        Static method. Return the standard Universal Integer type.
        """

    @property
    def p_universal_real_type(self) -> AdaNode:
        """
        Static method. Return the standard Universal Real type.
        """

    def p_top_level_decl(self, unit: AnalysisUnit) -> BasicDecl:
        """
        Static method. Get the top-level decl in ``unit``.  This is the body of
        a Subunit, or the item of a ``LibraryItem``.
        """

    def p_choice_match(self, value: int) -> bool:
        """
        Assuming that self is a choice expression (such as what can appear in
        an alternative of a case statement or in the RHS of a membership
        expression, this property returns whether the given value satisfies it.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    def p_gnat_xref(self, imprecise_fallback: bool = False) -> DefiningName:
        """
        Return a cross reference from this name to a defining identifier,
        trying to mimic GNAT's xrefs as much as possible.
        """

    @property
    def parent(self) -> AdaNode:
        """
        Return the lexical parent for this node. Return null for the root AST
        node or for AST nodes for which no one has a reference to the parent.
        """

    @property
    def parents(self) -> List[AdaNode]:
        """
        Return an array that contains the lexical parents (this node included).
        Nearer parents are first in the list.
        """

    @property
    def children(self) -> List[AdaNode]:
        """
        Return an array that contains the direct lexical children.
        """

    @property
    def token_start(self) -> Token:
        """
        Return the first token used to parse this node.
        """

    @property
    def token_end(self) -> Token:
        """
        Return the last token used to parse this node.
        """

    @property
    def child_index(self) -> int:
        """
        Return the 0-based index for Node in its parent's children.
        """

    @property
    def previous_sibling(self) -> AdaNode:
        """
        Return the node's previous sibling, if there is one.
        """

    @property
    def next_sibling(self) -> AdaNode:
        """
        Return the node's next sibling, if there is one.
        """

    @property
    def unit(self) -> AnalysisUnit:
        """
        Return the analysis unit owning this node.
        """

    @property
    def is_ghost(self) -> bool:
        """
        Return whether the node is a ghost.

        Unlike regular nodes, ghost nodes cover no token in the input source:
        they are logically located instead between two tokens. The
        "token_first" of all ghost nodes is the token right after this logical
        position, while they have no "token_last".
        """

    @property
    def full_sloc_image(self) -> str:
        """
        Return a string containing the filename + the sloc in GNU conformant
        format. Useful to create diagnostics from a node.
        """

    def __init__(
        self, c_value: Any, node_c_value: Any, metadata: Any, rebindings: Any
    ) -> None:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    def __ne__(self, other: Any) -> bool:
        ...

    def __hash__(self) -> int:
        ...

    @property
    def kind_name(self) -> str:
        """
        Return the kind of this node.
        """

    @property
    def is_token_node(self) -> bool:
        """
        Return whether this node is a node that contains only a single token.
        """

    @property
    def is_synthetic(self) -> bool:
        """
        Return whether this node is synthetic.
        """

    @property
    def sloc_range(self) -> SlocRange:
        """
        Return the spanning source location range for this node.

        Note that this returns the sloc of the parent for synthetic nodes.
        """

    @property
    def text(self) -> str:
        """
        Return the source buffer slice corresponding to the text that spans
        between the first and the last tokens of this node.

        Note that this returns the empty string for synthetic nodes.
        """

    @property
    def image(self) -> str:
        """
        Return a representation of this node as a string.
        """

    def lookup(self, sloc: Sloc) -> Opt[AdaNode]:
        """
        Return the bottom-most node from in ``Node`` and its children which
        contains ``Sloc``, or ``None`` if there is none.
        """

    def __bool__(self) -> bool:
        """
        Return always True so that checking a node against None can be done as
        simply as::

        if node: ...
        """

    def __nonzero__(self) -> bool:
        ...

    def __iter__(self) -> Iterator[AdaNode]:
        """
        Return an iterator on the children of this node.
        """

    def __len__(self) -> int:
        """
        Return the number of AdaNode children this node has.
        """

    def __getitem__(self, key: int) -> Opt[AdaNode]:
        """
        Return the Nth AdaNode child this node has.

        This handles negative indexes the same way Python lists do. Raise an
        IndexError if "key" is out of range.
        """

    def iter_fields(self) -> Iterator[Tuple[str, AdaNode]]:
        """
        Iterate through all the fields this node contains.

        Return an iterator that yields (name, value) couples for all abstract
        fields in this node. If "self" is a list, field names will be
        "item_{n}" with "n" being the index.
        """

    def dump_str(self) -> str:
        """
        Dump the sub-tree to a string in a human-readable format.
        """

    def dump(self, indent: str = "", file: IO[str] = sys.stdout) -> None:
        """
        Dump the sub-tree in a human-readable format on the given file.

        :param str indent: Prefix printed on each line during the dump. :param
        file file: File in which the dump must occur.
        """

    def findall(
        self,
        ast_type_or_pred: Union[Type[AdaNode], Callable[[AdaNode], bool]],
        **kwargs: Any
    ) -> List[AdaNode]:
        """
        Helper for finditer that will return all results as a list. See
        finditer's documentation for more details.
        """

    def find(
        self,
        ast_type_or_pred: Union[Type[AdaNode], Callable[[AdaNode], bool]],
        **kwargs: Any
    ) -> Opt[AdaNode]:
        """
        Helper for finditer that will return only the first result. See
        finditer's documentation for more details.
        """

    def finditer(
        self,
        ast_type_or_pred: Union[Type[AdaNode], Callable[[AdaNode], bool]],
        **kwargs: Any
    ) -> Iterator[AdaNode]:
        """
        Find every node corresponding to the passed predicates.

        :param ast_type_or_pred: If supplied with a subclass of AdaNode, will
        constrain the resulting collection to only the instances of this type
        or any subclass. If supplied with a predicate, it will apply the
        predicate on every node and keep only the ones for which it returns
        True. If supplied with a list of subclasses of AdaNode, it will match
        all instances of any of them. :type ast_type_or_pred: type|((AdaNode)
        -> bool)|list[type]

        :param kwargs: Allows the user to filter on attributes of the node. For
        every key value association, if the node has an attribute of name key
        that has the specified value, then the child is kept. :type kwargs:
        dict[str, Any]
        """

    @property
    def parent_chain(self) -> List[AdaNode]:
        """
        Return the parent chain of self. Self will be the first element,
        followed by the first parent, then this parent's parent, etc.
        """

    def __repr__(self) -> str:
        ...

    @property
    def entity_repr(self) -> str:
        ...

    @property
    def tokens(self) -> Iterator[Token]:
        """
        Return an iterator on the range of tokens that self encompasses.
        """

    def to_data(self) -> Union[list, dict]:
        """
        Return a nested python data-structure, constituted only of standard
        data types (dicts, lists, strings, ints, etc), and representing the
        portion of the AST corresponding to this node.
        """

    def to_json(self) -> str:
        """
        Return a JSON representation of this node.
        """

    def is_a(self, *types: Type[AdaNode]) -> bool:
        """
        Shortcut for isinstance(self, types). :rtype: bool
        """

    T = TypeVar("T", bound=AdaNode)

    def cast(self, typ: Type[T]) -> T:
        """
        Fluent interface style method. Return ``self``, raise an error if self
        is not of type ``typ``.

        :type typ: () -> T :rtype: T
        """


class AbortNode(AdaNode):
    """
    Qualifier for the ``abort`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of AbortPresent
        """


class AbortAbsent(AbortNode):
    """"""

    pass


class AbortPresent(AbortNode):
    """"""

    pass


class AbstractNode(AdaNode):
    """
    Qualifier for the ``abstract`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of AbstractPresent
        """


class AbstractAbsent(AbstractNode):
    """"""

    pass


class AbstractPresent(AbstractNode):
    """"""

    pass


class AdaList(AdaNode):
    """"""

    pass


class AdaNodeList(AdaList):
    """
    List of AdaNode.

    This list node can contain one of the following nodes:

    * AbstractSubpDecl

    * Allocator

    * AspectClause

    * AttributeRef

    * BaseAggregate

    * BinOp

    * BodyNode

    * CallExpr

    * CharLiteral

    * ComponentClause

    * ComponentDecl

    * CondExpr

    * DeclExpr

    * DottedName

    * EntryDecl

    * ErrorDecl

    * ExceptionDecl

    * ExceptionHandler

    * ExplicitDeref

    * GenericDecl

    * GenericFormal

    * GenericInstantiation

    * GenericRenamingDecl

    * Identifier

    * IncompleteTypeDecl

    * MembershipExpr

    * NullComponentDecl

    * NullLiteral

    * NumLiteral

    * NumberDecl

    * ObjectDecl

    * OthersDesignator

    * PackageDecl

    * PackageRenamingDecl

    * ParenExpr

    * PragmaNode

    * ProtectedTypeDecl

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * SingleProtectedDecl

    * SingleTaskDecl

    * Stmt

    * StringLiteral

    * SubpDecl

    * SubtypeDecl

    * SubtypeIndication

    * TargetName

    * TaskTypeDecl

    * TypeDecl

    * UnOp

    * UseClause

    * WithClause
    """

    def __iter__(self) -> Iterator[AdaNode]:
        ...

    def __getitem__(self, index: int) -> AdaNode:
        ...


class AlternativesList(AdaNodeList):
    """
    List of alternatives in a ``when ...`` clause.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeIndication

    * DottedName

    * ExplicitDeref

    * Identifier

    * MembershipExpr

    * NullLiteral

    * NumLiteral

    * OthersDesignator

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    def __iter__(self) -> Iterator[AdaNode]:
        ...

    def __getitem__(self, index: int) -> AdaNode:
        ...


class ConstraintList(AdaNodeList):
    """
    List of constraints.

    This list node can contain one of the following nodes:

    * AttributeRef

    * BinOp

    * CallExpr

    * CharLiteral

    * DottedName

    * ExplicitDeref

    * Identifier

    * QualExpr

    * StringLiteral

    * SubtypeIndication

    * TargetName
    """

    def __iter__(self) -> Iterator[AdaNode]:
        ...

    def __getitem__(self, index: int) -> AdaNode:
        ...


class DeclList(AdaNodeList):
    """
    List of declarations.

    This list node can contain one of the following nodes:

    * AbstractSubpDecl

    * AspectClause

    * ComponentDecl

    * EntryDecl

    * ExprFunction

    * NullSubpDecl

    * PragmaNode

    * SubpDecl

    * SubpRenamingDecl
    """

    def __iter__(self) -> Iterator[AdaNode]:
        ...

    def __getitem__(self, index: int) -> AdaNode:
        ...


class StmtList(AdaNodeList):
    """
    List of statements.

    This list node can contain one of the following nodes:

    * PragmaNode

    * Stmt
    """

    def __iter__(self) -> Iterator[AdaNode]:
        ...

    def __getitem__(self, index: int) -> AdaNode:
        ...


class AspectAssocList(AdaList):
    """
    List of AspectAssoc.
    """

    def __iter__(self) -> Iterator[AspectAssoc]:
        ...

    def __getitem__(self, index: int) -> AspectAssoc:
        ...


class BaseAssocList(AdaList):
    """
    List of BaseAssoc.
    """

    def __iter__(self) -> Iterator[BaseAssoc]:
        ...

    def __getitem__(self, index: int) -> BaseAssoc:
        ...


class BasicAssocList(AdaList):
    """
    List of BasicAssoc.
    """

    def __iter__(self) -> Iterator[BasicAssoc]:
        ...

    def __getitem__(self, index: int) -> BasicAssoc:
        ...


class AssocList(BasicAssocList):
    """
    List of associations.
    """

    def p_zip_with_params(self, imprecise_fallback: bool = False) -> List[ParamActual]:
        """
        Returns an array of pairs, associating formal parameters to actual
        expressions. The formals to match are retrieved by resolving the call
        which this AssocList represents the actuals of.
        """

    def __iter__(self) -> Iterator[BasicAssoc]:
        ...

    def __getitem__(self, index: int) -> BasicAssoc:
        ...


class BasicDeclList(AdaList):
    """
    List of BasicDecl.

    This list node can contain one of the following nodes:

    * NumberDecl

    * ObjectDecl

    * SingleProtectedDecl

    * SingleTaskDecl
    """

    def __iter__(self) -> Iterator[BasicDecl]:
        ...

    def __getitem__(self, index: int) -> BasicDecl:
        ...


class CaseExprAlternativeList(AdaList):
    """
    List of CaseExprAlternative.
    """

    def __iter__(self) -> Iterator[CaseExprAlternative]:
        ...

    def __getitem__(self, index: int) -> CaseExprAlternative:
        ...


class CaseStmtAlternativeList(AdaList):
    """
    List of CaseStmtAlternative.
    """

    def __iter__(self) -> Iterator[CaseStmtAlternative]:
        ...

    def __getitem__(self, index: int) -> CaseStmtAlternative:
        ...


class CompilationUnitList(AdaList):
    """
    List of CompilationUnit.
    """

    def __iter__(self) -> Iterator[CompilationUnit]:
        ...

    def __getitem__(self, index: int) -> CompilationUnit:
        ...


class ContractCaseAssocList(AdaList):
    """
    List of ContractCaseAssoc.
    """

    def __iter__(self) -> Iterator[ContractCaseAssoc]:
        ...

    def __getitem__(self, index: int) -> ContractCaseAssoc:
        ...


class DefiningNameList(AdaList):
    """
    List of DefiningName.
    """

    def __iter__(self) -> Iterator[DefiningName]:
        ...

    def __getitem__(self, index: int) -> DefiningName:
        ...


class DiscriminantSpecList(AdaList):
    """
    List of DiscriminantSpec.
    """

    def __iter__(self) -> Iterator[DiscriminantSpec]:
        ...

    def __getitem__(self, index: int) -> DiscriminantSpec:
        ...


class ElsifExprPartList(AdaList):
    """
    List of ElsifExprPart.
    """

    def __iter__(self) -> Iterator[ElsifExprPart]:
        ...

    def __getitem__(self, index: int) -> ElsifExprPart:
        ...


class ElsifStmtPartList(AdaList):
    """
    List of ElsifStmtPart.
    """

    def __iter__(self) -> Iterator[ElsifStmtPart]:
        ...

    def __getitem__(self, index: int) -> ElsifStmtPart:
        ...


class EnumLiteralDeclList(AdaList):
    """
    List of EnumLiteralDecl.
    """

    def __iter__(self) -> Iterator[EnumLiteralDecl]:
        ...

    def __getitem__(self, index: int) -> EnumLiteralDecl:
        ...


class ExprList(AdaList):
    """
    List of Expr.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeName

    * DottedName

    * ExplicitDeref

    * Identifier

    * NullLiteral

    * NumLiteral

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    def __iter__(self) -> Iterator[Expr]:
        ...

    def __getitem__(self, index: int) -> Expr:
        ...


class ExprAlternativesList(ExprList):
    """
    List of alternatives in a membership test expression.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeName

    * DottedName

    * ExplicitDeref

    * Identifier

    * NullLiteral

    * NumLiteral

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    def __iter__(self) -> Iterator[Expr]:
        ...

    def __getitem__(self, index: int) -> Expr:
        ...


class IdentifierList(AdaList):
    """
    List of Identifier.
    """

    def __iter__(self) -> Iterator[Identifier]:
        ...

    def __getitem__(self, index: int) -> Identifier:
        ...


class DiscriminantChoiceList(IdentifierList):
    """
    List of discriminant associations.
    """

    def __iter__(self) -> Iterator[Identifier]:
        ...

    def __getitem__(self, index: int) -> Identifier:
        ...


class NameList(AdaList):
    """
    List of Name.

    This list node can contain one of the following nodes:

    * AttributeRef

    * CallExpr

    * CharLiteral

    * DottedName

    * ExplicitDeref

    * Identifier

    * QualExpr

    * StringLiteral

    * TargetName
    """

    def __iter__(self) -> Iterator[Name]:
        ...

    def __getitem__(self, index: int) -> Name:
        ...


class ParentList(NameList):
    """
    List of parents in a type declaration.

    This list node can contain one of the following nodes:

    * CharLiteral

    * DottedName

    * Identifier

    * StringLiteral
    """

    def __iter__(self) -> Iterator[Name]:
        ...

    def __getitem__(self, index: int) -> Name:
        ...


class ParamSpecList(AdaList):
    """
    List of ParamSpec.
    """

    def __iter__(self) -> Iterator[ParamSpec]:
        ...

    def __getitem__(self, index: int) -> ParamSpec:
        ...


class PragmaNodeList(AdaList):
    """
    List of Pragma.
    """

    def __iter__(self) -> Iterator[PragmaNode]:
        ...

    def __getitem__(self, index: int) -> PragmaNode:
        ...


class SelectWhenPartList(AdaList):
    """
    List of SelectWhenPart.
    """

    def __iter__(self) -> Iterator[SelectWhenPart]:
        ...

    def __getitem__(self, index: int) -> SelectWhenPart:
        ...


class UnconstrainedArrayIndexList(AdaList):
    """
    List of UnconstrainedArrayIndex.
    """

    def __iter__(self) -> Iterator[UnconstrainedArrayIndex]:
        ...

    def __getitem__(self, index: int) -> UnconstrainedArrayIndex:
        ...


class VariantList(AdaList):
    """
    List of Variant.
    """

    def __iter__(self) -> Iterator[Variant]:
        ...

    def __getitem__(self, index: int) -> Variant:
        ...


class AliasedNode(AdaNode):
    """
    Qualifier for the ``aliased`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of AliasedPresent
        """


class AliasedAbsent(AliasedNode):
    """"""

    pass


class AliasedPresent(AliasedNode):
    """"""

    pass


class AllNode(AdaNode):
    """
    Qualifier for the ``all`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of AllPresent
        """


class AllAbsent(AllNode):
    """"""

    pass


class AllPresent(AllNode):
    """"""

    pass


class ArrayIndices(AdaNode):
    """
    Specification for array indexes.
    """

    pass


class ConstrainedArrayIndices(ArrayIndices):
    """
    Constrained specification for array indexes.
    """

    @property
    def f_list(self) -> ConstraintList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName
        """


class UnconstrainedArrayIndices(ArrayIndices):
    """
    Unconstrained specification for array indexes.
    """

    @property
    def f_types(self) -> UnconstrainedArrayIndexList:
        pass


class AspectAssoc(AdaNode):
    """
    Name/expression association in an aspect.
    """

    @property
    def f_id(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * ContractCases

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class AspectClause(AdaNode):
    """
    Base class for aspect clauses.
    """

    pass


class AtClause(AspectClause):
    """
    Representation clause (``for .. use at ...;``).
    """

    @property
    def f_name(self) -> BaseId:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * Identifier

        * StringLiteral
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class AttributeDefClause(AspectClause):
    """
    Clause for an attribute definition (``for ...'Attribute use ...;``).
    """

    @property
    def f_attribute_expr(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class EnumRepClause(AspectClause):
    """
    Representation clause for enumeration types.
    """

    @property
    def f_type_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_aggregate(self) -> BaseAggregate:
        pass


class RecordRepClause(AspectClause):
    """
    Representation clause for a record type.
    """

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_at_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_components(self) -> AdaNodeList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ComponentClause

        * PragmaNode
        """


class AspectSpec(AdaNode):
    """
    List of aspects in a declaration.
    """

    @property
    def f_aspect_assocs(self) -> AspectAssocList:
        pass


class BaseAssoc(AdaNode):
    """
    Abstract class for a key/value association, where the value is an
    expression.
    """

    @property
    def p_assoc_expr(self) -> Expr:
        """
        Returns the expression side of this assoc node.
        """


class ContractCaseAssoc(BaseAssoc):
    """
    Single association for the ``Contract_Case`` aspect.
    """

    @property
    def f_guard(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_consequence(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class PragmaArgumentAssoc(BaseAssoc):
    """
    Argument assocation in a pragma.
    """

    @property
    def f_id(self) -> Identifier:
        pass

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class BaseFormalParamHolder(AdaNode):
    """
    Base class for lists of formal parameters. This is used both for subprogram
    specifications and for records, so that we can share the matching and
    unpacking logic.
    """

    @property
    def p_abstract_formal_params(self) -> List[BaseFormalParamDecl]:
        """
        Return the list of abstract formal parameters for this holder.
        """

    @property
    def p_nb_min_params(self) -> int:
        """
        Return the minimum number of parameters this subprogram can be called
        while still being a legal call.
        """

    @property
    def p_nb_max_params(self) -> int:
        """
        Return the maximum number of parameters this subprogram can be called
        while still being a legal call.
        """

    def p_param_types(self, origin: AdaNode = None) -> List[BaseTypeDecl]:
        """
        Returns the type of each parameter of Self.
        """


class BaseSubpSpec(BaseFormalParamHolder):
    """
    Base class for subprogram specifications.
    """

    @property
    def p_returns(self) -> TypeExpr:
        """
        Syntax property. Return the type expression node corresponding to the
        return of this subprogram spec.
        """

    @property
    def p_params(self) -> List[ParamSpec]:
        """
        Returns the array of parameters specification for this subprogram spec.
        """

    def p_primitive_subp_types(
        self, imprecise_fallback: bool = False
    ) -> List[BaseTypeDecl]:
        """
        Return the types of which this subprogram is a primitive of.
        """

    def p_primitive_subp_first_type(
        self, imprecise_fallback: bool = False
    ) -> BaseTypeDecl:
        """
        Return the first type of which this subprogram is a primitive of.
        """

    def p_primitive_subp_tagged_type(
        self, imprecise_fallback: bool = False
    ) -> BaseTypeDecl:
        """
        If this subprogram is a primitive for a tagged type, then return this
        type.
        """

    def p_return_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Returns the return type of Self, if applicable (e.g. if Self is a
        subprogram). Else, returns null.
        """


class EntrySpec(BaseSubpSpec):
    """
    Entry specification.
    """

    @property
    def f_entry_name(self) -> DefiningName:
        pass

    @property
    def f_family_type(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName
        """

    @property
    def f_entry_params(self) -> Params:
        pass


class EnumSubpSpec(BaseSubpSpec):
    """
    Synthetic node for the abstract subprogram spec of an enum literal.

    NOTE: This has no existence in the ARM. While enum literals are functions
    semantically, they're not such syntactically.
    """

    pass


class SubpSpec(BaseSubpSpec):
    """
    Subprogram specification.
    """

    @property
    def f_subp_kind(self) -> SubpKind:
        pass

    @property
    def f_subp_name(self) -> DefiningName:
        pass

    @property
    def f_subp_params(self) -> Params:
        pass

    @property
    def f_subp_returns(self) -> TypeExpr:
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication
        """


class ComponentList(BaseFormalParamHolder):
    """
    List of component declarations.
    """

    @property
    def f_components(self) -> AdaNodeList:
        pass

    @property
    def f_variant_part(self) -> VariantPart:
        pass


class DiscriminantPart(BaseFormalParamHolder):
    """
    Specification for discriminants in type declarations.
    """

    pass


class KnownDiscriminantPart(DiscriminantPart):
    """
    Known list of discriminants in type declarations.
    """

    @property
    def f_discr_specs(self) -> DiscriminantSpecList:
        pass


class UnknownDiscriminantPart(DiscriminantPart):
    """
    Unknown list of discriminants in type declarations.
    """

    pass


class EntryCompletionFormalParams(BaseFormalParamHolder):
    """
    Formal parameters for the completion of an ``EntryDecl`` (either an
    ``EntryBody`` or an ``AcceptStmt``).
    """

    @property
    def f_params(self) -> Params:
        pass


class GenericFormalPart(BaseFormalParamHolder):
    """
    List of declaration for generic formals.
    """

    @property
    def f_decls(self) -> AdaNodeList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * GenericFormal

        * PragmaNode

        * UseClause
        """


class BaseRecordDef(AdaNode):
    """
    Base class for record definitions.
    """

    @property
    def f_components(self) -> ComponentList:
        pass


class NullRecordDef(BaseRecordDef):
    """
    Record definition for ``null record``.
    """

    pass


class RecordDef(BaseRecordDef):
    """
    Record definition that contains components (``record ... end record``).
    """

    pass


class BasicAssoc(AdaNode):
    """
    Association of one or several names to an expression.
    """

    def p_get_params(self, imprecise_fallback: bool = False) -> List[DefiningName]:
        """
        Return the list of parameters that this association refers to.
        """


class AggregateAssoc(BasicAssoc):
    """
    Assocation (X => Y) used for aggregates and parameter associations.
    """

    @property
    def f_designators(self) -> AlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_r_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class MultiDimArrayAssoc(AggregateAssoc):
    """
    Association used for multi-dimension array aggregates.
    """

    pass


class DiscriminantAssoc(BasicAssoc):
    """
    Association of discriminant names to an expression.
    """

    @property
    def f_ids(self) -> DiscriminantChoiceList:
        pass

    @property
    def f_discr_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class IteratedAssoc(BasicAssoc):
    """
    Iterated association (Ada 2020).
    """

    @property
    def f_spec(self) -> ForLoopSpec:
        pass

    @property
    def f_r_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class ParamAssoc(BasicAssoc):
    """
    Assocation (X => Y) used for aggregates and parameter associations.
    """

    @property
    def f_designator(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * Identifier

        * OthersDesignator

        * StringLiteral
        """

    @property
    def f_r_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class BasicDecl(AdaNode):
    """
    Root class for an Ada declaration (RM 3.1). A declaration associates a name
    with a language entity, for example a type or a variable.
    """

    @property
    def p_is_formal(self) -> bool:
        """
        Whether this decl is the nested decl of a generic formal declaration.
        """

    @property
    def p_doc_annotations(self) -> List[DocAnnotation]:
        """
        Return the documentation annotations associated with this decl.
        Annotations are any comment line of the form::

        --% [annotation_name]: [annotation]

        Raises a property error if the doc is incorrectly formatted.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    @property
    def p_doc(self) -> str:
        """
        Return the documentation associated with this decl. Raises a property
        error if the doc is incorrectly formatted.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    def p_previous_part_for_decl(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the previous part for this decl, if applicable.

        .. note:: It is not named previous_part, because BaseTypeDecl has a
        more precise version of previous_part that returns a BaseTypeDecl.
        Probably, we want to rename the specific versions, and have the root
        property be named previous_part. (TODO R925-008)
        """

    def p_canonical_part(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the canonical part for this decl. In the case of decls composed
        of several parts, the canonical part will be the first part.
        """

    def p_is_static_decl(self, imprecise_fallback: bool = False) -> bool:
        """
        Return whether this declaration is static.
        """

    @property
    def p_is_imported(self) -> bool:
        """
        Whether this declaration is imported from another language.
        """

    @property
    def f_aspects(self) -> AspectSpec:
        """
        Return the list of aspects that are attached to this node.
        """

    def p_get_aspect_assoc(self, name: str) -> AspectAssoc:
        """
        Return the aspect with name ``name`` for this entity.
        """

    def p_get_aspect_spec_expr(self, name: str) -> Expr:
        """
        Return the expression associated to the aspect with name ``name`` for
        this entity.
        """

    def p_get_aspect(self, name: str, imprecise_fallback: bool = False) -> Aspect:
        """
        Return the aspect with name ``name`` associated to this entity.

        Aspects are properties of entities that can be specified by the Ada
        program, either via aspect specifications, pragmas, or attributes.

        This will return the syntactic node corresponding to attribute
        directly.
        """

    def p_has_aspect(self, name: str, imprecise_fallback: bool = False) -> bool:
        """
        Returns whether the boolean aspect named ``name`` is set on the entity
        represented by this node.

        "Aspect" is used as in RM terminology (see RM 13).
        """

    def p_get_pragma(self, name: str) -> PragmaNode:
        """
        Return the pragma with name ``name`` associated to this entity.
        """

    def p_get_representation_clause(
        self, name: str, imprecise_fallback: bool = False
    ) -> AttributeDefClause:
        """
        Return the representation clause associated to this type decl that
        defines the given attribute name.
        """

    @property
    def p_is_compilation_unit_root(self) -> bool:
        """
        Whether a BasicDecl is the root decl for its unit.
        """

    def p_is_visible(self, from_node: AdaNode) -> bool:
        """
        Return whether this declaration is visible from the point of view of
        the given ``origin`` node.

        .. ATTENTION:: Only package-level (public or private) declarations are
        supported for now.
        """

    @property
    def p_base_subp_declarations(self) -> List[BasicDecl]:
        """
        If Self declares a primitive subprogram of some tagged type T, return
        the set of all subprogram declarations that it overrides (including
        itself).
        """

    def p_root_subp_declarations(self, origin: AdaNode = None) -> List[BasicDecl]:
        """
        If Self declares a primitive subprogram of some tagged type T, return
        the root subprogram declarations that it overrides. There can be
        several, as in the following scenario:

        - package Root defines the root tagged type T and subprogram Foo.

        - package Itf defines interface I and abstract subprogram Foo.

        - package D defines "type U is new Root.T and Itf.I" and an overriding
        subprogram Foo.

        Here, root_subp_declarations of Foo defined in package D will return
        both Foo from package Root and Foo from package Itf.
        """

    def p_find_all_overrides(
        self, units: List[AnalysisUnit], imprecise_fallback: bool = False
    ) -> List[BasicDecl]:
        """
        If Self is the declaration of a primitive of some type T, return the
        list of all subprogram that override this subprogram among the given
        units.
        """

    @property
    def p_defining_names(self) -> List[DefiningName]:
        """
        Get all the names of this basic declaration.
        """

    @property
    def p_defining_name(self) -> DefiningName:
        """
        Get the name of this declaration. If this declaration has several
        names, it will return the first one.
        """

    @property
    def p_type_expression(self) -> TypeExpr:
        """
        Return the type expression for this BasicDecl if applicable, a null
        otherwise.
        """

    def p_subp_spec_or_null(self, follow_generic: bool = False) -> BaseSubpSpec:
        """
        If Self is a Subp, returns the specification of this subprogram.

        If ``follow_generic`` is True, will also work for instances of
        ``GenericSubpDecl``.
        """

    @property
    def p_is_subprogram(self) -> bool:
        """
        Return True if self is a subprogram node in the general sense (which
        is, an entity that can be called). This includes separates and entries.
        """

    @property
    def p_relative_name(self) -> SingleTokNode:
        """
        Return the relative name for Self. If Self's defining name is
        ``A.B.C``, return C as a node.
        """

    @property
    def p_relative_name_text(self) -> str:
        """
        Return the relative name for Self, as text.
        """

    def p_next_part_for_decl(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the next part of this declaration, if applicable.

        .. note:: It is not named next_part, because BaseTypeDecl has a more
        precise version of next_part that returns a BaseTypeDecl. Probably, we
        want to rename the specific versions, and have the root property be
        named next_part. (TODO R925-008)
        """

    def p_body_part_for_decl(self, imprecise_fallback: bool = False) -> BodyNode:
        """
        Return the body corresponding to this declaration, if applicable.

        .. note:: It is not named body_part, subclasses have more precise
        versions named body_part and returning a more precise result. Probably,
        we want to rename the specific versions, and have the root property be
        named body_part. (TODO R925-008)
        """

    @property
    def p_fully_qualified_name_array(self) -> List[str]:
        """
        Return the fully qualified name corresponding to this declaration, as
        an array of symbols.
        """

    @property
    def p_fully_qualified_name(self) -> str:
        """
        Return the fully qualified name corresponding to this declaration.
        """

    @property
    def p_canonical_fully_qualified_name(self) -> str:
        """
        Return a canonical representation of the fully qualified name
        corresponding to this declaration.
        """

    @property
    def p_unique_identifying_name(self) -> str:
        """
        Return a unique identifying name for this declaration, provided this
        declaration is a public declaration. In the case of subprograms, this
        will include the profile.

        .. attention:: This will only return a unique name for public
        declarations. Notably, anything nested in an unnamed declare block
        won't be handled correctly.
        """


class BaseFormalParamDecl(BasicDecl):
    """
    Base class for formal parameter declarations. This is used both for records
    components and for subprogram parameters.

    This is a Libadalang abstraction, that has no ARM existence.
    """

    def p_formal_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Return the type for this formal.
        """


class ComponentDecl(BaseFormalParamDecl):
    """
    Declaration for a component.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_component_def(self) -> ComponentDef:
        pass

    @property
    def f_default_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class DiscriminantSpec(BaseFormalParamDecl):
    """
    Known list of discriminants in type declarations.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_type_expr(self) -> TypeExpr:
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication
        """

    @property
    def f_default_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class GenericFormal(BaseFormalParamDecl):
    """
    Enclosing declaration for a generic formal. The real declaration is
    accessible via the ``decl`` field.
    """

    @property
    def f_decl(self) -> BasicDecl:
        """
        This field can contain one of the following nodes:

        * FormalSubpDecl

        * GenericInstantiation

        * IncompleteTypeDecl

        * NumberDecl

        * ObjectDecl

        * SingleProtectedDecl

        * SingleTaskDecl

        * TypeDecl
        """


class GenericFormalObjDecl(GenericFormal):
    """
    Formal declaration for an object.
    """

    pass


class GenericFormalPackage(GenericFormal):
    """
    Formal declaration for a package.
    """

    pass


class GenericFormalSubpDecl(GenericFormal):
    """
    Formal declaration for a subprogram.
    """

    pass


class GenericFormalTypeDecl(GenericFormal):
    """
    Formal declaration for a type.
    """

    pass


class ParamSpec(BaseFormalParamDecl):
    """
    Specification for a parameter.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_has_aliased(self) -> AliasedNode:
        pass

    @property
    def f_mode(self) -> Mode:
        pass

    @property
    def f_type_expr(self) -> TypeExpr:
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication
        """

    @property
    def f_default_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class BasePackageDecl(BasicDecl):
    """
    Package declarations. Concrete instances of this class will be created in
    generic package declarations. Other non-generic package declarations will
    be instances of PackageDecl.

    The behavior is the same, the only difference is that BasePackageDecl and
    PackageDecl have different behavior regarding lexical environments. In the
    case of generic package declarations, we use BasePackageDecl which has no
    env_spec, and the environment behavior is handled by the GenericPackageDecl
    instance.
    """

    @property
    def f_package_name(self) -> DefiningName:
        pass

    @property
    def f_public_part(self) -> PublicPart:
        pass

    @property
    def f_private_part(self) -> PrivatePart:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass

    @property
    def p_body_part(self) -> PackageBody:
        """
        Return the PackageBody corresponding to this node.
        """


class GenericPackageInternal(BasePackageDecl):
    """
    This class denotes the internal package contained by a GenericPackageDecl.
    """

    pass


class PackageDecl(BasePackageDecl):
    """
    Non-generic package declarations.
    """

    pass


class BaseTypeDecl(BasicDecl):
    """
    Base class for type declarations.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    def p_base_subtype(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        If this type decl is a subtype decl, return the base subtype. If not,
        return ``Self``.
        """

    @property
    def p_private_completion(self) -> BaseTypeDecl:
        """
        Return the private completion for this type, if there is one.
        """

    def p_is_inherited_primitive(self, p: BasicDecl) -> bool:
        """
        Assuming that P is a primitive of Self, return whether the given
        primitive P is inherited from one of Self's parents.
        """

    def p_get_record_representation_clause(
        self, imprecise_fallback: bool = False
    ) -> RecordRepClause:
        """
        Return the record representation clause associated to this type decl,
        if applicable (i.e. this type decl defines a record type).
        """

    def p_get_enum_representation_clause(
        self, imprecise_fallback: bool = False
    ) -> EnumRepClause:
        """
        Return the enum representation clause associated to this type decl, if
        applicable (i.e. this type decl defines an enum type).
        """

    def p_is_record_type(self, origin: AdaNode = None) -> bool:
        """
        Return whether this type is a record type.
        """

    def p_is_array_type(self, origin: AdaNode = None) -> bool:
        """
        Return whether this type is an array type.
        """

    def p_find_derived_types(
        self, root: AdaNode, origin: AdaNode, imprecise_fallback: bool = False
    ) -> List[TypeDecl]:
        """
        Find types derived from self in the given ``root`` and its children.
        """

    def p_is_real_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is a real type or not.
        """

    def p_is_float_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is a float type or not.
        """

    def p_is_fixed_point(self, origin: AdaNode = None) -> bool:
        """
        Whether type is a fixed point type or not.
        """

    def p_is_enum_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is an enum type
        """

    def p_is_access_type(self, origin: AdaNode = None) -> bool:
        """
        Whether Self is an access type or not
        """

    def p_is_char_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is a character type or not
        """

    @property
    def p_discrete_range(self) -> DiscreteRange:
        """
        Return the discrete range for this type decl, if applicable.
        """

    def p_is_discrete_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is a discrete type or not.
        """

    def p_is_int_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is an integer type or not.
        """

    def p_accessed_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        If this type is an access type, or a type with an Implicit_Dereference
        aspect, return the type of a dereference of an instance of this type.
        """

    def p_is_tagged_type(self, origin: AdaNode = None) -> bool:
        """
        Whether type is tagged or not
        """

    def p_base_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Return the base type entity for this derived type declaration
        """

    def p_base_types(self, origin: AdaNode = None) -> List[BaseTypeDecl]:
        """
        Return the list of base types for Self.
        """

    def p_find_all_derived_types(
        self, units: List[AnalysisUnit], imprecise_fallback: bool = False
    ) -> List[TypeDecl]:
        """
        Return the list of all types that inherit (directly or inderictly) from
        Self among the given units.
        """

    def p_comp_type(
        self, is_subscript: bool = False, origin: AdaNode = None
    ) -> BaseTypeDecl:
        """
        Return the component type of `Self`, if applicable. The component type
        is the type you'll get if you call a value whose type is `Self`.  So it
        can either be:

        1. The component type for an array. 2. The return type for an access to
        function.
        """

    def p_index_type(self, dim: int, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Return the index type for dimension ``dim`` for this type, if
        applicable.
        """

    def p_is_derived_type(
        self, other_type: BaseTypeDecl, origin: AdaNode = None
    ) -> bool:
        """
        Whether Self is derived from other_type.
        """

    def p_is_interface_type(self, origin: AdaNode = None) -> bool:
        """
        Return True iff this type declaration is an interface definition.
        """

    def p_matching_type(
        self, expected_type: BaseTypeDecl, origin: AdaNode = None
    ) -> bool:
        """
        Return whether ``self`` matches ``expected_type``.
        """

    def p_canonical_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Return the canonical type declaration for this type declaration. For
        subtypes, it will return the base type declaration.
        """

    def p_previous_part(self, go_to_incomplete: bool = True) -> BaseTypeDecl:
        """
        Returns the previous part for this type decl.
        """

    @property
    def p_next_part(self) -> BaseTypeDecl:
        """
        Returns the next part for this type decl.
        """

    @property
    def p_full_view(self) -> BaseTypeDecl:
        """
        Return the full completion of this type.
        """

    def p_is_definite_subtype(self, origin: AdaNode = None) -> bool:
        """
        Returns whether this is a definite subtype.

        For convenience, this will return ``False`` for incomplete types, even
        though the correct answer is more akin to "non applicable".
        """

    @property
    def p_is_private(self) -> bool:
        """
        Whether node is a private view of corresponding type.
        """

    @property
    def p_discriminants_list(self) -> List[BaseFormalParamDecl]:
        """
        Return the list of all discriminants of this type. If this type has no
        discriminant or only unknown discriminants, an empty list is returned.
        """

    def p_root_type(self, origin: AdaNode = None) -> BaseTypeDecl:
        """
        Return the type that is at the root of the derivation hierarchy
        (ignoring secondary interfaces derivations for tagged types)
        """

    def p_shapes(
        self, include_discriminants: bool = True, origin: AdaNode = None
    ) -> List[Shape]:
        """
        Must be called on a record (sub-)type declaration. Return all the
        possible shapes that a value of this record type can take. For example,
        consider the following record definition:

        .. code::

        type R (A : Integer; B : Integer) is record X : Integer; case A is when
        1 .. 10 => Y_1 : Integer; case B is when 1 .. 10 => Z_1 : Integer; when
        others => null; end case; when 11 .. 20 => Y_2 : Integer; case B is
        when 1 .. 10 => Z_2 : Integer; when others => null; end case; when
        others => null; end case; end record;

        For this instance, this property will return the following results:

        .. code::

        [ [X, Y_1, Z_1], [X, Y_1], [X, Y_2, Z_2], [X, Y_2], [X] ]

        .. ATTENTION:: This property is inaccurate when called on a record
        extension which defines components under a certain condition C, and
        this same condition is used to define some components in the parent
        record: in that case, any feasible shape will in practice contain
        either both the components defined under condition C in the child
        record and the parent record, or none of them.

        However, due to the simplified algorithm we use here to compute the
        feasible shapes, we will also return shapes that include the components
        of the child record but not the parent record, and conversely.
        """


class BaseSubtypeDecl(BaseTypeDecl):
    """
    Base class for subtype declarations.
    """

    pass


class DiscreteBaseSubtypeDecl(BaseSubtypeDecl):
    """
    Specific ``BaseSubtypeDecl`` synthetic subclass for the base type of scalar
    types.
    """

    pass


class SubtypeDecl(BaseSubtypeDecl):
    """
    Subtype declaration.
    """

    @property
    def f_subtype(self) -> SubtypeIndication:
        pass


class ClasswideTypeDecl(BaseTypeDecl):
    """
    Synthetic node (not parsed, generated from a property call). Refers to the
    classwide type for a given tagged type. The aim is that those be mostly
    equivalent to their non-classwide type, except for some resolution rules.
    """

    pass


class IncompleteTypeDecl(BaseTypeDecl):
    """
    Incomplete declaration for a type.
    """

    @property
    def f_discriminants(self) -> DiscriminantPart:
        pass


class IncompleteTaggedTypeDecl(IncompleteTypeDecl):
    """
    Incomplete declaration for a tagged type.
    """

    @property
    def f_has_abstract(self) -> AbstractNode:
        pass


class ProtectedTypeDecl(BaseTypeDecl):
    """
    Declaration for a protected type.
    """

    @property
    def f_discriminants(self) -> DiscriminantPart:
        pass

    @property
    def f_interfaces(self) -> ParentList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_definition(self) -> ProtectedDef:
        pass


class TaskTypeDecl(BaseTypeDecl):
    """
    Declaration for a task type.
    """

    @property
    def f_discriminants(self) -> DiscriminantPart:
        pass

    @property
    def f_definition(self) -> TaskDef:
        pass


class SingleTaskTypeDecl(TaskTypeDecl):
    """
    Type declaration for a single task.
    """

    pass


class TypeDecl(BaseTypeDecl):
    """
    Type declarations that embed a type definition node.
    """

    @property
    def f_discriminants(self) -> DiscriminantPart:
        pass

    @property
    def f_type_def(self) -> TypeDef:
        pass

    def p_get_primitives(self, only_inherited: bool = False) -> List[BasicDecl]:
        """
        Return the list of all primitive operations that are available on this
        type. If `only_inherited` is True, it will only return the primitives
        that are implicitly inherited by this type, discarding those explicitly
        defined on this type.
        """


class AnonymousTypeDecl(TypeDecl):
    """
    Anonymous type declaration (for anonymous array or access types).
    """

    pass


class SynthAnonymousTypeDecl(AnonymousTypeDecl):
    """
    Synthetic anonymous type decl. Used to generate anonymous access types.
    """

    pass


class BasicSubpDecl(BasicDecl):
    """
    Base class for subprogram declarations.
    """

    @property
    def p_subp_decl_spec(self) -> BaseSubpSpec:
        """
        Return the specification for this subprogram
        """


class ClassicSubpDecl(BasicSubpDecl):
    """
    This is an intermediate abstract class for subprogram declarations with a
    common structure: overriding indicator, ``SubpSpec``, aspects, <other
    fields>.
    """

    @property
    def f_overriding(self) -> OverridingNode:
        pass

    @property
    def f_subp_spec(self) -> SubpSpec:
        pass

    def p_body_part(self, imprecise_fallback: bool = False) -> BaseSubpBody:
        """
        Return the BaseSubpBody corresponding to this node.
        """


class AbstractSubpDecl(ClassicSubpDecl):
    """
    Declaration for an abstract subprogram.
    """

    pass


class FormalSubpDecl(ClassicSubpDecl):
    """
    Formal subprogram declarations, in generic declarations formal parts.
    """

    @property
    def f_default_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BoxExpr

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * QualExpr

        * StringLiteral

        * TargetName
        """


class AbstractFormalSubpDecl(FormalSubpDecl):
    """
    Formal declaration for an abstract subprogram.
    """

    pass


class ConcreteFormalSubpDecl(FormalSubpDecl):
    """
    Formal declaration for a concrete subprogram.
    """

    pass


class SubpDecl(ClassicSubpDecl):
    """
    Regular subprogram declaration.
    """

    pass


class EntryDecl(BasicSubpDecl):
    """
    Entry declaration.
    """

    @property
    def f_overriding(self) -> OverridingNode:
        pass

    @property
    def f_spec(self) -> EntrySpec:
        pass

    def p_body_part(self, imprecise_fallback: bool = False) -> EntryBody:
        """
        Return the entry body associated to this entry declaration.
        """


class EnumLiteralDecl(BasicSubpDecl):
    """
    Declaration for an enumeration literal.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def p_enum_type(self) -> TypeDecl:
        """
        Return the enum type corresponding to this enum literal.
        """


class GenericSubpInternal(BasicSubpDecl):
    """
    Internal node for generic subprograms.
    """

    @property
    def f_subp_spec(self) -> SubpSpec:
        pass


class BodyNode(BasicDecl):
    """
    Base class for an Ada body (RM 3.11). A body is the completion of a
    declaration.
    """

    def p_previous_part(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the previous part for this body. Might be a declaration or a
        body stub.
        """

    def p_decl_part(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the decl corresponding to this node if applicable.
        """

    @property
    def p_subunit_root(self) -> BasicDecl:
        """
        If self is a subunit, return the body in which it is rooted.
        """


class BaseSubpBody(BodyNode):
    """
    Base class for subprogram bodies.
    """

    @property
    def f_overriding(self) -> OverridingNode:
        pass

    @property
    def f_subp_spec(self) -> SubpSpec:
        pass


class ExprFunction(BaseSubpBody):
    """
    Expression function.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * BaseAggregate

        * ParenExpr
        """


class NullSubpDecl(BaseSubpBody):
    """
    Declaration for a null subprogram.
    """

    pass


class SubpBody(BaseSubpBody):
    """
    Subprogram body.
    """

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class SubpRenamingDecl(BaseSubpBody):
    """
    Declaration for a subprogram renaming.
    """

    @property
    def f_renames(self) -> RenamingClause:
        pass


class BodyStub(BodyNode):
    """
    Base class for a body stub (RM 10.1.3). A body stub is meant to be
    completed by .
    """

    @property
    def p_syntactic_fully_qualified_name(self) -> List[str]:
        """
        Return the syntactic fully qualified name to refer to this body.

        Note that this can raise a Property_Error when the stub is in an
        illegal place (too nested, in a declare block, etc.).
        """


class PackageBodyStub(BodyStub):
    """
    Stub for a package body (``is separate``).
    """

    @property
    def f_name(self) -> DefiningName:
        pass


class ProtectedBodyStub(BodyStub):
    """
    Stub for a protected object body (``is separate``).
    """

    @property
    def f_name(self) -> DefiningName:
        pass


class SubpBodyStub(BodyStub):
    """
    Stub for a subprogram body (``is separate``).
    """

    @property
    def f_overriding(self) -> OverridingNode:
        pass

    @property
    def f_subp_spec(self) -> SubpSpec:
        pass


class TaskBodyStub(BodyStub):
    """
    Stub for a task body (``is separate``).
    """

    @property
    def f_name(self) -> DefiningName:
        pass


class EntryBody(BodyNode):
    """
    Entry body.
    """

    @property
    def f_entry_name(self) -> DefiningName:
        pass

    @property
    def f_index_spec(self) -> EntryIndexSpec:
        pass

    @property
    def f_params(self) -> EntryCompletionFormalParams:
        pass

    @property
    def f_barrier(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class PackageBody(BodyNode):
    """
    Package body.
    """

    @property
    def f_package_name(self) -> DefiningName:
        pass

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class ProtectedBody(BodyNode):
    """
    Protected object body.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class TaskBody(BodyNode):
    """
    Task body.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class EntryIndexSpec(BasicDecl):
    """
    Index specification for an entry body.
    """

    @property
    def f_id(self) -> DefiningName:
        pass

    @property
    def f_subtype(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName
        """


class ErrorDecl(BasicDecl):
    """
    Placeholder node for syntax errors in lists of declarations.
    """

    pass


class ExceptionDecl(BasicDecl):
    """
    Exception declarations.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_renames(self) -> RenamingClause:
        pass


class ExceptionHandler(BasicDecl):
    """
    Exception handler.
    """

    @property
    def f_exception_name(self) -> DefiningName:
        pass

    @property
    def f_handled_exceptions(self) -> AlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * OthersDesignator

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class ForLoopVarDecl(BasicDecl):
    """
    Declaration for the controlling variable in a ``for`` loop.
    """

    @property
    def f_id(self) -> DefiningName:
        pass

    @property
    def f_id_type(self) -> SubtypeIndication:
        pass


class GenericDecl(BasicDecl):
    """
    Base class for generic declarations.
    """

    @property
    def f_formal_part(self) -> GenericFormalPart:
        pass


class GenericPackageDecl(GenericDecl):
    """
    Generic package declaration.
    """

    @property
    def f_package_decl(self) -> GenericPackageInternal:
        pass

    @property
    def p_body_part(self) -> PackageBody:
        """
        Return the PackageBody corresponding to this node, or null if there is
        none.
        """


class GenericSubpDecl(GenericDecl):
    """
    Generic subprogram declaration.
    """

    @property
    def f_subp_decl(self) -> GenericSubpInternal:
        pass

    def p_body_part(self, imprecise_fallback: bool = False) -> BaseSubpBody:
        """
        Return the BaseSubpBody corresponding to this node.
        """


class GenericInstantiation(BasicDecl):
    """
    Instantiations of generics.
    """

    @property
    def p_designated_generic_decl(self) -> BasicDecl:
        """
        Return the generic decl entity designated by this instantiation,
        containing the generic context. This is equivalent to the expanded
        generic unit in GNAT.
        """


class GenericPackageInstantiation(GenericInstantiation):
    """
    Instantiations of a generic package.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_generic_pkg_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_params(self) -> AssocList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ParamAssoc
        """


class GenericSubpInstantiation(GenericInstantiation):
    """
    Instantiations of a generic subprogram.
    """

    @property
    def f_overriding(self) -> OverridingNode:
        pass

    @property
    def f_kind(self) -> SubpKind:
        pass

    @property
    def f_subp_name(self) -> DefiningName:
        pass

    @property
    def f_generic_subp_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_params(self) -> AssocList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ParamAssoc
        """

    @property
    def p_designated_subp(self) -> AdaNode:
        """
        Return the subprogram decl designated by this instantiation.
        """


class GenericRenamingDecl(BasicDecl):
    """
    Base node for all generic renaming declarations.
    """

    pass


class GenericPackageRenamingDecl(GenericRenamingDecl):
    """
    Declaration for a generic package renaming.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_renames(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class GenericSubpRenamingDecl(GenericRenamingDecl):
    """
    Declaration for a generic subprogram renaming.
    """

    @property
    def f_kind(self) -> SubpKind:
        pass

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_renames(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class LabelDecl(BasicDecl):
    """
    Declaration for a code label.
    """

    @property
    def f_name(self) -> DefiningName:
        pass


class NamedStmtDecl(BasicDecl):
    """
    BasicDecl that is always the declaration inside a named statement.
    """

    @property
    def f_name(self) -> DefiningName:
        pass


class NumberDecl(BasicDecl):
    """
    Declaration for a static constant number.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class ObjectDecl(BasicDecl):
    """
    Base class for Ada object declarations (RM 3.3.1). Ada object declarations
    are variables/constants declarations that can be declared in any
    declarative scope.
    """

    @property
    def f_ids(self) -> DefiningNameList:
        pass

    @property
    def f_has_aliased(self) -> AliasedNode:
        pass

    @property
    def f_has_constant(self) -> ConstantNode:
        pass

    @property
    def f_mode(self) -> Mode:
        pass

    @property
    def f_type_expr(self) -> TypeExpr:
        pass

    @property
    def f_default_expr(self) -> Expr:
        pass

    @property
    def f_renaming_clause(self) -> RenamingClause:
        pass

    @property
    def p_public_part_decl(self) -> BasicDecl:
        """
        If this object decl is the constant completion of an object decl in the
        public part, return the object decl from the public part.
        """

    @property
    def p_private_part_decl(self) -> BasicDecl:
        """
        If this object decl is the incomplete declaration of a constant in a
        public part, return its completion in the private part.
        """


class AnonymousObjectDecl(ObjectDecl):
    """"""

    pass


class ExtendedReturnStmtObjectDecl(ObjectDecl):
    """
    Object declaration that is part of an extended return statement.
    """

    pass


class PackageRenamingDecl(BasicDecl):
    """
    Declaration for a package renaming.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_renames(self) -> RenamingClause:
        pass

    @property
    def p_renamed_package(self) -> BasicDecl:
        """
        Return the declaration of the package that is renamed by Self.
        """

    @property
    def p_final_renamed_package(self) -> BasicDecl:
        """
        Return the declaration of the package that is ultimately renamed by
        Self, skipping through all intermediate package renamings.
        """


class SingleProtectedDecl(BasicDecl):
    """
    Declaration for a single protected object.
    """

    @property
    def f_name(self) -> DefiningName:
        pass

    @property
    def f_interfaces(self) -> ParentList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_definition(self) -> ProtectedDef:
        pass


class SingleTaskDecl(BasicDecl):
    """
    Declaration for a single task.
    """

    @property
    def f_task_type(self) -> SingleTaskTypeDecl:
        pass


class CaseStmtAlternative(AdaNode):
    """
    Alternative in a ``case`` statement (``when ... => ...``).
    """

    @property
    def f_choices(self) -> AlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class CompilationUnit(AdaNode):
    """
    Root node for all Ada analysis units.
    """

    @property
    def f_prelude(self) -> AdaNodeList:
        """
        ``with``, ``use`` or ``pragma`` statements.

        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * UseClause

        * WithClause
        """

    @property
    def f_body(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * LibraryItem

        * Subunit
        """

    @property
    def f_pragmas(self) -> PragmaNodeList:
        pass

    @property
    def p_syntactic_fully_qualified_name(self) -> List[str]:
        """
        Return the syntactic fully qualified name of this compilation unit.
        """

    @property
    def p_unit_kind(self) -> str:
        """
        Return the kind corresponding to this analysis unit.
        """

    @property
    def p_withed_units(self) -> List[CompilationUnit]:
        """
        Look for all "with" clauses at the top of this compilation unit and
        return all the compilation units designated by them.
        """

    @property
    def p_imported_units(self) -> List[CompilationUnit]:
        """
        Return all the compilation units that are directly imported by this
        one. This includes "with"ed units as well as the direct parent unit.
        """

    @property
    def p_unit_dependencies(self) -> List[CompilationUnit]:
        """
        Return the list of all the compilation units that are (direct and
        indirect) dependencies of this one.
        """

    @property
    def p_decl(self) -> BasicDecl:
        """
        Get the root basic decl defined in this compilation unit.
        """

    def p_is_preelaborable(self, imprecise_fallback: bool = False) -> bool:
        """
        Whether this compilation unit is preelaborable or not.
        """


class ComponentClause(AdaNode):
    """
    Representation clause for a single component.
    """

    @property
    def f_id(self) -> Identifier:
        pass

    @property
    def f_position(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class ComponentDef(AdaNode):
    """
    Definition for a component.
    """

    @property
    def f_has_aliased(self) -> AliasedNode:
        pass

    @property
    def f_has_constant(self) -> ConstantNode:
        pass

    @property
    def f_type_expr(self) -> TypeExpr:
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication
        """


class ConstantNode(AdaNode):
    """
    Qualifier for the ``constant`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of ConstantPresent
        """


class ConstantAbsent(ConstantNode):
    """"""

    pass


class ConstantPresent(ConstantNode):
    """"""

    pass


class Constraint(AdaNode):
    """
    Base class for type constraints.
    """

    pass


class DeltaConstraint(Constraint):
    """
    Delta and range type constraint.
    """

    @property
    def f_digits(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class DigitsConstraint(Constraint):
    """
    Digits and range type constraint.
    """

    @property
    def f_digits(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class DiscriminantConstraint(Constraint):
    """
    List of constraints that relate to type discriminants.
    """

    @property
    def f_constraints(self) -> AssocList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * DiscriminantAssoc
        """


class IndexConstraint(Constraint):
    """
    List of type constraints.
    """

    @property
    def f_constraints(self) -> ConstraintList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName
        """


class RangeConstraint(Constraint):
    """
    Range-based type constraint.
    """

    @property
    def f_range(self) -> RangeSpec:
        pass


class DeclarativePart(AdaNode):
    """
    List of declarations.
    """

    @property
    def f_decls(self) -> AdaNodeList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AbstractSubpDecl

        * AspectClause

        * BodyNode

        * ComponentDecl

        * EntryDecl

        * ErrorDecl

        * ExceptionDecl

        * GenericDecl

        * GenericInstantiation

        * GenericRenamingDecl

        * IncompleteTypeDecl

        * NumberDecl

        * ObjectDecl

        * PackageDecl

        * PackageRenamingDecl

        * PragmaNode

        * ProtectedTypeDecl

        * SingleProtectedDecl

        * SingleTaskDecl

        * SubpDecl

        * SubtypeDecl

        * TaskTypeDecl

        * TypeDecl

        * UseClause
        """


class PrivatePart(DeclarativePart):
    """
    List of declarations in a private part.
    """

    pass


class PublicPart(DeclarativePart):
    """
    List of declarations in a public part.
    """

    pass


class ElsifExprPart(AdaNode):
    """
    ``elsif`` block, part of an ``if`` expression.
    """

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_then_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class ElsifStmtPart(AdaNode):
    """
    ``elsif`` part in an ``if`` statement block.
    """

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class Expr(AdaNode):
    """
    Base class for expressions.
    """

    @property
    def p_expression_type(self) -> BaseTypeDecl:
        """
        Return the declaration corresponding to the type of this expression
        after name resolution.
        """

    def p_is_dynamically_tagged(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns whether this expression is dynamically tagged (See RM 3.9.2).
        """

    def p_is_static_expr(self, imprecise_fallback: bool = False) -> bool:
        """
        Return whether this expression is static according to the ARM
        definition of static. See RM 4.9.
        """

    @property
    def p_first_corresponding_decl(self) -> BasicDecl:
        """
        Return the first decl that is lexically named like self in self's
        scope.
        """

    @property
    def p_eval_as_int(self) -> int:
        """
        Statically evaluates self, and returns the value of the evaluation as
        an integer.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    def p_eval_as_int_in_env(self, env: List[Substitution]) -> int:
        """
        Statically evaluates self, and returns the value of the evaluation as
        an integer. The given environment is used to substitute references to
        declarations by actual values.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    @property
    def p_eval_as_string(self) -> str:
        """
        Statically evaluates self, and returns the value of the evaluation as a
        string.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    def p_eval_as_string_in_env(self, env: List[Substitution]) -> str:
        """
        Statically evaluates self, and returns the value of the evaluation as a
        string. The given environment is used to substitute references to
        declarations by actual values.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.
        """

    @property
    def p_matching_nodes(self) -> List[AdaNode]:
        """
        Return the list of AST nodes that can be a match for this expression
        before overloading analysis.
        """


class Allocator(Expr):
    """
    Allocator expression (``new ...``).
    """

    @property
    def f_subpool(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_type_or_expr(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * QualExpr

        * SubtypeIndication
        """

    @property
    def p_get_allocated_type(self) -> BaseTypeDecl:
        """
        Return the allocated type for this allocator.
        """


class BaseAggregate(Expr):
    """
    Base class for aggregates.
    """

    @property
    def f_ancestor_expr(self) -> Expr:
        pass

    @property
    def f_assocs(self) -> AssocList:
        pass


class Aggregate(BaseAggregate):
    """
    Aggregate that is not a ``null record`` aggregate.
    """

    pass


class BracketAggregate(Aggregate):
    """
    Bracket array or container aggregate (Ada 2020).
    """

    pass


class DeltaAggregate(BaseAggregate):
    """"""

    pass


class BracketDeltaAggregate(DeltaAggregate):
    """
    Bracket delta aggregate (Ada 2020).
    """

    pass


class NullRecordAggregate(BaseAggregate):
    """
    Aggregate for ``null record``.
    """

    pass


class BinOp(Expr):
    """
    Binary expression.
    """

    @property
    def f_left(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_op(self) -> Op:
        """
        This field can contain one of the following nodes:

        * OpAnd

        * OpAndThen

        * OpConcat

        * OpDiv

        * OpDoubleDot

        * OpEq

        * OpGt

        * OpGte

        * OpLt

        * OpLte

        * OpMinus

        * OpMod

        * OpMult

        * OpNeq

        * OpOr

        * OpOrElse

        * OpPlus

        * OpPow

        * OpRem

        * OpXor
        """

    @property
    def f_right(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class RelationOp(BinOp):
    """
    Binary operation that compares two value, producing a boolean.
    """

    pass


class BoxExpr(Expr):
    """
    Box expression (``<>``).
    """

    pass


class CaseExprAlternative(Expr):
    """
    Alternative in a ``case`` expression (``when ... => ...``).
    """

    @property
    def f_choices(self) -> AlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class CondExpr(Expr):
    """
    Base class for a conditional expressions (RM 4.5.7).
    """

    @property
    def p_dependent_exprs(self) -> List[Expr]:
        """
        Return the dependent expressions for this conditional expression.
        """


class CaseExpr(CondExpr):
    """
    ``case`` expression (RM 4.5.7).
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_cases(self) -> CaseExprAlternativeList:
        pass


class IfExpr(CondExpr):
    """
    ``if`` expression (RM 4.5.7).
    """

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_then_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_alternatives(self) -> ElsifExprPartList:
        pass

    @property
    def f_else_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class ContractCases(Expr):
    """
    List of associations for the ``Contract_Case`` aspect.
    """

    @property
    def f_contract_cases(self) -> ContractCaseAssocList:
        pass


class DeclExpr(Expr):
    """
    Declare expression (Ada 2020).
    """

    @property
    def f_decls(self) -> BasicDeclList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * NumberDecl

        * ObjectDecl

        * SingleProtectedDecl

        * SingleTaskDecl
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class MembershipExpr(Expr):
    """
    Represent a membership test (in/not in operators).

    Note that we don't consider them as binary operators since multiple
    expressions on the right hand side are allowed.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_op(self) -> Op:
        """
        This field can contain one of the following nodes:

        * OpIn

        * OpNotIn
        """

    @property
    def f_membership_exprs(self) -> ExprAlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeName

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class Name(Expr):
    """
    Base class for names.
    """

    @property
    def p_enclosing_defining_name(self) -> DefiningName:
        """
        If this name is part of a defining name, return the enclosing defining
        name node.
        """

    @property
    def p_is_defining(self) -> bool:
        """
        Return True if this name is part of a defining name.
        """

    def p_name_is(self, sym: str) -> bool:
        """
        Helper. Check that this name matches ``sym``.
        """

    @property
    def p_is_direct_call(self) -> bool:
        """
        Return True iff this name represents a call to a subprogram which is
        referred by its defining name. (i.e. not through a subprogram access).
        """

    @property
    def p_is_access_call(self) -> bool:
        """
        Return True iff this name represents a call to subprogram through an
        access type.
        """

    @property
    def p_is_call(self) -> bool:
        """
        Returns True if this Name corresponds to a call.
        """

    def p_is_dot_call(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns True if this Name corresponds to a dot notation call.
        """

    def p_failsafe_referenced_def_name(
        self, imprecise_fallback: bool = False
    ) -> RefdDef:
        """
        Failsafe version of ``referenced_defining_name``. Returns a
        ``RefdDef``, which can be precise, imprecise, or error.
        """

    def p_referenced_defining_name(
        self, imprecise_fallback: bool = False
    ) -> DefiningName:
        """
        Like ``referenced_decl``, but will return the defining identifier for
        the decl, rather than the basic declaration node itself.
        """

    def p_all_env_elements(
        self, seq: bool = True, seq_from: AdaNode = None
    ) -> List[AdaNode]:
        """
        Return all elements in self's scope that are lexically named like Self.
        """

    @property
    def p_called_subp_spec(self) -> BaseFormalParamHolder:
        """
        Return the subprogram specification of the subprogram or subprogram
        access that is being called by this exact Name, if relevant.
        """

    def p_referenced_decl(self, imprecise_fallback: bool = False) -> BasicDecl:
        """
        Return the declaration this node references after name resolution. If
        imprecise_fallback is True, errors raised during resolution of the xref
        equation are catched and a fallback mechanism is triggered, which tries
        to find the referenced declaration in an ad-hoc way.
        """

    def p_failsafe_referenced_decl(self, imprecise_fallback: bool = False) -> RefdDecl:
        """
        Failsafe version of ``referenced_decl``. Returns a ``RefdDecl``, which
        can be precise, imprecise, or error.
        """

    def p_referenced_decl_internal(self, imprecise_fallback: bool = False) -> RefdDecl:
        """
        Return the declaration this node references. Try not to run name res if
        already resolved. INTERNAL USE ONLY.
        """

    @property
    def p_name_designated_type(self) -> BaseTypeDecl:
        """
        Like SubtypeIndication.designated_type, but on names, since because of
        Ada's ambiguous grammar, some subtype indications will be parsed as
        names.
        """

    def p_is_static_subtype(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns whether Self denotes a static subtype or not.
        """

    def p_name_matches(self, n: Name) -> bool:
        """
        Return whether two names match each other.

        This compares the symbol for Identifier and StringLiteral nodes. We
        consider that there is no match for all other node kinds.
        """

    @property
    def p_relative_name(self) -> SingleTokNode:
        """
        Returns the relative name of this instance. For example, for a prefix
        A.B.C, this will return C.
        """

    @property
    def p_is_operator_name(self) -> bool:
        """
        Return whether the name that Self designates is an operator.
        """

    def p_is_write_reference(self, imprecise_fallback: bool = False) -> bool:
        """
        Whether this name is a write reference.

        For example, `X` is a write reference in the following cases::

        1. `X := 2;` 2. `X (2) := 2;` 3. `P(F => X)` where F is declared `out`
        or `in out`. 6. `X.P` where the formal for X is declared `out` or `in
        out`. 4. `X'Access`. 5. `X.C := 2`, `R.X := 2`

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "write reference".
        """

    def p_is_dispatching_call(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns True if this Name corresponds to a dispatching call, including:

        - calls done through subprogram access types.

        - calls to dispatching subprograms, in the object-oriented sense.

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "dispatching call".
        """

    def p_is_static_call(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns True if this Name corresponds to a static non-dispatching call.
        In other words, this will return True if and only if the target of the
        call is known statically.

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "static call".
        """

    @property
    def p_as_symbol_array(self) -> List[str]:
        """
        Turn this name into an array of symbols.

        For instance, a node with name ``A.B.C`` is turned into ``['A', 'B',
        'C']``.
        """


class AttributeRef(Name):
    """
    Expression to reference an attribute.
    """

    @property
    def f_prefix(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_attribute(self) -> Identifier:
        pass

    @property
    def f_args(self) -> AdaNode:
        pass


class UpdateAttributeRef(AttributeRef):
    """
    Reference to the ``Update`` attribute.
    """

    pass


class CallExpr(Name):
    """
    Represent a syntactic call expression.

    At the semantic level, this can be either a subprogram call, an array
    subcomponent access expression, an array slice or a type conversion.
    """

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_suffix(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BasicAssocList

        * BinOp

        * CallExpr

        * CharLiteral

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def p_is_array_slice(self) -> bool:
        """
        Return whether this CallExpr is actually an access to a slice of the
        array denoted by the prefix of this CallExpr.
        """


class DefiningName(Name):
    """
    Name that defines an entity.
    """

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def p_basic_decl(self) -> BasicDecl:
        """
        Returns this DefiningName's basic declaration
        """

    def p_find_refs(
        self, root: AdaNode, origin: AdaNode, imprecise_fallback: bool = False
    ) -> List[RefResult]:
        """
        Find all references to this defining name in the given ``root`` and its
        children.
        """

    def p_find_all_references(
        self, units: List[AnalysisUnit], imprecise_fallback: bool = False
    ) -> List[RefResult]:
        """
        Searches all references to this defining name in the given list of
        units.
        """

    def p_find_all_calls(
        self, units: List[AnalysisUnit], imprecise_fallback: bool = False
    ) -> List[RefResult]:
        """
        Return the list of all possible calls to the subprogram which Self is
        the defining name of.

        This will return the name corresponding to the call, excluding the
        parameters if there are any. For instance, it will return `A` for the
        `A (B)` call.

        .. note:: This does not yet support calls done inside generics.
        """

    def p_next_part(self, imprecise_fallback: bool = False) -> DefiningName:
        """
        Like ``BasicDecl.next_part_for_decl`` on a defining name
        """

    def p_previous_part(self, imprecise_fallback: bool = False) -> DefiningName:
        """
        Like ``BasicDecl.previous_part_for_decl`` on a defining name
        """

    def p_canonical_part(self, imprecise_fallback: bool = False) -> DefiningName:
        """
        Like ``BasicDecl.canonical_part`` on a defining name
        """


class DiscreteSubtypeName(Name):
    """
    Subtype name for membership test expressions.
    """

    @property
    def f_subtype(self) -> DiscreteSubtypeIndication:
        pass


class DottedName(Name):
    """
    Name to select a suffix in a prefix.
    """

    @property
    def f_prefix(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_suffix(self) -> BaseId:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * Identifier

        * StringLiteral
        """


class EndName(Name):
    """
    Entity name in ``end ...;`` syntactic constructs.
    """

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def p_basic_decl(self) -> BasicDecl:
        """
        Returns this EndName's basic declaration
        """


class ExplicitDeref(Name):
    """
    Explicit dereference expression (``.all``).
    """

    @property
    def f_prefix(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class QualExpr(Name):
    """
    Qualified expression (``...'(...)``).
    """

    @property
    def f_prefix(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_suffix(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * BaseAggregate

        * ParenExpr
        """


class SingleTokNode(Name):
    """
    Base class for nodes that are made up of a single token.
    """

    @property
    def p_canonical_text(self) -> str:
        """
        Return a canonicalized version of this node's text.
        """


class BaseId(SingleTokNode):
    """
    Base class for identifiers.
    """

    pass


class CharLiteral(BaseId):
    """
    Character literal.
    """

    @property
    def p_denoted_value(self) -> str:
        """
        Return the value that this literal denotes.
        """


class Identifier(BaseId):
    """
    Regular identifier.
    """

    pass


class Op(BaseId):
    """
    Operation in a binary expression.

    Note that the ARM does not consider "double_dot" ("..") as a binary
    operator, but we process it this way here anyway to keep things simple.
    """

    pass


class OpAbs(Op):
    """"""

    pass


class OpAnd(Op):
    """"""

    pass


class OpAndThen(Op):
    """"""

    pass


class OpConcat(Op):
    """"""

    pass


class OpDiv(Op):
    """"""

    pass


class OpDoubleDot(Op):
    """"""

    pass


class OpEq(Op):
    """"""

    pass


class OpGt(Op):
    """"""

    pass


class OpGte(Op):
    """"""

    pass


class OpIn(Op):
    """"""

    pass


class OpLt(Op):
    """"""

    pass


class OpLte(Op):
    """"""

    pass


class OpMinus(Op):
    """"""

    pass


class OpMod(Op):
    """"""

    pass


class OpMult(Op):
    """"""

    pass


class OpNeq(Op):
    """"""

    pass


class OpNot(Op):
    """"""

    pass


class OpNotIn(Op):
    """"""

    pass


class OpOr(Op):
    """"""

    pass


class OpOrElse(Op):
    """"""

    pass


class OpPlus(Op):
    """"""

    pass


class OpPow(Op):
    """"""

    pass


class OpRem(Op):
    """"""

    pass


class OpXor(Op):
    """"""

    pass


class StringLiteral(BaseId):
    """
    String literal.
    """

    @property
    def p_denoted_value(self) -> str:
        """
        Return the value that this literal denotes.
        """


class NullLiteral(SingleTokNode):
    """
    The ``null`` literal.
    """

    pass


class NumLiteral(SingleTokNode):
    """
    Base class for number literals.
    """

    pass


class IntLiteral(NumLiteral):
    """
    Literal for an integer.
    """

    @property
    def p_denoted_value(self) -> int:
        """
        Return the value that this literal denotes.
        """


class RealLiteral(NumLiteral):
    """
    Literal for a real number.
    """

    pass


class TargetName(Name):
    """
    Name for Ada 2020 ``@``.
    """

    pass


class ParenExpr(Expr):
    """
    Parenthesized expression.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class QuantifiedExpr(Expr):
    """
    Quantified expression.
    """

    @property
    def f_quantifier(self) -> Quantifier:
        pass

    @property
    def f_loop_spec(self) -> ForLoopSpec:
        pass

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class RaiseExpr(Expr):
    """
    Expression to raise an exception.
    """

    @property
    def f_exception_name(self) -> Name:
        pass

    @property
    def f_error_message(self) -> Expr:
        pass


class UnOp(Expr):
    """
    Unary expression.
    """

    @property
    def f_op(self) -> Op:
        """
        This field can contain one of the following nodes:

        * OpAbs

        * OpMinus

        * OpNot

        * OpPlus
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class HandledStmts(AdaNode):
    """
    List of statements, with optional exception handlers.
    """

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """

    @property
    def f_exceptions(self) -> AdaNodeList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ExceptionHandler

        * PragmaNode
        """


class InterfaceKind(AdaNode):
    """
    Kind of interface type.
    """

    pass


class InterfaceKindLimited(InterfaceKind):
    """"""

    pass


class InterfaceKindProtected(InterfaceKind):
    """"""

    pass


class InterfaceKindSynchronized(InterfaceKind):
    """"""

    pass


class InterfaceKindTask(InterfaceKind):
    """"""

    pass


class IterType(AdaNode):
    """
    Iteration type for ``for`` loops.
    """

    pass


class IterTypeIn(IterType):
    """"""

    pass


class IterTypeOf(IterType):
    """"""

    pass


class LibraryItem(AdaNode):
    """
    Library item in a compilation unit.
    """

    @property
    def f_has_private(self) -> PrivateNode:
        pass

    @property
    def f_item(self) -> BasicDecl:
        """
        This field can contain one of the following nodes:

        * AbstractSubpDecl

        * BaseSubpBody

        * ErrorDecl

        * GenericDecl

        * GenericInstantiation

        * GenericRenamingDecl

        * PackageBody

        * PackageDecl

        * PackageRenamingDecl

        * SubpDecl
        """


class LimitedNode(AdaNode):
    """
    Qualifier for the ``limited`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of LimitedPresent
        """


class LimitedAbsent(LimitedNode):
    """"""

    pass


class LimitedPresent(LimitedNode):
    """"""

    pass


class LoopSpec(AdaNode):
    """
    Base class for loop specifications.
    """

    pass


class ForLoopSpec(LoopSpec):
    """
    Specification for a ``for`` loop.
    """

    @property
    def f_var_decl(self) -> ForLoopVarDecl:
        pass

    @property
    def f_loop_type(self) -> IterType:
        pass

    @property
    def f_has_reverse(self) -> ReverseNode:
        pass

    @property
    def f_iter_expr(self) -> AdaNode:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class WhileLoopSpec(LoopSpec):
    """
    Specification for a ``while`` loop.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class Mode(AdaNode):
    """
    Syntactic indicators for passing modes in formals.
    """

    pass


class ModeDefault(Mode):
    """"""

    pass


class ModeIn(Mode):
    """"""

    pass


class ModeInOut(Mode):
    """"""

    pass


class ModeOut(Mode):
    """"""

    pass


class NotNull(AdaNode):
    """
    Qualifier for the ``not null`` keywords.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of NotNullPresent
        """


class NotNullAbsent(NotNull):
    """"""

    pass


class NotNullPresent(NotNull):
    """"""

    pass


class NullComponentDecl(AdaNode):
    """
    Placeholder for the ``null`` in lists of components.
    """

    pass


class OthersDesignator(AdaNode):
    """
    ``other`` designator.
    """

    pass


class OverridingNode(AdaNode):
    """
    Syntactic indicators for subprogram overriding modes.
    """

    pass


class OverridingNotOverriding(OverridingNode):
    """"""

    pass


class OverridingOverriding(OverridingNode):
    """"""

    pass


class OverridingUnspecified(OverridingNode):
    """"""

    pass


class Params(AdaNode):
    """
    List of parameter specifications.
    """

    @property
    def f_params(self) -> ParamSpecList:
        pass


class PragmaNode(AdaNode):
    """
    Class for pragmas (RM 2.8). Pragmas are compiler directives, that can be
    language or compiler defined.
    """

    @property
    def f_id(self) -> Identifier:
        pass

    @property
    def f_args(self) -> BaseAssocList:
        pass

    @property
    def p_associated_decls(self) -> List[BasicDecl]:
        """
        Return an array of ``BasicDecl`` instances associated with this pragma,
        or an empty array if non applicable.
        """


class PrimTypeAccessor(AdaNode):
    """
    Synthetic node wrapping around a primitive type entity. Used in metadata.
    """

    pass


class PrivateNode(AdaNode):
    """
    Qualifier for the ``private`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of PrivatePresent
        """


class PrivateAbsent(PrivateNode):
    """"""

    pass


class PrivatePresent(PrivateNode):
    """"""

    pass


class ProtectedDef(AdaNode):
    """
    Type definition for a protected object.
    """

    @property
    def f_public_part(self) -> PublicPart:
        pass

    @property
    def f_private_part(self) -> PrivatePart:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class ProtectedNode(AdaNode):
    """
    Qualifier for the ``protected`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of ProtectedPresent
        """


class ProtectedAbsent(ProtectedNode):
    """"""

    pass


class ProtectedPresent(ProtectedNode):
    """"""

    pass


class Quantifier(AdaNode):
    """
    Type for quantified expressions.
    """

    pass


class QuantifierAll(Quantifier):
    """"""

    pass


class QuantifierSome(Quantifier):
    """"""

    pass


class RangeSpec(AdaNode):
    """
    Range specification.
    """

    @property
    def f_range(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class RenamingClause(AdaNode):
    """
    Renaming clause, used everywhere renamings are valid.
    """

    @property
    def f_renamed_object(self) -> Name:
        pass


class SyntheticRenamingClause(RenamingClause):
    """
    Synthetic renaming clause. Used to synthesize object decls with renamings.
    (See to_anonymous_object_decl).
    """

    pass


class ReverseNode(AdaNode):
    """
    Qualifier for the ``reverse`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of ReversePresent
        """


class ReverseAbsent(ReverseNode):
    """"""

    pass


class ReversePresent(ReverseNode):
    """"""

    pass


class SelectWhenPart(AdaNode):
    """
    Alternative part in a ``select`` statements block.
    """

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class Stmt(AdaNode):
    """
    Bass class for statements.
    """

    pass


class CompositeStmt(Stmt):
    """
    Base class for composite statements.
    """

    pass


class AcceptStmt(CompositeStmt):
    """
    ``accept`` statement.
    """

    @property
    def f_name(self) -> Identifier:
        pass

    @property
    def f_entry_index_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_params(self) -> EntryCompletionFormalParams:
        pass


class AcceptStmtWithStmts(AcceptStmt):
    """
    Extended ``accept`` statement.
    """

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class BaseLoopStmt(CompositeStmt):
    """
    Base class for loop statements.
    """

    @property
    def f_spec(self) -> LoopSpec:
        pass

    @property
    def f_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """

    @property
    def f_end_name(self) -> EndName:
        pass


class ForLoopStmt(BaseLoopStmt):
    """
    Statement for ``for`` loops (``for ... loop ... end loop;``).
    """

    pass


class LoopStmt(BaseLoopStmt):
    """
    Statement for simple loops (``loop ... end loop;``).
    """

    pass


class WhileLoopStmt(BaseLoopStmt):
    """
    Statement for ``while`` loops (``while ... loop ... end loop;``).
    """

    pass


class BlockStmt(CompositeStmt):
    """
    Base class for statement blocks.
    """

    pass


class BeginBlock(BlockStmt):
    """
    Statement block with no declarative part.
    """

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class DeclBlock(BlockStmt):
    """
    Statement block with a declarative part.
    """

    @property
    def f_decls(self) -> DeclarativePart:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class CaseStmt(CompositeStmt):
    """
    ``case`` statement.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_alternatives(self) -> CaseStmtAlternativeList:
        pass


class ExtendedReturnStmt(CompositeStmt):
    """
    Extended ``return`` statement.
    """

    @property
    def f_decl(self) -> ExtendedReturnStmtObjectDecl:
        pass

    @property
    def f_stmts(self) -> HandledStmts:
        pass


class IfStmt(CompositeStmt):
    """
    ``if`` statement block.
    """

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_then_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """

    @property
    def f_alternatives(self) -> ElsifStmtPartList:
        pass

    @property
    def f_else_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class NamedStmt(CompositeStmt):
    """
    Wrapper class, used for composite statements that can be named (declare
    blocks, loops). This allows to both have a BasicDecl for the named entity
    declared, and a CompositeStmt for the statement hierarchy.
    """

    @property
    def f_decl(self) -> NamedStmtDecl:
        pass

    @property
    def f_stmt(self) -> CompositeStmt:
        """
        This field can contain one of the following nodes:

        * BaseLoopStmt

        * BlockStmt
        """


class SelectStmt(CompositeStmt):
    """
    ``select`` statements block.
    """

    @property
    def f_guards(self) -> SelectWhenPartList:
        pass

    @property
    def f_else_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """

    @property
    def f_abort_stmts(self) -> StmtList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt
        """


class ErrorStmt(Stmt):
    """
    Placeholder node for syntax errors in lists of statements.
    """

    pass


class SimpleStmt(Stmt):
    """
    Base class for simple statements.
    """

    pass


class AbortStmt(SimpleStmt):
    """
    ``abort`` statement.
    """

    @property
    def f_names(self) -> NameList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class AssignStmt(SimpleStmt):
    """
    Statement for assignments.
    """

    @property
    def f_dest(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class CallStmt(SimpleStmt):
    """
    Statement for entry or procedure calls.
    """

    @property
    def f_call(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class DelayStmt(SimpleStmt):
    """
    ``delay`` statement.
    """

    @property
    def f_has_until(self) -> UntilNode:
        pass

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class ExitStmt(SimpleStmt):
    """
    ``exit`` statement.
    """

    @property
    def f_loop_name(self) -> Identifier:
        pass

    @property
    def f_cond_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class GotoStmt(SimpleStmt):
    """
    ``goto`` statement.
    """

    @property
    def f_label_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class Label(SimpleStmt):
    """
    Statement to declare a code label.
    """

    @property
    def f_decl(self) -> LabelDecl:
        pass


class NullStmt(SimpleStmt):
    """
    ``null;`` statement.
    """

    pass


class RaiseStmt(SimpleStmt):
    """
    ``raise`` statement.
    """

    @property
    def f_exception_name(self) -> Name:
        pass

    @property
    def f_error_message(self) -> Expr:
        pass


class RequeueStmt(SimpleStmt):
    """
    ``requeue`` statement.
    """

    @property
    def f_call_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """

    @property
    def f_has_abort(self) -> AbortNode:
        pass


class ReturnStmt(SimpleStmt):
    """
    ``return`` statement.
    """

    @property
    def f_return_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class TerminateAlternative(SimpleStmt):
    """
    ``terminate`` alternative in a ``select`` statement.
    """

    pass


class SubpKind(AdaNode):
    """
    Qualifier for a subprogram kind.
    """

    pass


class SubpKindFunction(SubpKind):
    """"""

    pass


class SubpKindProcedure(SubpKind):
    """"""

    pass


class Subunit(AdaNode):
    """
    Subunit (``separate``).
    """

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_body(self) -> BodyNode:
        """
        This field can contain one of the following nodes:

        * PackageBody

        * ProtectedBody

        * SubpBody

        * TaskBody
        """

    @property
    def p_body_root(self) -> BasicDecl:
        """
        Return the body in which this subunit is rooted.
        """


class SynchronizedNode(AdaNode):
    """
    Qualifier for the ``synchronized`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of SynchronizedPresent
        """


class SynchronizedAbsent(SynchronizedNode):
    """"""

    pass


class SynchronizedPresent(SynchronizedNode):
    """"""

    pass


class TaggedNode(AdaNode):
    """
    Qualifier for the ``tagged`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of TaggedPresent
        """


class TaggedAbsent(TaggedNode):
    """"""

    pass


class TaggedPresent(TaggedNode):
    """"""

    pass


class TaskDef(AdaNode):
    """
    Type definition for a task type.
    """

    @property
    def f_interfaces(self) -> ParentList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_public_part(self) -> PublicPart:
        pass

    @property
    def f_private_part(self) -> PrivatePart:
        pass

    @property
    def f_end_name(self) -> EndName:
        pass


class TypeDef(AdaNode):
    """
    Base class for type definitions.
    """

    pass


class AccessDef(TypeDef):
    """
    Base class for access type definitions.
    """

    @property
    def f_has_not_null(self) -> NotNull:
        pass


class AccessToSubpDef(AccessDef):
    """
    Type definition for accesses to subprograms.
    """

    @property
    def f_has_protected(self) -> ProtectedNode:
        pass

    @property
    def f_subp_spec(self) -> SubpSpec:
        pass


class BaseTypeAccessDef(AccessDef):
    """
    Base class for access type definitions.
    """

    pass


class AnonymousTypeAccessDef(BaseTypeAccessDef):
    """
    Synthetic type access, that will directly reference a type decl. It is used
    to generate synthetic anonymous access types.
    """

    @property
    def f_type_decl(self) -> BaseTypeDecl:
        pass


class TypeAccessDef(BaseTypeAccessDef):
    """
    Syntactic type definition for accesses.
    """

    @property
    def f_has_all(self) -> AllNode:
        pass

    @property
    def f_has_constant(self) -> ConstantNode:
        pass

    @property
    def f_subtype_indication(self) -> SubtypeIndication:
        pass


class ArrayTypeDef(TypeDef):
    """
    Type definition for an array.
    """

    @property
    def f_indices(self) -> ArrayIndices:
        pass

    @property
    def f_component_type(self) -> ComponentDef:
        pass


class DerivedTypeDef(TypeDef):
    """
    Type definition for a derived type.
    """

    @property
    def f_has_abstract(self) -> AbstractNode:
        pass

    @property
    def f_has_limited(self) -> LimitedNode:
        pass

    @property
    def f_has_synchronized(self) -> SynchronizedNode:
        pass

    @property
    def f_subtype_indication(self) -> SubtypeIndication:
        pass

    @property
    def f_interfaces(self) -> ParentList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_record_extension(self) -> BaseRecordDef:
        pass

    @property
    def f_has_with_private(self) -> WithPrivate:
        pass


class EnumTypeDef(TypeDef):
    """
    Type definition for enumerations.
    """

    @property
    def f_enum_literals(self) -> EnumLiteralDeclList:
        pass


class FormalDiscreteTypeDef(TypeDef):
    """
    Type definition for discrete types in generic formals.
    """

    pass


class InterfaceTypeDef(TypeDef):
    """
    Type definition for an interface.
    """

    @property
    def f_interface_kind(self) -> InterfaceKind:
        pass

    @property
    def f_interfaces(self) -> ParentList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class ModIntTypeDef(TypeDef):
    """
    Type definition for a modular integer type.
    """

    @property
    def f_expr(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """


class PrivateTypeDef(TypeDef):
    """
    Type definition for a private type.
    """

    @property
    def f_has_abstract(self) -> AbstractNode:
        pass

    @property
    def f_has_tagged(self) -> TaggedNode:
        pass

    @property
    def f_has_limited(self) -> LimitedNode:
        pass


class RealTypeDef(TypeDef):
    """
    Type definition for real numbers.
    """

    pass


class DecimalFixedPointDef(RealTypeDef):
    """
    Type definition for decimal fixed-point numbers.
    """

    @property
    def f_delta(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_digits(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class FloatingPointDef(RealTypeDef):
    """
    Type definition for floating-point numbers.
    """

    @property
    def f_num_digits(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class OrdinaryFixedPointDef(RealTypeDef):
    """
    Type definition for ordinary fixed-point numbers.
    """

    @property
    def f_delta(self) -> Expr:
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_range(self) -> RangeSpec:
        pass


class RecordTypeDef(TypeDef):
    """
    Type definition for a record.
    """

    @property
    def f_has_abstract(self) -> AbstractNode:
        pass

    @property
    def f_has_tagged(self) -> TaggedNode:
        pass

    @property
    def f_has_limited(self) -> LimitedNode:
        pass

    @property
    def f_record_def(self) -> BaseRecordDef:
        pass


class SignedIntTypeDef(TypeDef):
    """
    Type definition for a signed integer type.
    """

    @property
    def f_range(self) -> RangeSpec:
        pass


class TypeExpr(AdaNode):
    """
    A type expression is an abstract node that embodies the concept of a
    reference to a type.

    Since Ada has both subtype_indications and anonymous (inline) type
    declarations, a type expression contains one or the other.
    """

    @property
    def p_type_name(self) -> Name:
        """
        Return the name node for this type expression, if applicable, else null
        """

    @property
    def p_designated_type_decl(self) -> BaseTypeDecl:
        """
        Returns the type declaration designated by this type expression.
        """

    def p_designated_type_decl_from(self, origin_node: AdaNode) -> BaseTypeDecl:
        """
        Return the type declaration designated by this type expression as
        viewed from the node given by origin_node.
        """


class AnonymousType(TypeExpr):
    """
    Container for inline anonymous array and access types declarations.
    """

    @property
    def f_type_decl(self) -> AnonymousTypeDecl:
        pass


class EnumLitSynthTypeExpr(TypeExpr):
    """
    Synthetic node. Represents the type expression for an enum literal.
    """

    pass


class SubtypeIndication(TypeExpr):
    """
    Reference to a type by name.
    """

    @property
    def f_has_not_null(self) -> NotNull:
        pass

    @property
    def f_name(self) -> Name:
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """

    @property
    def f_constraint(self) -> Constraint:
        pass

    def p_is_static_subtype(self, imprecise_fallback: bool = False) -> bool:
        """
        Returns whether Self denotes a static subtype or not.
        """


class ConstrainedSubtypeIndication(SubtypeIndication):
    """
    Reference to a type with a range constraint.
    """

    pass


class DiscreteSubtypeIndication(SubtypeIndication):
    """
    Reference to a type with a general constraint.
    """

    pass


class UnconstrainedArrayIndex(AdaNode):
    """
    List of unconstrained array indexes.
    """

    @property
    def f_subtype_indication(self) -> SubtypeIndication:
        pass


class UntilNode(AdaNode):
    """
    Qualifier for the ``until`` keyword.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of UntilPresent
        """


class UntilAbsent(UntilNode):
    """"""

    pass


class UntilPresent(UntilNode):
    """"""

    pass


class UseClause(AdaNode):
    """
    Base class for use clauses.
    """

    pass


class UsePackageClause(UseClause):
    """
    Use clause for packages.
    """

    @property
    def f_packages(self) -> NameList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class UseTypeClause(UseClause):
    """
    Use clause for types.
    """

    @property
    def f_has_all(self) -> AllNode:
        pass

    @property
    def f_types(self) -> NameList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName
        """


class Variant(AdaNode):
    """
    Single variant in a discriminated type record declaration.

    This corresponds to a ``when ... => ...`` section in a variant part.
    """

    @property
    def f_choices(self) -> AlternativesList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp
        """

    @property
    def f_components(self) -> ComponentList:
        pass


class VariantPart(AdaNode):
    """
    Variant part in a discriminated type record declaration.

    This corresponds to the whole ``case ... is ... end case;`` block.
    """

    @property
    def f_discr_name(self) -> Identifier:
        pass

    @property
    def f_variant(self) -> VariantList:
        pass


class WithClause(AdaNode):
    """
    With clause.
    """

    @property
    def f_has_limited(self) -> LimitedNode:
        pass

    @property
    def f_has_private(self) -> PrivateNode:
        pass

    @property
    def f_packages(self) -> NameList:
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral
        """


class WithPrivate(AdaNode):
    """
    Qualifier for the ``private`` keyword in ``with private`` record clauses.
    """

    @property
    def p_as_bool(self) -> bool:
        """
        Return whether this is an instance of WithPrivatePresent
        """


class WithPrivateAbsent(WithPrivate):
    """"""

    pass


class WithPrivatePresent(WithPrivate):
    """"""

    pass


class Aspect(object):
    """
    Composite field representing the aspect of an entity (RM 13).
    """

    def __init__(
        self,
        exists: bool,
        node: AdaNode,
        value: Expr,
    ) -> None:
        ...

    @property
    def exists(self) -> bool:
        """
        Whether the aspect is defined or not
        """

    @property
    def node(self) -> AdaNode:
        """
        Syntactic node that defines the aspect
        """

    @property
    def value(self) -> Expr:
        """
        Expr node defining the value of the aspect
        """


class CompletionItem(object):
    """"""

    def __init__(
        self,
        decl: BasicDecl,
        is_dot_call: bool,
        is_visible: bool,
    ) -> None:
        ...

    @property
    def decl(self) -> BasicDecl:
        pass

    @property
    def is_dot_call(self) -> bool:
        pass

    @property
    def is_visible(self) -> bool:
        pass


class DiscreteRange(object):
    """
    Represent the range of a discrete type or subtype. The bounds are not
    evaluated, you need to call ``eval_as_int`` on them, if they're static, to
    get their value.
    """

    def __init__(
        self,
        low_bound: Expr,
        high_bound: Expr,
    ) -> None:
        ...

    @property
    def low_bound(self) -> Expr:
        pass

    @property
    def high_bound(self) -> Expr:
        pass


class DiscriminantValues(object):
    """
    Represent a set of values (as a list of choices) on a discriminant.
    """

    def __init__(
        self,
        discriminant: Identifier,
        values: AlternativesList,
    ) -> None:
        ...

    @property
    def discriminant(self) -> Identifier:
        pass

    @property
    def values(self) -> AlternativesList:
        pass


class DocAnnotation(object):
    """
    Documentation annotation.
    """

    def __init__(
        self,
        key: str,
        value: str,
    ) -> None:
        ...

    @property
    def key(self) -> str:
        """
        Annotation key
        """

    @property
    def value(self) -> str:
        """
        Annotation value
        """


class ParamActual(object):
    """
    Data structure used by zip_with_params property. Associates an expression
    (the actual) to a formal param declaration (the parameter).
    """

    def __init__(
        self,
        param: DefiningName,
        actual: Expr,
    ) -> None:
        ...

    @property
    def param(self) -> DefiningName:
        pass

    @property
    def actual(self) -> Expr:
        pass


class RefResult(object):
    """
    Result for a cross reference query returning a reference.
    """

    def __init__(
        self,
        ref: BaseId,
        kind: str,
    ) -> None:
        ...

    @property
    def ref(self) -> BaseId:
        pass

    @property
    def kind(self) -> str:
        pass


class RefdDecl(object):
    """
    Result for a cross reference query returning a referenced decl.
    """

    def __init__(
        self,
        decl: BasicDecl,
        kind: str,
    ) -> None:
        ...

    @property
    def decl(self) -> BasicDecl:
        pass

    @property
    def kind(self) -> str:
        pass


class RefdDef(object):
    """
    Result for a cross reference query returning a referenced defining name.
    """

    def __init__(
        self,
        def_name: DefiningName,
        kind: str,
    ) -> None:
        ...

    @property
    def def_name(self) -> DefiningName:
        pass

    @property
    def kind(self) -> str:
        pass


class Shape(object):
    """
    Represent one of the shapes that a variant record can have, as a list of
    the available components.
    """

    def __init__(
        self,
        components: List[BaseFormalParamDecl],
        discriminants_values: List[DiscriminantValues],
    ) -> None:
        ...

    @property
    def components(self) -> List[BaseFormalParamDecl]:
        pass

    @property
    def discriminants_values(self) -> List[DiscriminantValues]:
        pass


class Substitution(object):
    """
    Represent a substitution of a BasicDecl by a given value. This can then be
    used as part of an environment in the eval_as_*_in_env property. See the
    declaration of those properties for more details.
    """

    def __init__(
        self,
        from_decl: BasicDecl,
        to_value: int,
        value_type: BaseTypeDecl,
    ) -> None:
        ...

    @property
    def from_decl(self) -> BasicDecl:
        """
        The declaration to substitute.
        """

    @property
    def to_value(self) -> int:
        """
        The value by which to substitute the declaration.
        """

    @property
    def value_type(self) -> BaseTypeDecl:
        """
        The type of the substituted value.
        """


class App(object):
    parser: argparse.ArgumentParser
    args: argparse.Namespace
    u: AnalysisUnit
    units: Dict[str, AnalysisUnit]
    ctx: AnalysisContext

    @property
    def description(self) -> str:
        ...

    def __init__(self, args: Opt[List[str]]) -> None:
        ...

    def add_arguments(self) -> None:
        ...

    def create_unit_provider(self) -> Opt[UnitProvider]:
        ...

    def process_unit(self, unit: AnalysisUnit) -> None:
        ...

    @classmethod
    def run(cls, args: Opt[List[str]] = None) -> None:
        ...
