"""
Python binding of the Libadalang API.

Please consider all exported entities whose names that start with an underscore
("_") as internal implementation details. They are not meant to be used
directly.
"""


from __future__ import absolute_import, division, print_function


import argparse
import collections
import ctypes
import json
import os
import sys
import weakref

import libadalang._py2to3 as _py2to3


#
# Low-level binding - First part
#

_so_ext = {
    "win32": "dll",
    "darwin": "dylib",
}.get(sys.platform, "so")

# Loading the shared library here is quite involved as we want to support
# Python packages that embed all the required shared libraries: if we can
# find the shared library in the package directory, import it from there
# directly.

# Directory that contains this __init__.py module
_self_path = os.path.dirname(os.path.abspath(__file__))

# Base and full names for the shared library to load. Full name assumes the
# shared lib is in the package directory.
_c_lib_name = "libadalang.{}".format(_so_ext)
_c_lib_path = os.path.join(_self_path, _c_lib_name)

# If we can find the shared lirbray in the package directory, load it from
# here, otherwise let the dynamic loader find it in the environment. On
# Windows, there is no RPATH trick, so we need to temporarily alter the PATH
# environment variable in order to import the whole closure of DLLs.
_old_env_path = None
if os.path.exists(_c_lib_path):
    if sys.platform == "win32":
        _old_env_path = os.environ["PATH"]
        os.environ["PATH"] = "{}{}{}".format(
            _self_path, os.path.pathsep, os.environ["PATH"]
        )
else:
    _c_lib_path = _c_lib_name

# Finally load the library
_c_lib = ctypes.cdll.LoadLibrary(_c_lib_path)

# Restore the PATH environment variable if we altered it
if _old_env_path is not None:
    os.environ["PATH"] = _old_env_path


def _import_func(name, argtypes, restype, exc_wrap=True):
    """
    Import "name" from the C library, set its arguments/return types and return
    the binding.

    :param str name: Name of the symbol for the function to import.
    :param list[ctypes._CData] argtypes: Types for function argruments.
    :param None|ctypes._CData restype: Function return type, or None if it
        does not return anything.
    :param bool exc_wrap: If True, wrap the returned function to check for
      exceptions.
    """
    func = getattr(_c_lib, name)
    func.argtypes = argtypes
    func.restype = restype

    def check_argcount(args, kwargs):
        argcount = len(args) + len(kwargs)
        if argcount != len(argtypes):
            raise TypeError(
                "{} takes {} positional arguments but {} was given".format(
                    name, len(argtypes), argcount
                )
            )

    # Wrapper for "func" that raises a NativeException in case of internal
    # error.

    if exc_wrap:

        def wrapper(*args, **kwargs):
            check_argcount(args, kwargs)
            result = func(*args, **kwargs)
            exc = _get_last_exception()
            if exc:
                raise exc.contents._wrap()
            return result

    else:

        def wrapper(*args, **kwargs):
            check_argcount(args, kwargs)
            return func(*args, **kwargs)

    return wrapper


class _Exception(ctypes.Structure):
    _fields_ = [("kind", ctypes.c_int), ("information", ctypes.c_char_p)]

    def _wrap(self):
        # In Python3, turn information into native strings, i.e. decode bytes.
        # These strings are only informative, so do not raise an error if
        # decoding fails: do best effort decoding instead to be as helpful as
        # possible.
        info = self.information
        if not _py2to3.python2:
            info = info.decode(errors="replace")
        return _exception_kind_to_type[self.kind](info)


def _type_fullname(t):
    """
    Return the fully qualified name for the given `t` type.
    """
    name = t.__name__
    module = t.__module__
    return (
        name
        if module in (None, object.__class__.__module__)
        else "{}.{}".format(module, name)
    )


def _raise_type_error(expected_type_name, actual_value):
    raise TypeError(
        "{} instance expected, got {} instead".format(
            expected_type_name, _type_fullname(type(actual_value))
        )
    )


_get_last_exception = _import_func(
    "ada_get_last_exception", [], ctypes.POINTER(_Exception), exc_wrap=False
)


def _hashable_c_pointer(pointed_type=None):
    """
    Create a "pointer to `pointed_type` type and make it hashable.

    :param pointed_type: ctypes type class. If left to `None`, we return a
        subclass of `ctypes.c_void_p`.
    :rtype: ctypes.POINTER
    """

    if pointed_type is None:

        class _c_type(ctypes.c_void_p):
            @property
            def _pointer_value(self):
                return self.value or 0

    else:

        @property
        def _pointer_value(self):
            return ctypes.addressof(self.contents)

        _c_type = ctypes.POINTER(pointed_type)
        _c_type._pointer_value = _pointer_value

    def __hash__(self):
        return self._pointer_value

    def __eq__(self, other):
        return self._pointer_value == other._pointer_value

    def __ne__(self, other):
        return not (self == other)

    _c_type.__hash__ = __hash__
    _c_type.__eq__ = __eq__
    _c_type.__ne__ = __ne__
    return _c_type


class _text(ctypes.Structure):
    """
    C value for unicode strings. This object is the owner of the underlying
    buffer, so the string will be deallocated when ``self`` is destroyed.

    ``_unwrap`` takes a string/unicode object and returns a ``_text`` instance,
    while ``_wrap`` retuns an unicode instance.
    """

    # The chars field really is a uint32_t* but considering it as a char* here
    # is more convenient for conversion in this binding layer. On the other
    # side, we have to be careful about converting the length when retrieving
    # the chars.
    _fields_ = [
        ("chars", ctypes.POINTER(ctypes.c_char)),
        ("length", ctypes.c_size_t),
        ("is_allocated", ctypes.c_int),
    ]

    encoding = "utf-32le" if sys.byteorder == "little" else "utf-32be"

    # Instances can hold buffers that they own. In this case, the buffer must
    # be deallocated when the instance is destroyed. Thus instances will hold
    # a "text_buffer" attribute that will be automatically destroyed.
    text_buffer = None

    @classmethod
    def _unwrap(cls, value):
        value = cls.cast(value)

        text = value.encode(cls.encoding)
        text_buffer = ctypes.create_string_buffer(text)
        text_buffer_ptr = ctypes.cast(
            ctypes.pointer(text_buffer), ctypes.POINTER(ctypes.c_char)
        )
        result = _text(text_buffer_ptr, len(value))
        result.text_buffer = text_buffer
        return result

    def _wrap(self):
        if self.length > 0:
            # self.length tells how much UTF-32 chars there are in self.chars
            # but self.chars is a char* so we have to fetch 4 times more bytes
            # than characters.
            return self.chars[: 4 * self.length].decode(self.encoding)
        else:
            return u""

    @classmethod
    def cast(cls, value):
        """
        Try to cast ``value`` into an unicode object. Raise a TypeError, or
        raise a string decoding error when this is not possible.
        """
        if isinstance(value, _py2to3.bytes_type):
            return value.decode("ascii")
        elif not isinstance(value, _py2to3.text_type):
            _raise_type_error("text string", value)
        else:
            return value

    def __del__(self):
        _destroy_text(ctypes.byref(self))


class _symbol_type(ctypes.Structure):
    _fields_ = [("data", ctypes.c_void_p), ("bounds", ctypes.c_void_p)]

    @classmethod
    def wrap(cls, c_value):
        # First extract the text associated to this symbol in "text"
        text = _text()
        _symbol_text(ctypes.byref(c_value), ctypes.byref(text))

        # Then wrap this text
        return text._wrap()

    @classmethod
    def unwrap(cls, py_value, context):
        # First turn the given symbol into a low-level text object
        text = _text._unwrap(py_value)

        # Then convert it to a symbol
        result = cls()
        if not _context_symbol(context, ctypes.byref(text), ctypes.byref(result)):
            raise InvalidSymbolError(py_value)
        return result


class _big_integer(object):
    class c_type(ctypes.c_void_p):
        pass

    def __init__(self, c_value):
        self.c_value = c_value

    @classmethod
    def unwrap(cls, value):
        if not _py2to3.is_int(value):
            _raise_type_error("int or long", value)

        text = _text._unwrap(str(value))
        c_value = cls.create(ctypes.byref(text))
        return cls(c_value)

    @classmethod
    def wrap(cls, c_value):
        helper = cls(c_value)
        text = _text()
        cls.text(helper.c_value, ctypes.byref(text))
        return int(text._wrap())

    def clear(self):
        self.c_value = None

    def __del__(self):
        self.decref(self.c_value)
        self.clear()

    create = staticmethod(
        _import_func("ada_create_big_integer", [ctypes.POINTER(_text)], c_type)
    )
    text = staticmethod(
        _import_func("ada_big_integer_text", [c_type, ctypes.POINTER(_text)], None)
    )
    decref = staticmethod(_import_func("ada_big_integer_decref", [c_type], None))


class _Enum(object):

    _name = None
    """
    Name for this enumeration type.
    :type: str
    """

    _c_to_py = None
    """
    Mapping from C values to user-level Python values.
    :type: list[str]
    """

    _py_to_c = None
    """
    Mapping from user-level Python values to C values.
    :type: dict[str, int]
    """

    @classmethod
    def _unwrap(cls, py_value):
        if not isinstance(py_value, str):
            _raise_type_error("str", py_value)
        try:
            return _py2to3.text_to_bytes(cls._py_to_c[py_value])
        except KeyError:
            raise ValueError("Invalid {}: {}".format(cls._name, py_value))

    @classmethod
    def _wrap(cls, c_value):
        if isinstance(c_value, ctypes.c_int):
            c_value = c_value.value
        return cls._c_to_py[c_value]


class AnalysisUnitKind(_Enum):
    """
    Specify a kind of analysis unit. Specification units provide an interface
    to the outer world while body units provide an implementation for the
    corresponding interface.
    """

    unit_specification = "unit_specification"
    unit_body = "unit_body"

    _name = "AnalysisUnitKind"
    _c_to_py = [unit_specification, unit_body]
    _py_to_c = {name: index for index, name in enumerate(_c_to_py)}


class LookupKind(_Enum):
    """"""

    recursive = "recursive"
    flat = "flat"
    minimal = "minimal"

    _name = "LookupKind"
    _c_to_py = [recursive, flat, minimal]
    _py_to_c = {name: index for index, name in enumerate(_c_to_py)}


class FindAllMode(_Enum):
    """"""

    references = "references"
    derivedtypes = "derivedtypes"

    _name = "FindAllMode"
    _c_to_py = [references, derivedtypes]
    _py_to_c = {name: index for index, name in enumerate(_c_to_py)}


class RefResultKind(_Enum):
    """
    Kind for the result of a cross reference operation.

    - ``NoRef`` is for no reference, it is the null value for this enum.

    - ``Precise`` is when the reference result is precise.

    - ``Imprecise`` is when there was an error computing the precise result,
    and a result was gotten in an imprecise fashion.

    - ``Error`` is for unrecoverable errors (either there is no imprecise path
    for the request you made, or the imprecise path errored out too.
    """

    noref = "noref"
    precise = "precise"
    imprecise = "imprecise"
    error = "error"

    _name = "RefResultKind"
    _c_to_py = [noref, precise, imprecise, error]
    _py_to_c = {name: index for index, name in enumerate(_c_to_py)}


class GrammarRule(_Enum):
    """
    Gramar rule to use for parsing.
    """

    parent_list_rule = "parent_list_rule"
    protected_type_decl_rule = "protected_type_decl_rule"
    protected_op_rule = "protected_op_rule"
    protected_el_rule = "protected_el_rule"
    protected_def_rule = "protected_def_rule"
    protected_decl_rule = "protected_decl_rule"
    task_item_rule = "task_item_rule"
    task_def_rule = "task_def_rule"
    task_type_decl_rule = "task_type_decl_rule"
    subtype_decl_rule = "subtype_decl_rule"
    interface_type_def_rule = "interface_type_def_rule"
    unconstrained_index_rule = "unconstrained_index_rule"
    array_type_def_rule = "array_type_def_rule"
    discrete_subtype_definition_rule = "discrete_subtype_definition_rule"
    constraint_list_rule = "constraint_list_rule"
    signed_int_type_def_rule = "signed_int_type_def_rule"
    mod_int_type_def_rule = "mod_int_type_def_rule"
    derived_type_def_rule = "derived_type_def_rule"
    discriminant_assoc_rule = "discriminant_assoc_rule"
    discriminant_constraint_rule = "discriminant_constraint_rule"
    index_constraint_rule = "index_constraint_rule"
    digits_constraint_rule = "digits_constraint_rule"
    delta_constraint_rule = "delta_constraint_rule"
    range_constraint_rule = "range_constraint_rule"
    constraint_rule = "constraint_rule"
    discriminant_spec_rule = "discriminant_spec_rule"
    discr_spec_list_rule = "discr_spec_list_rule"
    discriminant_part_rule = "discriminant_part_rule"
    enum_literal_decl_rule = "enum_literal_decl_rule"
    formal_discrete_type_def_rule = "formal_discrete_type_def_rule"
    record_def_rule = "record_def_rule"
    range_spec_rule = "range_spec_rule"
    real_type_def_rule = "real_type_def_rule"
    sexpr_or_box_rule = "sexpr_or_box_rule"
    ordinary_fixed_point_def_rule = "ordinary_fixed_point_def_rule"
    decimal_fixed_point_def_rule = "decimal_fixed_point_def_rule"
    floating_point_def_rule = "floating_point_def_rule"
    record_type_def_rule = "record_type_def_rule"
    access_def_rule = "access_def_rule"
    enum_type_def_rule = "enum_type_def_rule"
    type_def_rule = "type_def_rule"
    variant_rule = "variant_rule"
    anonymous_type_decl_rule = "anonymous_type_decl_rule"
    type_decl_rule = "type_decl_rule"
    variant_part_rule = "variant_part_rule"
    component_def_rule = "component_def_rule"
    component_item_rule = "component_item_rule"
    component_decl_rule = "component_decl_rule"
    component_list_rule = "component_list_rule"
    generic_decl_rule = "generic_decl_rule"
    generic_formal_part_rule = "generic_formal_part_rule"
    generic_formal_decl_rule = "generic_formal_decl_rule"
    formal_subp_decl_rule = "formal_subp_decl_rule"
    renaming_clause_rule = "renaming_clause_rule"
    generic_renaming_decl_rule = "generic_renaming_decl_rule"
    generic_instantiation_rule = "generic_instantiation_rule"
    exception_decl_rule = "exception_decl_rule"
    basic_decls_rule = "basic_decls_rule"
    package_renaming_decl_rule = "package_renaming_decl_rule"
    package_decl_rule = "package_decl_rule"
    basic_decl_rule = "basic_decl_rule"
    object_decl_rule = "object_decl_rule"
    sub_object_decl_rule = "sub_object_decl_rule"
    ext_ret_stmt_object_decl_rule = "ext_ret_stmt_object_decl_rule"
    defining_id_list_rule = "defining_id_list_rule"
    number_decl_rule = "number_decl_rule"
    contract_case_assoc_rule = "contract_case_assoc_rule"
    contract_cases_expr_rule = "contract_cases_expr_rule"
    aspect_assoc_rule = "aspect_assoc_rule"
    aspect_spec_rule = "aspect_spec_rule"
    single_task_decl_rule = "single_task_decl_rule"
    overriding_indicator_rule = "overriding_indicator_rule"
    entry_decl_rule = "entry_decl_rule"
    component_clause_rule = "component_clause_rule"
    aspect_clause_rule = "aspect_clause_rule"
    param_spec_rule = "param_spec_rule"
    param_specs_rule = "param_specs_rule"
    subp_spec_rule = "subp_spec_rule"
    expr_fn_rule = "expr_fn_rule"
    null_subp_decl_rule = "null_subp_decl_rule"
    abstract_subp_decl_rule = "abstract_subp_decl_rule"
    subp_renaming_decl_rule = "subp_renaming_decl_rule"
    simple_subp_decl_rule = "simple_subp_decl_rule"
    subp_decl_rule = "subp_decl_rule"
    with_clause_rule = "with_clause_rule"
    context_item_rule = "context_item_rule"
    use_clause_rule = "use_clause_rule"
    use_package_clause_rule = "use_package_clause_rule"
    use_type_clause_rule = "use_type_clause_rule"
    subtype_indication_rule = "subtype_indication_rule"
    discrete_subtype_indication_rule = "discrete_subtype_indication_rule"
    constrained_subtype_indication_rule = "constrained_subtype_indication_rule"
    type_expr_rule = "type_expr_rule"
    anonymous_type_rule = "anonymous_type_rule"
    mode_rule = "mode_rule"
    pragma_argument_rule = "pragma_argument_rule"
    pragma_rule = "pragma_rule"
    subunit_rule = "subunit_rule"
    library_unit_body_rule = "library_unit_body_rule"
    library_unit_renaming_decl_rule = "library_unit_renaming_decl_rule"
    library_item_rule = "library_item_rule"
    compilation_unit_rule = "compilation_unit_rule"
    compilation_rule = "compilation_rule"
    decl_part_rule = "decl_part_rule"
    entry_body_rule = "entry_body_rule"
    protected_body_rule = "protected_body_rule"
    protected_body_stub_rule = "protected_body_stub_rule"
    task_body_rule = "task_body_rule"
    task_body_stub_rule = "task_body_stub_rule"
    package_body_stub_rule = "package_body_stub_rule"
    package_body_rule = "package_body_rule"
    terminate_alternative_rule = "terminate_alternative_rule"
    select_stmt_rule = "select_stmt_rule"
    accept_stmt_rule = "accept_stmt_rule"
    case_alt_rule = "case_alt_rule"
    case_stmt_rule = "case_stmt_rule"
    ext_return_stmt_rule = "ext_return_stmt_rule"
    iblock_stmt_rule = "iblock_stmt_rule"
    block_stmt_rule = "block_stmt_rule"
    iloop_stmt_rule = "iloop_stmt_rule"
    loop_stmt_rule = "loop_stmt_rule"
    compound_stmt_rule = "compound_stmt_rule"
    if_stmt_rule = "if_stmt_rule"
    raise_stmt_rule = "raise_stmt_rule"
    delay_stmt_rule = "delay_stmt_rule"
    abort_stmt_rule = "abort_stmt_rule"
    body_rule = "body_rule"
    body_stub_rule = "body_stub_rule"
    subp_body_stub_rule = "subp_body_stub_rule"
    recov_decl_part_rule = "recov_decl_part_rule"
    subp_body_rule = "subp_body_rule"
    handled_stmts_rule = "handled_stmts_rule"
    exception_handler_rule = "exception_handler_rule"
    stmts_rule = "stmts_rule"
    label_rule = "label_rule"
    stmt_rule = "stmt_rule"
    call_stmt_rule = "call_stmt_rule"
    simple_stmt_rule = "simple_stmt_rule"
    null_stmt_rule = "null_stmt_rule"
    assignment_stmt_rule = "assignment_stmt_rule"
    goto_stmt_rule = "goto_stmt_rule"
    exit_stmt_rule = "exit_stmt_rule"
    return_stmt_rule = "return_stmt_rule"
    requeue_stmt_rule = "requeue_stmt_rule"
    identifier_rule = "identifier_rule"
    char_literal_rule = "char_literal_rule"
    string_literal_rule = "string_literal_rule"
    defining_id_rule = "defining_id_rule"
    dec_literal_rule = "dec_literal_rule"
    int_literal_rule = "int_literal_rule"
    num_literal_rule = "num_literal_rule"
    null_literal_rule = "null_literal_rule"
    allocator_rule = "allocator_rule"
    for_loop_param_spec_rule = "for_loop_param_spec_rule"
    quantified_expr_rule = "quantified_expr_rule"
    case_expr_rule = "case_expr_rule"
    case_expr_alt_rule = "case_expr_alt_rule"
    raise_expr_rule = "raise_expr_rule"
    if_expr_rule = "if_expr_rule"
    conditional_expr_rule = "conditional_expr_rule"
    box_expr_rule = "box_expr_rule"
    others_designator_rule = "others_designator_rule"
    aggregate_assoc_rule = "aggregate_assoc_rule"
    regular_aggregate_rule = "regular_aggregate_rule"
    bracket_aggregate_rule = "bracket_aggregate_rule"
    aggregate_rule = "aggregate_rule"
    direct_name_rule = "direct_name_rule"
    param_assoc_rule = "param_assoc_rule"
    call_suffix_rule = "call_suffix_rule"
    qualified_name_rule = "qualified_name_rule"
    qual_name_internal_rule = "qual_name_internal_rule"
    name_rule = "name_rule"
    defining_name_rule = "defining_name_rule"
    direct_name_or_target_name_rule = "direct_name_or_target_name_rule"
    target_name_rule = "target_name_rule"
    update_attr_aggregate_rule = "update_attr_aggregate_rule"
    update_attr_content_rule = "update_attr_content_rule"
    multidim_array_assoc_rule = "multidim_array_assoc_rule"
    subtype_name_rule = "subtype_name_rule"
    static_name_rule = "static_name_rule"
    primary_rule = "primary_rule"
    paren_expr_rule = "paren_expr_rule"
    declare_expr_rule = "declare_expr_rule"
    factor_rule = "factor_rule"
    term_rule = "term_rule"
    unop_term_rule = "unop_term_rule"
    simple_expr_rule = "simple_expr_rule"
    boolean_op_rule = "boolean_op_rule"
    discrete_range_rule = "discrete_range_rule"
    choice_rule = "choice_rule"
    choice_list_rule = "choice_list_rule"
    rel_op_rule = "rel_op_rule"
    membership_choice_rule = "membership_choice_rule"
    membership_choice_list_rule = "membership_choice_list_rule"
    relation_rule = "relation_rule"
    expr_rule = "expr_rule"

    _name = "GrammarRule"
    _c_to_py = [
        parent_list_rule,
        protected_type_decl_rule,
        protected_op_rule,
        protected_el_rule,
        protected_def_rule,
        protected_decl_rule,
        task_item_rule,
        task_def_rule,
        task_type_decl_rule,
        subtype_decl_rule,
        interface_type_def_rule,
        unconstrained_index_rule,
        array_type_def_rule,
        discrete_subtype_definition_rule,
        constraint_list_rule,
        signed_int_type_def_rule,
        mod_int_type_def_rule,
        derived_type_def_rule,
        discriminant_assoc_rule,
        discriminant_constraint_rule,
        index_constraint_rule,
        digits_constraint_rule,
        delta_constraint_rule,
        range_constraint_rule,
        constraint_rule,
        discriminant_spec_rule,
        discr_spec_list_rule,
        discriminant_part_rule,
        enum_literal_decl_rule,
        formal_discrete_type_def_rule,
        record_def_rule,
        range_spec_rule,
        real_type_def_rule,
        sexpr_or_box_rule,
        ordinary_fixed_point_def_rule,
        decimal_fixed_point_def_rule,
        floating_point_def_rule,
        record_type_def_rule,
        access_def_rule,
        enum_type_def_rule,
        type_def_rule,
        variant_rule,
        anonymous_type_decl_rule,
        type_decl_rule,
        variant_part_rule,
        component_def_rule,
        component_item_rule,
        component_decl_rule,
        component_list_rule,
        generic_decl_rule,
        generic_formal_part_rule,
        generic_formal_decl_rule,
        formal_subp_decl_rule,
        renaming_clause_rule,
        generic_renaming_decl_rule,
        generic_instantiation_rule,
        exception_decl_rule,
        basic_decls_rule,
        package_renaming_decl_rule,
        package_decl_rule,
        basic_decl_rule,
        object_decl_rule,
        sub_object_decl_rule,
        ext_ret_stmt_object_decl_rule,
        defining_id_list_rule,
        number_decl_rule,
        contract_case_assoc_rule,
        contract_cases_expr_rule,
        aspect_assoc_rule,
        aspect_spec_rule,
        single_task_decl_rule,
        overriding_indicator_rule,
        entry_decl_rule,
        component_clause_rule,
        aspect_clause_rule,
        param_spec_rule,
        param_specs_rule,
        subp_spec_rule,
        expr_fn_rule,
        null_subp_decl_rule,
        abstract_subp_decl_rule,
        subp_renaming_decl_rule,
        simple_subp_decl_rule,
        subp_decl_rule,
        with_clause_rule,
        context_item_rule,
        use_clause_rule,
        use_package_clause_rule,
        use_type_clause_rule,
        subtype_indication_rule,
        discrete_subtype_indication_rule,
        constrained_subtype_indication_rule,
        type_expr_rule,
        anonymous_type_rule,
        mode_rule,
        pragma_argument_rule,
        pragma_rule,
        subunit_rule,
        library_unit_body_rule,
        library_unit_renaming_decl_rule,
        library_item_rule,
        compilation_unit_rule,
        compilation_rule,
        decl_part_rule,
        entry_body_rule,
        protected_body_rule,
        protected_body_stub_rule,
        task_body_rule,
        task_body_stub_rule,
        package_body_stub_rule,
        package_body_rule,
        terminate_alternative_rule,
        select_stmt_rule,
        accept_stmt_rule,
        case_alt_rule,
        case_stmt_rule,
        ext_return_stmt_rule,
        iblock_stmt_rule,
        block_stmt_rule,
        iloop_stmt_rule,
        loop_stmt_rule,
        compound_stmt_rule,
        if_stmt_rule,
        raise_stmt_rule,
        delay_stmt_rule,
        abort_stmt_rule,
        body_rule,
        body_stub_rule,
        subp_body_stub_rule,
        recov_decl_part_rule,
        subp_body_rule,
        handled_stmts_rule,
        exception_handler_rule,
        stmts_rule,
        label_rule,
        stmt_rule,
        call_stmt_rule,
        simple_stmt_rule,
        null_stmt_rule,
        assignment_stmt_rule,
        goto_stmt_rule,
        exit_stmt_rule,
        return_stmt_rule,
        requeue_stmt_rule,
        identifier_rule,
        char_literal_rule,
        string_literal_rule,
        defining_id_rule,
        dec_literal_rule,
        int_literal_rule,
        num_literal_rule,
        null_literal_rule,
        allocator_rule,
        for_loop_param_spec_rule,
        quantified_expr_rule,
        case_expr_rule,
        case_expr_alt_rule,
        raise_expr_rule,
        if_expr_rule,
        conditional_expr_rule,
        box_expr_rule,
        others_designator_rule,
        aggregate_assoc_rule,
        regular_aggregate_rule,
        bracket_aggregate_rule,
        aggregate_rule,
        direct_name_rule,
        param_assoc_rule,
        call_suffix_rule,
        qualified_name_rule,
        qual_name_internal_rule,
        name_rule,
        defining_name_rule,
        direct_name_or_target_name_rule,
        target_name_rule,
        update_attr_aggregate_rule,
        update_attr_content_rule,
        multidim_array_assoc_rule,
        subtype_name_rule,
        static_name_rule,
        primary_rule,
        paren_expr_rule,
        declare_expr_rule,
        factor_rule,
        term_rule,
        unop_term_rule,
        simple_expr_rule,
        boolean_op_rule,
        discrete_range_rule,
        choice_rule,
        choice_list_rule,
        rel_op_rule,
        membership_choice_rule,
        membership_choice_list_rule,
        relation_rule,
        expr_rule,
    ]
    _py_to_c = {name: index for index, name in enumerate(_c_to_py)}


default_grammar_rule = GrammarRule.compilation_rule


_unit_provider = _hashable_c_pointer()


def _canonicalize_buffer(buffer, charset):
    """Canonicalize source buffers to be bytes buffers."""
    if isinstance(buffer, _py2to3.text_type):
        if charset:
            raise TypeError("`charset` must be null when the buffer is" " Unicode")
        buffer = buffer.encode("utf-8")
        charset = b"utf-8"
    elif not isinstance(buffer, _py2to3.bytes_type):
        raise TypeError("`buffer` must be a string")
    return (buffer, charset)


#
# High-level binding
#


class InvalidField(Exception):
    """
    Raised when introspection functions (``Libadalang.Introspection``) are
    requested an invalid field.
    """

    pass


class NodeDataEvaluationError(Exception):
    """
    Raised when introspection functions (``Libadalang.Introspection``) are
    improperly used to evaluate a node data (field or property).
    """

    pass


class InvalidInput(Exception):
    """
    Raised by lexing functions (``Libadalang.Lexer``) when the input contains
    an invalid byte sequence.
    """

    pass


class InvalidSymbolError(Exception):
    """
    Exception raise when an invalid symbol is passed to a subprogram.
    """

    pass


class InvalidUnitNameError(Exception):
    """
    Raised when an invalid unit name is provided.
    """

    pass


class NativeException(Exception):
    """
    Exception raised in language bindings when the underlying C API reports an
    unexpected error that occurred in the library.

    This kind of exception is raised for internal errors: they should never
    happen in normal situations and if they are raised at some point, it means
    the library state is potentially corrupted.

    Nevertheless, the library does its best not to crash the program,
    materializing internal errors using this kind of exception.
    """

    pass


class PreconditionFailure(Exception):
    """
    Exception raised when an API is called while its preconditions are not
    satisfied.
    """

    pass


class PropertyError(Exception):
    """
    Exception that is raised when an error occurs while evaluating any AST node
    method whose name starts with ``p_``. This is the only exceptions that such
    functions can raise.
    """

    pass


class TemplateArgsError(Exception):
    """
    Exception raised when the provided arguments for a template don't match
    what the template expects.
    """

    pass


class TemplateFormatError(Exception):
    """
    Exception raised when a template has an invalid syntax, such as badly
    formatted placeholders.
    """

    pass


class TemplateInstantiationError(Exception):
    """
    Exception raised when the instantiation of a template cannot be parsed.
    """

    pass


class StaleReferenceError(Exception):
    """
    Exception raised while trying to access data that was deallocated. This
    happens when one tries to use a node whose unit has been reparsed, for
    instance.
    """

    pass


class UnknownCharset(Exception):
    """
    Raised by lexing functions (``Libadalang.Lexer``) when the input charset is
    not supported.
    """

    pass


_exception_kind_to_type = [
    InvalidField,
    NodeDataEvaluationError,
    InvalidInput,
    InvalidSymbolError,
    InvalidUnitNameError,
    NativeException,
    PreconditionFailure,
    PropertyError,
    TemplateArgsError,
    TemplateFormatError,
    TemplateInstantiationError,
    StaleReferenceError,
    UnknownCharset,
]


class InvalidProjectError(Exception):
    """
    Raised when an error occurs while loading a project file.
    """

    pass


class AnalysisContext(object):
    """
    This type represents a context for all source analysis. This is the first
    type you need to create to use Libadalang. It will contain the results of
    all analysis, and is the main holder for all the data.

    You can create several analysis contexts if you need to, which enables you,
    for example to:

    * analyze several different projects at the same time;

    * analyze different parts of the same projects in parallel.

    In the current design, contexts always keep all of their analysis units
    allocated. If you need to get this memory released, the only option at your
    disposal is to destroy your analysis context instance.
    """

    __slots__ = (
        "_c_value",
        "_unit_provider",
        "_serial_number",
        "_unit_cache",
        "__weakref__",
    )

    _context_cache = weakref.WeakValueDictionary()
    """
    Cache for analysis context wrappers. Indexed by analysis context addresses,
    which are known to stay valid forever (and re-used).

    Unlike unit and node caches, this one should contain weak references so
    that analysis contexts (and their units/nodes) can be free'd when user code
    does not reference them anymore.

    :type: dict[AnalysisContext._c_type, AnalysisContext]
    """

    def __init__(
        self,
        charset=None,
        unit_provider=None,
        with_trivia=True,
        tab_stop=8,
        _c_value=None,
    ):
        """
        Create a new analysis context.

        ``Charset`` will be used as a default charset to decode input sources
        in analysis units. Please see ``GNATCOLL.Iconv`` for several supported
        charsets. Be careful: passing an unsupported charset is not guaranteed
        to raise an error here. If no charset is provided, ``"iso-8859-1"`` is
        the default.

        .. todo:: Passing an unsupported charset here is not guaranteed to
        raise an error right here, but this would be really helpful for users.

        When ``With_Trivia`` is true, the parsed analysis units will contain
        trivias.

        If provided, ``Unit_Provider`` will be used to query the file name that
        corresponds to a unit reference during semantic analysis. If it is
        ``None``, the default one is used instead.

        ``Tab_Stop`` is a positive number to describe the effect of tabulation
        characters on the column number in source files.
        """

        # Initialize this field in case we raise an exception during
        # construction, so that the destructor can run later on.
        self._c_value = None

        if _c_value is None:
            charset = _py2to3.text_to_bytes(charset)
            if not isinstance(tab_stop, int) or tab_stop < 1:
                raise ValueError("Invalid tab_stop (positive integer expected)")
            c_unit_provider = unit_provider._c_value if unit_provider else None
            self._c_value = _create_analysis_context(
                charset, c_unit_provider, with_trivia, tab_stop
            )
        else:
            self._c_value = _context_incref(_c_value)
        assert self._c_value not in self._context_cache
        self._context_cache[self._c_value] = self

        # Keep a reference to the unit provider so that it is live at least as
        # long as the analysis context is live.
        self._unit_provider = unit_provider

        self._serial_number = None
        self._unit_cache = {}
        """
        Cache for AnalysisUnit wrappers, indexed by analysis unit addresses,
        which are known to stay valid as long as the context is alive.

        :type: dict[str, AnalysisUnit]
        """

        self._check_unit_cache()

    def __del__(self):
        if self._c_value:
            _context_decref(self._c_value)

    def __eq__(self, other):
        return self._c_value == other._c_value

    def __hash__(self):
        return hash(self._c_value)

    def get_from_file(
        self, filename, charset=None, reparse=False, rule=default_grammar_rule
    ):
        """
        Create a new analysis unit for ``Filename`` or return the existing one
        if any. If ``Reparse`` is true and the analysis unit already exists,
        reparse it from ``Filename``.

        ``Rule`` controls which grammar rule is used to parse the unit.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If any failure occurs, such as file opening, decoding, lexing or
        parsing failure, return an analysis unit anyway: errors are described
        as diagnostics of the returned analysis unit.
        """
        filename = _py2to3.text_to_bytes(filename)
        charset = _py2to3.text_to_bytes(charset or "")
        c_value = _get_analysis_unit_from_file(
            self._c_value, filename, charset, reparse, GrammarRule._unwrap(rule)
        )
        return AnalysisUnit._wrap(c_value)

    def get_from_buffer(
        self, filename, buffer, charset=None, reparse=False, rule=default_grammar_rule
    ):
        """
        Create a new analysis unit for ``Filename`` or return the existing one
        if any. Whether the analysis unit already exists or not, (re)parse it
        from the source code in ``Buffer``.

        ``Rule`` controls which grammar rule is used to parse the unit.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If any failure occurs, such as file opening, decoding, lexing or
        parsing failure, return an analysis unit anyway: errors are described
        as diagnostics of the returned analysis unit.
        """
        filename = _py2to3.text_to_bytes(filename)
        charset = _py2to3.text_to_bytes(charset or "")
        buffer, charset = _canonicalize_buffer(buffer, charset)
        c_value = _get_analysis_unit_from_buffer(
            self._c_value,
            filename,
            charset,
            buffer,
            len(buffer),
            GrammarRule._unwrap(rule),
        )
        return AnalysisUnit._wrap(c_value)

    def get_from_provider(self, name, kind, charset=None, reparse=False):
        """
        Create a new analysis unit for ``Name``/``Kind`` or return the existing
        one if any. If ``Reparse`` is true and the analysis unit already
        exists, reparse it from ``Filename``.

        Use ``Charset`` in order to decode the source. If ``Charset`` is empty
        then use the context's default charset.

        If the unit name cannot be tuned into a file name, raise an
        ``InvalidUnitNameError`` exception. If any other failure occurs, such
        as file opening, decoding, lexing or parsing failure, return an
        analysis unit anyway: errors are described as diagnostics of the
        returned analysis unit.
        """
        name = _py2to3.bytes_to_text(name)
        charset = _py2to3.text_to_bytes(charset or "")

        _name = _text._unwrap(name)
        _kind = AnalysisUnitKind._unwrap(kind)
        c_value = _get_analysis_unit_from_provider(
            self._c_value, ctypes.byref(_name), _kind, charset, reparse
        )
        if c_value:
            return AnalysisUnit._wrap(c_value)
        else:
            raise InvalidUnitNameError(
                "Invalid unit name: {} ({})".format(repr(name), kind)
            )

    def discard_errors_in_populate_lexical_env(self, discard):
        """
        Debug helper. Set whether ``Property_Error`` exceptions raised in
        ``Populate_Lexical_Env`` should be discarded. They are by default.
        """
        _discard_errors_in_populate_lexical_env(self._c_value, bool(discard))

    class _c_struct(ctypes.Structure):
        _fields_ = [("serial_number", ctypes.c_uint64)]

    _c_type = _hashable_c_pointer(_c_struct)

    @classmethod
    def _wrap(cls, c_value):
        try:
            return cls._context_cache[c_value]
        except KeyError:
            return cls(_c_value=c_value)

    def _check_unit_cache(self):
        """
        If this context has been re-used, invalidate its unit cache.
        """
        serial_number = self._c_value.contents.serial_number
        if self._serial_number != serial_number:
            self._unit_cache = {}
            self._serial_number = serial_number


class AnalysisUnit(object):
    """
    This type represents the analysis of a single file.
    """

    __slots__ = ("_c_value", "_context_link", "_cache_version_number", "_node_cache")

    class TokenIterator(object):
        """
        Iterator over the tokens in an analysis unit.
        """

        def __init__(self, first):
            self.first = first

        def __iter__(self):
            return self

        def __next__(self):
            if not self.first:
                raise StopIteration()
            result = self.first
            self.first = self.first.next
            return result

        next = __next__

    def __init__(self, context, c_value):
        """
        This constructor is an implementation detail, and is not meant to be
        used directly. Please use AnalysisContext.get_from_* methods to create
        analysis unit instances instead.
        """
        self._c_value = c_value

        # Keep a reference on the owning context so that we keep it alive at
        # least as long as this unit is alive.
        self._context_link = context

        # Store this wrapper in caches for later re-use
        assert c_value not in context._unit_cache
        context._unit_cache[c_value] = self

        self._cache_version_number = None
        """
        Last version number we saw for this analysis unit wrapper. If it's
        different from `self._unit_version`, it means that the unit was
        reparsed: in this case we need to clear the node cache below (see the
        `_check_node_cache` method).

        :type: int
        """

        self._node_cache = {}
        """
        Cache for all node wrappers in this unit. Indexed by couples:
        (c_value, metadata, rebindings).

        :type: dict[T, AdaNode]
        """

        self._check_node_cache()

    def __eq__(self, other):
        return self._c_value == other._c_value

    def __hash__(self):
        return hash(self._c_value)

    @property
    def context(self):
        """
        Return the context that owns this unit.
        """
        return self._context_link

    def reparse(self, buffer=None, charset=None):
        """
        Reparse an analysis unit from a buffer, if provided, or from the
        original file otherwise. If ``Charset`` is empty or ``None``, use the
        last charset successfuly used for this unit, otherwise use it to decode
        the content of the source file.

        If any failure occurs, such as decoding, lexing or parsing failure,
        diagnostic are emitted to explain what happened.
        """
        charset = _py2to3.text_to_bytes(charset or "")
        if buffer is None:
            _unit_reparse_from_file(self._c_value, charset)
        else:
            buffer, charset = _canonicalize_buffer(buffer, charset)
            _unit_reparse_from_buffer(self._c_value, charset, buffer, len(buffer))

    def populate_lexical_env(self):
        """
        Create lexical environments for this analysis unit, according to the
        specifications given in the language spec.

        If not done before, it will be automatically called during semantic
        analysis. Calling it before enables one to control where the latency
        occurs.

        Depending on whether errors are discarded (see
        ``Discard_Errors_In_Populate_Lexical_Env``), raise a ``Property_Error``
        on failure.
        """
        if not _unit_populate_lexical_env(self._c_value):
            raise PropertyError()

    @property
    def root(self):
        """
        Return the root node for this unit, or ``None`` if there is none.

        :rtype: AdaNode
        """
        result = _Entity_c_type()
        _unit_root(self._c_value, ctypes.byref(result))
        return AdaNode._wrap(result)

    @property
    def first_token(self):
        """
        Return a reference to the first token scanned in this unit.
        """
        result = Token()
        _unit_first_token(self._c_value, ctypes.byref(result))
        return result._wrap()

    @property
    def last_token(self):
        """
        Return a reference to the last token scanned in this unit.
        """
        result = Token()
        _unit_last_token(self._c_value, ctypes.byref(result))
        return result._wrap()

    @property
    def text(self):
        """
        Return the source buffer associated to this unit.
        """
        return Token.text_range(self.first_token, self.last_token)

    @property
    def token_count(self):
        """
        Return the number of tokens in this unit.
        """
        return _unit_token_count(self._c_value)

    @property
    def trivia_count(self):
        """
        Return the number of trivias in this unit. This is 0 for units that
        were parsed with trivia analysis disabled.
        """
        return _unit_trivia_count(self._c_value)

    def lookup_token(self, sloc):
        """
        Look for a token in this unit that contains the given source location.
        If this falls before the first token, return the first token. If this
        falls between two tokens, return the token that appears before. If this
        falls after the last token, return the last token. If there is no token
        in this unit, return no token.
        """
        unit = AnalysisUnit._unwrap(self)
        _sloc = Sloc._c_type._unwrap(sloc)
        tok = Token()
        _unit_lookup_token(unit, ctypes.byref(_sloc), ctypes.byref(tok))
        return tok._wrap()

    def _dump_lexical_env(self):
        """
        Debug helper: output the lexical envs for the given analysis unit.
        """
        unit = AnalysisUnit._unwrap(self)
        _unit_dump_lexical_env(unit)

    def iter_tokens(self):
        """
        Iterator over the tokens in an analysis unit.
        """
        return self.TokenIterator(self.first_token)

    @property
    def filename(self):
        """
        Return the filename this unit is associated to.
        """
        filename = _unit_filename(self._c_value)
        return _unwrap_str(filename)

    @property
    def diagnostics(self):
        """
        Diagnostics for this unit.
        """
        count = _unit_diagnostic_count(self._c_value)
        result = []
        diag = Diagnostic._c_type()
        for i in range(count):
            assert _unit_diagnostic(self._c_value, i, ctypes.byref(diag))
            result.append(diag._wrap())
        return result

    def __repr__(self):
        return "<AnalysisUnit {}>".format(repr(os.path.basename(self.filename)))

    class _c_struct(ctypes.Structure):
        _fields_ = [("unit_version", ctypes.c_uint64)]

    _c_type = _hashable_c_pointer(_c_struct)

    @classmethod
    def _wrap(cls, c_value):
        if not c_value:
            return None

        # Invalidate the unit cache if needed, then look for an existing
        # wrapper for this unit.
        context = cls._context(c_value)
        context._check_unit_cache()

        try:
            return context._unit_cache[c_value]
        except KeyError:
            return cls(context, c_value)

    @classmethod
    def _unwrap(cls, value):
        if value is None:
            return value
        elif not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)
        else:
            return value._c_value

    @classmethod
    def _context(cls, c_value):
        ctx = _unit_context(c_value)
        return AnalysisContext._wrap(ctx)

    @property
    def _unit_version(self):
        return self._c_value.contents.unit_version

    def _check_node_cache(self):
        """
        If this unit has been reparsed, invalidate its node cache.
        """
        if self._cache_version_number != self._unit_version:
            self._node_cache = {}
            self._cache_version_number = self._unit_version


class Sloc(object):
    """
    Location in a source file. Line and column numbers are one-based.
    """

    def __init__(self, line, column):
        assert line >= 0 and column >= 0
        self.line = line
        self.column = column

    def __bool__(self):
        return bool(self.line or self.column)

    __nonzero__ = __bool__

    def __lt__(self, other):
        # First compare line numbers...
        if self.line < other.line:
            return True
        elif self.line > other.line:
            return False

        # Past this point, we know that both are on the same line, so now
        # compare column numbers.
        else:
            return self.column < other.column

    def __eq__(self, other):
        return self.line == other.line and self.column == other.column

    def __hash__(self):
        return hash((self.line, self.column))

    def __str__(self):
        return "{}:{}".format(self.line, self.column)

    def __repr__(self):
        return "<Sloc {} at {:#x}>".format(self, id(self))

    class _c_type(ctypes.Structure):
        _fields_ = [("line", ctypes.c_uint32), ("column", ctypes.c_uint16)]

        def _wrap(self):
            return Sloc(self.line, self.column)

        @classmethod
        def _unwrap(cls, sloc):
            return cls(sloc.line, sloc.column)


class SlocRange(object):
    """
    Location of a span of text in a source file.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __bool__(self):
        return bool(self.start or self.end)

    __nonzero__ = __bool__

    def __lt__(self, other):
        raise NotImplementedError("SlocRange comparison not supported")

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end

    def __hash__(self):
        return hash((self.start, self.end))

    def __str__(self):
        return "{}-{}".format(self.start, self.end)

    def __repr__(self):
        return "<SlocRange {}:{}-{}:{}>".format(
            self.start.line, self.start.column, self.end.line, self.end.column
        )

    class _c_type(ctypes.Structure):
        _fields_ = [("start", Sloc._c_type), ("end", Sloc._c_type)]

        def _wrap(self):
            return SlocRange(self.start._wrap(), self.end._wrap())


class Diagnostic(object):
    """
    Diagnostic for an analysis unit: cannot open the source file, parsing
    error, ...
    """

    def __init__(self, sloc_range, message):
        self.sloc_range = sloc_range
        self.message = message

    @property
    def as_text(self):
        return (
            u"{}: {}".format(self.sloc_range, self.message)
            if self.sloc_range
            else self.message
        )

    def __str__(self):
        result = self.as_text
        if _py2to3.python2:
            result = result.encode("ascii", errors="replace")
        return result

    def __repr__(self):
        return "<Diagnostic {}>".format(self)

    class _c_type(ctypes.Structure):
        _fields_ = [("sloc_range", SlocRange._c_type), ("message", _text)]

        def _wrap(self):
            return Diagnostic(self.sloc_range._wrap(), self.message._wrap())


class Token(ctypes.Structure):
    """
    Reference to a token in an analysis unit.
    """

    _tdh_c_type = _hashable_c_pointer()

    _fields_ = [
        ("_token_data", _tdh_c_type),
        ("_token_index", ctypes.c_int),
        ("_trivia_index", ctypes.c_int),
        ("_kind", ctypes.c_int),
        ("_text", _text),
        ("_sloc_range", SlocRange._c_type),
    ]

    def _wrap(self):
        return self if self._token_data else None

    @staticmethod
    def _check_token(value):
        if not isinstance(value, Token):
            raise TypeError("invalid token: {}".format(value))

    def _check_same_unit(self, other):
        if self._token_data != other._token_data:
            raise ValueError(
                "{} and {} come from different analysis units".format(self, other)
            )

    @property
    def next(self):
        """
        Return a reference to the next token in the corresponding analysis
        unit.
        """
        t = Token()
        _token_next(ctypes.byref(self), ctypes.byref(t))
        return t._wrap()

    @property
    def previous(self):
        """
        Return a reference to the previous token in the corresponding analysis
        unit.
        """
        t = Token()
        _token_previous(ctypes.byref(self), ctypes.byref(t))
        return t._wrap()

    def range_until(self, other):
        """
        Return an iterator on the list of tokens that spans between `self` and
        `other` (included). This returns an empty list if the first token
        appears after the other one in the source code. Raise a ``ValueError``
        if both tokens come from different analysis units.
        """
        self._check_token(other)
        self._check_same_unit(other)

        # Keep the generator as a nested function so that the above checks are
        # executed when the generator is created, instead of only when its
        # first item is requested.
        def generator():
            if other < self:
                return

            yield self
            current = self
            while current < other:
                current = current.next
                yield current

        return generator()

    def is_equivalent(self, other):
        """
        Return whether ``L`` and ``R`` are structurally equivalent tokens. This
        means that their position in the stream won't be taken into account,
        only the kind and text of the token.
        """
        self._check_token(other)
        return bool(_token_is_equivalent(ctypes.byref(self), ctypes.byref(other)))

    @property
    def kind(self):
        """
        Kind for this token.
        """
        name = _token_kind_name(self._kind)
        # The _token_kind_name wrapper is already supposed to handle exceptions
        # so this should always return a non-null value.
        assert name
        return _unwrap_str(name)

    @property
    def is_trivia(self):
        """
        Return whether this token is a trivia. If it's not, it's a regular
        token.
        """
        return self._trivia_index != 0

    @property
    def index(self):
        """
        Zero-based index for this token/trivia. Tokens and trivias get their
        own index space.
        """
        return (
            self._token_index - 1 if self._trivia_index == 0 else self._trivia_index - 1
        )

    @property
    def text(self):
        """
        Return the text of the given token.
        """
        return self._text._wrap()

    @classmethod
    def text_range(cls, first, last):
        """
        Compute the source buffer slice corresponding to the text that spans
        between the ``First`` and ``Last`` tokens (both included). This yields
        an empty slice if ``Last`` actually appears before ``First``.

        This raises a ``ValueError`` if ``First`` and ``Last`` don't belong to
        the same analysis unit.
        """
        cls._check_token(first)
        cls._check_token(last)
        first._check_same_unit(last)
        result = _text()
        assert _token_range_text(
            ctypes.byref(first), ctypes.byref(last), ctypes.byref(result)
        )
        return result._wrap() or u""

    @property
    def sloc_range(self):
        """
        Return the source location range of the given token.
        """
        return self._sloc_range._wrap()

    def __eq__(self, other):
        """
        Return whether the two tokens refer to the same token in the same unit.

        Note that this does not actually compares the token data.
        """
        return (
            isinstance(other, Token) and self._identity_tuple == other._identity_tuple
        )

    def __hash__(self):
        return hash(self._identity_tuple)

    def __repr__(self):
        return "<Token {}{} at {}>".format(
            self.kind,
            " {}".format(_py2to3.text_repr(self.text)) if self.text else "",
            self.sloc_range,
        )

    def __lt__(self, other):
        """
        Consider that None comes before all tokens. Then, sort by unit, token
        index, and trivia index.
        """

        # None always comes first
        if other is None:
            return False

        self._check_token(other)
        self._check_same_unit(other)
        return self._identity_tuple < other._identity_tuple

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        return not (self <= other)

    def __ge__(self, other):
        return not (self < other)

    def to_data(self):
        """
        Return a dict representation of this Token.
        """
        return {"kind": "Token", "token_kind": self.kind, "text": self.text}

    @property
    def _identity_tuple(self):
        """
        Return a tuple that return a tuple that contains "identity" information
        for this token. Think of it as a database primary key.

        This property is for internal use only.
        """
        return (self._token_data, self._token_index, self._trivia_index)


class UnitProvider(object):
    """
    Interface to fetch analysis units from a name and a unit kind.

    The unit provider mechanism provides an abstraction which assumes that to
    any couple (unit name, unit kind) we can associate at most one source file.
    This means that several couples can be associated to the same source file,
    but on the other hand, only one one source file can be associated to a
    couple.

    This is used to make the semantic analysis able to switch from one analysis
    units to another.
    """

    def __init__(self, c_value):
        """
        This constructor is an implementation detail, and is not meant to be
        used directly.
        """
        self._c_value = c_value

    def __del__(self):
        _dec_ref_unit_provider(self._c_value)

    # No matter the Python version, take bytes (forwarded as-is to C) but
    # also accept text (encoded using the system encoding).

    @staticmethod
    def _coerce_bytes(label, value, what="a bytes string", or_none=False):
        if value is None and or_none:
            return None
        elif isinstance(value, _py2to3.bytes_type):
            return value
        elif isinstance(value, _py2to3.text_type):
            return value.encode()
        else:
            raise TypeError(
                "`{}` argument must be {} (got {})".format(
                    label, what, _type_fullname(type(value))
                )
            )

    @classmethod
    def for_project(
        cls, project_file, project=None, scenario_vars=None, target=None, runtime=None
    ):
        """
        Load the project file at ``Project_File`` and return a unit provider
        that uses it.

        If ``Project`` is passed, use it to provide units, otherwise, try use
        the whole project tree.

        As unit providers must guarantee that there exists at most one source
        file for each couple (unit name, unit kind), aggregate projects that
        contains several conflicting units are not supported: trying to load
        one will yield an error (see below).

        If provided, ``Scenario_Vars`` must be a dict with key strings and key
        values to describe the set of scenario variables for this project.

        In order to load the given project with non-default target and
        runtimes, pass these as strings to the ``target`` and ``runtime``
        arguments.

        If the requested project is invalid (error while opening the file,
        error while analysing its syntax, ...), or if it is an unsupported
        aggregate project, this raises an ``InvalidProjectError`` exception.
        """

        project_file = cls._coerce_bytes("project_file", project_file)
        project_file_text = _py2to3.bytes_to_text(project_file)
        project = cls._coerce_bytes("project", project, or_none=True)
        target = cls._coerce_bytes("target", target, or_none=True)
        runtime = cls._coerce_bytes("runtime", runtime, or_none=True)

        if scenario_vars:
            items = scenario_vars.items()
            scn_vars_array_type = _project_scenario_variable * (len(items) + 1)
            scn_vars_array = scn_vars_array_type()
            for i, (name, value) in enumerate(items):
                what = "a dict mapping bytes strings to bytes strings"
                name = cls._coerce_bytes("scenario_vars", name, what)
                value = cls._coerce_bytes("scenario_vars", value, what)
                scn_vars_array[i] = _project_scenario_variable(name, value)
            scn_vars_array[-1] = _project_scenario_variable(None, None)
        else:
            scn_vars_array = None

        c_value = _create_project_unit_provider(
            project_file, project, scn_vars_array, target, runtime
        )
        if c_value:
            return cls(c_value)
        else:
            raise InvalidProjectError(
                "Cannot open project {}".format(project_file_text)
            )

    @classmethod
    def auto(cls, input_files, charset=None):
        """
        Return a unit provider that knows which compilation units are to be
        found in the given list of source files.

        This knowledge is built trying to parse all given input files as Ada
        source files and listing the compilation units found there. Files that
        cannot be parsed properly are discarded. If two compilation units are
        found for the same unit, the first that is found in the given input
        files is taken and the other ones are discarded.

        Source files are decoded using the given charset. If it is ``None``,
        the default charset (ISO-8859-1) is used.

        .. todo:: Find a way to report discarded source files/compilation
        units.
        """

        # Create a NULL-terminated array of strings
        c_strings = [
            ctypes.c_char_p(
                cls._coerce_bytes("input_files", f, "a list of bytes strings")
            )
            for f in input_files
        ]
        c_array_type = ctypes.c_char_p * (len(input_files) + 1)
        c_array = c_array_type()
        for i, c_str in enumerate(c_strings):
            c_array[i] = c_str
        c_array[-1] = None

        c_array_ptr = ctypes.pointer(c_array)
        input_files_arg = ctypes.cast(c_array_ptr, ctypes.POINTER(ctypes.c_char_p))
        c_value = _create_auto_provider(input_files_arg, charset)
        return cls(c_value)


class AdaNode(object):
    """
    Root node class for the Ada syntax tree.
    """

    is_list_type = False
    __slots__ = (
        "_unprotected_c_value",
        "_node_c_value",
        "_metadata",
        "_rebindings",
        "_unprotected_getitem_cache",
        "_unit",
        "_unit_version",
    )

    @property
    def p_declarative_scope(self):
        """
        Return the scope of definition of this basic declaration.

        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_ada_node_p_declarative_scope)

        return result

    @property
    def p_complete(self):
        """
        Return possible completions at this point in the file.

        :rtype: List[CompletionItem]
        """

        c_result = self._eval_field(
            _CompletionItemArrayConverter.c_type(), _ada_node_p_complete
        )
        result = _CompletionItemArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_valid_keywords(self):
        """
        Return the list of keywords that are valid at this point in the file.

        .. note:: This is work in progress. It will return all keywords for
        now, without looking at the context.

        :rtype: List[str]
        """

        c_result = self._eval_field(
            _UnboundedTextTypeArrayConverter.c_type(), _ada_node_p_valid_keywords
        )
        result = _UnboundedTextTypeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_generic_instantiations(self):
        """
        Return the potentially empty list of generic package/subprogram
        instantiations that led to the creation of this entity. Outer-most
        instantiations appear last.

        :rtype: List[GenericInstantiation]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _ada_node_p_generic_instantiations
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_semantic_parent(self):
        """
        Return the semantic parent for this node, if applicable, null
        otherwise.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_semantic_parent)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_parent_basic_decl(self):
        """
        Return the parent basic decl for this node, if applicable, null
        otherwise.

        .. note:: If the parent BasicDecl of the given node is a generic
        declaration, this call will return the instantiation from which the
        node was retrieved instead, if any.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_parent_basic_decl)
        result = AdaNode._wrap(c_result)

        return result

    def p_filter_is_imported_by(self, units, transitive):
        """
        Filters out among the list of given units those that cannot refer to
        the unit in which this node lies. If transitive is True, the whole
        transitive closure of imports will be used to find a reference to the
        unit of this node.

        :param units:
        :type units: List[AnalysisUnit]
        :param transitive:
        :type transitive: bool
        :rtype: List[AnalysisUnit]
        """

        unwrapped_units = _AnalysisUnitArrayConverter.unwrap(units)
        unwrapped_transitive = bool(transitive)

        c_result = self._eval_field(
            _AnalysisUnitArrayConverter.c_type(),
            _ada_node_p_filter_is_imported_by,
            unwrapped_units.c_value,
            unwrapped_transitive,
        )
        result = _AnalysisUnitArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_xref_entry_point(self):
        """
        Designates entities that are entry point for the xref solving
        infrastructure. If this returns true, then resolve_names can be called
        on it.

        .. note:: For convenience, and unlike what is defined in the ARM wrt.
        complete contexts for name resolution, ``xref_entry_points`` can be
        nested.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _ada_node_p_xref_entry_point)
        result = bool(c_result.value)

        return result

    @property
    def p_resolve_names(self):
        """
        This will resolve names for this node. If the operation is successful,
        then type_var and ref_var will be bound on appropriate subnodes of the
        statement.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _ada_node_p_resolve_names)
        result = bool(c_result.value)

        return result

    @property
    def p_standard_unit(self):
        """
        Static method. Return the analysis unit corresponding to the Standard
        package.

        :rtype: AnalysisUnit
        """

        c_result = self._eval_field(AnalysisUnit._c_type(), _ada_node_p_standard_unit)
        result = AnalysisUnit._wrap(c_result)

        return result

    def p_std_entity(self, sym):
        """
        Static property. Return an entity from the standard package with name
        `sym`.

        :param sym:
        :type sym: str
        :rtype: AdaNode
        """

        _context = self.unit.context._c_value
        unwrapped_sym = _symbol_type.unwrap(sym, _context)

        c_result = self._eval_field(
            _Entity_c_type(), _ada_node_p_std_entity, unwrapped_sym
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_bool_type(self):
        """
        Static method. Return the standard Boolean type.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_bool_type)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_int_type(self):
        """
        Static method. Return the standard Integer type.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_int_type)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_universal_int_type(self):
        """
        Static method. Return the standard Universal Integer type.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_universal_int_type)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_universal_real_type(self):
        """
        Static method. Return the standard Universal Real type.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_p_universal_real_type)
        result = AdaNode._wrap(c_result)

        return result

    def p_top_level_decl(self, unit):
        """
        Static method. Get the top-level decl in ``unit``.  This is the body of
        a Subunit, or the item of a ``LibraryItem``.

        :param unit:
        :type unit: AnalysisUnit
        :rtype: BasicDecl
        """

        unwrapped_unit = AnalysisUnit._unwrap(unit)

        c_result = self._eval_field(
            _Entity_c_type(), _ada_node_p_top_level_decl, unwrapped_unit
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_choice_match(self, value):
        """
        Assuming that self is a choice expression (such as what can appear in
        an alternative of a case statement or in the RHS of a membership
        expression, this property returns whether the given value satisfies it.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :param value:
        :type value: int
        :rtype: bool
        """

        unwrapped_value = _big_integer.unwrap(value)

        c_result = self._eval_field(
            ctypes.c_uint8(), _ada_node_p_choice_match, unwrapped_value.c_value
        )
        result = bool(c_result.value)

        return result

    def p_gnat_xref(self, imprecise_fallback=False):
        """
        Return a cross reference from this name to a defining identifier,
        trying to mimic GNAT's xrefs as much as possible.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: DefiningName
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _ada_node_p_gnat_xref, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def parent(self):
        """
        Return the lexical parent for this node. Return null for the root AST
        node or for AST nodes for which no one has a reference to the parent.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_parent)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def parents(self):
        """
        Return an array that contains the lexical parents (this node included).
        Nearer parents are first in the list.

        :rtype: List[AdaNode]
        """

        c_result = self._eval_field(_AdaNodeArrayConverter.c_type(), _ada_node_parents)
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def children(self):
        """
        Return an array that contains the direct lexical children.

        :rtype: List[AdaNode]
        """

        c_result = self._eval_field(_AdaNodeArrayConverter.c_type(), _ada_node_children)
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def token_start(self):
        """
        Return the first token used to parse this node.

        :rtype: Token
        """

        c_result = self._eval_field(Token(), _ada_node_token_start)
        result = c_result

        return result

    @property
    def token_end(self):
        """
        Return the last token used to parse this node.

        :rtype: Token
        """

        c_result = self._eval_field(Token(), _ada_node_token_end)
        result = c_result

        return result

    @property
    def child_index(self):
        """
        Return the 0-based index for Node in its parent's children.

        :rtype: int
        """

        c_result = self._eval_field(ctypes.c_int(), _ada_node_child_index)
        result = c_result.value

        return result

    @property
    def previous_sibling(self):
        """
        Return the node's previous sibling, if there is one.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_previous_sibling)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def next_sibling(self):
        """
        Return the node's next sibling, if there is one.

        :rtype: AdaNode
        """

        c_result = self._eval_field(_Entity_c_type(), _ada_node_next_sibling)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def unit(self):
        """
        Return the analysis unit owning this node.

        :rtype: AnalysisUnit
        """

        c_result = self._eval_field(AnalysisUnit._c_type(), _ada_node_unit)
        result = AnalysisUnit._wrap(c_result)

        return result

    @property
    def is_ghost(self):
        """
        Return whether the node is a ghost.

        Unlike regular nodes, ghost nodes cover no token in the input source:
        they are logically located instead between two tokens. The
        "token_first" of all ghost nodes is the token right after this logical
        position, while they have no "token_last".

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _ada_node_is_ghost)
        result = bool(c_result.value)

        return result

    @property
    def full_sloc_image(self):
        """
        Return a string containing the filename + the sloc in GNU conformant
        format. Useful to create diagnostics from a node.

        :rtype: str
        """

        c_result = self._eval_field(
            _TextTypeConverter.c_type(), _ada_node_full_sloc_image
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    _field_names = () + ()

    def __init__(self, c_value, node_c_value, metadata, rebindings):
        """
        This constructor is an implementation detail, and is not meant to be
        used directly. For now, the creation of AST nodes can happen only as
        part of the parsing of an analysis unit.
        """

        self._unprotected_c_value = c_value

        # Access to these fields is unprotected from stale references, but it
        # is supposed to be used only in _id_tuple, which itself should not be
        # used outside of hashing/equality use cases.
        self._node_c_value = node_c_value
        self._rebindings = rebindings
        self._metadata = metadata

        self._unprotected_getitem_cache = {}
        """
        Cache for the __getitem__ override.

        :type: dict[int, AdaNode]
        """

        # Information to check before accessing node data that it is still
        # valid.
        self._unit = self._fetch_unit(c_value)
        self._unit_version = self._unit._unit_version

    def _check_stale_reference(self):
        # We have a reference to the owning unit, so there is no need to
        # check that the unit and the context are still valid. Just check that
        # the unit has not been reparsed.
        if self._unit._unit_version != self._unit_version:
            raise StaleReferenceError()

    @property
    def _c_value(self):
        self._check_stale_reference()
        return self._unprotected_c_value

    @property
    def _getitem_cache(self):
        self._check_stale_reference()
        return self._unprotected_getitem_cache

    @property
    def _id_tuple(self):
        return (self._node_c_value, self._rebindings)

    def __eq__(self, other):
        return isinstance(other, AdaNode) and self._id_tuple == other._id_tuple

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self._id_tuple)

    @property
    def kind_name(self):
        """
        Return the kind of this node.
        """
        return self._kind_name

    @property
    def is_token_node(self):
        """
        Return whether this node is a node that contains only a single token.
        """
        node = self._unwrap(self)
        return bool(_node_is_token_node(ctypes.byref(node)))

    @property
    def is_synthetic(self):
        """
        Return whether this node is synthetic.
        """
        node = self._unwrap(self)
        return bool(_node_is_synthetic(ctypes.byref(node)))

    @property
    def sloc_range(self):
        """
        Return the spanning source location range for this node.

        Note that this returns the sloc of the parent for synthetic nodes.
        """
        node = self._unwrap(self)
        result = SlocRange._c_type()
        _node_sloc_range(ctypes.byref(node), ctypes.byref(result))
        return result._wrap()

    @property
    def text(self):
        """
        Return the source buffer slice corresponding to the text that spans
        between the first and the last tokens of this node.

        Note that this returns the empty string for synthetic nodes.
        """
        node = self._unwrap(self)
        result = _text()
        _node_text(ctypes.byref(node), ctypes.byref(result))
        return result._wrap()

    @property
    def image(self):
        """
        Return a representation of this node as a string.
        """
        c_node = self._unwrap(self)
        c_result = _text()
        _node_image(ctypes.byref(c_node), ctypes.byref(c_result))
        return c_result._wrap()

    def lookup(self, sloc):
        """
        Return the bottom-most node from in ``Node`` and its children which
        contains ``Sloc``, or ``None`` if there is none.
        """
        node = self._unwrap(self)
        c_sloc = Sloc._c_type._unwrap(sloc)
        result = _Entity_c_type()
        _lookup_in_node(ctypes.byref(node), ctypes.byref(c_sloc), ctypes.byref(result))
        return AdaNode._wrap(result)

    def __bool__(self):
        """
        Return always True so that checking a node against None can be done as
        simply as::

        if node: ...
        """
        return True

    __nonzero__ = __bool__

    def __iter__(self):
        """
        Return an iterator on the children of this node.
        """
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        """
        Return the number of AdaNode children this node has.
        """
        node = self._unwrap(self)
        return _node_children_count(ctypes.byref(node))

    def __getitem__(self, key):
        """
        Return the Nth AdaNode child this node has.

        This handles negative indexes the same way Python lists do. Raise an
        IndexError if "key" is out of range.
        """
        if not isinstance(key, int):
            msg = ("AdaNode children are integer-indexed" " (got {})").format(type(key))
            raise TypeError(msg)

        if key < 0:
            key += len(self)

        if key in self._getitem_cache:
            return self._getitem_cache[key]

        node = self._unwrap(self)
        result = _Entity_c_type()
        success = _node_child(ctypes.byref(node), key, ctypes.byref(result))
        if not success:
            raise IndexError("child index out of range")
        else:
            result = AdaNode._wrap(result)
            self._getitem_cache[key] = result
            return result

    def iter_fields(self):
        """
        Iterate through all the fields this node contains.

        Return an iterator that yields (name, value) couples for all abstract
        fields in this node. If "self" is a list, field names will be
        "item_{n}" with "n" being the index.
        """
        if self.is_list_type:
            for i, value in enumerate(self):
                yield ("item_{}".format(i), value)
        else:
            for field_name in self._field_names:
                yield (field_name, getattr(self, "{}".format(field_name)))

    def dump_str(self):
        """
        Dump the sub-tree to a string in a human-readable format.
        """
        output = _py2to3.StringIO()
        self.dump(file=output)
        ret = output.getvalue()
        output.close()
        return ret

    def dump(self, indent="", file=sys.stdout):
        """
        Dump the sub-tree in a human-readable format on the given file.

        :param str indent: Prefix printed on each line during the dump. :param
        file file: File in which the dump must occur.
        """

        def print_node(name, value):
            if isinstance(value, AdaNode):
                print("{}{}:".format(indent, name), file=file)
                value.dump(indent + "  ", file)
            else:
                print("{}{}: {}".format(indent, name, value), file=file)

        erepr = self.entity_repr[1:-1]
        print(
            "{}{}{}".format(
                indent, erepr, ": {}".format(self.text) if self.is_token_node else ""
            ),
            file=file,
        )
        indent = indent + "|"
        if self.is_list_type:
            for i, value in enumerate(self):
                print_node("item_{}".format(i), value)
        else:
            for name, value in self.iter_fields():
                # Remove the f_ prefix to have the same behavior as the Ada
                # dumper.
                print_node(name[2:], value)

    def findall(self, ast_type_or_pred, **kwargs):
        """
        Helper for finditer that will return all results as a list. See
        finditer's documentation for more details.
        """
        return list(self.finditer(ast_type_or_pred, **kwargs))

    def find(self, ast_type_or_pred, **kwargs):
        """
        Helper for finditer that will return only the first result. See
        finditer's documentation for more details.
        """
        try:
            return next(self.finditer(ast_type_or_pred, **kwargs))
        except StopIteration:
            return None

    def finditer(self, ast_type_or_pred, **kwargs):
        """
        Find every node corresponding to the passed predicates.

        :param ast_type_or_pred: If supplied with a subclass of AdaNode, will
        constrain the resulting collection to only the instances of this type
        or any subclass. If supplied with a predicate, it will apply the
        predicate on every node and keep only the ones for which it returns
        True. If supplied with a list of subclasses of AdaNode, it will match
        all instances of any of them. :type ast_type_or_pred: type|((AdaNode)
        -> bool)|list[type]

        :param kwargs: Allows the user to filter on attributes of the node. For
        every key value association, if the node has an attribute of name key
        that has the specified value, then the child is kept. :type kwargs:
        dict[str, Any]
        """
        # Create a "pred" function to use as the node filter during the
        # traversal.
        if isinstance(ast_type_or_pred, type):
            sought_type = ast_type_or_pred
            pred = lambda node: isinstance(node, sought_type)
        elif isinstance(ast_type_or_pred, collections.Sequence):
            sought_types = ast_type_or_pred
            pred = lambda node: isinstance(node, tuple(sought_types))
        else:
            pred = ast_type_or_pred

        def match(left, right):
            """
            :param left: Node child to match.
            :param right: Matcher, coming from ``kwargs``.
            """
            if left is None:
                return
            if hasattr(left, "match"):
                return left.match(right)
            else:
                return left == right

        def helper(node):
            for child in node:
                if child is not None:
                    if pred(child):
                        if not kwargs:
                            yield child
                        elif all(
                            [
                                match(getattr(child, key, None), val)
                                for key, val in kwargs.items()
                            ]
                        ):
                            yield child
                    for c in helper(child):
                        if c is not None:
                            yield c

        return helper(self)

    @property
    def parent_chain(self):
        """
        Return the parent chain of self. Self will be the first element,
        followed by the first parent, then this parent's parent, etc.
        """

        def _parent_chain(node):
            yield node
            if node.parent is not None:
                for p in _parent_chain(node.parent):
                    yield p

        return list(_parent_chain(self))

    def __repr__(self):
        return self.image

    @property
    def entity_repr(self):
        c_value = self._unwrap(self)
        c_result = _text()
        _entity_image(ctypes.byref(c_value), ctypes.byref(c_result))
        return c_result._wrap()

    @property
    def tokens(self):
        """
        Return an iterator on the range of tokens that self encompasses.
        """
        start = self.token_start
        end = self.token_end
        while not start == end:
            yield start
            start = start.next
        yield end

    def to_data(self):
        """
        Return a nested python data-structure, constituted only of standard
        data types (dicts, lists, strings, ints, etc), and representing the
        portion of the AST corresponding to this node.
        """
        if self.is_list_type:
            return [i.to_data() for i in self if i is not None]
        else:
            return {n: v.to_data() for n, v in self.iter_fields() if v is not None}

    def to_json(self):
        """
        Return a JSON representation of this node.
        """
        return json.dumps(self.to_data())

    def is_a(self, *types):
        """
        Shortcut for isinstance(self, types).
        :rtype: bool
        """
        return isinstance(self, tuple(types))

    def cast(self, typ):
        """
        Fluent interface style method. Return ``self``, raise an error if self
        is not of type ``typ``.

        :type typ: () -> T
        :rtype: T
        """
        assert isinstance(self, typ)
        return self

    _node_c_type = _hashable_c_pointer()

    @classmethod
    def _wrap(cls, c_value):
        """
        Internal helper to wrap a low-level entity value into an instance of
        the the appropriate high-level Python wrapper subclass.
        """
        node_c_value = c_value.node
        if not node_c_value:
            return None

        rebindings = c_value.info.rebindings
        metadata = c_value.info.md

        # Look for an already existing wrapper for this node
        cache_key = (node_c_value, metadata, rebindings)
        unit = cls._fetch_unit(c_value)
        unit._check_node_cache()
        try:
            return unit._node_cache[cache_key]
        except KeyError:
            pass

        # Pick the right subclass to materialize this node in Python
        kind = _node_kind(ctypes.byref(c_value))
        result = _kind_to_astnode_cls[kind](c_value, node_c_value, metadata, rebindings)
        unit._node_cache[cache_key] = result
        return result

    @classmethod
    def _wrap_bare_node(cls, c_value):
        return cls._wrap(_Entity_c_type.from_bare_node(c_value))

    @classmethod
    def _unwrap(cls, py_value):
        """
        Internal helper to unwrap a high-level ASTNode instance into a
        low-level value. Raise a TypeError if the input value has unexpected
        type.
        """
        if py_value is None:
            return _Entity_c_type._null_value
        elif not isinstance(py_value, AdaNode):
            _raise_type_error("AdaNode", py_value)
        else:
            return py_value._c_value

    @property
    def _unwrap_einfo(self):
        return self._c_value.info

    @classmethod
    def _fetch_unit(cls, c_value):
        return AnalysisUnit._wrap(_node_unit(ctypes.byref(c_value)))

    def _eval_field(self, c_result, c_accessor, *c_args):
        """
        Internal helper to evaluate low-level field accessors/properties.

        This calls "c_accessor" on this node with the input arguments and puts
        the result in "c_result". This raises a PropertyError if the evaluation
        failed. Return "c_result" for convenience.
        """
        args = (self._unwrap(self),) + c_args + (ctypes.byref(c_result),)
        if not c_accessor(*args):
            raise PropertyError()
        return c_result

    def _eval_astnode_field(self, c_accessor):
        """
        Internal helper. Wrapper around _eval_field for fields that return an
        AST node and that accept no explicit argument. This is useful as it's
        the most common case of field, so using this wrapper reduces generated
        code length.
        """
        return AdaNode._wrap(self._eval_field(_Entity_c_type(), c_accessor))


class AbortNode(AdaNode):
    """
    Qualifier for the ``abort`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of AbortPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _abort_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class AbortAbsent(AbortNode):
    """"""

    __slots__ = []

    _field_names = AbortNode._field_names + ()

    _kind_name = "AbortAbsent"


class AbortPresent(AbortNode):
    """"""

    __slots__ = []

    _field_names = AbortNode._field_names + ()

    _kind_name = "AbortPresent"


class AbstractNode(AdaNode):
    """
    Qualifier for the ``abstract`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of AbstractPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _abstract_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class AbstractAbsent(AbstractNode):
    """"""

    __slots__ = []

    _field_names = AbstractNode._field_names + ()

    _kind_name = "AbstractAbsent"


class AbstractPresent(AbstractNode):
    """"""

    __slots__ = []

    _field_names = AbstractNode._field_names + ()

    _kind_name = "AbstractPresent"


class AdaList(AdaNode):
    """"""

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class AdaNodeList(AdaList):
    """
    List of AdaNode.

    This list node can contain one of the following nodes:

    * AbstractSubpDecl

    * Allocator

    * AspectClause

    * AttributeRef

    * BaseAggregate

    * BinOp

    * BodyNode

    * CallExpr

    * CharLiteral

    * ComponentClause

    * ComponentDecl

    * CondExpr

    * DeclExpr

    * DottedName

    * EntryDecl

    * ErrorDecl

    * ExceptionDecl

    * ExceptionHandler

    * ExplicitDeref

    * GenericDecl

    * GenericFormal

    * GenericInstantiation

    * GenericRenamingDecl

    * Identifier

    * IncompleteTypeDecl

    * MembershipExpr

    * NullComponentDecl

    * NullLiteral

    * NumLiteral

    * NumberDecl

    * ObjectDecl

    * OthersDesignator

    * PackageDecl

    * PackageRenamingDecl

    * ParenExpr

    * PragmaNode

    * ProtectedTypeDecl

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * SingleProtectedDecl

    * SingleTaskDecl

    * Stmt

    * StringLiteral

    * SubpDecl

    * SubtypeDecl

    * SubtypeIndication

    * TargetName

    * TaskTypeDecl

    * TypeDecl

    * UnOp

    * UseClause

    * WithClause
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "AdaNodeList"

    is_list_type = True


class AlternativesList(AdaNodeList):
    """
    List of alternatives in a ``when ...`` clause.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeIndication

    * DottedName

    * ExplicitDeref

    * Identifier

    * MembershipExpr

    * NullLiteral

    * NumLiteral

    * OthersDesignator

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    __slots__ = []

    _field_names = AdaNodeList._field_names + ()

    _kind_name = "AlternativesList"

    is_list_type = True


class ConstraintList(AdaNodeList):
    """
    List of constraints.

    This list node can contain one of the following nodes:

    * AttributeRef

    * BinOp

    * CallExpr

    * CharLiteral

    * DottedName

    * ExplicitDeref

    * Identifier

    * QualExpr

    * StringLiteral

    * SubtypeIndication

    * TargetName
    """

    __slots__ = []

    _field_names = AdaNodeList._field_names + ()

    _kind_name = "ConstraintList"

    is_list_type = True


class DeclList(AdaNodeList):
    """
    List of declarations.

    This list node can contain one of the following nodes:

    * AbstractSubpDecl

    * AspectClause

    * ComponentDecl

    * EntryDecl

    * ExprFunction

    * NullSubpDecl

    * PragmaNode

    * SubpDecl

    * SubpRenamingDecl
    """

    __slots__ = []

    _field_names = AdaNodeList._field_names + ()

    _kind_name = "DeclList"

    is_list_type = True


class StmtList(AdaNodeList):
    """
    List of statements.

    This list node can contain one of the following nodes:

    * PragmaNode

    * Stmt
    """

    __slots__ = []

    _field_names = AdaNodeList._field_names + ()

    _kind_name = "StmtList"

    is_list_type = True


class AspectAssocList(AdaList):
    """
    List of AspectAssoc.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "AspectAssocList"

    is_list_type = True


class BaseAssocList(AdaList):
    """
    List of BaseAssoc.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "BaseAssocList"

    is_list_type = True


class BasicAssocList(AdaList):
    """
    List of BasicAssoc.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    is_list_type = True


class AssocList(BasicAssocList):
    """
    List of associations.
    """

    __slots__ = []

    def p_zip_with_params(self, imprecise_fallback=False):
        """
        Returns an array of pairs, associating formal parameters to actual
        expressions. The formals to match are retrieved by resolving the call
        which this AssocList represents the actuals of.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[ParamActual]
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _ParamActualArrayConverter.c_type(),
            _assoc_list_p_zip_with_params,
            unwrapped_imprecise_fallback,
        )
        result = _ParamActualArrayConverter.wrap(c_result, False)

        return result

    _field_names = BasicAssocList._field_names + ()

    _kind_name = "AssocList"

    is_list_type = True


class BasicDeclList(AdaList):
    """
    List of BasicDecl.

    This list node can contain one of the following nodes:

    * NumberDecl

    * ObjectDecl

    * SingleProtectedDecl

    * SingleTaskDecl
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "BasicDeclList"

    is_list_type = True


class CaseExprAlternativeList(AdaList):
    """
    List of CaseExprAlternative.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "CaseExprAlternativeList"

    is_list_type = True


class CaseStmtAlternativeList(AdaList):
    """
    List of CaseStmtAlternative.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "CaseStmtAlternativeList"

    is_list_type = True


class CompilationUnitList(AdaList):
    """
    List of CompilationUnit.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "CompilationUnitList"

    is_list_type = True


class ContractCaseAssocList(AdaList):
    """
    List of ContractCaseAssoc.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "ContractCaseAssocList"

    is_list_type = True


class DefiningNameList(AdaList):
    """
    List of DefiningName.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "DefiningNameList"

    is_list_type = True


class DiscriminantSpecList(AdaList):
    """
    List of DiscriminantSpec.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "DiscriminantSpecList"

    is_list_type = True


class ElsifExprPartList(AdaList):
    """
    List of ElsifExprPart.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "ElsifExprPartList"

    is_list_type = True


class ElsifStmtPartList(AdaList):
    """
    List of ElsifStmtPart.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "ElsifStmtPartList"

    is_list_type = True


class EnumLiteralDeclList(AdaList):
    """
    List of EnumLiteralDecl.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "EnumLiteralDeclList"

    is_list_type = True


class ExprList(AdaList):
    """
    List of Expr.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeName

    * DottedName

    * ExplicitDeref

    * Identifier

    * NullLiteral

    * NumLiteral

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    is_list_type = True


class ExprAlternativesList(ExprList):
    """
    List of alternatives in a membership test expression.

    This list node can contain one of the following nodes:

    * Allocator

    * AttributeRef

    * BaseAggregate

    * BinOp

    * CallExpr

    * CharLiteral

    * CondExpr

    * DeclExpr

    * DiscreteSubtypeName

    * DottedName

    * ExplicitDeref

    * Identifier

    * NullLiteral

    * NumLiteral

    * ParenExpr

    * QualExpr

    * QuantifiedExpr

    * RaiseExpr

    * StringLiteral

    * TargetName

    * UnOp
    """

    __slots__ = []

    _field_names = ExprList._field_names + ()

    _kind_name = "ExprAlternativesList"

    is_list_type = True


class IdentifierList(AdaList):
    """
    List of Identifier.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    is_list_type = True


class DiscriminantChoiceList(IdentifierList):
    """
    List of discriminant associations.
    """

    __slots__ = []

    _field_names = IdentifierList._field_names + ()

    _kind_name = "DiscriminantChoiceList"

    is_list_type = True


class NameList(AdaList):
    """
    List of Name.

    This list node can contain one of the following nodes:

    * AttributeRef

    * CallExpr

    * CharLiteral

    * DottedName

    * ExplicitDeref

    * Identifier

    * QualExpr

    * StringLiteral

    * TargetName
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "NameList"

    is_list_type = True


class ParentList(NameList):
    """
    List of parents in a type declaration.

    This list node can contain one of the following nodes:

    * CharLiteral

    * DottedName

    * Identifier

    * StringLiteral
    """

    __slots__ = []

    _field_names = NameList._field_names + ()

    _kind_name = "ParentList"

    is_list_type = True


class ParamSpecList(AdaList):
    """
    List of ParamSpec.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "ParamSpecList"

    is_list_type = True


class PragmaNodeList(AdaList):
    """
    List of Pragma.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "PragmaNodeList"

    is_list_type = True


class SelectWhenPartList(AdaList):
    """
    List of SelectWhenPart.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "SelectWhenPartList"

    is_list_type = True


class UnconstrainedArrayIndexList(AdaList):
    """
    List of UnconstrainedArrayIndex.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "UnconstrainedArrayIndexList"

    is_list_type = True


class VariantList(AdaList):
    """
    List of Variant.
    """

    __slots__ = []

    _field_names = AdaList._field_names + ()

    _kind_name = "VariantList"

    is_list_type = True


class AliasedNode(AdaNode):
    """
    Qualifier for the ``aliased`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of AliasedPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _aliased_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class AliasedAbsent(AliasedNode):
    """"""

    __slots__ = []

    _field_names = AliasedNode._field_names + ()

    _kind_name = "AliasedAbsent"


class AliasedPresent(AliasedNode):
    """"""

    __slots__ = []

    _field_names = AliasedNode._field_names + ()

    _kind_name = "AliasedPresent"


class AllNode(AdaNode):
    """
    Qualifier for the ``all`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of AllPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _all_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class AllAbsent(AllNode):
    """"""

    __slots__ = []

    _field_names = AllNode._field_names + ()

    _kind_name = "AllAbsent"


class AllPresent(AllNode):
    """"""

    __slots__ = []

    _field_names = AllNode._field_names + ()

    _kind_name = "AllPresent"


class ArrayIndices(AdaNode):
    """
    Specification for array indexes.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class ConstrainedArrayIndices(ArrayIndices):
    """
    Constrained specification for array indexes.
    """

    __slots__ = []

    @property
    def f_list(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName

        :rtype: ConstraintList
        """

        result = self._eval_astnode_field(_constrained_array_indices_f_list)

        return result

    _field_names = ArrayIndices._field_names + ("f_list",)

    _kind_name = "ConstrainedArrayIndices"


class UnconstrainedArrayIndices(ArrayIndices):
    """
    Unconstrained specification for array indexes.
    """

    __slots__ = []

    @property
    def f_types(self):
        """
        :rtype: UnconstrainedArrayIndexList
        """

        result = self._eval_astnode_field(_unconstrained_array_indices_f_types)

        return result

    _field_names = ArrayIndices._field_names + ("f_types",)

    _kind_name = "UnconstrainedArrayIndices"


class AspectAssoc(AdaNode):
    """
    Name/expression association in an aspect.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_aspect_assoc_f_id)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * ContractCases

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_aspect_assoc_f_expr)

        return result

    _field_names = AdaNode._field_names + (
        "f_id",
        "f_expr",
    )

    _kind_name = "AspectAssoc"


class AspectClause(AdaNode):
    """
    Base class for aspect clauses.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class AtClause(AspectClause):
    """
    Representation clause (``for .. use at ...;``).
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * Identifier

        * StringLiteral

        :rtype: BaseId
        """

        result = self._eval_astnode_field(_at_clause_f_name)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_at_clause_f_expr)

        return result

    _field_names = AspectClause._field_names + (
        "f_name",
        "f_expr",
    )

    _kind_name = "AtClause"


class AttributeDefClause(AspectClause):
    """
    Clause for an attribute definition (``for ...'Attribute use ...;``).
    """

    __slots__ = []

    @property
    def f_attribute_expr(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_attribute_def_clause_f_attribute_expr)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_attribute_def_clause_f_expr)

        return result

    _field_names = AspectClause._field_names + (
        "f_attribute_expr",
        "f_expr",
    )

    _kind_name = "AttributeDefClause"


class EnumRepClause(AspectClause):
    """
    Representation clause for enumeration types.
    """

    __slots__ = []

    @property
    def f_type_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_enum_rep_clause_f_type_name)

        return result

    @property
    def f_aggregate(self):
        """
        :rtype: BaseAggregate
        """

        result = self._eval_astnode_field(_enum_rep_clause_f_aggregate)

        return result

    _field_names = AspectClause._field_names + (
        "f_type_name",
        "f_aggregate",
    )

    _kind_name = "EnumRepClause"


class RecordRepClause(AspectClause):
    """
    Representation clause for a record type.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_record_rep_clause_f_name)

        return result

    @property
    def f_at_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_record_rep_clause_f_at_expr)

        return result

    @property
    def f_components(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ComponentClause

        * PragmaNode

        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_record_rep_clause_f_components)

        return result

    _field_names = AspectClause._field_names + (
        "f_name",
        "f_at_expr",
        "f_components",
    )

    _kind_name = "RecordRepClause"


class AspectSpec(AdaNode):
    """
    List of aspects in a declaration.
    """

    __slots__ = []

    @property
    def f_aspect_assocs(self):
        """
        :rtype: AspectAssocList
        """

        result = self._eval_astnode_field(_aspect_spec_f_aspect_assocs)

        return result

    _field_names = AdaNode._field_names + ("f_aspect_assocs",)

    _kind_name = "AspectSpec"


class BaseAssoc(AdaNode):
    """
    Abstract class for a key/value association, where the value is an
    expression.
    """

    __slots__ = []

    @property
    def p_assoc_expr(self):
        """
        Returns the expression side of this assoc node.

        :rtype: Expr
        """

        c_result = self._eval_field(_Entity_c_type(), _base_assoc_p_assoc_expr)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = AdaNode._field_names + ()


class ContractCaseAssoc(BaseAssoc):
    """
    Single association for the ``Contract_Case`` aspect.
    """

    __slots__ = []

    @property
    def f_guard(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_contract_case_assoc_f_guard)

        return result

    @property
    def f_consequence(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_contract_case_assoc_f_consequence)

        return result

    _field_names = BaseAssoc._field_names + (
        "f_guard",
        "f_consequence",
    )

    _kind_name = "ContractCaseAssoc"


class PragmaArgumentAssoc(BaseAssoc):
    """
    Argument assocation in a pragma.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_pragma_argument_assoc_f_id)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_pragma_argument_assoc_f_expr)

        return result

    _field_names = BaseAssoc._field_names + (
        "f_id",
        "f_expr",
    )

    _kind_name = "PragmaArgumentAssoc"


class BaseFormalParamHolder(AdaNode):
    """
    Base class for lists of formal parameters. This is used both for subprogram
    specifications and for records, so that we can share the matching and
    unpacking logic.
    """

    __slots__ = []

    @property
    def p_abstract_formal_params(self):
        """
        Return the list of abstract formal parameters for this holder.

        :rtype: List[BaseFormalParamDecl]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_formal_param_holder_p_abstract_formal_params,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_nb_min_params(self):
        """
        Return the minimum number of parameters this subprogram can be called
        while still being a legal call.

        :rtype: int
        """

        c_result = self._eval_field(
            ctypes.c_int(), _base_formal_param_holder_p_nb_min_params
        )
        result = c_result.value

        return result

    @property
    def p_nb_max_params(self):
        """
        Return the maximum number of parameters this subprogram can be called
        while still being a legal call.

        :rtype: int
        """

        c_result = self._eval_field(
            ctypes.c_int(), _base_formal_param_holder_p_nb_max_params
        )
        result = c_result.value

        return result

    def p_param_types(self, origin=None):
        """
        Returns the type of each parameter of Self.

        :param origin:
        :type origin: AdaNode
        :rtype: List[BaseTypeDecl]
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_formal_param_holder_p_param_types,
            unwrapped_origin,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = AdaNode._field_names + ()


class BaseSubpSpec(BaseFormalParamHolder):
    """
    Base class for subprogram specifications.
    """

    __slots__ = []

    @property
    def p_returns(self):
        """
        Syntax property. Return the type expression node corresponding to the
        return of this subprogram spec.

        :rtype: TypeExpr
        """

        c_result = self._eval_field(_Entity_c_type(), _base_subp_spec_p_returns)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_params(self):
        """
        Returns the array of parameters specification for this subprogram spec.

        :rtype: List[ParamSpec]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _base_subp_spec_p_params
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_primitive_subp_types(self, imprecise_fallback=False):
        """
        Return the types of which this subprogram is a primitive of.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[BaseTypeDecl]
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_subp_spec_p_primitive_subp_types,
            unwrapped_imprecise_fallback,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_primitive_subp_first_type(self, imprecise_fallback=False):
        """
        Return the first type of which this subprogram is a primitive of.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BaseTypeDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_subp_spec_p_primitive_subp_first_type,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_primitive_subp_tagged_type(self, imprecise_fallback=False):
        """
        If this subprogram is a primitive for a tagged type, then return this
        type.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BaseTypeDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_subp_spec_p_primitive_subp_tagged_type,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_return_type(self, origin=None):
        """
        Returns the return type of Self, if applicable (e.g. if Self is a
        subprogram). Else, returns null.

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_subp_spec_p_return_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BaseFormalParamHolder._field_names + ()


class EntrySpec(BaseSubpSpec):
    """
    Entry specification.
    """

    __slots__ = []

    @property
    def f_entry_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_entry_spec_f_entry_name)

        return result

    @property
    def f_family_type(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_entry_spec_f_family_type)

        return result

    @property
    def f_entry_params(self):
        """
        :rtype: Params
        """

        result = self._eval_astnode_field(_entry_spec_f_entry_params)

        return result

    _field_names = BaseSubpSpec._field_names + (
        "f_entry_name",
        "f_family_type",
        "f_entry_params",
    )

    _kind_name = "EntrySpec"


class EnumSubpSpec(BaseSubpSpec):
    """
    Synthetic node for the abstract subprogram spec of an enum literal.

    NOTE: This has no existence in the ARM. While enum literals are functions
    semantically, they're not such syntactically.
    """

    __slots__ = []

    _field_names = BaseSubpSpec._field_names + ()

    _kind_name = "EnumSubpSpec"


class SubpSpec(BaseSubpSpec):
    """
    Subprogram specification.
    """

    __slots__ = []

    @property
    def f_subp_kind(self):
        """
        :rtype: SubpKind
        """

        result = self._eval_astnode_field(_subp_spec_f_subp_kind)

        return result

    @property
    def f_subp_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_subp_spec_f_subp_name)

        return result

    @property
    def f_subp_params(self):
        """
        :rtype: Params
        """

        result = self._eval_astnode_field(_subp_spec_f_subp_params)

        return result

    @property
    def f_subp_returns(self):
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication

        :rtype: TypeExpr
        """

        result = self._eval_astnode_field(_subp_spec_f_subp_returns)

        return result

    _field_names = BaseSubpSpec._field_names + (
        "f_subp_kind",
        "f_subp_name",
        "f_subp_params",
        "f_subp_returns",
    )

    _kind_name = "SubpSpec"


class ComponentList(BaseFormalParamHolder):
    """
    List of component declarations.
    """

    __slots__ = []

    @property
    def f_components(self):
        """
        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_component_list_f_components)

        return result

    @property
    def f_variant_part(self):
        """
        :rtype: VariantPart
        """

        result = self._eval_astnode_field(_component_list_f_variant_part)

        return result

    _field_names = BaseFormalParamHolder._field_names + (
        "f_components",
        "f_variant_part",
    )

    _kind_name = "ComponentList"


class DiscriminantPart(BaseFormalParamHolder):
    """
    Specification for discriminants in type declarations.
    """

    __slots__ = []

    _field_names = BaseFormalParamHolder._field_names + ()


class KnownDiscriminantPart(DiscriminantPart):
    """
    Known list of discriminants in type declarations.
    """

    __slots__ = []

    @property
    def f_discr_specs(self):
        """
        :rtype: DiscriminantSpecList
        """

        result = self._eval_astnode_field(_known_discriminant_part_f_discr_specs)

        return result

    _field_names = DiscriminantPart._field_names + ("f_discr_specs",)

    _kind_name = "KnownDiscriminantPart"


class UnknownDiscriminantPart(DiscriminantPart):
    """
    Unknown list of discriminants in type declarations.
    """

    __slots__ = []

    _field_names = DiscriminantPart._field_names + ()

    _kind_name = "UnknownDiscriminantPart"


class EntryCompletionFormalParams(BaseFormalParamHolder):
    """
    Formal parameters for the completion of an ``EntryDecl`` (either an
    ``EntryBody`` or an ``AcceptStmt``).
    """

    __slots__ = []

    @property
    def f_params(self):
        """
        :rtype: Params
        """

        result = self._eval_astnode_field(_entry_completion_formal_params_f_params)

        return result

    _field_names = BaseFormalParamHolder._field_names + ("f_params",)

    _kind_name = "EntryCompletionFormalParams"


class GenericFormalPart(BaseFormalParamHolder):
    """
    List of declaration for generic formals.
    """

    __slots__ = []

    @property
    def f_decls(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * GenericFormal

        * PragmaNode

        * UseClause

        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_generic_formal_part_f_decls)

        return result

    _field_names = BaseFormalParamHolder._field_names + ("f_decls",)

    _kind_name = "GenericFormalPart"


class BaseRecordDef(AdaNode):
    """
    Base class for record definitions.
    """

    __slots__ = []

    @property
    def f_components(self):
        """
        :rtype: ComponentList
        """

        result = self._eval_astnode_field(_base_record_def_f_components)

        return result

    _field_names = AdaNode._field_names + ("f_components",)


class NullRecordDef(BaseRecordDef):
    """
    Record definition for ``null record``.
    """

    __slots__ = []

    _field_names = BaseRecordDef._field_names + ()

    _kind_name = "NullRecordDef"


class RecordDef(BaseRecordDef):
    """
    Record definition that contains components (``record ... end record``).
    """

    __slots__ = []

    _field_names = BaseRecordDef._field_names + ()

    _kind_name = "RecordDef"


class BasicAssoc(AdaNode):
    """
    Association of one or several names to an expression.
    """

    __slots__ = []

    def p_get_params(self, imprecise_fallback=False):
        """
        Return the list of parameters that this association refers to.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[DefiningName]
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _basic_assoc_p_get_params,
            unwrapped_imprecise_fallback,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = AdaNode._field_names + ()


class AggregateAssoc(BasicAssoc):
    """
    Assocation (X => Y) used for aggregates and parameter associations.
    """

    __slots__ = []

    @property
    def f_designators(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: AlternativesList
        """

        result = self._eval_astnode_field(_aggregate_assoc_f_designators)

        return result

    @property
    def f_r_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_aggregate_assoc_f_r_expr)

        return result

    _field_names = BasicAssoc._field_names + (
        "f_designators",
        "f_r_expr",
    )

    _kind_name = "AggregateAssoc"


class MultiDimArrayAssoc(AggregateAssoc):
    """
    Association used for multi-dimension array aggregates.
    """

    __slots__ = []

    _field_names = AggregateAssoc._field_names + ()

    _kind_name = "MultiDimArrayAssoc"


class DiscriminantAssoc(BasicAssoc):
    """
    Association of discriminant names to an expression.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DiscriminantChoiceList
        """

        result = self._eval_astnode_field(_discriminant_assoc_f_ids)

        return result

    @property
    def f_discr_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_discriminant_assoc_f_discr_expr)

        return result

    _field_names = BasicAssoc._field_names + (
        "f_ids",
        "f_discr_expr",
    )

    _kind_name = "DiscriminantAssoc"


class IteratedAssoc(BasicAssoc):
    """
    Iterated association (Ada 2020).
    """

    __slots__ = []

    @property
    def f_spec(self):
        """
        :rtype: ForLoopSpec
        """

        result = self._eval_astnode_field(_iterated_assoc_f_spec)

        return result

    @property
    def f_r_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_iterated_assoc_f_r_expr)

        return result

    _field_names = BasicAssoc._field_names + (
        "f_spec",
        "f_r_expr",
    )

    _kind_name = "IteratedAssoc"


class ParamAssoc(BasicAssoc):
    """
    Assocation (X => Y) used for aggregates and parameter associations.
    """

    __slots__ = []

    @property
    def f_designator(self):
        """
        This field can contain one of the following nodes:

        * Identifier

        * OthersDesignator

        * StringLiteral

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_param_assoc_f_designator)

        return result

    @property
    def f_r_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_param_assoc_f_r_expr)

        return result

    _field_names = BasicAssoc._field_names + (
        "f_designator",
        "f_r_expr",
    )

    _kind_name = "ParamAssoc"


class BasicDecl(AdaNode):
    """
    Root class for an Ada declaration (RM 3.1). A declaration associates a name
    with a language entity, for example a type or a variable.
    """

    __slots__ = []

    @property
    def p_is_formal(self):
        """
        Whether this decl is the nested decl of a generic formal declaration.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _basic_decl_p_is_formal)
        result = bool(c_result.value)

        return result

    @property
    def p_doc_annotations(self):
        """
        Return the documentation annotations associated with this decl.
        Annotations are any comment line of the form::

        --% [annotation_name]: [annotation]

        Raises a property error if the doc is incorrectly formatted.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :rtype: List[DocAnnotation]
        """

        c_result = self._eval_field(
            _DocAnnotationArrayConverter.c_type(), _basic_decl_p_doc_annotations
        )
        result = _DocAnnotationArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_doc(self):
        """
        Return the documentation associated with this decl. Raises a property
        error if the doc is incorrectly formatted.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :rtype: str
        """

        c_result = self._eval_field(_TextTypeConverter.c_type(), _basic_decl_p_doc)
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    def p_previous_part_for_decl(self, imprecise_fallback=False):
        """
        Return the previous part for this decl, if applicable.

        .. note:: It is not named previous_part, because BaseTypeDecl has a
        more precise version of previous_part that returns a BaseTypeDecl.
        Probably, we want to rename the specific versions, and have the root
        property be named previous_part. (TODO R925-008)

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _basic_decl_p_previous_part_for_decl,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_canonical_part(self, imprecise_fallback=False):
        """
        Return the canonical part for this decl. In the case of decls composed
        of several parts, the canonical part will be the first part.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _basic_decl_p_canonical_part, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_is_static_decl(self, imprecise_fallback=False):
        """
        Return whether this declaration is static.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _basic_decl_p_is_static_decl, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    @property
    def p_is_imported(self):
        """
        Whether this declaration is imported from another language.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _basic_decl_p_is_imported)
        result = bool(c_result.value)

        return result

    @property
    def f_aspects(self):
        """
        Return the list of aspects that are attached to this node.

        :rtype: AspectSpec
        """

        result = self._eval_astnode_field(_basic_decl_f_aspects)

        return result

    def p_get_aspect_assoc(self, name):
        """
        Return the aspect with name ``name`` for this entity.

        :param name:
        :type name: str
        :rtype: AspectAssoc
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)

        c_result = self._eval_field(
            _Entity_c_type(), _basic_decl_p_get_aspect_assoc, unwrapped_name
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_get_aspect_spec_expr(self, name):
        """
        Return the expression associated to the aspect with name ``name`` for
        this entity.

        :param name:
        :type name: str
        :rtype: Expr
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)

        c_result = self._eval_field(
            _Entity_c_type(), _basic_decl_p_get_aspect_spec_expr, unwrapped_name
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_get_aspect(self, name, imprecise_fallback=False):
        """
        Return the aspect with name ``name`` associated to this entity.

        Aspects are properties of entities that can be specified by the Ada
        program, either via aspect specifications, pragmas, or attributes.

        This will return the syntactic node corresponding to attribute
        directly.

        :param name:
        :type name: str
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: Aspect
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            Aspect._c_type(),
            _basic_decl_p_get_aspect,
            unwrapped_name,
            unwrapped_imprecise_fallback,
        )
        result = Aspect._wrap(c_result)

        return result

    def p_has_aspect(self, name, imprecise_fallback=False):
        """
        Returns whether the boolean aspect named ``name`` is set on the entity
        represented by this node.

        "Aspect" is used as in RM terminology (see RM 13).

        :param name:
        :type name: str
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _basic_decl_p_has_aspect,
            unwrapped_name,
            unwrapped_imprecise_fallback,
        )
        result = bool(c_result.value)

        return result

    def p_get_pragma(self, name):
        """
        Return the pragma with name ``name`` associated to this entity.

        :param name:
        :type name: str
        :rtype: PragmaNode
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)

        c_result = self._eval_field(
            _Entity_c_type(), _basic_decl_p_get_pragma, unwrapped_name
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_get_representation_clause(self, name, imprecise_fallback=False):
        """
        Return the representation clause associated to this type decl that
        defines the given attribute name.

        :param name:
        :type name: str
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: AttributeDefClause
        """

        _context = self.unit.context._c_value
        unwrapped_name = _symbol_type.unwrap(name, _context)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _basic_decl_p_get_representation_clause,
            unwrapped_name,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_is_compilation_unit_root(self):
        """
        Whether a BasicDecl is the root decl for its unit.

        :rtype: bool
        """

        c_result = self._eval_field(
            ctypes.c_uint8(), _basic_decl_p_is_compilation_unit_root
        )
        result = bool(c_result.value)

        return result

    def p_is_visible(self, from_node):
        """
        Return whether this declaration is visible from the point of view of
        the given ``origin`` node.

        .. ATTENTION:: Only package-level (public or private) declarations are
        supported for now.

        :param from_node:
        :type from_node: AdaNode
        :rtype: bool
        """

        unwrapped_from_node = AdaNode._unwrap(from_node)

        c_result = self._eval_field(
            ctypes.c_uint8(), _basic_decl_p_is_visible, unwrapped_from_node
        )
        result = bool(c_result.value)

        return result

    @property
    def p_base_subp_declarations(self):
        """
        If Self declares a primitive subprogram of some tagged type T, return
        the set of all subprogram declarations that it overrides (including
        itself).

        :rtype: List[BasicDecl]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _basic_decl_p_base_subp_declarations
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_root_subp_declarations(self, origin=None):
        """
        If Self declares a primitive subprogram of some tagged type T, return
        the root subprogram declarations that it overrides. There can be
        several, as in the following scenario:

        - package Root defines the root tagged type T and subprogram Foo.

        - package Itf defines interface I and abstract subprogram Foo.

        - package D defines "type U is new Root.T and Itf.I" and an overriding
        subprogram Foo.

        Here, root_subp_declarations of Foo defined in package D will return
        both Foo from package Root and Foo from package Itf.

        :param origin:
        :type origin: AdaNode
        :rtype: List[BasicDecl]
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _basic_decl_p_root_subp_declarations,
            unwrapped_origin,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_find_all_overrides(self, units, imprecise_fallback=False):
        """
        If Self is the declaration of a primitive of some type T, return the
        list of all subprogram that override this subprogram among the given
        units.

        :param units:
        :type units: List[AnalysisUnit]
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[BasicDecl]
        """

        unwrapped_units = _AnalysisUnitArrayConverter.unwrap(units)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _basic_decl_p_find_all_overrides,
            unwrapped_units.c_value,
            unwrapped_imprecise_fallback,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_defining_names(self):
        """
        Get all the names of this basic declaration.

        :rtype: List[DefiningName]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _basic_decl_p_defining_names
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_defining_name(self):
        """
        Get the name of this declaration. If this declaration has several
        names, it will return the first one.

        :rtype: DefiningName
        """

        c_result = self._eval_field(_Entity_c_type(), _basic_decl_p_defining_name)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_type_expression(self):
        """
        Return the type expression for this BasicDecl if applicable, a null
        otherwise.

        :rtype: TypeExpr
        """

        c_result = self._eval_field(_Entity_c_type(), _basic_decl_p_type_expression)
        result = AdaNode._wrap(c_result)

        return result

    def p_subp_spec_or_null(self, follow_generic=False):
        """
        If Self is a Subp, returns the specification of this subprogram.

        If ``follow_generic`` is True, will also work for instances of
        ``GenericSubpDecl``.

        :param follow_generic:
        :type follow_generic: bool
        :rtype: BaseSubpSpec
        """

        unwrapped_follow_generic = bool(follow_generic)

        c_result = self._eval_field(
            _Entity_c_type(), _basic_decl_p_subp_spec_or_null, unwrapped_follow_generic
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_is_subprogram(self):
        """
        Return True if self is a subprogram node in the general sense (which
        is, an entity that can be called). This includes separates and entries.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _basic_decl_p_is_subprogram)
        result = bool(c_result.value)

        return result

    @property
    def p_relative_name(self):
        """
        Return the relative name for Self. If Self's defining name is
        ``A.B.C``, return C as a node.

        :rtype: SingleTokNode
        """

        c_result = self._eval_field(_Entity_c_type(), _basic_decl_p_relative_name)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_relative_name_text(self):
        """
        Return the relative name for Self, as text.

        :rtype: str
        """

        c_result = self._eval_field(_symbol_type(), _basic_decl_p_relative_name_text)
        result = _symbol_type.wrap(c_result)

        return result

    def p_next_part_for_decl(self, imprecise_fallback=False):
        """
        Return the next part of this declaration, if applicable.

        .. note:: It is not named next_part, because BaseTypeDecl has a more
        precise version of next_part that returns a BaseTypeDecl. Probably, we
        want to rename the specific versions, and have the root property be
        named next_part. (TODO R925-008)

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _basic_decl_p_next_part_for_decl,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_body_part_for_decl(self, imprecise_fallback=False):
        """
        Return the body corresponding to this declaration, if applicable.

        .. note:: It is not named body_part, subclasses have more precise
        versions named body_part and returning a more precise result. Probably,
        we want to rename the specific versions, and have the root property be
        named body_part. (TODO R925-008)

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BodyNode
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _basic_decl_p_body_part_for_decl,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_fully_qualified_name_array(self):
        """
        Return the fully qualified name corresponding to this declaration, as
        an array of symbols.

        :rtype: List[str]
        """

        c_result = self._eval_field(
            _UnboundedTextTypeArrayConverter.c_type(),
            _basic_decl_p_fully_qualified_name_array,
        )
        result = _UnboundedTextTypeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_fully_qualified_name(self):
        """
        Return the fully qualified name corresponding to this declaration.

        :rtype: str
        """

        c_result = self._eval_field(
            _TextTypeConverter.c_type(), _basic_decl_p_fully_qualified_name
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    @property
    def p_canonical_fully_qualified_name(self):
        """
        Return a canonical representation of the fully qualified name
        corresponding to this declaration.

        :rtype: str
        """

        c_result = self._eval_field(
            _TextTypeConverter.c_type(), _basic_decl_p_canonical_fully_qualified_name
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    @property
    def p_unique_identifying_name(self):
        """
        Return a unique identifying name for this declaration, provided this
        declaration is a public declaration. In the case of subprograms, this
        will include the profile.

        .. attention:: This will only return a unique name for public
        declarations. Notably, anything nested in an unnamed declare block
        won't be handled correctly.

        :rtype: str
        """

        c_result = self._eval_field(
            _TextTypeConverter.c_type(), _basic_decl_p_unique_identifying_name
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    _field_names = AdaNode._field_names + ()


class BaseFormalParamDecl(BasicDecl):
    """
    Base class for formal parameter declarations. This is used both for records
    components and for subprogram parameters.

    This is a Libadalang abstraction, that has no ARM existence.
    """

    __slots__ = []

    def p_formal_type(self, origin=None):
        """
        Return the type for this formal.

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_formal_param_decl_p_formal_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + ()


class ComponentDecl(BaseFormalParamDecl):
    """
    Declaration for a component.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_component_decl_f_ids)

        return result

    @property
    def f_component_def(self):
        """
        :rtype: ComponentDef
        """

        result = self._eval_astnode_field(_component_decl_f_component_def)

        return result

    @property
    def f_default_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_component_decl_f_default_expr)

        return result

    _field_names = BaseFormalParamDecl._field_names + (
        "f_ids",
        "f_component_def",
        "f_default_expr",
        "f_aspects",
    )

    _kind_name = "ComponentDecl"


class DiscriminantSpec(BaseFormalParamDecl):
    """
    Known list of discriminants in type declarations.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_discriminant_spec_f_ids)

        return result

    @property
    def f_type_expr(self):
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication

        :rtype: TypeExpr
        """

        result = self._eval_astnode_field(_discriminant_spec_f_type_expr)

        return result

    @property
    def f_default_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_discriminant_spec_f_default_expr)

        return result

    _field_names = BaseFormalParamDecl._field_names + (
        "f_ids",
        "f_type_expr",
        "f_default_expr",
    )

    _kind_name = "DiscriminantSpec"


class GenericFormal(BaseFormalParamDecl):
    """
    Enclosing declaration for a generic formal. The real declaration is
    accessible via the ``decl`` field.
    """

    __slots__ = []

    @property
    def f_decl(self):
        """
        This field can contain one of the following nodes:

        * FormalSubpDecl

        * GenericInstantiation

        * IncompleteTypeDecl

        * NumberDecl

        * ObjectDecl

        * SingleProtectedDecl

        * SingleTaskDecl

        * TypeDecl

        :rtype: BasicDecl
        """

        result = self._eval_astnode_field(_generic_formal_f_decl)

        return result

    _field_names = BaseFormalParamDecl._field_names + ("f_decl",)


class GenericFormalObjDecl(GenericFormal):
    """
    Formal declaration for an object.
    """

    __slots__ = []

    _field_names = GenericFormal._field_names + ()

    _kind_name = "GenericFormalObjDecl"


class GenericFormalPackage(GenericFormal):
    """
    Formal declaration for a package.
    """

    __slots__ = []

    _field_names = GenericFormal._field_names + ()

    _kind_name = "GenericFormalPackage"


class GenericFormalSubpDecl(GenericFormal):
    """
    Formal declaration for a subprogram.
    """

    __slots__ = []

    _field_names = GenericFormal._field_names + ()

    _kind_name = "GenericFormalSubpDecl"


class GenericFormalTypeDecl(GenericFormal):
    """
    Formal declaration for a type.
    """

    __slots__ = []

    _field_names = GenericFormal._field_names + ()

    _kind_name = "GenericFormalTypeDecl"


class ParamSpec(BaseFormalParamDecl):
    """
    Specification for a parameter.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_param_spec_f_ids)

        return result

    @property
    def f_has_aliased(self):
        """
        :rtype: AliasedNode
        """

        result = self._eval_astnode_field(_param_spec_f_has_aliased)

        return result

    @property
    def f_mode(self):
        """
        :rtype: Mode
        """

        result = self._eval_astnode_field(_param_spec_f_mode)

        return result

    @property
    def f_type_expr(self):
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication

        :rtype: TypeExpr
        """

        result = self._eval_astnode_field(_param_spec_f_type_expr)

        return result

    @property
    def f_default_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_param_spec_f_default_expr)

        return result

    _field_names = BaseFormalParamDecl._field_names + (
        "f_ids",
        "f_has_aliased",
        "f_mode",
        "f_type_expr",
        "f_default_expr",
    )

    _kind_name = "ParamSpec"


class BasePackageDecl(BasicDecl):
    """
    Package declarations. Concrete instances of this class will be created in
    generic package declarations. Other non-generic package declarations will
    be instances of PackageDecl.

    The behavior is the same, the only difference is that BasePackageDecl and
    PackageDecl have different behavior regarding lexical environments. In the
    case of generic package declarations, we use BasePackageDecl which has no
    env_spec, and the environment behavior is handled by the GenericPackageDecl
    instance.
    """

    __slots__ = []

    @property
    def f_package_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_base_package_decl_f_package_name)

        return result

    @property
    def f_public_part(self):
        """
        :rtype: PublicPart
        """

        result = self._eval_astnode_field(_base_package_decl_f_public_part)

        return result

    @property
    def f_private_part(self):
        """
        :rtype: PrivatePart
        """

        result = self._eval_astnode_field(_base_package_decl_f_private_part)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_base_package_decl_f_end_name)

        return result

    @property
    def p_body_part(self):
        """
        Return the PackageBody corresponding to this node.

        :rtype: PackageBody
        """

        c_result = self._eval_field(_Entity_c_type(), _base_package_decl_p_body_part)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + (
        "f_package_name",
        "f_aspects",
        "f_public_part",
        "f_private_part",
        "f_end_name",
    )


class GenericPackageInternal(BasePackageDecl):
    """
    This class denotes the internal package contained by a GenericPackageDecl.
    """

    __slots__ = []

    _field_names = BasePackageDecl._field_names + ()

    _kind_name = "GenericPackageInternal"


class PackageDecl(BasePackageDecl):
    """
    Non-generic package declarations.
    """

    __slots__ = []

    _field_names = BasePackageDecl._field_names + ()

    _kind_name = "PackageDecl"


class BaseTypeDecl(BasicDecl):
    """
    Base class for type declarations.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_base_type_decl_f_name)

        return result

    def p_base_subtype(self, origin=None):
        """
        If this type decl is a subtype decl, return the base subtype. If not,
        return ``Self``.

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_base_subtype, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_private_completion(self):
        """
        Return the private completion for this type, if there is one.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_private_completion
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_is_inherited_primitive(self, p):
        """
        Assuming that P is a primitive of Self, return whether the given
        primitive P is inherited from one of Self's parents.

        :param p:
        :type p: BasicDecl
        :rtype: bool
        """

        unwrapped_p = AdaNode._unwrap(p)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_inherited_primitive, unwrapped_p
        )
        result = bool(c_result.value)

        return result

    def p_get_record_representation_clause(self, imprecise_fallback=False):
        """
        Return the record representation clause associated to this type decl,
        if applicable (i.e. this type decl defines a record type).

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: RecordRepClause
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_type_decl_p_get_record_representation_clause,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_get_enum_representation_clause(self, imprecise_fallback=False):
        """
        Return the enum representation clause associated to this type decl, if
        applicable (i.e. this type decl defines an enum type).

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: EnumRepClause
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_type_decl_p_get_enum_representation_clause,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_is_record_type(self, origin=None):
        """
        Return whether this type is a record type.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_record_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_array_type(self, origin=None):
        """
        Return whether this type is an array type.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_array_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_find_derived_types(self, root, origin, imprecise_fallback=False):
        """
        Find types derived from self in the given ``root`` and its children.

        :param root:
        :type root: AdaNode
        :param origin:
        :type origin: AdaNode
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[TypeDecl]
        """

        unwrapped_root = AdaNode._unwrap(root)
        unwrapped_origin = AdaNode._unwrap(origin)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_type_decl_p_find_derived_types,
            unwrapped_root,
            unwrapped_origin,
            unwrapped_imprecise_fallback,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_is_real_type(self, origin=None):
        """
        Whether type is a real type or not.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_real_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_float_type(self, origin=None):
        """
        Whether type is a float type or not.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_float_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_fixed_point(self, origin=None):
        """
        Whether type is a fixed point type or not.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_fixed_point, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_enum_type(self, origin=None):
        """
        Whether type is an enum type

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_enum_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_access_type(self, origin=None):
        """
        Whether Self is an access type or not

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_access_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_char_type(self, origin=None):
        """
        Whether type is a character type or not

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_char_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    @property
    def p_discrete_range(self):
        """
        Return the discrete range for this type decl, if applicable.

        :rtype: DiscreteRange
        """

        c_result = self._eval_field(
            DiscreteRange._c_type(), _base_type_decl_p_discrete_range
        )
        result = DiscreteRange._wrap(c_result)

        return result

    def p_is_discrete_type(self, origin=None):
        """
        Whether type is a discrete type or not.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_discrete_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_is_int_type(self, origin=None):
        """
        Whether type is an integer type or not.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_int_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_accessed_type(self, origin=None):
        """
        If this type is an access type, or a type with an Implicit_Dereference
        aspect, return the type of a dereference of an instance of this type.

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_accessed_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_is_tagged_type(self, origin=None):
        """
        Whether type is tagged or not

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_tagged_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_base_type(self, origin=None):
        """
        Return the base type entity for this derived type declaration

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_base_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_base_types(self, origin=None):
        """
        Return the list of base types for Self.

        :param origin:
        :type origin: AdaNode
        :rtype: List[BaseTypeDecl]
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_type_decl_p_base_types,
            unwrapped_origin,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_find_all_derived_types(self, units, imprecise_fallback=False):
        """
        Return the list of all types that inherit (directly or inderictly) from
        Self among the given units.

        :param units:
        :type units: List[AnalysisUnit]
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[TypeDecl]
        """

        unwrapped_units = _AnalysisUnitArrayConverter.unwrap(units)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _base_type_decl_p_find_all_derived_types,
            unwrapped_units.c_value,
            unwrapped_imprecise_fallback,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_comp_type(self, is_subscript=False, origin=None):
        """
        Return the component type of `Self`, if applicable. The component type
        is the type you'll get if you call a value whose type is `Self`.  So it
        can either be:

        1. The component type for an array. 2. The return type for an access to
        function.

        :param is_subscript:
        :type is_subscript: bool
        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_is_subscript = bool(is_subscript)
        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_type_decl_p_comp_type,
            unwrapped_is_subscript,
            unwrapped_origin,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_index_type(self, dim, origin=None):
        """
        Return the index type for dimension ``dim`` for this type, if
        applicable.

        :param dim:
        :type dim: int
        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_dim = int(dim)
        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_type_decl_p_index_type,
            unwrapped_dim,
            unwrapped_origin,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_is_derived_type(self, other_type, origin=None):
        """
        Whether Self is derived from other_type.

        :param other_type:
        :type other_type: BaseTypeDecl
        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_other_type = AdaNode._unwrap(other_type)
        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _base_type_decl_p_is_derived_type,
            unwrapped_other_type,
            unwrapped_origin,
        )
        result = bool(c_result.value)

        return result

    def p_is_interface_type(self, origin=None):
        """
        Return True iff this type declaration is an interface definition.

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_interface_type, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    def p_matching_type(self, expected_type, origin=None):
        """
        Return whether ``self`` matches ``expected_type``.

        :param expected_type:
        :type expected_type: BaseTypeDecl
        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_expected_type = AdaNode._unwrap(expected_type)
        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _base_type_decl_p_matching_type,
            unwrapped_expected_type,
            unwrapped_origin,
        )
        result = bool(c_result.value)

        return result

    def p_canonical_type(self, origin=None):
        """
        Return the canonical type declaration for this type declaration. For
        subtypes, it will return the base type declaration.

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_canonical_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_previous_part(self, go_to_incomplete=True):
        """
        Returns the previous part for this type decl.

        :param go_to_incomplete:
        :type go_to_incomplete: bool
        :rtype: BaseTypeDecl
        """

        unwrapped_go_to_incomplete = bool(go_to_incomplete)

        c_result = self._eval_field(
            _Entity_c_type(),
            _base_type_decl_p_previous_part,
            unwrapped_go_to_incomplete,
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_next_part(self):
        """
        Returns the next part for this type decl.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _base_type_decl_p_next_part)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_full_view(self):
        """
        Return the full completion of this type.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _base_type_decl_p_full_view)
        result = AdaNode._wrap(c_result)

        return result

    def p_is_definite_subtype(self, origin=None):
        """
        Returns whether this is a definite subtype.

        For convenience, this will return ``False`` for incomplete types, even
        though the correct answer is more akin to "non applicable".

        :param origin:
        :type origin: AdaNode
        :rtype: bool
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            ctypes.c_uint8(), _base_type_decl_p_is_definite_subtype, unwrapped_origin
        )
        result = bool(c_result.value)

        return result

    @property
    def p_is_private(self):
        """
        Whether node is a private view of corresponding type.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _base_type_decl_p_is_private)
        result = bool(c_result.value)

        return result

    @property
    def p_discriminants_list(self):
        """
        Return the list of all discriminants of this type. If this type has no
        discriminant or only unknown discriminants, an empty list is returned.

        :rtype: List[BaseFormalParamDecl]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _base_type_decl_p_discriminants_list
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    def p_root_type(self, origin=None):
        """
        Return the type that is at the root of the derivation hierarchy
        (ignoring secondary interfaces derivations for tagged types)

        :param origin:
        :type origin: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _Entity_c_type(), _base_type_decl_p_root_type, unwrapped_origin
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_shapes(self, include_discriminants=True, origin=None):
        """
        Must be called on a record (sub-)type declaration. Return all the
        possible shapes that a value of this record type can take. For example,
        consider the following record definition:

        .. code::

        type R (A : Integer; B : Integer) is record X : Integer; case A is when
        1 .. 10 => Y_1 : Integer; case B is when 1 .. 10 => Z_1 : Integer; when
        others => null; end case; when 11 .. 20 => Y_2 : Integer; case B is
        when 1 .. 10 => Z_2 : Integer; when others => null; end case; when
        others => null; end case; end record;

        For this instance, this property will return the following results:

        .. code::

        [ [X, Y_1, Z_1], [X, Y_1], [X, Y_2, Z_2], [X, Y_2], [X] ]

        .. ATTENTION:: This property is inaccurate when called on a record
        extension which defines components under a certain condition C, and
        this same condition is used to define some components in the parent
        record: in that case, any feasible shape will in practice contain
        either both the components defined under condition C in the child
        record and the parent record, or none of them.

        However, due to the simplified algorithm we use here to compute the
        feasible shapes, we will also return shapes that include the components
        of the child record but not the parent record, and conversely.

        :param include_discriminants:
        :type include_discriminants: bool
        :param origin:
        :type origin: AdaNode
        :rtype: List[Shape]
        """

        unwrapped_include_discriminants = bool(include_discriminants)
        unwrapped_origin = AdaNode._unwrap(origin)

        c_result = self._eval_field(
            _ShapeArrayConverter.c_type(),
            _base_type_decl_p_shapes,
            unwrapped_include_discriminants,
            unwrapped_origin,
        )
        result = _ShapeArrayConverter.wrap(c_result, False)

        return result

    _field_names = BasicDecl._field_names + ("f_name",)


class BaseSubtypeDecl(BaseTypeDecl):
    """
    Base class for subtype declarations.
    """

    __slots__ = []

    _field_names = BaseTypeDecl._field_names + ()


class DiscreteBaseSubtypeDecl(BaseSubtypeDecl):
    """
    Specific ``BaseSubtypeDecl`` synthetic subclass for the base type of scalar
    types.
    """

    __slots__ = []

    _field_names = BaseSubtypeDecl._field_names + ()

    _kind_name = "DiscreteBaseSubtypeDecl"


class SubtypeDecl(BaseSubtypeDecl):
    """
    Subtype declaration.
    """

    __slots__ = []

    @property
    def f_subtype(self):
        """
        :rtype: SubtypeIndication
        """

        result = self._eval_astnode_field(_subtype_decl_f_subtype)

        return result

    _field_names = BaseSubtypeDecl._field_names + (
        "f_subtype",
        "f_aspects",
    )

    _kind_name = "SubtypeDecl"


class ClasswideTypeDecl(BaseTypeDecl):
    """
    Synthetic node (not parsed, generated from a property call). Refers to the
    classwide type for a given tagged type. The aim is that those be mostly
    equivalent to their non-classwide type, except for some resolution rules.
    """

    __slots__ = []

    _field_names = BaseTypeDecl._field_names + ()

    _kind_name = "ClasswideTypeDecl"


class IncompleteTypeDecl(BaseTypeDecl):
    """
    Incomplete declaration for a type.
    """

    __slots__ = []

    @property
    def f_discriminants(self):
        """
        :rtype: DiscriminantPart
        """

        result = self._eval_astnode_field(_incomplete_type_decl_f_discriminants)

        return result

    _field_names = BaseTypeDecl._field_names + ("f_discriminants",)

    _kind_name = "IncompleteTypeDecl"


class IncompleteTaggedTypeDecl(IncompleteTypeDecl):
    """
    Incomplete declaration for a tagged type.
    """

    __slots__ = []

    @property
    def f_has_abstract(self):
        """
        :rtype: AbstractNode
        """

        result = self._eval_astnode_field(_incomplete_tagged_type_decl_f_has_abstract)

        return result

    _field_names = IncompleteTypeDecl._field_names + ("f_has_abstract",)

    _kind_name = "IncompleteTaggedTypeDecl"


class ProtectedTypeDecl(BaseTypeDecl):
    """
    Declaration for a protected type.
    """

    __slots__ = []

    @property
    def f_discriminants(self):
        """
        :rtype: DiscriminantPart
        """

        result = self._eval_astnode_field(_protected_type_decl_f_discriminants)

        return result

    @property
    def f_interfaces(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: ParentList
        """

        result = self._eval_astnode_field(_protected_type_decl_f_interfaces)

        return result

    @property
    def f_definition(self):
        """
        :rtype: ProtectedDef
        """

        result = self._eval_astnode_field(_protected_type_decl_f_definition)

        return result

    _field_names = BaseTypeDecl._field_names + (
        "f_discriminants",
        "f_aspects",
        "f_interfaces",
        "f_definition",
    )

    _kind_name = "ProtectedTypeDecl"


class TaskTypeDecl(BaseTypeDecl):
    """
    Declaration for a task type.
    """

    __slots__ = []

    @property
    def f_discriminants(self):
        """
        :rtype: DiscriminantPart
        """

        result = self._eval_astnode_field(_task_type_decl_f_discriminants)

        return result

    @property
    def f_definition(self):
        """
        :rtype: TaskDef
        """

        result = self._eval_astnode_field(_task_type_decl_f_definition)

        return result

    _field_names = BaseTypeDecl._field_names + (
        "f_discriminants",
        "f_aspects",
        "f_definition",
    )

    _kind_name = "TaskTypeDecl"


class SingleTaskTypeDecl(TaskTypeDecl):
    """
    Type declaration for a single task.
    """

    __slots__ = []

    _field_names = TaskTypeDecl._field_names + ()

    _kind_name = "SingleTaskTypeDecl"


class TypeDecl(BaseTypeDecl):
    """
    Type declarations that embed a type definition node.
    """

    __slots__ = []

    @property
    def f_discriminants(self):
        """
        :rtype: DiscriminantPart
        """

        result = self._eval_astnode_field(_type_decl_f_discriminants)

        return result

    @property
    def f_type_def(self):
        """
        :rtype: TypeDef
        """

        result = self._eval_astnode_field(_type_decl_f_type_def)

        return result

    def p_get_primitives(self, only_inherited=False):
        """
        Return the list of all primitive operations that are available on this
        type. If `only_inherited` is True, it will only return the primitives
        that are implicitly inherited by this type, discarding those explicitly
        defined on this type.

        :param only_inherited:
        :type only_inherited: bool
        :rtype: List[BasicDecl]
        """

        unwrapped_only_inherited = bool(only_inherited)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _type_decl_p_get_primitives,
            unwrapped_only_inherited,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = BaseTypeDecl._field_names + (
        "f_discriminants",
        "f_type_def",
        "f_aspects",
    )

    _kind_name = "TypeDecl"


class AnonymousTypeDecl(TypeDecl):
    """
    Anonymous type declaration (for anonymous array or access types).
    """

    __slots__ = []

    _field_names = TypeDecl._field_names + ()

    _kind_name = "AnonymousTypeDecl"


class SynthAnonymousTypeDecl(AnonymousTypeDecl):
    """
    Synthetic anonymous type decl. Used to generate anonymous access types.
    """

    __slots__ = []

    _field_names = AnonymousTypeDecl._field_names + ()

    _kind_name = "SynthAnonymousTypeDecl"


class BasicSubpDecl(BasicDecl):
    """
    Base class for subprogram declarations.
    """

    __slots__ = []

    @property
    def p_subp_decl_spec(self):
        """
        Return the specification for this subprogram

        :rtype: BaseSubpSpec
        """

        c_result = self._eval_field(_Entity_c_type(), _basic_subp_decl_p_subp_decl_spec)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + ()


class ClassicSubpDecl(BasicSubpDecl):
    """
    This is an intermediate abstract class for subprogram declarations with a
    common structure: overriding indicator, ``SubpSpec``, aspects, <other
    fields>.
    """

    __slots__ = []

    @property
    def f_overriding(self):
        """
        :rtype: OverridingNode
        """

        result = self._eval_astnode_field(_classic_subp_decl_f_overriding)

        return result

    @property
    def f_subp_spec(self):
        """
        :rtype: SubpSpec
        """

        result = self._eval_astnode_field(_classic_subp_decl_f_subp_spec)

        return result

    def p_body_part(self, imprecise_fallback=False):
        """
        Return the BaseSubpBody corresponding to this node.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BaseSubpBody
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _classic_subp_decl_p_body_part,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicSubpDecl._field_names + (
        "f_overriding",
        "f_subp_spec",
    )


class AbstractSubpDecl(ClassicSubpDecl):
    """
    Declaration for an abstract subprogram.
    """

    __slots__ = []

    _field_names = ClassicSubpDecl._field_names + ("f_aspects",)

    _kind_name = "AbstractSubpDecl"


class FormalSubpDecl(ClassicSubpDecl):
    """
    Formal subprogram declarations, in generic declarations formal parts.
    """

    __slots__ = []

    @property
    def f_default_expr(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BoxExpr

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Expr
        """

        result = self._eval_astnode_field(_formal_subp_decl_f_default_expr)

        return result

    _field_names = ClassicSubpDecl._field_names + (
        "f_default_expr",
        "f_aspects",
    )


class AbstractFormalSubpDecl(FormalSubpDecl):
    """
    Formal declaration for an abstract subprogram.
    """

    __slots__ = []

    _field_names = FormalSubpDecl._field_names + ()

    _kind_name = "AbstractFormalSubpDecl"


class ConcreteFormalSubpDecl(FormalSubpDecl):
    """
    Formal declaration for a concrete subprogram.
    """

    __slots__ = []

    _field_names = FormalSubpDecl._field_names + ()

    _kind_name = "ConcreteFormalSubpDecl"


class SubpDecl(ClassicSubpDecl):
    """
    Regular subprogram declaration.
    """

    __slots__ = []

    _field_names = ClassicSubpDecl._field_names + ("f_aspects",)

    _kind_name = "SubpDecl"


class EntryDecl(BasicSubpDecl):
    """
    Entry declaration.
    """

    __slots__ = []

    @property
    def f_overriding(self):
        """
        :rtype: OverridingNode
        """

        result = self._eval_astnode_field(_entry_decl_f_overriding)

        return result

    @property
    def f_spec(self):
        """
        :rtype: EntrySpec
        """

        result = self._eval_astnode_field(_entry_decl_f_spec)

        return result

    def p_body_part(self, imprecise_fallback=False):
        """
        Return the entry body associated to this entry declaration.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: EntryBody
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _entry_decl_p_body_part, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicSubpDecl._field_names + (
        "f_overriding",
        "f_spec",
        "f_aspects",
    )

    _kind_name = "EntryDecl"


class EnumLiteralDecl(BasicSubpDecl):
    """
    Declaration for an enumeration literal.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_enum_literal_decl_f_name)

        return result

    @property
    def p_enum_type(self):
        """
        Return the enum type corresponding to this enum literal.

        :rtype: TypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _enum_literal_decl_p_enum_type)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicSubpDecl._field_names + ("f_name",)

    _kind_name = "EnumLiteralDecl"


class GenericSubpInternal(BasicSubpDecl):
    """
    Internal node for generic subprograms.
    """

    __slots__ = []

    @property
    def f_subp_spec(self):
        """
        :rtype: SubpSpec
        """

        result = self._eval_astnode_field(_generic_subp_internal_f_subp_spec)

        return result

    _field_names = BasicSubpDecl._field_names + (
        "f_subp_spec",
        "f_aspects",
    )

    _kind_name = "GenericSubpInternal"


class BodyNode(BasicDecl):
    """
    Base class for an Ada body (RM 3.11). A body is the completion of a
    declaration.
    """

    __slots__ = []

    def p_previous_part(self, imprecise_fallback=False):
        """
        Return the previous part for this body. Might be a declaration or a
        body stub.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _body_node_p_previous_part, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_decl_part(self, imprecise_fallback=False):
        """
        Return the decl corresponding to this node if applicable.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _body_node_p_decl_part, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_subunit_root(self):
        """
        If self is a subunit, return the body in which it is rooted.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _body_node_p_subunit_root)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + ()


class BaseSubpBody(BodyNode):
    """
    Base class for subprogram bodies.
    """

    __slots__ = []

    @property
    def f_overriding(self):
        """
        :rtype: OverridingNode
        """

        result = self._eval_astnode_field(_base_subp_body_f_overriding)

        return result

    @property
    def f_subp_spec(self):
        """
        :rtype: SubpSpec
        """

        result = self._eval_astnode_field(_base_subp_body_f_subp_spec)

        return result

    _field_names = BodyNode._field_names + (
        "f_overriding",
        "f_subp_spec",
    )


class ExprFunction(BaseSubpBody):
    """
    Expression function.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * BaseAggregate

        * ParenExpr

        :rtype: Expr
        """

        result = self._eval_astnode_field(_expr_function_f_expr)

        return result

    _field_names = BaseSubpBody._field_names + (
        "f_expr",
        "f_aspects",
    )

    _kind_name = "ExprFunction"


class NullSubpDecl(BaseSubpBody):
    """
    Declaration for a null subprogram.
    """

    __slots__ = []

    _field_names = BaseSubpBody._field_names + ("f_aspects",)

    _kind_name = "NullSubpDecl"


class SubpBody(BaseSubpBody):
    """
    Subprogram body.
    """

    __slots__ = []

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_subp_body_f_decls)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_subp_body_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_subp_body_f_end_name)

        return result

    _field_names = BaseSubpBody._field_names + (
        "f_aspects",
        "f_decls",
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "SubpBody"


class SubpRenamingDecl(BaseSubpBody):
    """
    Declaration for a subprogram renaming.
    """

    __slots__ = []

    @property
    def f_renames(self):
        """
        :rtype: RenamingClause
        """

        result = self._eval_astnode_field(_subp_renaming_decl_f_renames)

        return result

    _field_names = BaseSubpBody._field_names + (
        "f_renames",
        "f_aspects",
    )

    _kind_name = "SubpRenamingDecl"


class BodyStub(BodyNode):
    """
    Base class for a body stub (RM 10.1.3). A body stub is meant to be
    completed by .
    """

    __slots__ = []

    @property
    def p_syntactic_fully_qualified_name(self):
        """
        Return the syntactic fully qualified name to refer to this body.

        Note that this can raise a Property_Error when the stub is in an
        illegal place (too nested, in a declare block, etc.).

        :rtype: List[str]
        """

        c_result = self._eval_field(
            _UnboundedTextTypeArrayConverter.c_type(),
            _body_stub_p_syntactic_fully_qualified_name,
        )
        result = _UnboundedTextTypeArrayConverter.wrap(c_result, False)

        return result

    _field_names = BodyNode._field_names + ()


class PackageBodyStub(BodyStub):
    """
    Stub for a package body (``is separate``).
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_package_body_stub_f_name)

        return result

    _field_names = BodyStub._field_names + (
        "f_name",
        "f_aspects",
    )

    _kind_name = "PackageBodyStub"


class ProtectedBodyStub(BodyStub):
    """
    Stub for a protected object body (``is separate``).
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_protected_body_stub_f_name)

        return result

    _field_names = BodyStub._field_names + (
        "f_name",
        "f_aspects",
    )

    _kind_name = "ProtectedBodyStub"


class SubpBodyStub(BodyStub):
    """
    Stub for a subprogram body (``is separate``).
    """

    __slots__ = []

    @property
    def f_overriding(self):
        """
        :rtype: OverridingNode
        """

        result = self._eval_astnode_field(_subp_body_stub_f_overriding)

        return result

    @property
    def f_subp_spec(self):
        """
        :rtype: SubpSpec
        """

        result = self._eval_astnode_field(_subp_body_stub_f_subp_spec)

        return result

    _field_names = BodyStub._field_names + (
        "f_overriding",
        "f_subp_spec",
        "f_aspects",
    )

    _kind_name = "SubpBodyStub"


class TaskBodyStub(BodyStub):
    """
    Stub for a task body (``is separate``).
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_task_body_stub_f_name)

        return result

    _field_names = BodyStub._field_names + (
        "f_name",
        "f_aspects",
    )

    _kind_name = "TaskBodyStub"


class EntryBody(BodyNode):
    """
    Entry body.
    """

    __slots__ = []

    @property
    def f_entry_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_entry_body_f_entry_name)

        return result

    @property
    def f_index_spec(self):
        """
        :rtype: EntryIndexSpec
        """

        result = self._eval_astnode_field(_entry_body_f_index_spec)

        return result

    @property
    def f_params(self):
        """
        :rtype: EntryCompletionFormalParams
        """

        result = self._eval_astnode_field(_entry_body_f_params)

        return result

    @property
    def f_barrier(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_entry_body_f_barrier)

        return result

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_entry_body_f_decls)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_entry_body_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_entry_body_f_end_name)

        return result

    _field_names = BodyNode._field_names + (
        "f_entry_name",
        "f_index_spec",
        "f_params",
        "f_barrier",
        "f_decls",
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "EntryBody"


class PackageBody(BodyNode):
    """
    Package body.
    """

    __slots__ = []

    @property
    def f_package_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_package_body_f_package_name)

        return result

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_package_body_f_decls)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_package_body_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_package_body_f_end_name)

        return result

    _field_names = BodyNode._field_names + (
        "f_package_name",
        "f_aspects",
        "f_decls",
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "PackageBody"


class ProtectedBody(BodyNode):
    """
    Protected object body.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_protected_body_f_name)

        return result

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_protected_body_f_decls)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_protected_body_f_end_name)

        return result

    _field_names = BodyNode._field_names + (
        "f_name",
        "f_aspects",
        "f_decls",
        "f_end_name",
    )

    _kind_name = "ProtectedBody"


class TaskBody(BodyNode):
    """
    Task body.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_task_body_f_name)

        return result

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_task_body_f_decls)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_task_body_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_task_body_f_end_name)

        return result

    _field_names = BodyNode._field_names + (
        "f_name",
        "f_aspects",
        "f_decls",
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "TaskBody"


class EntryIndexSpec(BasicDecl):
    """
    Index specification for an entry body.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_entry_index_spec_f_id)

        return result

    @property
    def f_subtype(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_entry_index_spec_f_subtype)

        return result

    _field_names = BasicDecl._field_names + (
        "f_id",
        "f_subtype",
    )

    _kind_name = "EntryIndexSpec"


class ErrorDecl(BasicDecl):
    """
    Placeholder node for syntax errors in lists of declarations.
    """

    __slots__ = []

    _field_names = BasicDecl._field_names + ()

    _kind_name = "ErrorDecl"


class ExceptionDecl(BasicDecl):
    """
    Exception declarations.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_exception_decl_f_ids)

        return result

    @property
    def f_renames(self):
        """
        :rtype: RenamingClause
        """

        result = self._eval_astnode_field(_exception_decl_f_renames)

        return result

    _field_names = BasicDecl._field_names + (
        "f_ids",
        "f_renames",
        "f_aspects",
    )

    _kind_name = "ExceptionDecl"


class ExceptionHandler(BasicDecl):
    """
    Exception handler.
    """

    __slots__ = []

    @property
    def f_exception_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_exception_handler_f_exception_name)

        return result

    @property
    def f_handled_exceptions(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * OthersDesignator

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: AlternativesList
        """

        result = self._eval_astnode_field(_exception_handler_f_handled_exceptions)

        return result

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_exception_handler_f_stmts)

        return result

    _field_names = BasicDecl._field_names + (
        "f_exception_name",
        "f_handled_exceptions",
        "f_stmts",
    )

    _kind_name = "ExceptionHandler"


class ForLoopVarDecl(BasicDecl):
    """
    Declaration for the controlling variable in a ``for`` loop.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_for_loop_var_decl_f_id)

        return result

    @property
    def f_id_type(self):
        """
        :rtype: SubtypeIndication
        """

        result = self._eval_astnode_field(_for_loop_var_decl_f_id_type)

        return result

    _field_names = BasicDecl._field_names + (
        "f_id",
        "f_id_type",
    )

    _kind_name = "ForLoopVarDecl"


class GenericDecl(BasicDecl):
    """
    Base class for generic declarations.
    """

    __slots__ = []

    @property
    def f_formal_part(self):
        """
        :rtype: GenericFormalPart
        """

        result = self._eval_astnode_field(_generic_decl_f_formal_part)

        return result

    _field_names = BasicDecl._field_names + ("f_formal_part",)


class GenericPackageDecl(GenericDecl):
    """
    Generic package declaration.
    """

    __slots__ = []

    @property
    def f_package_decl(self):
        """
        :rtype: GenericPackageInternal
        """

        result = self._eval_astnode_field(_generic_package_decl_f_package_decl)

        return result

    @property
    def p_body_part(self):
        """
        Return the PackageBody corresponding to this node, or null if there is
        none.

        :rtype: PackageBody
        """

        c_result = self._eval_field(_Entity_c_type(), _generic_package_decl_p_body_part)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = GenericDecl._field_names + ("f_package_decl",)

    _kind_name = "GenericPackageDecl"


class GenericSubpDecl(GenericDecl):
    """
    Generic subprogram declaration.
    """

    __slots__ = []

    @property
    def f_subp_decl(self):
        """
        :rtype: GenericSubpInternal
        """

        result = self._eval_astnode_field(_generic_subp_decl_f_subp_decl)

        return result

    def p_body_part(self, imprecise_fallback=False):
        """
        Return the BaseSubpBody corresponding to this node.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BaseSubpBody
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _generic_subp_decl_p_body_part,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = GenericDecl._field_names + ("f_subp_decl",)

    _kind_name = "GenericSubpDecl"


class GenericInstantiation(BasicDecl):
    """
    Instantiations of generics.
    """

    __slots__ = []

    @property
    def p_designated_generic_decl(self):
        """
        Return the generic decl entity designated by this instantiation,
        containing the generic context. This is equivalent to the expanded
        generic unit in GNAT.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(
            _Entity_c_type(), _generic_instantiation_p_designated_generic_decl
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + ()


class GenericPackageInstantiation(GenericInstantiation):
    """
    Instantiations of a generic package.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_generic_package_instantiation_f_name)

        return result

    @property
    def f_generic_pkg_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(
            _generic_package_instantiation_f_generic_pkg_name
        )

        return result

    @property
    def f_params(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ParamAssoc

        :rtype: AssocList
        """

        result = self._eval_astnode_field(_generic_package_instantiation_f_params)

        return result

    _field_names = GenericInstantiation._field_names + (
        "f_name",
        "f_generic_pkg_name",
        "f_params",
        "f_aspects",
    )

    _kind_name = "GenericPackageInstantiation"


class GenericSubpInstantiation(GenericInstantiation):
    """
    Instantiations of a generic subprogram.
    """

    __slots__ = []

    @property
    def f_overriding(self):
        """
        :rtype: OverridingNode
        """

        result = self._eval_astnode_field(_generic_subp_instantiation_f_overriding)

        return result

    @property
    def f_kind(self):
        """
        :rtype: SubpKind
        """

        result = self._eval_astnode_field(_generic_subp_instantiation_f_kind)

        return result

    @property
    def f_subp_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_generic_subp_instantiation_f_subp_name)

        return result

    @property
    def f_generic_subp_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(
            _generic_subp_instantiation_f_generic_subp_name
        )

        return result

    @property
    def f_params(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ParamAssoc

        :rtype: AssocList
        """

        result = self._eval_astnode_field(_generic_subp_instantiation_f_params)

        return result

    @property
    def p_designated_subp(self):
        """
        Return the subprogram decl designated by this instantiation.

        :rtype: AdaNode
        """

        c_result = self._eval_field(
            _Entity_c_type(), _generic_subp_instantiation_p_designated_subp
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = GenericInstantiation._field_names + (
        "f_overriding",
        "f_kind",
        "f_subp_name",
        "f_generic_subp_name",
        "f_params",
        "f_aspects",
    )

    _kind_name = "GenericSubpInstantiation"


class GenericRenamingDecl(BasicDecl):
    """
    Base node for all generic renaming declarations.
    """

    __slots__ = []

    _field_names = BasicDecl._field_names + ()


class GenericPackageRenamingDecl(GenericRenamingDecl):
    """
    Declaration for a generic package renaming.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_generic_package_renaming_decl_f_name)

        return result

    @property
    def f_renames(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_generic_package_renaming_decl_f_renames)

        return result

    _field_names = GenericRenamingDecl._field_names + (
        "f_name",
        "f_renames",
        "f_aspects",
    )

    _kind_name = "GenericPackageRenamingDecl"


class GenericSubpRenamingDecl(GenericRenamingDecl):
    """
    Declaration for a generic subprogram renaming.
    """

    __slots__ = []

    @property
    def f_kind(self):
        """
        :rtype: SubpKind
        """

        result = self._eval_astnode_field(_generic_subp_renaming_decl_f_kind)

        return result

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_generic_subp_renaming_decl_f_name)

        return result

    @property
    def f_renames(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_generic_subp_renaming_decl_f_renames)

        return result

    _field_names = GenericRenamingDecl._field_names + (
        "f_kind",
        "f_name",
        "f_renames",
        "f_aspects",
    )

    _kind_name = "GenericSubpRenamingDecl"


class LabelDecl(BasicDecl):
    """
    Declaration for a code label.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_label_decl_f_name)

        return result

    _field_names = BasicDecl._field_names + ("f_name",)

    _kind_name = "LabelDecl"


class NamedStmtDecl(BasicDecl):
    """
    BasicDecl that is always the declaration inside a named statement.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_named_stmt_decl_f_name)

        return result

    _field_names = BasicDecl._field_names + ("f_name",)

    _kind_name = "NamedStmtDecl"


class NumberDecl(BasicDecl):
    """
    Declaration for a static constant number.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_number_decl_f_ids)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_number_decl_f_expr)

        return result

    _field_names = BasicDecl._field_names + (
        "f_ids",
        "f_expr",
    )

    _kind_name = "NumberDecl"


class ObjectDecl(BasicDecl):
    """
    Base class for Ada object declarations (RM 3.3.1). Ada object declarations
    are variables/constants declarations that can be declared in any
    declarative scope.
    """

    __slots__ = []

    @property
    def f_ids(self):
        """
        :rtype: DefiningNameList
        """

        result = self._eval_astnode_field(_object_decl_f_ids)

        return result

    @property
    def f_has_aliased(self):
        """
        :rtype: AliasedNode
        """

        result = self._eval_astnode_field(_object_decl_f_has_aliased)

        return result

    @property
    def f_has_constant(self):
        """
        :rtype: ConstantNode
        """

        result = self._eval_astnode_field(_object_decl_f_has_constant)

        return result

    @property
    def f_mode(self):
        """
        :rtype: Mode
        """

        result = self._eval_astnode_field(_object_decl_f_mode)

        return result

    @property
    def f_type_expr(self):
        """
        :rtype: TypeExpr
        """

        result = self._eval_astnode_field(_object_decl_f_type_expr)

        return result

    @property
    def f_default_expr(self):
        """
        :rtype: Expr
        """

        result = self._eval_astnode_field(_object_decl_f_default_expr)

        return result

    @property
    def f_renaming_clause(self):
        """
        :rtype: RenamingClause
        """

        result = self._eval_astnode_field(_object_decl_f_renaming_clause)

        return result

    @property
    def p_public_part_decl(self):
        """
        If this object decl is the constant completion of an object decl in the
        public part, return the object decl from the public part.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _object_decl_p_public_part_decl)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_private_part_decl(self):
        """
        If this object decl is the incomplete declaration of a constant in a
        public part, return its completion in the private part.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _object_decl_p_private_part_decl)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + (
        "f_ids",
        "f_has_aliased",
        "f_has_constant",
        "f_mode",
        "f_type_expr",
        "f_default_expr",
        "f_renaming_clause",
        "f_aspects",
    )

    _kind_name = "ObjectDecl"


class AnonymousObjectDecl(ObjectDecl):
    """"""

    __slots__ = []

    _field_names = ObjectDecl._field_names + ()

    _kind_name = "AnonymousObjectDecl"


class ExtendedReturnStmtObjectDecl(ObjectDecl):
    """
    Object declaration that is part of an extended return statement.
    """

    __slots__ = []

    _field_names = ObjectDecl._field_names + ()

    _kind_name = "ExtendedReturnStmtObjectDecl"


class PackageRenamingDecl(BasicDecl):
    """
    Declaration for a package renaming.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_package_renaming_decl_f_name)

        return result

    @property
    def f_renames(self):
        """
        :rtype: RenamingClause
        """

        result = self._eval_astnode_field(_package_renaming_decl_f_renames)

        return result

    @property
    def p_renamed_package(self):
        """
        Return the declaration of the package that is renamed by Self.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(
            _Entity_c_type(), _package_renaming_decl_p_renamed_package
        )
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_final_renamed_package(self):
        """
        Return the declaration of the package that is ultimately renamed by
        Self, skipping through all intermediate package renamings.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(
            _Entity_c_type(), _package_renaming_decl_p_final_renamed_package
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = BasicDecl._field_names + (
        "f_name",
        "f_renames",
        "f_aspects",
    )

    _kind_name = "PackageRenamingDecl"


class SingleProtectedDecl(BasicDecl):
    """
    Declaration for a single protected object.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: DefiningName
        """

        result = self._eval_astnode_field(_single_protected_decl_f_name)

        return result

    @property
    def f_interfaces(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: ParentList
        """

        result = self._eval_astnode_field(_single_protected_decl_f_interfaces)

        return result

    @property
    def f_definition(self):
        """
        :rtype: ProtectedDef
        """

        result = self._eval_astnode_field(_single_protected_decl_f_definition)

        return result

    _field_names = BasicDecl._field_names + (
        "f_name",
        "f_aspects",
        "f_interfaces",
        "f_definition",
    )

    _kind_name = "SingleProtectedDecl"


class SingleTaskDecl(BasicDecl):
    """
    Declaration for a single task.
    """

    __slots__ = []

    @property
    def f_task_type(self):
        """
        :rtype: SingleTaskTypeDecl
        """

        result = self._eval_astnode_field(_single_task_decl_f_task_type)

        return result

    _field_names = BasicDecl._field_names + ("f_task_type",)

    _kind_name = "SingleTaskDecl"


class CaseStmtAlternative(AdaNode):
    """
    Alternative in a ``case`` statement (``when ... => ...``).
    """

    __slots__ = []

    @property
    def f_choices(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: AlternativesList
        """

        result = self._eval_astnode_field(_case_stmt_alternative_f_choices)

        return result

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_case_stmt_alternative_f_stmts)

        return result

    _field_names = AdaNode._field_names + (
        "f_choices",
        "f_stmts",
    )

    _kind_name = "CaseStmtAlternative"


class CompilationUnit(AdaNode):
    """
    Root node for all Ada analysis units.
    """

    __slots__ = []

    @property
    def f_prelude(self):
        """
        ``with``, ``use`` or ``pragma`` statements.

        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * UseClause

        * WithClause

        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_compilation_unit_f_prelude)

        return result

    @property
    def f_body(self):
        """
        This field can contain one of the following nodes:

        * LibraryItem

        * Subunit

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_compilation_unit_f_body)

        return result

    @property
    def f_pragmas(self):
        """
        :rtype: PragmaNodeList
        """

        result = self._eval_astnode_field(_compilation_unit_f_pragmas)

        return result

    @property
    def p_syntactic_fully_qualified_name(self):
        """
        Return the syntactic fully qualified name of this compilation unit.

        :rtype: List[str]
        """

        c_result = self._eval_field(
            _UnboundedTextTypeArrayConverter.c_type(),
            _compilation_unit_p_syntactic_fully_qualified_name,
        )
        result = _UnboundedTextTypeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_unit_kind(self):
        """
        Return the kind corresponding to this analysis unit.

        :rtype: str
        """

        c_result = self._eval_field(ctypes.c_int(), _compilation_unit_p_unit_kind)
        result = AnalysisUnitKind._wrap(c_result)

        return result

    @property
    def p_withed_units(self):
        """
        Look for all "with" clauses at the top of this compilation unit and
        return all the compilation units designated by them.

        :rtype: List[CompilationUnit]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _compilation_unit_p_withed_units
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_imported_units(self):
        """
        Return all the compilation units that are directly imported by this
        one. This includes "with"ed units as well as the direct parent unit.

        :rtype: List[CompilationUnit]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _compilation_unit_p_imported_units
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_unit_dependencies(self):
        """
        Return the list of all the compilation units that are (direct and
        indirect) dependencies of this one.

        :rtype: List[CompilationUnit]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _compilation_unit_p_unit_dependencies
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_decl(self):
        """
        Get the root basic decl defined in this compilation unit.

        :rtype: BasicDecl
        """

        result = self._eval_astnode_field(_compilation_unit_p_decl)

        return result

    def p_is_preelaborable(self, imprecise_fallback=False):
        """
        Whether this compilation unit is preelaborable or not.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _compilation_unit_p_is_preelaborable,
            unwrapped_imprecise_fallback,
        )
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + (
        "f_prelude",
        "f_body",
        "f_pragmas",
    )

    _kind_name = "CompilationUnit"


class ComponentClause(AdaNode):
    """
    Representation clause for a single component.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_component_clause_f_id)

        return result

    @property
    def f_position(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_component_clause_f_position)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_component_clause_f_range)

        return result

    _field_names = AdaNode._field_names + (
        "f_id",
        "f_position",
        "f_range",
    )

    _kind_name = "ComponentClause"


class ComponentDef(AdaNode):
    """
    Definition for a component.
    """

    __slots__ = []

    @property
    def f_has_aliased(self):
        """
        :rtype: AliasedNode
        """

        result = self._eval_astnode_field(_component_def_f_has_aliased)

        return result

    @property
    def f_has_constant(self):
        """
        :rtype: ConstantNode
        """

        result = self._eval_astnode_field(_component_def_f_has_constant)

        return result

    @property
    def f_type_expr(self):
        """
        This field can contain one of the following nodes:

        * AnonymousType

        * SubtypeIndication

        :rtype: TypeExpr
        """

        result = self._eval_astnode_field(_component_def_f_type_expr)

        return result

    _field_names = AdaNode._field_names + (
        "f_has_aliased",
        "f_has_constant",
        "f_type_expr",
    )

    _kind_name = "ComponentDef"


class ConstantNode(AdaNode):
    """
    Qualifier for the ``constant`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of ConstantPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _constant_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class ConstantAbsent(ConstantNode):
    """"""

    __slots__ = []

    _field_names = ConstantNode._field_names + ()

    _kind_name = "ConstantAbsent"


class ConstantPresent(ConstantNode):
    """"""

    __slots__ = []

    _field_names = ConstantNode._field_names + ()

    _kind_name = "ConstantPresent"


class Constraint(AdaNode):
    """
    Base class for type constraints.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class DeltaConstraint(Constraint):
    """
    Delta and range type constraint.
    """

    __slots__ = []

    @property
    def f_digits(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_delta_constraint_f_digits)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_delta_constraint_f_range)

        return result

    _field_names = Constraint._field_names + (
        "f_digits",
        "f_range",
    )

    _kind_name = "DeltaConstraint"


class DigitsConstraint(Constraint):
    """
    Digits and range type constraint.
    """

    __slots__ = []

    @property
    def f_digits(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_digits_constraint_f_digits)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_digits_constraint_f_range)

        return result

    _field_names = Constraint._field_names + (
        "f_digits",
        "f_range",
    )

    _kind_name = "DigitsConstraint"


class DiscriminantConstraint(Constraint):
    """
    List of constraints that relate to type discriminants.
    """

    __slots__ = []

    @property
    def f_constraints(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * DiscriminantAssoc

        :rtype: AssocList
        """

        result = self._eval_astnode_field(_discriminant_constraint_f_constraints)

        return result

    _field_names = Constraint._field_names + ("f_constraints",)

    _kind_name = "DiscriminantConstraint"


class IndexConstraint(Constraint):
    """
    List of type constraints.
    """

    __slots__ = []

    @property
    def f_constraints(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * SubtypeIndication

        * TargetName

        :rtype: ConstraintList
        """

        result = self._eval_astnode_field(_index_constraint_f_constraints)

        return result

    _field_names = Constraint._field_names + ("f_constraints",)

    _kind_name = "IndexConstraint"


class RangeConstraint(Constraint):
    """
    Range-based type constraint.
    """

    __slots__ = []

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_range_constraint_f_range)

        return result

    _field_names = Constraint._field_names + ("f_range",)

    _kind_name = "RangeConstraint"


class DeclarativePart(AdaNode):
    """
    List of declarations.
    """

    __slots__ = []

    @property
    def f_decls(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AbstractSubpDecl

        * AspectClause

        * BodyNode

        * ComponentDecl

        * EntryDecl

        * ErrorDecl

        * ExceptionDecl

        * GenericDecl

        * GenericInstantiation

        * GenericRenamingDecl

        * IncompleteTypeDecl

        * NumberDecl

        * ObjectDecl

        * PackageDecl

        * PackageRenamingDecl

        * PragmaNode

        * ProtectedTypeDecl

        * SingleProtectedDecl

        * SingleTaskDecl

        * SubpDecl

        * SubtypeDecl

        * TaskTypeDecl

        * TypeDecl

        * UseClause

        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_declarative_part_f_decls)

        return result

    _field_names = AdaNode._field_names + ("f_decls",)

    _kind_name = "DeclarativePart"


class PrivatePart(DeclarativePart):
    """
    List of declarations in a private part.
    """

    __slots__ = []

    _field_names = DeclarativePart._field_names + ()

    _kind_name = "PrivatePart"


class PublicPart(DeclarativePart):
    """
    List of declarations in a public part.
    """

    __slots__ = []

    _field_names = DeclarativePart._field_names + ()

    _kind_name = "PublicPart"


class ElsifExprPart(AdaNode):
    """
    ``elsif`` block, part of an ``if`` expression.
    """

    __slots__ = []

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_elsif_expr_part_f_cond_expr)

        return result

    @property
    def f_then_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_elsif_expr_part_f_then_expr)

        return result

    _field_names = AdaNode._field_names + (
        "f_cond_expr",
        "f_then_expr",
    )

    _kind_name = "ElsifExprPart"


class ElsifStmtPart(AdaNode):
    """
    ``elsif`` part in an ``if`` statement block.
    """

    __slots__ = []

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_elsif_stmt_part_f_cond_expr)

        return result

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_elsif_stmt_part_f_stmts)

        return result

    _field_names = AdaNode._field_names + (
        "f_cond_expr",
        "f_stmts",
    )

    _kind_name = "ElsifStmtPart"


class Expr(AdaNode):
    """
    Base class for expressions.
    """

    __slots__ = []

    @property
    def p_expression_type(self):
        """
        Return the declaration corresponding to the type of this expression
        after name resolution.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _expr_p_expression_type)
        result = AdaNode._wrap(c_result)

        return result

    def p_is_dynamically_tagged(self, imprecise_fallback=False):
        """
        Returns whether this expression is dynamically tagged (See RM 3.9.2).

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _expr_p_is_dynamically_tagged,
            unwrapped_imprecise_fallback,
        )
        result = bool(c_result.value)

        return result

    def p_is_static_expr(self, imprecise_fallback=False):
        """
        Return whether this expression is static according to the ARM
        definition of static. See RM 4.9.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _expr_p_is_static_expr, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    @property
    def p_first_corresponding_decl(self):
        """
        Return the first decl that is lexically named like self in self's
        scope.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _expr_p_first_corresponding_decl)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_eval_as_int(self):
        """
        Statically evaluates self, and returns the value of the evaluation as
        an integer.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :rtype: int
        """

        c_result = self._eval_field(_big_integer.c_type(), _expr_p_eval_as_int)
        result = _big_integer.wrap(c_result)

        return result

    def p_eval_as_int_in_env(self, env):
        """
        Statically evaluates self, and returns the value of the evaluation as
        an integer. The given environment is used to substitute references to
        declarations by actual values.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :param env:
        :type env: List[Substitution]
        :rtype: int
        """

        unwrapped_env = _SubstitutionArrayConverter.unwrap(env)

        c_result = self._eval_field(
            _big_integer.c_type(), _expr_p_eval_as_int_in_env, unwrapped_env.c_value
        )
        result = _big_integer.wrap(c_result)

        return result

    @property
    def p_eval_as_string(self):
        """
        Statically evaluates self, and returns the value of the evaluation as a
        string.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :rtype: str
        """

        c_result = self._eval_field(_TextTypeConverter.c_type(), _expr_p_eval_as_string)
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    def p_eval_as_string_in_env(self, env):
        """
        Statically evaluates self, and returns the value of the evaluation as a
        string. The given environment is used to substitute references to
        declarations by actual values.

        .. note:: In order for a call to this not to raise, the expression
        needs to be a static expression, as specified in the ARM section 4.9.
        You can verify whether an expression is static with the
        ``is_static_expr`` property.

        .. ATTENTION:: This is an experimental feature, so even if it is
        exposed to allow experiments, it is totally unsupported and the API and
        behavior are very likely to change in the future.

        :param env:
        :type env: List[Substitution]
        :rtype: str
        """

        unwrapped_env = _SubstitutionArrayConverter.unwrap(env)

        c_result = self._eval_field(
            _TextTypeConverter.c_type(),
            _expr_p_eval_as_string_in_env,
            unwrapped_env.c_value,
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    @property
    def p_matching_nodes(self):
        """
        Return the list of AST nodes that can be a match for this expression
        before overloading analysis.

        :rtype: List[AdaNode]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _expr_p_matching_nodes
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = AdaNode._field_names + ()


class Allocator(Expr):
    """
    Allocator expression (``new ...``).
    """

    __slots__ = []

    @property
    def f_subpool(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_allocator_f_subpool)

        return result

    @property
    def f_type_or_expr(self):
        """
        This field can contain one of the following nodes:

        * QualExpr

        * SubtypeIndication

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_allocator_f_type_or_expr)

        return result

    @property
    def p_get_allocated_type(self):
        """
        Return the allocated type for this allocator.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _allocator_p_get_allocated_type)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = Expr._field_names + (
        "f_subpool",
        "f_type_or_expr",
    )

    _kind_name = "Allocator"


class BaseAggregate(Expr):
    """
    Base class for aggregates.
    """

    __slots__ = []

    @property
    def f_ancestor_expr(self):
        """
        :rtype: Expr
        """

        result = self._eval_astnode_field(_base_aggregate_f_ancestor_expr)

        return result

    @property
    def f_assocs(self):
        """
        :rtype: AssocList
        """

        result = self._eval_astnode_field(_base_aggregate_f_assocs)

        return result

    _field_names = Expr._field_names + (
        "f_ancestor_expr",
        "f_assocs",
    )


class Aggregate(BaseAggregate):
    """
    Aggregate that is not a ``null record`` aggregate.
    """

    __slots__ = []

    _field_names = BaseAggregate._field_names + ()

    _kind_name = "Aggregate"


class BracketAggregate(Aggregate):
    """
    Bracket array or container aggregate (Ada 2020).
    """

    __slots__ = []

    _field_names = Aggregate._field_names + ()

    _kind_name = "BracketAggregate"


class DeltaAggregate(BaseAggregate):
    """"""

    __slots__ = []

    _field_names = BaseAggregate._field_names + ()

    _kind_name = "DeltaAggregate"


class BracketDeltaAggregate(DeltaAggregate):
    """
    Bracket delta aggregate (Ada 2020).
    """

    __slots__ = []

    _field_names = DeltaAggregate._field_names + ()

    _kind_name = "BracketDeltaAggregate"


class NullRecordAggregate(BaseAggregate):
    """
    Aggregate for ``null record``.
    """

    __slots__ = []

    _field_names = BaseAggregate._field_names + ()

    _kind_name = "NullRecordAggregate"


class BinOp(Expr):
    """
    Binary expression.
    """

    __slots__ = []

    @property
    def f_left(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_bin_op_f_left)

        return result

    @property
    def f_op(self):
        """
        This field can contain one of the following nodes:

        * OpAnd

        * OpAndThen

        * OpConcat

        * OpDiv

        * OpDoubleDot

        * OpEq

        * OpGt

        * OpGte

        * OpLt

        * OpLte

        * OpMinus

        * OpMod

        * OpMult

        * OpNeq

        * OpOr

        * OpOrElse

        * OpPlus

        * OpPow

        * OpRem

        * OpXor

        :rtype: Op
        """

        result = self._eval_astnode_field(_bin_op_f_op)

        return result

    @property
    def f_right(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_bin_op_f_right)

        return result

    _field_names = Expr._field_names + (
        "f_left",
        "f_op",
        "f_right",
    )

    _kind_name = "BinOp"


class RelationOp(BinOp):
    """
    Binary operation that compares two value, producing a boolean.
    """

    __slots__ = []

    _field_names = BinOp._field_names + ()

    _kind_name = "RelationOp"


class BoxExpr(Expr):
    """
    Box expression (``<>``).
    """

    __slots__ = []

    _field_names = Expr._field_names + ()

    _kind_name = "BoxExpr"


class CaseExprAlternative(Expr):
    """
    Alternative in a ``case`` expression (``when ... => ...``).
    """

    __slots__ = []

    @property
    def f_choices(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: AlternativesList
        """

        result = self._eval_astnode_field(_case_expr_alternative_f_choices)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_case_expr_alternative_f_expr)

        return result

    _field_names = Expr._field_names + (
        "f_choices",
        "f_expr",
    )

    _kind_name = "CaseExprAlternative"


class CondExpr(Expr):
    """
    Base class for a conditional expressions (RM 4.5.7).
    """

    __slots__ = []

    @property
    def p_dependent_exprs(self):
        """
        Return the dependent expressions for this conditional expression.

        :rtype: List[Expr]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _cond_expr_p_dependent_exprs
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = Expr._field_names + ()


class CaseExpr(CondExpr):
    """
    ``case`` expression (RM 4.5.7).
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_case_expr_f_expr)

        return result

    @property
    def f_cases(self):
        """
        :rtype: CaseExprAlternativeList
        """

        result = self._eval_astnode_field(_case_expr_f_cases)

        return result

    _field_names = CondExpr._field_names + (
        "f_expr",
        "f_cases",
    )

    _kind_name = "CaseExpr"


class IfExpr(CondExpr):
    """
    ``if`` expression (RM 4.5.7).
    """

    __slots__ = []

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_if_expr_f_cond_expr)

        return result

    @property
    def f_then_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_if_expr_f_then_expr)

        return result

    @property
    def f_alternatives(self):
        """
        :rtype: ElsifExprPartList
        """

        result = self._eval_astnode_field(_if_expr_f_alternatives)

        return result

    @property
    def f_else_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_if_expr_f_else_expr)

        return result

    _field_names = CondExpr._field_names + (
        "f_cond_expr",
        "f_then_expr",
        "f_alternatives",
        "f_else_expr",
    )

    _kind_name = "IfExpr"


class ContractCases(Expr):
    """
    List of associations for the ``Contract_Case`` aspect.
    """

    __slots__ = []

    @property
    def f_contract_cases(self):
        """
        :rtype: ContractCaseAssocList
        """

        result = self._eval_astnode_field(_contract_cases_f_contract_cases)

        return result

    _field_names = Expr._field_names + ("f_contract_cases",)

    _kind_name = "ContractCases"


class DeclExpr(Expr):
    """
    Declare expression (Ada 2020).
    """

    __slots__ = []

    @property
    def f_decls(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * NumberDecl

        * ObjectDecl

        * SingleProtectedDecl

        * SingleTaskDecl

        :rtype: BasicDeclList
        """

        result = self._eval_astnode_field(_decl_expr_f_decls)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_decl_expr_f_expr)

        return result

    _field_names = Expr._field_names + (
        "f_decls",
        "f_expr",
    )

    _kind_name = "DeclExpr"


class MembershipExpr(Expr):
    """
    Represent a membership test (in/not in operators).

    Note that we don't consider them as binary operators since multiple
    expressions on the right hand side are allowed.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_membership_expr_f_expr)

        return result

    @property
    def f_op(self):
        """
        This field can contain one of the following nodes:

        * OpIn

        * OpNotIn

        :rtype: Op
        """

        result = self._eval_astnode_field(_membership_expr_f_op)

        return result

    @property
    def f_membership_exprs(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeName

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: ExprAlternativesList
        """

        result = self._eval_astnode_field(_membership_expr_f_membership_exprs)

        return result

    _field_names = Expr._field_names + (
        "f_expr",
        "f_op",
        "f_membership_exprs",
    )

    _kind_name = "MembershipExpr"


class Name(Expr):
    """
    Base class for names.
    """

    __slots__ = []

    @property
    def p_enclosing_defining_name(self):
        """
        If this name is part of a defining name, return the enclosing defining
        name node.

        :rtype: DefiningName
        """

        c_result = self._eval_field(_Entity_c_type(), _name_p_enclosing_defining_name)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_is_defining(self):
        """
        Return True if this name is part of a defining name.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_is_defining)
        result = bool(c_result.value)

        return result

    def p_name_is(self, sym):
        """
        Helper. Check that this name matches ``sym``.

        :param sym:
        :type sym: str
        :rtype: bool
        """

        _context = self.unit.context._c_value
        unwrapped_sym = _symbol_type.unwrap(sym, _context)

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_name_is, unwrapped_sym)
        result = bool(c_result.value)

        return result

    @property
    def p_is_direct_call(self):
        """
        Return True iff this name represents a call to a subprogram which is
        referred by its defining name. (i.e. not through a subprogram access).

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_is_direct_call)
        result = bool(c_result.value)

        return result

    @property
    def p_is_access_call(self):
        """
        Return True iff this name represents a call to subprogram through an
        access type.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_is_access_call)
        result = bool(c_result.value)

        return result

    @property
    def p_is_call(self):
        """
        Returns True if this Name corresponds to a call.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_is_call)
        result = bool(c_result.value)

        return result

    def p_is_dot_call(self, imprecise_fallback=False):
        """
        Returns True if this Name corresponds to a dot notation call.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _name_p_is_dot_call, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    def p_failsafe_referenced_def_name(self, imprecise_fallback=False):
        """
        Failsafe version of ``referenced_defining_name``. Returns a
        ``RefdDef``, which can be precise, imprecise, or error.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: RefdDef
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            RefdDef._c_type(),
            _name_p_failsafe_referenced_def_name,
            unwrapped_imprecise_fallback,
        )
        result = RefdDef._wrap(c_result)

        return result

    def p_referenced_defining_name(self, imprecise_fallback=False):
        """
        Like ``referenced_decl``, but will return the defining identifier for
        the decl, rather than the basic declaration node itself.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: DefiningName
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _name_p_referenced_defining_name,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_all_env_elements(self, seq=True, seq_from=None):
        """
        Return all elements in self's scope that are lexically named like Self.

        :param seq:
        :type seq: bool
        :param seq_from:
        :type seq_from: AdaNode
        :rtype: List[AdaNode]
        """

        unwrapped_seq = bool(seq)
        unwrapped_seq_from = AdaNode._unwrap(seq_from)

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(),
            _name_p_all_env_elements,
            unwrapped_seq,
            unwrapped_seq_from,
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    @property
    def p_called_subp_spec(self):
        """
        Return the subprogram specification of the subprogram or subprogram
        access that is being called by this exact Name, if relevant.

        :rtype: BaseFormalParamHolder
        """

        c_result = self._eval_field(_Entity_c_type(), _name_p_called_subp_spec)
        result = AdaNode._wrap(c_result)

        return result

    def p_referenced_decl(self, imprecise_fallback=False):
        """
        Return the declaration this node references after name resolution. If
        imprecise_fallback is True, errors raised during resolution of the xref
        equation are catched and a fallback mechanism is triggered, which tries
        to find the referenced declaration in an ad-hoc way.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: BasicDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _name_p_referenced_decl, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_failsafe_referenced_decl(self, imprecise_fallback=False):
        """
        Failsafe version of ``referenced_decl``. Returns a ``RefdDecl``, which
        can be precise, imprecise, or error.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: RefdDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            RefdDecl._c_type(),
            _name_p_failsafe_referenced_decl,
            unwrapped_imprecise_fallback,
        )
        result = RefdDecl._wrap(c_result)

        return result

    def p_referenced_decl_internal(self, imprecise_fallback=False):
        """
        Return the declaration this node references. Try not to run name res if
        already resolved. INTERNAL USE ONLY.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: RefdDecl
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            RefdDecl._c_type(),
            _name_p_referenced_decl_internal,
            unwrapped_imprecise_fallback,
        )
        result = RefdDecl._wrap(c_result)

        return result

    @property
    def p_name_designated_type(self):
        """
        Like SubtypeIndication.designated_type, but on names, since because of
        Ada's ambiguous grammar, some subtype indications will be parsed as
        names.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _name_p_name_designated_type)
        result = AdaNode._wrap(c_result)

        return result

    def p_is_static_subtype(self, imprecise_fallback=False):
        """
        Returns whether Self denotes a static subtype or not.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _name_p_is_static_subtype, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    def p_name_matches(self, n):
        """
        Return whether two names match each other.

        This compares the symbol for Identifier and StringLiteral nodes. We
        consider that there is no match for all other node kinds.

        :param n:
        :type n: Name
        :rtype: bool
        """

        unwrapped_n = AdaNode._unwrap(n)

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_name_matches, unwrapped_n)
        result = bool(c_result.value)

        return result

    @property
    def p_relative_name(self):
        """
        Returns the relative name of this instance. For example, for a prefix
        A.B.C, this will return C.

        :rtype: SingleTokNode
        """

        c_result = self._eval_field(_Entity_c_type(), _name_p_relative_name)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_is_operator_name(self):
        """
        Return whether the name that Self designates is an operator.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _name_p_is_operator_name)
        result = bool(c_result.value)

        return result

    def p_is_write_reference(self, imprecise_fallback=False):
        """
        Whether this name is a write reference.

        For example, `X` is a write reference in the following cases::

        1. `X := 2;` 2. `X (2) := 2;` 3. `P(F => X)` where F is declared `out`
        or `in out`. 6. `X.P` where the formal for X is declared `out` or `in
        out`. 4. `X'Access`. 5. `X.C := 2`, `R.X := 2`

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "write reference".

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _name_p_is_write_reference, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    def p_is_dispatching_call(self, imprecise_fallback=False):
        """
        Returns True if this Name corresponds to a dispatching call, including:

        - calls done through subprogram access types.

        - calls to dispatching subprograms, in the object-oriented sense.

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "dispatching call".

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _name_p_is_dispatching_call, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    def p_is_static_call(self, imprecise_fallback=False):
        """
        Returns True if this Name corresponds to a static non-dispatching call.
        In other words, this will return True if and only if the target of the
        call is known statically.

        .. note:: This is an experimental feature. There might be some
        discrepancy with the GNAT concept of "static call".

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(), _name_p_is_static_call, unwrapped_imprecise_fallback
        )
        result = bool(c_result.value)

        return result

    @property
    def p_as_symbol_array(self):
        """
        Turn this name into an array of symbols.

        For instance, a node with name ``A.B.C`` is turned into ``['A', 'B',
        'C']``.

        :rtype: List[str]
        """

        c_result = self._eval_field(
            _UnboundedTextTypeArrayConverter.c_type(), _name_p_as_symbol_array
        )
        result = _UnboundedTextTypeArrayConverter.wrap(c_result, False)

        return result

    _field_names = Expr._field_names + ()


class AttributeRef(Name):
    """
    Expression to reference an attribute.
    """

    __slots__ = []

    @property
    def f_prefix(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_attribute_ref_f_prefix)

        return result

    @property
    def f_attribute(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_attribute_ref_f_attribute)

        return result

    @property
    def f_args(self):
        """
        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_attribute_ref_f_args)

        return result

    _field_names = Name._field_names + (
        "f_prefix",
        "f_attribute",
        "f_args",
    )

    _kind_name = "AttributeRef"


class UpdateAttributeRef(AttributeRef):
    """
    Reference to the ``Update`` attribute.
    """

    __slots__ = []

    _field_names = AttributeRef._field_names + ()

    _kind_name = "UpdateAttributeRef"


class CallExpr(Name):
    """
    Represent a syntactic call expression.

    At the semantic level, this can be either a subprogram call, an array
    subcomponent access expression, an array slice or a type conversion.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_call_expr_f_name)

        return result

    @property
    def f_suffix(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BasicAssocList

        * BinOp

        * CallExpr

        * CharLiteral

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_call_expr_f_suffix)

        return result

    @property
    def p_is_array_slice(self):
        """
        Return whether this CallExpr is actually an access to a slice of the
        array denoted by the prefix of this CallExpr.

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _call_expr_p_is_array_slice)
        result = bool(c_result.value)

        return result

    _field_names = Name._field_names + (
        "f_name",
        "f_suffix",
    )

    _kind_name = "CallExpr"


class DefiningName(Name):
    """
    Name that defines an entity.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_defining_name_f_name)

        return result

    @property
    def p_basic_decl(self):
        """
        Returns this DefiningName's basic declaration

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _defining_name_p_basic_decl)
        result = AdaNode._wrap(c_result)

        return result

    def p_find_refs(self, root, origin, imprecise_fallback=False):
        """
        Find all references to this defining name in the given ``root`` and its
        children.

        :param root:
        :type root: AdaNode
        :param origin:
        :type origin: AdaNode
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[RefResult]
        """

        unwrapped_root = AdaNode._unwrap(root)
        unwrapped_origin = AdaNode._unwrap(origin)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _RefResultArrayConverter.c_type(),
            _defining_name_p_find_refs,
            unwrapped_root,
            unwrapped_origin,
            unwrapped_imprecise_fallback,
        )
        result = _RefResultArrayConverter.wrap(c_result, False)

        return result

    def p_find_all_references(self, units, imprecise_fallback=False):
        """
        Searches all references to this defining name in the given list of
        units.

        :param units:
        :type units: List[AnalysisUnit]
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[RefResult]
        """

        unwrapped_units = _AnalysisUnitArrayConverter.unwrap(units)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _RefResultArrayConverter.c_type(),
            _defining_name_p_find_all_references,
            unwrapped_units.c_value,
            unwrapped_imprecise_fallback,
        )
        result = _RefResultArrayConverter.wrap(c_result, False)

        return result

    def p_find_all_calls(self, units, imprecise_fallback=False):
        """
        Return the list of all possible calls to the subprogram which Self is
        the defining name of.

        This will return the name corresponding to the call, excluding the
        parameters if there are any. For instance, it will return `A` for the
        `A (B)` call.

        .. note:: This does not yet support calls done inside generics.

        :param units:
        :type units: List[AnalysisUnit]
        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: List[RefResult]
        """

        unwrapped_units = _AnalysisUnitArrayConverter.unwrap(units)
        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _RefResultArrayConverter.c_type(),
            _defining_name_p_find_all_calls,
            unwrapped_units.c_value,
            unwrapped_imprecise_fallback,
        )
        result = _RefResultArrayConverter.wrap(c_result, False)

        return result

    def p_next_part(self, imprecise_fallback=False):
        """
        Like ``BasicDecl.next_part_for_decl`` on a defining name

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: DefiningName
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(), _defining_name_p_next_part, unwrapped_imprecise_fallback
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_previous_part(self, imprecise_fallback=False):
        """
        Like ``BasicDecl.previous_part_for_decl`` on a defining name

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: DefiningName
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _defining_name_p_previous_part,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    def p_canonical_part(self, imprecise_fallback=False):
        """
        Like ``BasicDecl.canonical_part`` on a defining name

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: DefiningName
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            _Entity_c_type(),
            _defining_name_p_canonical_part,
            unwrapped_imprecise_fallback,
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = Name._field_names + ("f_name",)

    _kind_name = "DefiningName"


class DiscreteSubtypeName(Name):
    """
    Subtype name for membership test expressions.
    """

    __slots__ = []

    @property
    def f_subtype(self):
        """
        :rtype: DiscreteSubtypeIndication
        """

        result = self._eval_astnode_field(_discrete_subtype_name_f_subtype)

        return result

    _field_names = Name._field_names + ("f_subtype",)

    _kind_name = "DiscreteSubtypeName"


class DottedName(Name):
    """
    Name to select a suffix in a prefix.
    """

    __slots__ = []

    @property
    def f_prefix(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_dotted_name_f_prefix)

        return result

    @property
    def f_suffix(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * Identifier

        * StringLiteral

        :rtype: BaseId
        """

        result = self._eval_astnode_field(_dotted_name_f_suffix)

        return result

    _field_names = Name._field_names + (
        "f_prefix",
        "f_suffix",
    )

    _kind_name = "DottedName"


class EndName(Name):
    """
    Entity name in ``end ...;`` syntactic constructs.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_end_name_f_name)

        return result

    @property
    def p_basic_decl(self):
        """
        Returns this EndName's basic declaration

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _end_name_p_basic_decl)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = Name._field_names + ("f_name",)

    _kind_name = "EndName"


class ExplicitDeref(Name):
    """
    Explicit dereference expression (``.all``).
    """

    __slots__ = []

    @property
    def f_prefix(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_explicit_deref_f_prefix)

        return result

    _field_names = Name._field_names + ("f_prefix",)

    _kind_name = "ExplicitDeref"


class QualExpr(Name):
    """
    Qualified expression (``...'(...)``).
    """

    __slots__ = []

    @property
    def f_prefix(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_qual_expr_f_prefix)

        return result

    @property
    def f_suffix(self):
        """
        This field can contain one of the following nodes:

        * BaseAggregate

        * ParenExpr

        :rtype: Expr
        """

        result = self._eval_astnode_field(_qual_expr_f_suffix)

        return result

    _field_names = Name._field_names + (
        "f_prefix",
        "f_suffix",
    )

    _kind_name = "QualExpr"


class SingleTokNode(Name):
    """
    Base class for nodes that are made up of a single token.
    """

    __slots__ = []

    @property
    def p_canonical_text(self):
        """
        Return a canonicalized version of this node's text.

        :rtype: str
        """

        c_result = self._eval_field(_symbol_type(), _single_tok_node_p_canonical_text)
        result = _symbol_type.wrap(c_result)

        return result

    _field_names = Name._field_names + ()


class BaseId(SingleTokNode):
    """
    Base class for identifiers.
    """

    __slots__ = []

    _field_names = SingleTokNode._field_names + ()


class CharLiteral(BaseId):
    """
    Character literal.
    """

    __slots__ = []

    @property
    def p_denoted_value(self):
        """
        Return the value that this literal denotes.

        :rtype: str
        """

        c_result = self._eval_field(ctypes.c_uint32(), _char_literal_p_denoted_value)
        result = _py2to3.unicode_character(c_result.value)

        return result

    _field_names = BaseId._field_names + ()

    _kind_name = "CharLiteral"


class Identifier(BaseId):
    """
    Regular identifier.
    """

    __slots__ = []

    _field_names = BaseId._field_names + ()

    _kind_name = "Identifier"


class Op(BaseId):
    """
    Operation in a binary expression.

    Note that the ARM does not consider "double_dot" ("..") as a binary
    operator, but we process it this way here anyway to keep things simple.
    """

    __slots__ = []

    _field_names = BaseId._field_names + ()


class OpAbs(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpAbs"


class OpAnd(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpAnd"


class OpAndThen(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpAndThen"


class OpConcat(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpConcat"


class OpDiv(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpDiv"


class OpDoubleDot(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpDoubleDot"


class OpEq(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpEq"


class OpGt(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpGt"


class OpGte(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpGte"


class OpIn(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpIn"


class OpLt(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpLt"


class OpLte(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpLte"


class OpMinus(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpMinus"


class OpMod(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpMod"


class OpMult(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpMult"


class OpNeq(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpNeq"


class OpNot(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpNot"


class OpNotIn(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpNotIn"


class OpOr(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpOr"


class OpOrElse(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpOrElse"


class OpPlus(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpPlus"


class OpPow(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpPow"


class OpRem(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpRem"


class OpXor(Op):
    """"""

    __slots__ = []

    _field_names = Op._field_names + ()

    _kind_name = "OpXor"


class StringLiteral(BaseId):
    """
    String literal.
    """

    __slots__ = []

    @property
    def p_denoted_value(self):
        """
        Return the value that this literal denotes.

        :rtype: str
        """

        c_result = self._eval_field(
            _TextTypeConverter.c_type(), _string_literal_p_denoted_value
        )
        result = _TextTypeConverter.wrap(c_result, False)

        return result

    _field_names = BaseId._field_names + ()

    _kind_name = "StringLiteral"


class NullLiteral(SingleTokNode):
    """
    The ``null`` literal.
    """

    __slots__ = []

    _field_names = SingleTokNode._field_names + ()

    _kind_name = "NullLiteral"


class NumLiteral(SingleTokNode):
    """
    Base class for number literals.
    """

    __slots__ = []

    _field_names = SingleTokNode._field_names + ()


class IntLiteral(NumLiteral):
    """
    Literal for an integer.
    """

    __slots__ = []

    @property
    def p_denoted_value(self):
        """
        Return the value that this literal denotes.

        :rtype: int
        """

        c_result = self._eval_field(_big_integer.c_type(), _int_literal_p_denoted_value)
        result = _big_integer.wrap(c_result)

        return result

    _field_names = NumLiteral._field_names + ()

    _kind_name = "IntLiteral"


class RealLiteral(NumLiteral):
    """
    Literal for a real number.
    """

    __slots__ = []

    _field_names = NumLiteral._field_names + ()

    _kind_name = "RealLiteral"


class TargetName(Name):
    """
    Name for Ada 2020 ``@``.
    """

    __slots__ = []

    _field_names = Name._field_names + ()

    _kind_name = "TargetName"


class ParenExpr(Expr):
    """
    Parenthesized expression.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_paren_expr_f_expr)

        return result

    _field_names = Expr._field_names + ("f_expr",)

    _kind_name = "ParenExpr"


class QuantifiedExpr(Expr):
    """
    Quantified expression.
    """

    __slots__ = []

    @property
    def f_quantifier(self):
        """
        :rtype: Quantifier
        """

        result = self._eval_astnode_field(_quantified_expr_f_quantifier)

        return result

    @property
    def f_loop_spec(self):
        """
        :rtype: ForLoopSpec
        """

        result = self._eval_astnode_field(_quantified_expr_f_loop_spec)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_quantified_expr_f_expr)

        return result

    _field_names = Expr._field_names + (
        "f_quantifier",
        "f_loop_spec",
        "f_expr",
    )

    _kind_name = "QuantifiedExpr"


class RaiseExpr(Expr):
    """
    Expression to raise an exception.
    """

    __slots__ = []

    @property
    def f_exception_name(self):
        """
        :rtype: Name
        """

        result = self._eval_astnode_field(_raise_expr_f_exception_name)

        return result

    @property
    def f_error_message(self):
        """
        :rtype: Expr
        """

        result = self._eval_astnode_field(_raise_expr_f_error_message)

        return result

    _field_names = Expr._field_names + (
        "f_exception_name",
        "f_error_message",
    )

    _kind_name = "RaiseExpr"


class UnOp(Expr):
    """
    Unary expression.
    """

    __slots__ = []

    @property
    def f_op(self):
        """
        This field can contain one of the following nodes:

        * OpAbs

        * OpMinus

        * OpNot

        * OpPlus

        :rtype: Op
        """

        result = self._eval_astnode_field(_un_op_f_op)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_un_op_f_expr)

        return result

    _field_names = Expr._field_names + (
        "f_op",
        "f_expr",
    )

    _kind_name = "UnOp"


class HandledStmts(AdaNode):
    """
    List of statements, with optional exception handlers.
    """

    __slots__ = []

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_handled_stmts_f_stmts)

        return result

    @property
    def f_exceptions(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * ExceptionHandler

        * PragmaNode

        :rtype: AdaNodeList
        """

        result = self._eval_astnode_field(_handled_stmts_f_exceptions)

        return result

    _field_names = AdaNode._field_names + (
        "f_stmts",
        "f_exceptions",
    )

    _kind_name = "HandledStmts"


class InterfaceKind(AdaNode):
    """
    Kind of interface type.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class InterfaceKindLimited(InterfaceKind):
    """"""

    __slots__ = []

    _field_names = InterfaceKind._field_names + ()

    _kind_name = "InterfaceKindLimited"


class InterfaceKindProtected(InterfaceKind):
    """"""

    __slots__ = []

    _field_names = InterfaceKind._field_names + ()

    _kind_name = "InterfaceKindProtected"


class InterfaceKindSynchronized(InterfaceKind):
    """"""

    __slots__ = []

    _field_names = InterfaceKind._field_names + ()

    _kind_name = "InterfaceKindSynchronized"


class InterfaceKindTask(InterfaceKind):
    """"""

    __slots__ = []

    _field_names = InterfaceKind._field_names + ()

    _kind_name = "InterfaceKindTask"


class IterType(AdaNode):
    """
    Iteration type for ``for`` loops.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class IterTypeIn(IterType):
    """"""

    __slots__ = []

    _field_names = IterType._field_names + ()

    _kind_name = "IterTypeIn"


class IterTypeOf(IterType):
    """"""

    __slots__ = []

    _field_names = IterType._field_names + ()

    _kind_name = "IterTypeOf"


class LibraryItem(AdaNode):
    """
    Library item in a compilation unit.
    """

    __slots__ = []

    @property
    def f_has_private(self):
        """
        :rtype: PrivateNode
        """

        result = self._eval_astnode_field(_library_item_f_has_private)

        return result

    @property
    def f_item(self):
        """
        This field can contain one of the following nodes:

        * AbstractSubpDecl

        * BaseSubpBody

        * ErrorDecl

        * GenericDecl

        * GenericInstantiation

        * GenericRenamingDecl

        * PackageBody

        * PackageDecl

        * PackageRenamingDecl

        * SubpDecl

        :rtype: BasicDecl
        """

        result = self._eval_astnode_field(_library_item_f_item)

        return result

    _field_names = AdaNode._field_names + (
        "f_has_private",
        "f_item",
    )

    _kind_name = "LibraryItem"


class LimitedNode(AdaNode):
    """
    Qualifier for the ``limited`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of LimitedPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _limited_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class LimitedAbsent(LimitedNode):
    """"""

    __slots__ = []

    _field_names = LimitedNode._field_names + ()

    _kind_name = "LimitedAbsent"


class LimitedPresent(LimitedNode):
    """"""

    __slots__ = []

    _field_names = LimitedNode._field_names + ()

    _kind_name = "LimitedPresent"


class LoopSpec(AdaNode):
    """
    Base class for loop specifications.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class ForLoopSpec(LoopSpec):
    """
    Specification for a ``for`` loop.
    """

    __slots__ = []

    @property
    def f_var_decl(self):
        """
        :rtype: ForLoopVarDecl
        """

        result = self._eval_astnode_field(_for_loop_spec_f_var_decl)

        return result

    @property
    def f_loop_type(self):
        """
        :rtype: IterType
        """

        result = self._eval_astnode_field(_for_loop_spec_f_loop_type)

        return result

    @property
    def f_has_reverse(self):
        """
        :rtype: ReverseNode
        """

        result = self._eval_astnode_field(_for_loop_spec_f_has_reverse)

        return result

    @property
    def f_iter_expr(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * CallExpr

        * CharLiteral

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: AdaNode
        """

        result = self._eval_astnode_field(_for_loop_spec_f_iter_expr)

        return result

    _field_names = LoopSpec._field_names + (
        "f_var_decl",
        "f_loop_type",
        "f_has_reverse",
        "f_iter_expr",
    )

    _kind_name = "ForLoopSpec"


class WhileLoopSpec(LoopSpec):
    """
    Specification for a ``while`` loop.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_while_loop_spec_f_expr)

        return result

    _field_names = LoopSpec._field_names + ("f_expr",)

    _kind_name = "WhileLoopSpec"


class Mode(AdaNode):
    """
    Syntactic indicators for passing modes in formals.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class ModeDefault(Mode):
    """"""

    __slots__ = []

    _field_names = Mode._field_names + ()

    _kind_name = "ModeDefault"


class ModeIn(Mode):
    """"""

    __slots__ = []

    _field_names = Mode._field_names + ()

    _kind_name = "ModeIn"


class ModeInOut(Mode):
    """"""

    __slots__ = []

    _field_names = Mode._field_names + ()

    _kind_name = "ModeInOut"


class ModeOut(Mode):
    """"""

    __slots__ = []

    _field_names = Mode._field_names + ()

    _kind_name = "ModeOut"


class NotNull(AdaNode):
    """
    Qualifier for the ``not null`` keywords.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of NotNullPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _not_null_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class NotNullAbsent(NotNull):
    """"""

    __slots__ = []

    _field_names = NotNull._field_names + ()

    _kind_name = "NotNullAbsent"


class NotNullPresent(NotNull):
    """"""

    __slots__ = []

    _field_names = NotNull._field_names + ()

    _kind_name = "NotNullPresent"


class NullComponentDecl(AdaNode):
    """
    Placeholder for the ``null`` in lists of components.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()

    _kind_name = "NullComponentDecl"


class OthersDesignator(AdaNode):
    """
    ``other`` designator.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()

    _kind_name = "OthersDesignator"


class OverridingNode(AdaNode):
    """
    Syntactic indicators for subprogram overriding modes.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class OverridingNotOverriding(OverridingNode):
    """"""

    __slots__ = []

    _field_names = OverridingNode._field_names + ()

    _kind_name = "OverridingNotOverriding"


class OverridingOverriding(OverridingNode):
    """"""

    __slots__ = []

    _field_names = OverridingNode._field_names + ()

    _kind_name = "OverridingOverriding"


class OverridingUnspecified(OverridingNode):
    """"""

    __slots__ = []

    _field_names = OverridingNode._field_names + ()

    _kind_name = "OverridingUnspecified"


class Params(AdaNode):
    """
    List of parameter specifications.
    """

    __slots__ = []

    @property
    def f_params(self):
        """
        :rtype: ParamSpecList
        """

        result = self._eval_astnode_field(_params_f_params)

        return result

    _field_names = AdaNode._field_names + ("f_params",)

    _kind_name = "Params"


class PragmaNode(AdaNode):
    """
    Class for pragmas (RM 2.8). Pragmas are compiler directives, that can be
    language or compiler defined.
    """

    __slots__ = []

    @property
    def f_id(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_pragma_node_f_id)

        return result

    @property
    def f_args(self):
        """
        :rtype: BaseAssocList
        """

        result = self._eval_astnode_field(_pragma_node_f_args)

        return result

    @property
    def p_associated_decls(self):
        """
        Return an array of ``BasicDecl`` instances associated with this pragma,
        or an empty array if non applicable.

        :rtype: List[BasicDecl]
        """

        c_result = self._eval_field(
            _AdaNodeArrayConverter.c_type(), _pragma_node_p_associated_decls
        )
        result = _AdaNodeArrayConverter.wrap(c_result, False)

        return result

    _field_names = AdaNode._field_names + (
        "f_id",
        "f_args",
    )

    _kind_name = "PragmaNode"


class PrimTypeAccessor(AdaNode):
    """
    Synthetic node wrapping around a primitive type entity. Used in metadata.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()

    _kind_name = "PrimTypeAccessor"


class PrivateNode(AdaNode):
    """
    Qualifier for the ``private`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of PrivatePresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _private_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class PrivateAbsent(PrivateNode):
    """"""

    __slots__ = []

    _field_names = PrivateNode._field_names + ()

    _kind_name = "PrivateAbsent"


class PrivatePresent(PrivateNode):
    """"""

    __slots__ = []

    _field_names = PrivateNode._field_names + ()

    _kind_name = "PrivatePresent"


class ProtectedDef(AdaNode):
    """
    Type definition for a protected object.
    """

    __slots__ = []

    @property
    def f_public_part(self):
        """
        :rtype: PublicPart
        """

        result = self._eval_astnode_field(_protected_def_f_public_part)

        return result

    @property
    def f_private_part(self):
        """
        :rtype: PrivatePart
        """

        result = self._eval_astnode_field(_protected_def_f_private_part)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_protected_def_f_end_name)

        return result

    _field_names = AdaNode._field_names + (
        "f_public_part",
        "f_private_part",
        "f_end_name",
    )

    _kind_name = "ProtectedDef"


class ProtectedNode(AdaNode):
    """
    Qualifier for the ``protected`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of ProtectedPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _protected_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class ProtectedAbsent(ProtectedNode):
    """"""

    __slots__ = []

    _field_names = ProtectedNode._field_names + ()

    _kind_name = "ProtectedAbsent"


class ProtectedPresent(ProtectedNode):
    """"""

    __slots__ = []

    _field_names = ProtectedNode._field_names + ()

    _kind_name = "ProtectedPresent"


class Quantifier(AdaNode):
    """
    Type for quantified expressions.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class QuantifierAll(Quantifier):
    """"""

    __slots__ = []

    _field_names = Quantifier._field_names + ()

    _kind_name = "QuantifierAll"


class QuantifierSome(Quantifier):
    """"""

    __slots__ = []

    _field_names = Quantifier._field_names + ()

    _kind_name = "QuantifierSome"


class RangeSpec(AdaNode):
    """
    Range specification.
    """

    __slots__ = []

    @property
    def f_range(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Expr
        """

        result = self._eval_astnode_field(_range_spec_f_range)

        return result

    _field_names = AdaNode._field_names + ("f_range",)

    _kind_name = "RangeSpec"


class RenamingClause(AdaNode):
    """
    Renaming clause, used everywhere renamings are valid.
    """

    __slots__ = []

    @property
    def f_renamed_object(self):
        """
        :rtype: Name
        """

        result = self._eval_astnode_field(_renaming_clause_f_renamed_object)

        return result

    _field_names = AdaNode._field_names + ("f_renamed_object",)

    _kind_name = "RenamingClause"


class SyntheticRenamingClause(RenamingClause):
    """
    Synthetic renaming clause. Used to synthesize object decls with renamings.
    (See to_anonymous_object_decl).
    """

    __slots__ = []

    _field_names = RenamingClause._field_names + ()

    _kind_name = "SyntheticRenamingClause"


class ReverseNode(AdaNode):
    """
    Qualifier for the ``reverse`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of ReversePresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _reverse_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class ReverseAbsent(ReverseNode):
    """"""

    __slots__ = []

    _field_names = ReverseNode._field_names + ()

    _kind_name = "ReverseAbsent"


class ReversePresent(ReverseNode):
    """"""

    __slots__ = []

    _field_names = ReverseNode._field_names + ()

    _kind_name = "ReversePresent"


class SelectWhenPart(AdaNode):
    """
    Alternative part in a ``select`` statements block.
    """

    __slots__ = []

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_select_when_part_f_cond_expr)

        return result

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_select_when_part_f_stmts)

        return result

    _field_names = AdaNode._field_names + (
        "f_cond_expr",
        "f_stmts",
    )

    _kind_name = "SelectWhenPart"


class Stmt(AdaNode):
    """
    Bass class for statements.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class CompositeStmt(Stmt):
    """
    Base class for composite statements.
    """

    __slots__ = []

    _field_names = Stmt._field_names + ()


class AcceptStmt(CompositeStmt):
    """
    ``accept`` statement.
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_accept_stmt_f_name)

        return result

    @property
    def f_entry_index_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_accept_stmt_f_entry_index_expr)

        return result

    @property
    def f_params(self):
        """
        :rtype: EntryCompletionFormalParams
        """

        result = self._eval_astnode_field(_accept_stmt_f_params)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_name",
        "f_entry_index_expr",
        "f_params",
    )

    _kind_name = "AcceptStmt"


class AcceptStmtWithStmts(AcceptStmt):
    """
    Extended ``accept`` statement.
    """

    __slots__ = []

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_accept_stmt_with_stmts_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_accept_stmt_with_stmts_f_end_name)

        return result

    _field_names = AcceptStmt._field_names + (
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "AcceptStmtWithStmts"


class BaseLoopStmt(CompositeStmt):
    """
    Base class for loop statements.
    """

    __slots__ = []

    @property
    def f_spec(self):
        """
        :rtype: LoopSpec
        """

        result = self._eval_astnode_field(_base_loop_stmt_f_spec)

        return result

    @property
    def f_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_base_loop_stmt_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_base_loop_stmt_f_end_name)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_spec",
        "f_stmts",
        "f_end_name",
    )


class ForLoopStmt(BaseLoopStmt):
    """
    Statement for ``for`` loops (``for ... loop ... end loop;``).
    """

    __slots__ = []

    _field_names = BaseLoopStmt._field_names + ()

    _kind_name = "ForLoopStmt"


class LoopStmt(BaseLoopStmt):
    """
    Statement for simple loops (``loop ... end loop;``).
    """

    __slots__ = []

    _field_names = BaseLoopStmt._field_names + ()

    _kind_name = "LoopStmt"


class WhileLoopStmt(BaseLoopStmt):
    """
    Statement for ``while`` loops (``while ... loop ... end loop;``).
    """

    __slots__ = []

    _field_names = BaseLoopStmt._field_names + ()

    _kind_name = "WhileLoopStmt"


class BlockStmt(CompositeStmt):
    """
    Base class for statement blocks.
    """

    __slots__ = []

    _field_names = CompositeStmt._field_names + ()


class BeginBlock(BlockStmt):
    """
    Statement block with no declarative part.
    """

    __slots__ = []

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_begin_block_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_begin_block_f_end_name)

        return result

    _field_names = BlockStmt._field_names + (
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "BeginBlock"


class DeclBlock(BlockStmt):
    """
    Statement block with a declarative part.
    """

    __slots__ = []

    @property
    def f_decls(self):
        """
        :rtype: DeclarativePart
        """

        result = self._eval_astnode_field(_decl_block_f_decls)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_decl_block_f_stmts)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_decl_block_f_end_name)

        return result

    _field_names = BlockStmt._field_names + (
        "f_decls",
        "f_stmts",
        "f_end_name",
    )

    _kind_name = "DeclBlock"


class CaseStmt(CompositeStmt):
    """
    ``case`` statement.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_case_stmt_f_expr)

        return result

    @property
    def f_alternatives(self):
        """
        :rtype: CaseStmtAlternativeList
        """

        result = self._eval_astnode_field(_case_stmt_f_alternatives)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_expr",
        "f_alternatives",
    )

    _kind_name = "CaseStmt"


class ExtendedReturnStmt(CompositeStmt):
    """
    Extended ``return`` statement.
    """

    __slots__ = []

    @property
    def f_decl(self):
        """
        :rtype: ExtendedReturnStmtObjectDecl
        """

        result = self._eval_astnode_field(_extended_return_stmt_f_decl)

        return result

    @property
    def f_stmts(self):
        """
        :rtype: HandledStmts
        """

        result = self._eval_astnode_field(_extended_return_stmt_f_stmts)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_decl",
        "f_stmts",
    )

    _kind_name = "ExtendedReturnStmt"


class IfStmt(CompositeStmt):
    """
    ``if`` statement block.
    """

    __slots__ = []

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_if_stmt_f_cond_expr)

        return result

    @property
    def f_then_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_if_stmt_f_then_stmts)

        return result

    @property
    def f_alternatives(self):
        """
        :rtype: ElsifStmtPartList
        """

        result = self._eval_astnode_field(_if_stmt_f_alternatives)

        return result

    @property
    def f_else_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_if_stmt_f_else_stmts)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_cond_expr",
        "f_then_stmts",
        "f_alternatives",
        "f_else_stmts",
    )

    _kind_name = "IfStmt"


class NamedStmt(CompositeStmt):
    """
    Wrapper class, used for composite statements that can be named (declare
    blocks, loops). This allows to both have a BasicDecl for the named entity
    declared, and a CompositeStmt for the statement hierarchy.
    """

    __slots__ = []

    @property
    def f_decl(self):
        """
        :rtype: NamedStmtDecl
        """

        result = self._eval_astnode_field(_named_stmt_f_decl)

        return result

    @property
    def f_stmt(self):
        """
        This field can contain one of the following nodes:

        * BaseLoopStmt

        * BlockStmt

        :rtype: CompositeStmt
        """

        result = self._eval_astnode_field(_named_stmt_f_stmt)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_decl",
        "f_stmt",
    )

    _kind_name = "NamedStmt"


class SelectStmt(CompositeStmt):
    """
    ``select`` statements block.
    """

    __slots__ = []

    @property
    def f_guards(self):
        """
        :rtype: SelectWhenPartList
        """

        result = self._eval_astnode_field(_select_stmt_f_guards)

        return result

    @property
    def f_else_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_select_stmt_f_else_stmts)

        return result

    @property
    def f_abort_stmts(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * PragmaNode

        * Stmt

        :rtype: StmtList
        """

        result = self._eval_astnode_field(_select_stmt_f_abort_stmts)

        return result

    _field_names = CompositeStmt._field_names + (
        "f_guards",
        "f_else_stmts",
        "f_abort_stmts",
    )

    _kind_name = "SelectStmt"


class ErrorStmt(Stmt):
    """
    Placeholder node for syntax errors in lists of statements.
    """

    __slots__ = []

    _field_names = Stmt._field_names + ()

    _kind_name = "ErrorStmt"


class SimpleStmt(Stmt):
    """
    Base class for simple statements.
    """

    __slots__ = []

    _field_names = Stmt._field_names + ()


class AbortStmt(SimpleStmt):
    """
    ``abort`` statement.
    """

    __slots__ = []

    @property
    def f_names(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: NameList
        """

        result = self._eval_astnode_field(_abort_stmt_f_names)

        return result

    _field_names = SimpleStmt._field_names + ("f_names",)

    _kind_name = "AbortStmt"


class AssignStmt(SimpleStmt):
    """
    Statement for assignments.
    """

    __slots__ = []

    @property
    def f_dest(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_assign_stmt_f_dest)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_assign_stmt_f_expr)

        return result

    _field_names = SimpleStmt._field_names + (
        "f_dest",
        "f_expr",
    )

    _kind_name = "AssignStmt"


class CallStmt(SimpleStmt):
    """
    Statement for entry or procedure calls.
    """

    __slots__ = []

    @property
    def f_call(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_call_stmt_f_call)

        return result

    _field_names = SimpleStmt._field_names + ("f_call",)

    _kind_name = "CallStmt"


class DelayStmt(SimpleStmt):
    """
    ``delay`` statement.
    """

    __slots__ = []

    @property
    def f_has_until(self):
        """
        :rtype: UntilNode
        """

        result = self._eval_astnode_field(_delay_stmt_f_has_until)

        return result

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_delay_stmt_f_expr)

        return result

    _field_names = SimpleStmt._field_names + (
        "f_has_until",
        "f_expr",
    )

    _kind_name = "DelayStmt"


class ExitStmt(SimpleStmt):
    """
    ``exit`` statement.
    """

    __slots__ = []

    @property
    def f_loop_name(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_exit_stmt_f_loop_name)

        return result

    @property
    def f_cond_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_exit_stmt_f_cond_expr)

        return result

    _field_names = SimpleStmt._field_names + (
        "f_loop_name",
        "f_cond_expr",
    )

    _kind_name = "ExitStmt"


class GotoStmt(SimpleStmt):
    """
    ``goto`` statement.
    """

    __slots__ = []

    @property
    def f_label_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_goto_stmt_f_label_name)

        return result

    _field_names = SimpleStmt._field_names + ("f_label_name",)

    _kind_name = "GotoStmt"


class Label(SimpleStmt):
    """
    Statement to declare a code label.
    """

    __slots__ = []

    @property
    def f_decl(self):
        """
        :rtype: LabelDecl
        """

        result = self._eval_astnode_field(_label_f_decl)

        return result

    _field_names = SimpleStmt._field_names + ("f_decl",)

    _kind_name = "Label"


class NullStmt(SimpleStmt):
    """
    ``null;`` statement.
    """

    __slots__ = []

    _field_names = SimpleStmt._field_names + ()

    _kind_name = "NullStmt"


class RaiseStmt(SimpleStmt):
    """
    ``raise`` statement.
    """

    __slots__ = []

    @property
    def f_exception_name(self):
        """
        :rtype: Name
        """

        result = self._eval_astnode_field(_raise_stmt_f_exception_name)

        return result

    @property
    def f_error_message(self):
        """
        :rtype: Expr
        """

        result = self._eval_astnode_field(_raise_stmt_f_error_message)

        return result

    _field_names = SimpleStmt._field_names + (
        "f_exception_name",
        "f_error_message",
    )

    _kind_name = "RaiseStmt"


class RequeueStmt(SimpleStmt):
    """
    ``requeue`` statement.
    """

    __slots__ = []

    @property
    def f_call_name(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: Name
        """

        result = self._eval_astnode_field(_requeue_stmt_f_call_name)

        return result

    @property
    def f_has_abort(self):
        """
        :rtype: AbortNode
        """

        result = self._eval_astnode_field(_requeue_stmt_f_has_abort)

        return result

    _field_names = SimpleStmt._field_names + (
        "f_call_name",
        "f_has_abort",
    )

    _kind_name = "RequeueStmt"


class ReturnStmt(SimpleStmt):
    """
    ``return`` statement.
    """

    __slots__ = []

    @property
    def f_return_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_return_stmt_f_return_expr)

        return result

    _field_names = SimpleStmt._field_names + ("f_return_expr",)

    _kind_name = "ReturnStmt"


class TerminateAlternative(SimpleStmt):
    """
    ``terminate`` alternative in a ``select`` statement.
    """

    __slots__ = []

    _field_names = SimpleStmt._field_names + ()

    _kind_name = "TerminateAlternative"


class SubpKind(AdaNode):
    """
    Qualifier for a subprogram kind.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class SubpKindFunction(SubpKind):
    """"""

    __slots__ = []

    _field_names = SubpKind._field_names + ()

    _kind_name = "SubpKindFunction"


class SubpKindProcedure(SubpKind):
    """"""

    __slots__ = []

    _field_names = SubpKind._field_names + ()

    _kind_name = "SubpKindProcedure"


class Subunit(AdaNode):
    """
    Subunit (``separate``).
    """

    __slots__ = []

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_subunit_f_name)

        return result

    @property
    def f_body(self):
        """
        This field can contain one of the following nodes:

        * PackageBody

        * ProtectedBody

        * SubpBody

        * TaskBody

        :rtype: BodyNode
        """

        result = self._eval_astnode_field(_subunit_f_body)

        return result

    @property
    def p_body_root(self):
        """
        Return the body in which this subunit is rooted.

        :rtype: BasicDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _subunit_p_body_root)
        result = AdaNode._wrap(c_result)

        return result

    _field_names = AdaNode._field_names + (
        "f_name",
        "f_body",
    )

    _kind_name = "Subunit"


class SynchronizedNode(AdaNode):
    """
    Qualifier for the ``synchronized`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of SynchronizedPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _synchronized_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class SynchronizedAbsent(SynchronizedNode):
    """"""

    __slots__ = []

    _field_names = SynchronizedNode._field_names + ()

    _kind_name = "SynchronizedAbsent"


class SynchronizedPresent(SynchronizedNode):
    """"""

    __slots__ = []

    _field_names = SynchronizedNode._field_names + ()

    _kind_name = "SynchronizedPresent"


class TaggedNode(AdaNode):
    """
    Qualifier for the ``tagged`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of TaggedPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _tagged_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class TaggedAbsent(TaggedNode):
    """"""

    __slots__ = []

    _field_names = TaggedNode._field_names + ()

    _kind_name = "TaggedAbsent"


class TaggedPresent(TaggedNode):
    """"""

    __slots__ = []

    _field_names = TaggedNode._field_names + ()

    _kind_name = "TaggedPresent"


class TaskDef(AdaNode):
    """
    Type definition for a task type.
    """

    __slots__ = []

    @property
    def f_interfaces(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: ParentList
        """

        result = self._eval_astnode_field(_task_def_f_interfaces)

        return result

    @property
    def f_public_part(self):
        """
        :rtype: PublicPart
        """

        result = self._eval_astnode_field(_task_def_f_public_part)

        return result

    @property
    def f_private_part(self):
        """
        :rtype: PrivatePart
        """

        result = self._eval_astnode_field(_task_def_f_private_part)

        return result

    @property
    def f_end_name(self):
        """
        :rtype: EndName
        """

        result = self._eval_astnode_field(_task_def_f_end_name)

        return result

    _field_names = AdaNode._field_names + (
        "f_interfaces",
        "f_public_part",
        "f_private_part",
        "f_end_name",
    )

    _kind_name = "TaskDef"


class TypeDef(AdaNode):
    """
    Base class for type definitions.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class AccessDef(TypeDef):
    """
    Base class for access type definitions.
    """

    __slots__ = []

    @property
    def f_has_not_null(self):
        """
        :rtype: NotNull
        """

        result = self._eval_astnode_field(_access_def_f_has_not_null)

        return result

    _field_names = TypeDef._field_names + ("f_has_not_null",)


class AccessToSubpDef(AccessDef):
    """
    Type definition for accesses to subprograms.
    """

    __slots__ = []

    @property
    def f_has_protected(self):
        """
        :rtype: ProtectedNode
        """

        result = self._eval_astnode_field(_access_to_subp_def_f_has_protected)

        return result

    @property
    def f_subp_spec(self):
        """
        :rtype: SubpSpec
        """

        result = self._eval_astnode_field(_access_to_subp_def_f_subp_spec)

        return result

    _field_names = AccessDef._field_names + (
        "f_has_protected",
        "f_subp_spec",
    )

    _kind_name = "AccessToSubpDef"


class BaseTypeAccessDef(AccessDef):
    """
    Base class for access type definitions.
    """

    __slots__ = []

    _field_names = AccessDef._field_names + ()


class AnonymousTypeAccessDef(BaseTypeAccessDef):
    """
    Synthetic type access, that will directly reference a type decl. It is used
    to generate synthetic anonymous access types.
    """

    __slots__ = []

    @property
    def f_type_decl(self):
        """
        :rtype: BaseTypeDecl
        """

        result = self._eval_astnode_field(_anonymous_type_access_def_f_type_decl)

        return result

    _field_names = BaseTypeAccessDef._field_names + ("f_type_decl",)

    _kind_name = "AnonymousTypeAccessDef"


class TypeAccessDef(BaseTypeAccessDef):
    """
    Syntactic type definition for accesses.
    """

    __slots__ = []

    @property
    def f_has_all(self):
        """
        :rtype: AllNode
        """

        result = self._eval_astnode_field(_type_access_def_f_has_all)

        return result

    @property
    def f_has_constant(self):
        """
        :rtype: ConstantNode
        """

        result = self._eval_astnode_field(_type_access_def_f_has_constant)

        return result

    @property
    def f_subtype_indication(self):
        """
        :rtype: SubtypeIndication
        """

        result = self._eval_astnode_field(_type_access_def_f_subtype_indication)

        return result

    _field_names = BaseTypeAccessDef._field_names + (
        "f_has_all",
        "f_has_constant",
        "f_subtype_indication",
    )

    _kind_name = "TypeAccessDef"


class ArrayTypeDef(TypeDef):
    """
    Type definition for an array.
    """

    __slots__ = []

    @property
    def f_indices(self):
        """
        :rtype: ArrayIndices
        """

        result = self._eval_astnode_field(_array_type_def_f_indices)

        return result

    @property
    def f_component_type(self):
        """
        :rtype: ComponentDef
        """

        result = self._eval_astnode_field(_array_type_def_f_component_type)

        return result

    _field_names = TypeDef._field_names + (
        "f_indices",
        "f_component_type",
    )

    _kind_name = "ArrayTypeDef"


class DerivedTypeDef(TypeDef):
    """
    Type definition for a derived type.
    """

    __slots__ = []

    @property
    def f_has_abstract(self):
        """
        :rtype: AbstractNode
        """

        result = self._eval_astnode_field(_derived_type_def_f_has_abstract)

        return result

    @property
    def f_has_limited(self):
        """
        :rtype: LimitedNode
        """

        result = self._eval_astnode_field(_derived_type_def_f_has_limited)

        return result

    @property
    def f_has_synchronized(self):
        """
        :rtype: SynchronizedNode
        """

        result = self._eval_astnode_field(_derived_type_def_f_has_synchronized)

        return result

    @property
    def f_subtype_indication(self):
        """
        :rtype: SubtypeIndication
        """

        result = self._eval_astnode_field(_derived_type_def_f_subtype_indication)

        return result

    @property
    def f_interfaces(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: ParentList
        """

        result = self._eval_astnode_field(_derived_type_def_f_interfaces)

        return result

    @property
    def f_record_extension(self):
        """
        :rtype: BaseRecordDef
        """

        result = self._eval_astnode_field(_derived_type_def_f_record_extension)

        return result

    @property
    def f_has_with_private(self):
        """
        :rtype: WithPrivate
        """

        result = self._eval_astnode_field(_derived_type_def_f_has_with_private)

        return result

    _field_names = TypeDef._field_names + (
        "f_has_abstract",
        "f_has_limited",
        "f_has_synchronized",
        "f_subtype_indication",
        "f_interfaces",
        "f_record_extension",
        "f_has_with_private",
    )

    _kind_name = "DerivedTypeDef"


class EnumTypeDef(TypeDef):
    """
    Type definition for enumerations.
    """

    __slots__ = []

    @property
    def f_enum_literals(self):
        """
        :rtype: EnumLiteralDeclList
        """

        result = self._eval_astnode_field(_enum_type_def_f_enum_literals)

        return result

    _field_names = TypeDef._field_names + ("f_enum_literals",)

    _kind_name = "EnumTypeDef"


class FormalDiscreteTypeDef(TypeDef):
    """
    Type definition for discrete types in generic formals.
    """

    __slots__ = []

    _field_names = TypeDef._field_names + ()

    _kind_name = "FormalDiscreteTypeDef"


class InterfaceTypeDef(TypeDef):
    """
    Type definition for an interface.
    """

    __slots__ = []

    @property
    def f_interface_kind(self):
        """
        :rtype: InterfaceKind
        """

        result = self._eval_astnode_field(_interface_type_def_f_interface_kind)

        return result

    @property
    def f_interfaces(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: ParentList
        """

        result = self._eval_astnode_field(_interface_type_def_f_interfaces)

        return result

    _field_names = TypeDef._field_names + (
        "f_interface_kind",
        "f_interfaces",
    )

    _kind_name = "InterfaceTypeDef"


class ModIntTypeDef(TypeDef):
    """
    Type definition for a modular integer type.
    """

    __slots__ = []

    @property
    def f_expr(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_mod_int_type_def_f_expr)

        return result

    _field_names = TypeDef._field_names + ("f_expr",)

    _kind_name = "ModIntTypeDef"


class PrivateTypeDef(TypeDef):
    """
    Type definition for a private type.
    """

    __slots__ = []

    @property
    def f_has_abstract(self):
        """
        :rtype: AbstractNode
        """

        result = self._eval_astnode_field(_private_type_def_f_has_abstract)

        return result

    @property
    def f_has_tagged(self):
        """
        :rtype: TaggedNode
        """

        result = self._eval_astnode_field(_private_type_def_f_has_tagged)

        return result

    @property
    def f_has_limited(self):
        """
        :rtype: LimitedNode
        """

        result = self._eval_astnode_field(_private_type_def_f_has_limited)

        return result

    _field_names = TypeDef._field_names + (
        "f_has_abstract",
        "f_has_tagged",
        "f_has_limited",
    )

    _kind_name = "PrivateTypeDef"


class RealTypeDef(TypeDef):
    """
    Type definition for real numbers.
    """

    __slots__ = []

    _field_names = TypeDef._field_names + ()


class DecimalFixedPointDef(RealTypeDef):
    """
    Type definition for decimal fixed-point numbers.
    """

    __slots__ = []

    @property
    def f_delta(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_decimal_fixed_point_def_f_delta)

        return result

    @property
    def f_digits(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_decimal_fixed_point_def_f_digits)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_decimal_fixed_point_def_f_range)

        return result

    _field_names = RealTypeDef._field_names + (
        "f_delta",
        "f_digits",
        "f_range",
    )

    _kind_name = "DecimalFixedPointDef"


class FloatingPointDef(RealTypeDef):
    """
    Type definition for floating-point numbers.
    """

    __slots__ = []

    @property
    def f_num_digits(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_floating_point_def_f_num_digits)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_floating_point_def_f_range)

        return result

    _field_names = RealTypeDef._field_names + (
        "f_num_digits",
        "f_range",
    )

    _kind_name = "FloatingPointDef"


class OrdinaryFixedPointDef(RealTypeDef):
    """
    Type definition for ordinary fixed-point numbers.
    """

    __slots__ = []

    @property
    def f_delta(self):
        """
        This field can contain one of the following nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * BoxExpr

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DottedName

        * ExplicitDeref

        * Identifier

        * NullLiteral

        * NumLiteral

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: Expr
        """

        result = self._eval_astnode_field(_ordinary_fixed_point_def_f_delta)

        return result

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_ordinary_fixed_point_def_f_range)

        return result

    _field_names = RealTypeDef._field_names + (
        "f_delta",
        "f_range",
    )

    _kind_name = "OrdinaryFixedPointDef"


class RecordTypeDef(TypeDef):
    """
    Type definition for a record.
    """

    __slots__ = []

    @property
    def f_has_abstract(self):
        """
        :rtype: AbstractNode
        """

        result = self._eval_astnode_field(_record_type_def_f_has_abstract)

        return result

    @property
    def f_has_tagged(self):
        """
        :rtype: TaggedNode
        """

        result = self._eval_astnode_field(_record_type_def_f_has_tagged)

        return result

    @property
    def f_has_limited(self):
        """
        :rtype: LimitedNode
        """

        result = self._eval_astnode_field(_record_type_def_f_has_limited)

        return result

    @property
    def f_record_def(self):
        """
        :rtype: BaseRecordDef
        """

        result = self._eval_astnode_field(_record_type_def_f_record_def)

        return result

    _field_names = TypeDef._field_names + (
        "f_has_abstract",
        "f_has_tagged",
        "f_has_limited",
        "f_record_def",
    )

    _kind_name = "RecordTypeDef"


class SignedIntTypeDef(TypeDef):
    """
    Type definition for a signed integer type.
    """

    __slots__ = []

    @property
    def f_range(self):
        """
        :rtype: RangeSpec
        """

        result = self._eval_astnode_field(_signed_int_type_def_f_range)

        return result

    _field_names = TypeDef._field_names + ("f_range",)

    _kind_name = "SignedIntTypeDef"


class TypeExpr(AdaNode):
    """
    A type expression is an abstract node that embodies the concept of a
    reference to a type.

    Since Ada has both subtype_indications and anonymous (inline) type
    declarations, a type expression contains one or the other.
    """

    __slots__ = []

    @property
    def p_type_name(self):
        """
        Return the name node for this type expression, if applicable, else null

        :rtype: Name
        """

        c_result = self._eval_field(_Entity_c_type(), _type_expr_p_type_name)
        result = AdaNode._wrap(c_result)

        return result

    @property
    def p_designated_type_decl(self):
        """
        Returns the type declaration designated by this type expression.

        :rtype: BaseTypeDecl
        """

        c_result = self._eval_field(_Entity_c_type(), _type_expr_p_designated_type_decl)
        result = AdaNode._wrap(c_result)

        return result

    def p_designated_type_decl_from(self, origin_node):
        """
        Return the type declaration designated by this type expression as
        viewed from the node given by origin_node.

        :param origin_node:
        :type origin_node: AdaNode
        :rtype: BaseTypeDecl
        """

        unwrapped_origin_node = AdaNode._unwrap(origin_node)

        c_result = self._eval_field(
            _Entity_c_type(),
            _type_expr_p_designated_type_decl_from,
            unwrapped_origin_node,
        )
        result = AdaNode._wrap(c_result)

        return result

    _field_names = AdaNode._field_names + ()


class AnonymousType(TypeExpr):
    """
    Container for inline anonymous array and access types declarations.
    """

    __slots__ = []

    @property
    def f_type_decl(self):
        """
        :rtype: AnonymousTypeDecl
        """

        result = self._eval_astnode_field(_anonymous_type_f_type_decl)

        return result

    _field_names = TypeExpr._field_names + ("f_type_decl",)

    _kind_name = "AnonymousType"


class EnumLitSynthTypeExpr(TypeExpr):
    """
    Synthetic node. Represents the type expression for an enum literal.
    """

    __slots__ = []

    _field_names = TypeExpr._field_names + ()

    _kind_name = "EnumLitSynthTypeExpr"


class SubtypeIndication(TypeExpr):
    """
    Reference to a type by name.
    """

    __slots__ = []

    @property
    def f_has_not_null(self):
        """
        :rtype: NotNull
        """

        result = self._eval_astnode_field(_subtype_indication_f_has_not_null)

        return result

    @property
    def f_name(self):
        """
        This field can contain one of the following nodes:

        * AttributeRef

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: Name
        """

        result = self._eval_astnode_field(_subtype_indication_f_name)

        return result

    @property
    def f_constraint(self):
        """
        :rtype: Constraint
        """

        result = self._eval_astnode_field(_subtype_indication_f_constraint)

        return result

    def p_is_static_subtype(self, imprecise_fallback=False):
        """
        Returns whether Self denotes a static subtype or not.

        :param imprecise_fallback:
        :type imprecise_fallback: bool
        :rtype: bool
        """

        unwrapped_imprecise_fallback = bool(imprecise_fallback)

        c_result = self._eval_field(
            ctypes.c_uint8(),
            _subtype_indication_p_is_static_subtype,
            unwrapped_imprecise_fallback,
        )
        result = bool(c_result.value)

        return result

    _field_names = TypeExpr._field_names + (
        "f_has_not_null",
        "f_name",
        "f_constraint",
    )

    _kind_name = "SubtypeIndication"


class ConstrainedSubtypeIndication(SubtypeIndication):
    """
    Reference to a type with a range constraint.
    """

    __slots__ = []

    _field_names = SubtypeIndication._field_names + ()

    _kind_name = "ConstrainedSubtypeIndication"


class DiscreteSubtypeIndication(SubtypeIndication):
    """
    Reference to a type with a general constraint.
    """

    __slots__ = []

    _field_names = SubtypeIndication._field_names + ()

    _kind_name = "DiscreteSubtypeIndication"


class UnconstrainedArrayIndex(AdaNode):
    """
    List of unconstrained array indexes.
    """

    __slots__ = []

    @property
    def f_subtype_indication(self):
        """
        :rtype: SubtypeIndication
        """

        result = self._eval_astnode_field(
            _unconstrained_array_index_f_subtype_indication
        )

        return result

    _field_names = AdaNode._field_names + ("f_subtype_indication",)

    _kind_name = "UnconstrainedArrayIndex"


class UntilNode(AdaNode):
    """
    Qualifier for the ``until`` keyword.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of UntilPresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _until_node_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class UntilAbsent(UntilNode):
    """"""

    __slots__ = []

    _field_names = UntilNode._field_names + ()

    _kind_name = "UntilAbsent"


class UntilPresent(UntilNode):
    """"""

    __slots__ = []

    _field_names = UntilNode._field_names + ()

    _kind_name = "UntilPresent"


class UseClause(AdaNode):
    """
    Base class for use clauses.
    """

    __slots__ = []

    _field_names = AdaNode._field_names + ()


class UsePackageClause(UseClause):
    """
    Use clause for packages.
    """

    __slots__ = []

    @property
    def f_packages(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: NameList
        """

        result = self._eval_astnode_field(_use_package_clause_f_packages)

        return result

    _field_names = UseClause._field_names + ("f_packages",)

    _kind_name = "UsePackageClause"


class UseTypeClause(UseClause):
    """
    Use clause for types.
    """

    __slots__ = []

    @property
    def f_has_all(self):
        """
        :rtype: AllNode
        """

        result = self._eval_astnode_field(_use_type_clause_f_has_all)

        return result

    @property
    def f_types(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * AttributeRef

        * CallExpr

        * CharLiteral

        * DottedName

        * ExplicitDeref

        * Identifier

        * QualExpr

        * StringLiteral

        * TargetName

        :rtype: NameList
        """

        result = self._eval_astnode_field(_use_type_clause_f_types)

        return result

    _field_names = UseClause._field_names + (
        "f_has_all",
        "f_types",
    )

    _kind_name = "UseTypeClause"


class Variant(AdaNode):
    """
    Single variant in a discriminated type record declaration.

    This corresponds to a ``when ... => ...`` section in a variant part.
    """

    __slots__ = []

    @property
    def f_choices(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * Allocator

        * AttributeRef

        * BaseAggregate

        * BinOp

        * CallExpr

        * CharLiteral

        * CondExpr

        * DeclExpr

        * DiscreteSubtypeIndication

        * DottedName

        * ExplicitDeref

        * Identifier

        * MembershipExpr

        * NullLiteral

        * NumLiteral

        * OthersDesignator

        * ParenExpr

        * QualExpr

        * QuantifiedExpr

        * RaiseExpr

        * StringLiteral

        * TargetName

        * UnOp

        :rtype: AlternativesList
        """

        result = self._eval_astnode_field(_variant_f_choices)

        return result

    @property
    def f_components(self):
        """
        :rtype: ComponentList
        """

        result = self._eval_astnode_field(_variant_f_components)

        return result

    _field_names = AdaNode._field_names + (
        "f_choices",
        "f_components",
    )

    _kind_name = "Variant"


class VariantPart(AdaNode):
    """
    Variant part in a discriminated type record declaration.

    This corresponds to the whole ``case ... is ... end case;`` block.
    """

    __slots__ = []

    @property
    def f_discr_name(self):
        """
        :rtype: Identifier
        """

        result = self._eval_astnode_field(_variant_part_f_discr_name)

        return result

    @property
    def f_variant(self):
        """
        :rtype: VariantList
        """

        result = self._eval_astnode_field(_variant_part_f_variant)

        return result

    _field_names = AdaNode._field_names + (
        "f_discr_name",
        "f_variant",
    )

    _kind_name = "VariantPart"


class WithClause(AdaNode):
    """
    With clause.
    """

    __slots__ = []

    @property
    def f_has_limited(self):
        """
        :rtype: LimitedNode
        """

        result = self._eval_astnode_field(_with_clause_f_has_limited)

        return result

    @property
    def f_has_private(self):
        """
        :rtype: PrivateNode
        """

        result = self._eval_astnode_field(_with_clause_f_has_private)

        return result

    @property
    def f_packages(self):
        """
        This field contains a list that itself contains one of the following
        nodes:

        * CharLiteral

        * DottedName

        * Identifier

        * StringLiteral

        :rtype: NameList
        """

        result = self._eval_astnode_field(_with_clause_f_packages)

        return result

    _field_names = AdaNode._field_names + (
        "f_has_limited",
        "f_has_private",
        "f_packages",
    )

    _kind_name = "WithClause"


class WithPrivate(AdaNode):
    """
    Qualifier for the ``private`` keyword in ``with private`` record clauses.
    """

    __slots__ = []

    @property
    def p_as_bool(self):
        """
        Return whether this is an instance of WithPrivatePresent

        :rtype: bool
        """

        c_result = self._eval_field(ctypes.c_uint8(), _with_private_p_as_bool)
        result = bool(c_result.value)

        return result

    _field_names = AdaNode._field_names + ()


class WithPrivateAbsent(WithPrivate):
    """"""

    __slots__ = []

    _field_names = WithPrivate._field_names + ()

    _kind_name = "WithPrivateAbsent"


class WithPrivatePresent(WithPrivate):
    """"""

    __slots__ = []

    _field_names = WithPrivate._field_names + ()

    _kind_name = "WithPrivatePresent"


_EnvRebindings_c_type = _hashable_c_pointer()


class _BaseStruct(object):
    """
    Mixin for Ada struct wrappers.
    """

    # Subclasses will override this to a subclass of ctypes.Structure
    _c_type = None

    def __getitem__(self, key):
        if not isinstance(key, int):
            raise TypeError(
                "Tuples items are indexed by integers, not {}".format(type(key))
            )

        fields = self._c_type._fields_
        if 0 <= key < len(fields):
            field_name, _ = fields[key]
            return getattr(self, field_name)
        else:
            raise IndexError("There is no {}th field".format(key))

    def __repr__(self):
        field_names = [name for name, _ in self._c_type._fields_]
        return "<{} {}>".format(
            type(self).__name__,
            " ".join("{}={}".format(name, getattr(self, name)) for name in field_names),
        )

    @property
    def as_tuple(self):
        return tuple(getattr(self, f) for f, _ in self._c_type._fields_)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.as_tuple == other.as_tuple

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.as_tuple)


class _Metadata_c_type(ctypes.Structure):
    _fields_ = [
        ("dottable_subp", ctypes.c_uint8),
        ("access_entity", ctypes.c_uint8),
        ("primitive", AdaNode._node_c_type),
        ("primitive_real_type", AdaNode._node_c_type),
    ]

    @property
    def as_tuple(self):
        return tuple(getattr(self, f) for f, _ in self._fields_)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.as_tuple == other.as_tuple

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.as_tuple)


class _EntityInfo_c_type(ctypes.Structure):
    _fields_ = [
        ("md", _Metadata_c_type),
        ("rebindings", _EnvRebindings_c_type),
        ("from_rebound", ctypes.c_uint8),
    ]


class _Entity_c_type(ctypes.Structure):
    _fields_ = [
        ("node", AdaNode._node_c_type),
        ("info", _EntityInfo_c_type),
    ]

    @classmethod
    def from_bare_node(cls, node_c_value):
        return cls(node_c_value, _EntityInfo_c_type._null_value)


class Aspect(_BaseStruct):
    """
    Composite field representing the aspect of an entity (RM 13).
    """

    __slots__ = ("_exists", "_node", "_value")

    def __init__(self, exists, node, value):
        self._exists = exists
        self._node = node
        self._value = value

    @property
    def exists(self):
        """
        Whether the aspect is defined or not
        """
        return self._exists

    @property
    def node(self):
        """
        Syntactic node that defines the aspect
        """
        return self._node

    @property
    def value(self):
        """
        Expr node defining the value of the aspect
        """
        return self._value

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("exists", ctypes.c_uint8),
            ("node", _Entity_c_type),
            ("value", _Entity_c_type),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            bool(c_value.exists),
            AdaNode._wrap(c_value.node),
            AdaNode._wrap(c_value.value),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        exists = bool(value.exists)

        node = AdaNode._unwrap(value.node)

        value = AdaNode._unwrap(value.value)

        result = cls._Holder(
            cls._c_type(
                exists=exists,
                node=node,
                value=value,
            )
        )

        return result


class CompletionItem(_BaseStruct):
    """"""

    __slots__ = ("_decl", "_is_dot_call", "_is_visible")

    def __init__(self, decl, is_dot_call, is_visible):
        self._decl = decl
        self._is_dot_call = is_dot_call
        self._is_visible = is_visible

    @property
    def decl(self):
        """"""
        return self._decl

    @property
    def is_dot_call(self):
        """"""
        return self._is_dot_call

    @property
    def is_visible(self):
        """"""
        return self._is_visible

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("decl", _Entity_c_type),
            ("is_dot_call", ctypes.c_uint8),
            ("is_visible", ctypes.c_uint8),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.decl),
            bool(c_value.is_dot_call),
            bool(c_value.is_visible),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        decl = AdaNode._unwrap(value.decl)

        is_dot_call = bool(value.is_dot_call)

        is_visible = bool(value.is_visible)

        result = cls._Holder(
            cls._c_type(
                decl=decl,
                is_dot_call=is_dot_call,
                is_visible=is_visible,
            )
        )

        return result


class DiscreteRange(_BaseStruct):
    """
    Represent the range of a discrete type or subtype. The bounds are not
    evaluated, you need to call ``eval_as_int`` on them, if they're static, to
    get their value.
    """

    __slots__ = ("_low_bound", "_high_bound")

    def __init__(self, low_bound, high_bound):
        self._low_bound = low_bound
        self._high_bound = high_bound

    @property
    def low_bound(self):
        """"""
        return self._low_bound

    @property
    def high_bound(self):
        """"""
        return self._high_bound

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("low_bound", _Entity_c_type),
            ("high_bound", _Entity_c_type),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.low_bound),
            AdaNode._wrap(c_value.high_bound),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        low_bound = AdaNode._unwrap(value.low_bound)

        high_bound = AdaNode._unwrap(value.high_bound)

        result = cls._Holder(
            cls._c_type(
                low_bound=low_bound,
                high_bound=high_bound,
            )
        )

        return result


class DiscriminantValues(_BaseStruct):
    """
    Represent a set of values (as a list of choices) on a discriminant.
    """

    __slots__ = ("_discriminant", "_values")

    def __init__(self, discriminant, values):
        self._discriminant = discriminant
        self._values = values

    @property
    def discriminant(self):
        """"""
        return self._discriminant

    @property
    def values(self):
        """"""
        return self._values

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("discriminant", _Entity_c_type),
            ("values", _Entity_c_type),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.discriminant),
            AdaNode._wrap(c_value.values),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        discriminant = AdaNode._unwrap(value.discriminant)

        values = AdaNode._unwrap(value.values)

        result = cls._Holder(
            cls._c_type(
                discriminant=discriminant,
                values=values,
            )
        )

        return result


class DocAnnotation(_BaseStruct):
    """
    Documentation annotation.
    """

    __slots__ = ("_key", "_value")

    def __init__(self, key, value):
        self._key = key
        self._value = value

    @property
    def key(self):
        """
        Annotation key
        """
        return self._key

    @property
    def value(self):
        """
        Annotation value
        """
        return self._value

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("key", ctypes.c_void_p),
            ("value", ctypes.c_void_p),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            DocAnnotation._dec_ref(self.c_value)
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            _TextTypeConverter.wrap(
                ctypes.cast(c_value.key, _TextTypeConverter.c_type), True
            ),
            _TextTypeConverter.wrap(
                ctypes.cast(c_value.value, _TextTypeConverter.c_type), True
            ),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        key = _TextTypeConverter.unwrap(value.key)

        value = _TextTypeConverter.unwrap(value.value)

        result = cls._Holder(
            cls._c_type(
                key=ctypes.cast(key.c_value, ctypes.c_void_p),
                value=ctypes.cast(value.c_value, ctypes.c_void_p),
            )
        )

        cls._inc_ref(result.c_value)

        return result

    _c_ptr_type = ctypes.POINTER(_c_type)
    _inc_ref = staticmethod(
        _import_func("ada_internal_doc_annotation_inc_ref", [_c_ptr_type], None)
    )
    _dec_ref = staticmethod(
        _import_func("ada_internal_doc_annotation_dec_ref", [_c_ptr_type], None)
    )


class ParamActual(_BaseStruct):
    """
    Data structure used by zip_with_params property. Associates an expression
    (the actual) to a formal param declaration (the parameter).
    """

    __slots__ = ("_param", "_actual")

    def __init__(self, param, actual):
        self._param = param
        self._actual = actual

    @property
    def param(self):
        """"""
        return self._param

    @property
    def actual(self):
        """"""
        return self._actual

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("param", _Entity_c_type),
            ("actual", _Entity_c_type),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.param),
            AdaNode._wrap(c_value.actual),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        param = AdaNode._unwrap(value.param)

        actual = AdaNode._unwrap(value.actual)

        result = cls._Holder(
            cls._c_type(
                param=param,
                actual=actual,
            )
        )

        return result


class RefResult(_BaseStruct):
    """
    Result for a cross reference query returning a reference.
    """

    __slots__ = ("_ref", "_kind")

    def __init__(self, ref, kind):
        self._ref = ref
        self._kind = kind

    @property
    def ref(self):
        """"""
        return self._ref

    @property
    def kind(self):
        """"""
        return self._kind

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("ref", _Entity_c_type),
            ("kind", ctypes.c_int),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.ref),
            RefResultKind._wrap(c_value.kind),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        ref = AdaNode._unwrap(value.ref)

        kind = RefResultKind._unwrap(value.kind)

        result = cls._Holder(
            cls._c_type(
                ref=ref,
                kind=kind,
            )
        )

        return result


class RefdDecl(_BaseStruct):
    """
    Result for a cross reference query returning a referenced decl.
    """

    __slots__ = ("_decl", "_kind")

    def __init__(self, decl, kind):
        self._decl = decl
        self._kind = kind

    @property
    def decl(self):
        """"""
        return self._decl

    @property
    def kind(self):
        """"""
        return self._kind

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("decl", _Entity_c_type),
            ("kind", ctypes.c_int),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.decl),
            RefResultKind._wrap(c_value.kind),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        decl = AdaNode._unwrap(value.decl)

        kind = RefResultKind._unwrap(value.kind)

        result = cls._Holder(
            cls._c_type(
                decl=decl,
                kind=kind,
            )
        )

        return result


class RefdDef(_BaseStruct):
    """
    Result for a cross reference query returning a referenced defining name.
    """

    __slots__ = ("_def_name", "_kind")

    def __init__(self, def_name, kind):
        self._def_name = def_name
        self._kind = kind

    @property
    def def_name(self):
        """"""
        return self._def_name

    @property
    def kind(self):
        """"""
        return self._kind

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("def_name", _Entity_c_type),
            ("kind", ctypes.c_int),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.def_name),
            RefResultKind._wrap(c_value.kind),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        def_name = AdaNode._unwrap(value.def_name)

        kind = RefResultKind._unwrap(value.kind)

        result = cls._Holder(
            cls._c_type(
                def_name=def_name,
                kind=kind,
            )
        )

        return result


class Shape(_BaseStruct):
    """
    Represent one of the shapes that a variant record can have, as a list of
    the available components.
    """

    __slots__ = ("_components", "_discriminants_values")

    def __init__(self, components, discriminants_values):
        self._components = components
        self._discriminants_values = discriminants_values

    @property
    def components(self):
        """"""
        return self._components

    @property
    def discriminants_values(self):
        """"""
        return self._discriminants_values

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("components", ctypes.c_void_p),
            ("discriminants_values", ctypes.c_void_p),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            Shape._dec_ref(self.c_value)
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            _AdaNodeArrayConverter.wrap(
                ctypes.cast(c_value.components, _AdaNodeArrayConverter.c_type), True
            ),
            _DiscriminantValuesArrayConverter.wrap(
                ctypes.cast(
                    c_value.discriminants_values,
                    _DiscriminantValuesArrayConverter.c_type,
                ),
                True,
            ),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        components = _AdaNodeArrayConverter.unwrap(value.components)

        discriminants_values = _DiscriminantValuesArrayConverter.unwrap(
            value.discriminants_values
        )

        result = cls._Holder(
            cls._c_type(
                components=ctypes.cast(components.c_value, ctypes.c_void_p),
                discriminants_values=ctypes.cast(
                    discriminants_values.c_value, ctypes.c_void_p
                ),
            )
        )

        cls._inc_ref(result.c_value)

        return result

    _c_ptr_type = ctypes.POINTER(_c_type)
    _inc_ref = staticmethod(
        _import_func("ada_internal_shape_inc_ref", [_c_ptr_type], None)
    )
    _dec_ref = staticmethod(
        _import_func("ada_internal_shape_dec_ref", [_c_ptr_type], None)
    )


class Substitution(_BaseStruct):
    """
    Represent a substitution of a BasicDecl by a given value. This can then be
    used as part of an environment in the eval_as_*_in_env property. See the
    declaration of those properties for more details.
    """

    __slots__ = ("_from_decl", "_to_value", "_value_type")

    def __init__(self, from_decl, to_value, value_type):
        self._from_decl = from_decl
        self._to_value = to_value
        self._value_type = value_type

    @property
    def from_decl(self):
        """
        The declaration to substitute.
        """
        return self._from_decl

    @property
    def to_value(self):
        """
        The value by which to substitute the declaration.
        """
        return self._to_value

    @property
    def value_type(self):
        """
        The type of the substituted value.
        """
        return self._value_type

    class _c_type(ctypes.Structure):
        _fields_ = [
            ("from_decl", _Entity_c_type),
            ("to_value", _big_integer.c_type),
            ("value_type", _Entity_c_type),
        ]

    class _Holder(object):
        def __init__(self, c_value):
            self.c_value = c_value

        def clear(self):
            self.c_value = None

        def __del__(self):
            Substitution._dec_ref(self.c_value)
            self.clear()

    @classmethod
    def _wrap(cls, c_value):
        return cls(
            AdaNode._wrap(c_value.from_decl),
            _big_integer.wrap(c_value.to_value),
            AdaNode._wrap(c_value.value_type),
        )

    @classmethod
    def _unwrap(cls, value, context=None):
        if not isinstance(value, cls):
            _raise_type_error(cls.__name__, value)

        from_decl = AdaNode._unwrap(value.from_decl)

        to_value = _big_integer.unwrap(value.to_value)

        value_type = AdaNode._unwrap(value.value_type)

        result = cls._Holder(
            cls._c_type(
                from_decl=from_decl,
                to_value=to_value.c_value,
                value_type=value_type,
            )
        )

        cls._inc_ref(result.c_value)

        return result

    _c_ptr_type = ctypes.POINTER(_c_type)
    _inc_ref = staticmethod(
        _import_func("ada_internal_substitution_inc_ref", [_c_ptr_type], None)
    )
    _dec_ref = staticmethod(
        _import_func("ada_internal_substitution_dec_ref", [_c_ptr_type], None)
    )


_Metadata_c_type._null_value = _Metadata_c_type(False, False, False, False)
_EntityInfo_c_type._null_value = _EntityInfo_c_type(_Metadata_c_type._null_value, None)


#
# Low-level binding - Second part
#

# For performance, allocate a single C API entity for all uses of null
# entities.
_Entity_c_type._null_value = _Entity_c_type()
_Entity_c_type._null_value.node = None


class _BaseArray(object):
    """
    Base class for Ada arrays bindings.
    """

    c_element_type = None
    """
    Ctype class for array elements.
    """

    items_refcounted = False
    """
    Whether items for this arrays are ref-counted.
    """

    __slots__ = ("c_value", "length", "items")

    def __init__(self, c_value):
        self.c_value = c_value

        self.length = c_value.contents.n

        items_addr = _field_address(c_value.contents, "items")
        items = self.c_element_type.from_address(items_addr)
        self.items = ctypes.pointer(items)

    def __repr__(self):
        return "<{} {}>".format(type(self).__name__, list(self))

    def clear(self):
        self.c_value = None
        self.length = None
        self.items = None

    def __del__(self):
        self.dec_ref(self.c_value)
        self.clear()

    @classmethod
    def wrap(cls, c_value, from_field_access):
        helper = cls(c_value)

        result = []
        for i in range(helper.length):
            # In ctypes, accessing an array element does not copy it, which
            # means the the array must live at least as long as the accessed
            # element. We cannot guarantee that, so we must copy the element so
            # that it is independent of the array it comes from.
            #
            # The try/except block tries to do a copy if "item" is indeed a
            # buffer to be copied, and will fail if it's a mere integer, which
            # does not need the buffer copy anyway, hence the "pass".
            item = helper.items[i]
            try:
                item = cls.c_element_type.from_buffer_copy(item)
            except TypeError:
                pass
            result.append(helper.wrap_item(item))

        # If this array value comes from a structure field, we must not call
        # its dec_ref primitive, as it is up to the structure's dec_ref
        # primitive to take care of it.
        if from_field_access:
            helper.clear()

        return result

    @classmethod
    def unwrap(cls, value, context=None):
        if not isinstance(value, list):
            _raise_type_error("list", value)

        # Create a holder for the result
        result = cls(cls.create(len(value)))

        # Unwrap all items at once, preserving their holder so that resources
        # are deallocated if there is an error at some point.
        items = [result.unwrap_item(item, context) for item in value]

        # Initialize the resulting array
        for i, (_, item) in enumerate(items):
            result.items[i] = item

        # At this point, we know that this is successful. We don't want
        # holders to dec-ref the content so that the return array takes over
        # the corresponding refcounting shares.
        if cls.items_refcounted:
            for holder, _ in items:
                holder.clear()

        return result


class _BareAdaNodeArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of BareAdaNode.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return AdaNode._wrap_bare_node(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = item._node_c_value
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = AdaNode._node_c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", AdaNode._node_c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_bare_ada_node_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_bare_ada_node_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_bare_ada_node_array_dec_ref", [c_type], None)
    )


class _AdaNodeArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalEntity.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return AdaNode._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = AdaNode._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = _Entity_c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", _Entity_c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_ada_node_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(_import_func("ada_ada_node_array_inc_ref", [c_type], None))
    dec_ref = staticmethod(_import_func("ada_ada_node_array_dec_ref", [c_type], None))


class _TextTypeConverter(_BaseArray):
    """
    Wrapper class for arrays of CharacterType.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return _py2to3.unicode_character(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = ord(item)
        c_value = c_holder
        return (c_holder, c_value)

    @classmethod
    def wrap(cls, c_value, from_field_access):
        # Reinterpret this array of uint32_t values as the equivalent array of
        # characters, then decode it using the appropriate UTF-32 encoding.
        chars = ctypes.cast(
            ctypes.pointer(c_value.contents.items), ctypes.POINTER(ctypes.c_char)
        )
        return chars[: 4 * c_value.contents.n].decode(_text.encoding)

    @classmethod
    def unwrap(cls, value, context=None):
        # If `value` is not a list, assume it's a string, and convert it to the
        # expected list.
        if not isinstance(value, list):
            value = list(_text.cast(value))

        return super(_TextTypeConverter, cls).unwrap(value, context)

    c_element_type = ctypes.c_uint32

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", ctypes.c_uint32 * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(_import_func("ada_text_type_create", [ctypes.c_int], c_type))
    inc_ref = staticmethod(_import_func("ada_text_type_inc_ref", [c_type], None))
    dec_ref = staticmethod(_import_func("ada_text_type_dec_ref", [c_type], None))


class _CompletionItemArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalCompletionItem.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return CompletionItem._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = CompletionItem._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = CompletionItem._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", CompletionItem._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_completion_item_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_completion_item_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_completion_item_array_dec_ref", [c_type], None)
    )


class _DiscriminantValuesArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalDiscriminantValues.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return DiscriminantValues._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = DiscriminantValues._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = DiscriminantValues._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", DiscriminantValues._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_discriminant_values_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_discriminant_values_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_discriminant_values_array_dec_ref", [c_type], None)
    )


class _DocAnnotationArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalDocAnnotation.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = True

    @staticmethod
    def wrap_item(item):
        return DocAnnotation._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = DocAnnotation._unwrap(item)
        c_value = c_holder.c_value
        return (c_holder, c_value)

    c_element_type = DocAnnotation._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", DocAnnotation._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_doc_annotation_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_doc_annotation_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_doc_annotation_array_dec_ref", [c_type], None)
    )


class _AdaNodeArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalEntity.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return AdaNode._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = AdaNode._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = _Entity_c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", _Entity_c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_ada_node_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(_import_func("ada_ada_node_array_inc_ref", [c_type], None))
    dec_ref = staticmethod(_import_func("ada_ada_node_array_dec_ref", [c_type], None))


class _ParamActualArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalParamActual.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return ParamActual._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = ParamActual._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = ParamActual._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", ParamActual._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_param_actual_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_param_actual_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_param_actual_array_dec_ref", [c_type], None)
    )


class _RefResultArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalRefResult.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return RefResult._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = RefResult._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = RefResult._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", RefResult._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_ref_result_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(_import_func("ada_ref_result_array_inc_ref", [c_type], None))
    dec_ref = staticmethod(_import_func("ada_ref_result_array_dec_ref", [c_type], None))


class _ShapeArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalShape.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = True

    @staticmethod
    def wrap_item(item):
        return Shape._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = Shape._unwrap(item)
        c_value = c_holder.c_value
        return (c_holder, c_value)

    c_element_type = Shape._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", Shape._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_shape_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(_import_func("ada_shape_array_inc_ref", [c_type], None))
    dec_ref = staticmethod(_import_func("ada_shape_array_dec_ref", [c_type], None))


class _SubstitutionArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalSubstitution.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = True

    @staticmethod
    def wrap_item(item):
        return Substitution._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = Substitution._unwrap(item)
        c_value = c_holder.c_value
        return (c_holder, c_value)

    c_element_type = Substitution._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", Substitution._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_substitution_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_substitution_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_substitution_array_dec_ref", [c_type], None)
    )


class _AnalysisUnitArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of InternalUnit.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return AnalysisUnit._wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = AnalysisUnit._unwrap(item)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = AnalysisUnit._c_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", AnalysisUnit._c_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_analysis_unit_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_analysis_unit_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_analysis_unit_array_dec_ref", [c_type], None)
    )


class _UnboundedTextTypeArrayConverter(_BaseArray):
    """
    Wrapper class for arrays of SymbolType.

    This class is not meant to be directly instantiated: it is only used to
    convert values that various methods take/return.
    """

    __slots__ = _BaseArray.__slots__
    items_refcounted = False

    @staticmethod
    def wrap_item(item):
        return _symbol_type.wrap(item)

    @staticmethod
    def unwrap_item(item, context=None):
        c_holder = _symbol_type.unwrap(item, context)
        c_value = c_holder
        return (c_holder, c_value)

    c_element_type = _symbol_type

    class c_struct(ctypes.Structure):
        _fields_ = [
            ("n", ctypes.c_int),
            ("ref_count", ctypes.c_int),
            ("items", _symbol_type * 1),
        ]

    c_type = ctypes.POINTER(c_struct)

    create = staticmethod(
        _import_func("ada_unbounded_text_type_array_create", [ctypes.c_int], c_type)
    )
    inc_ref = staticmethod(
        _import_func("ada_unbounded_text_type_array_inc_ref", [c_type], None)
    )
    dec_ref = staticmethod(
        _import_func("ada_unbounded_text_type_array_dec_ref", [c_type], None)
    )


_free = _import_func("ada_free", [ctypes.c_void_p], None)

_destroy_text = _import_func("ada_destroy_text", [ctypes.POINTER(_text)], None)

_symbol_text = _import_func(
    "ada_symbol_text", [ctypes.POINTER(_symbol_type), ctypes.POINTER(_text)], None
)

# Analysis primitives
_create_analysis_context = _import_func(
    "ada_create_analysis_context",
    [
        ctypes.c_char_p,  # charset
        _unit_provider,  # unit_provider
        ctypes.c_int,  # with_trivia
        ctypes.c_int,
    ],  # tab_stop
    AnalysisContext._c_type,
)
_context_incref = _import_func(
    "ada_context_incref", [AnalysisContext._c_type], AnalysisContext._c_type
)
_context_decref = _import_func("ada_context_decref", [AnalysisContext._c_type], None)
_context_symbol = _import_func(
    "ada_context_symbol",
    [AnalysisContext._c_type, ctypes.POINTER(_text), ctypes.POINTER(_symbol_type)],
    ctypes.c_int,
)
_discard_errors_in_populate_lexical_env = _import_func(
    "ada_context_discard_errors_in_populate_lexical_env",
    [AnalysisContext._c_type, ctypes.c_int],
    None,
)
_get_analysis_unit_from_file = _import_func(
    "ada_get_analysis_unit_from_file",
    [
        AnalysisContext._c_type,  # context
        ctypes.c_char_p,  # filename
        ctypes.c_char_p,  # charset
        ctypes.c_int,  # reparse
        ctypes.c_int,
    ],  # grammar rule
    AnalysisUnit._c_type,
)
_get_analysis_unit_from_buffer = _import_func(
    "ada_get_analysis_unit_from_buffer",
    [
        AnalysisContext._c_type,  # context
        ctypes.c_char_p,  # filename
        ctypes.c_char_p,  # charset
        ctypes.c_char_p,  # buffer
        ctypes.c_size_t,  # buffer_size
        ctypes.c_int,
    ],  # grammar rule
    AnalysisUnit._c_type,
)
_get_analysis_unit_from_provider = _import_func(
    "ada_get_analysis_unit_from_provider",
    [
        AnalysisContext._c_type,  # context
        ctypes.POINTER(_text),  # name
        ctypes.c_int,  # kind
        ctypes.c_char_p,  # charset
        ctypes.c_int,
    ],  # reparse
    AnalysisUnit._c_type,
)
_unit_root = _import_func(
    "ada_unit_root", [AnalysisUnit._c_type, ctypes.POINTER(_Entity_c_type)], None
)
_unit_first_token = _import_func(
    "ada_unit_first_token", [AnalysisUnit._c_type, ctypes.POINTER(Token)], None
)
_unit_last_token = _import_func(
    "ada_unit_last_token", [AnalysisUnit._c_type, ctypes.POINTER(Token)], None
)
_unit_token_count = _import_func(
    "ada_unit_token_count", [AnalysisUnit._c_type], ctypes.c_int
)
_unit_trivia_count = _import_func(
    "ada_unit_trivia_count", [AnalysisUnit._c_type], ctypes.c_int
)
_unit_lookup_token = _import_func(
    "ada_unit_lookup_token",
    [AnalysisUnit._c_type, ctypes.POINTER(Sloc._c_type), ctypes.POINTER(Token)],
    None,
)
_unit_dump_lexical_env = _import_func(
    "ada_unit_dump_lexical_env", [AnalysisUnit._c_type], None
)
_unit_filename = _import_func(
    "ada_unit_filename", [AnalysisUnit._c_type], ctypes.POINTER(ctypes.c_char)
)
_unit_diagnostic_count = _import_func(
    "ada_unit_diagnostic_count", [AnalysisUnit._c_type], ctypes.c_uint
)
_unit_diagnostic = _import_func(
    "ada_unit_diagnostic",
    [AnalysisUnit._c_type, ctypes.c_uint, ctypes.POINTER(Diagnostic._c_type)],
    ctypes.c_int,
)
_unit_context = _import_func(
    "ada_unit_context", [AnalysisUnit._c_type], AnalysisContext._c_type
)
_unit_reparse_from_file = _import_func(
    "ada_unit_reparse_from_file",
    [AnalysisUnit._c_type, ctypes.c_char_p],  # unit  # charset
    ctypes.c_int,
)
_unit_reparse_from_buffer = _import_func(
    "ada_unit_reparse_from_buffer",
    [
        AnalysisUnit._c_type,  # unit
        ctypes.c_char_p,  # charset
        ctypes.c_char_p,  # buffer
        ctypes.c_size_t,
    ],  # buffer_size
    None,
)
_unit_populate_lexical_env = _import_func(
    "ada_unit_populate_lexical_env", [AnalysisUnit._c_type], ctypes.c_int
)

# General AST node primitives
_node_kind = _import_func(
    "ada_node_kind", [ctypes.POINTER(_Entity_c_type)], ctypes.c_int
)
_node_unit = _import_func(
    "ada_node_unit", [ctypes.POINTER(_Entity_c_type)], AnalysisUnit._c_type
)
_node_is_token_node = _import_func(
    "ada_node_is_token_node", [ctypes.POINTER(_Entity_c_type)], ctypes.c_int
)
_node_is_synthetic = _import_func(
    "ada_node_is_synthetic", [ctypes.POINTER(_Entity_c_type)], ctypes.c_int
)
_node_image = _import_func(
    "ada_node_image", [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_text)], None
)
_node_text = _import_func(
    "ada_node_text", [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_text)], None
)
_node_sloc_range = _import_func(
    "ada_node_sloc_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(SlocRange._c_type)],
    None,
)
_lookup_in_node = _import_func(
    "ada_lookup_in_node",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(Sloc._c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    None,
)
_node_children_count = _import_func(
    "ada_node_children_count", [ctypes.POINTER(_Entity_c_type)], ctypes.c_uint
)
_node_child = _import_func(
    "ada_node_child",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)

_ada_node_p_declarative_scope = _import_func(
    "ada_ada_node_p_declarative_scope",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_complete = _import_func(
    "ada_ada_node_p_complete",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_CompletionItemArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_ada_node_p_valid_keywords = _import_func(
    "ada_ada_node_p_valid_keywords",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_UnboundedTextTypeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_ada_node_p_generic_instantiations = _import_func(
    "ada_ada_node_p_generic_instantiations",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_ada_node_p_semantic_parent = _import_func(
    "ada_ada_node_p_semantic_parent",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_parent_basic_decl = _import_func(
    "ada_ada_node_p_parent_basic_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_filter_is_imported_by = _import_func(
    "ada_ada_node_p_filter_is_imported_by",
    [
        ctypes.POINTER(_Entity_c_type),
        _AnalysisUnitArrayConverter.c_type,
        ctypes.c_uint8,
        ctypes.POINTER(_AnalysisUnitArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_ada_node_p_xref_entry_point = _import_func(
    "ada_ada_node_p_xref_entry_point",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_ada_node_p_resolve_names = _import_func(
    "ada_ada_node_p_resolve_names",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_ada_node_p_standard_unit = _import_func(
    "ada_ada_node_p_standard_unit",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(AnalysisUnit._c_type)],
    ctypes.c_int,
)
_ada_node_p_std_entity = _import_func(
    "ada_ada_node_p_std_entity",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_ada_node_p_bool_type = _import_func(
    "ada_ada_node_p_bool_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_int_type = _import_func(
    "ada_ada_node_p_int_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_universal_int_type = _import_func(
    "ada_ada_node_p_universal_int_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_universal_real_type = _import_func(
    "ada_ada_node_p_universal_real_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_p_top_level_decl = _import_func(
    "ada_ada_node_p_top_level_decl",
    [
        ctypes.POINTER(_Entity_c_type),
        AnalysisUnit._c_type,
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_ada_node_p_choice_match = _import_func(
    "ada_ada_node_p_choice_match",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_big_integer.c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_ada_node_p_gnat_xref = _import_func(
    "ada_ada_node_p_gnat_xref",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_parent = _import_func(
    "ada_ada_node_parent",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_parents = _import_func(
    "ada_ada_node_parents",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_ada_node_children = _import_func(
    "ada_ada_node_children",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_ada_node_token_start = _import_func(
    "ada_ada_node_token_start",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(Token)],
    ctypes.c_int,
)
_ada_node_token_end = _import_func(
    "ada_ada_node_token_end",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(Token)],
    ctypes.c_int,
)
_ada_node_child_index = _import_func(
    "ada_ada_node_child_index",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_int)],
    ctypes.c_int,
)
_ada_node_previous_sibling = _import_func(
    "ada_ada_node_previous_sibling",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_next_sibling = _import_func(
    "ada_ada_node_next_sibling",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ada_node_unit = _import_func(
    "ada_ada_node_unit",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(AnalysisUnit._c_type)],
    ctypes.c_int,
)
_ada_node_is_ghost = _import_func(
    "ada_ada_node_is_ghost",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_ada_node_full_sloc_image = _import_func(
    "ada_ada_node_full_sloc_image",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_abort_node_p_as_bool = _import_func(
    "ada_abort_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_abstract_node_p_as_bool = _import_func(
    "ada_abstract_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_assoc_list_p_zip_with_params = _import_func(
    "ada_assoc_list_p_zip_with_params",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_ParamActualArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_aliased_node_p_as_bool = _import_func(
    "ada_aliased_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_all_node_p_as_bool = _import_func(
    "ada_all_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_constrained_array_indices_f_list = _import_func(
    "ada_constrained_array_indices_f_list",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_unconstrained_array_indices_f_types = _import_func(
    "ada_unconstrained_array_indices_f_types",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_aspect_assoc_f_id = _import_func(
    "ada_aspect_assoc_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_aspect_assoc_f_expr = _import_func(
    "ada_aspect_assoc_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_at_clause_f_name = _import_func(
    "ada_at_clause_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_at_clause_f_expr = _import_func(
    "ada_at_clause_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_attribute_def_clause_f_attribute_expr = _import_func(
    "ada_attribute_def_clause_f_attribute_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_attribute_def_clause_f_expr = _import_func(
    "ada_attribute_def_clause_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_enum_rep_clause_f_type_name = _import_func(
    "ada_enum_rep_clause_f_type_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_enum_rep_clause_f_aggregate = _import_func(
    "ada_enum_rep_clause_f_aggregate",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_rep_clause_f_name = _import_func(
    "ada_record_rep_clause_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_rep_clause_f_at_expr = _import_func(
    "ada_record_rep_clause_f_at_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_rep_clause_f_components = _import_func(
    "ada_record_rep_clause_f_components",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_aspect_spec_f_aspect_assocs = _import_func(
    "ada_aspect_spec_f_aspect_assocs",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_assoc_p_assoc_expr = _import_func(
    "ada_base_assoc_p_assoc_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_contract_case_assoc_f_guard = _import_func(
    "ada_contract_case_assoc_f_guard",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_contract_case_assoc_f_consequence = _import_func(
    "ada_contract_case_assoc_f_consequence",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_pragma_argument_assoc_f_id = _import_func(
    "ada_pragma_argument_assoc_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_pragma_argument_assoc_f_expr = _import_func(
    "ada_pragma_argument_assoc_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_formal_param_holder_p_abstract_formal_params = _import_func(
    "ada_base_formal_param_holder_p_abstract_formal_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_base_formal_param_holder_p_nb_min_params = _import_func(
    "ada_base_formal_param_holder_p_nb_min_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_int)],
    ctypes.c_int,
)
_base_formal_param_holder_p_nb_max_params = _import_func(
    "ada_base_formal_param_holder_p_nb_max_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_int)],
    ctypes.c_int,
)
_base_formal_param_holder_p_param_types = _import_func(
    "ada_base_formal_param_holder_p_param_types",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_base_subp_spec_p_returns = _import_func(
    "ada_base_subp_spec_p_returns",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_subp_spec_p_params = _import_func(
    "ada_base_subp_spec_p_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_base_subp_spec_p_primitive_subp_types = _import_func(
    "ada_base_subp_spec_p_primitive_subp_types",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_base_subp_spec_p_primitive_subp_first_type = _import_func(
    "ada_base_subp_spec_p_primitive_subp_first_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_subp_spec_p_primitive_subp_tagged_type = _import_func(
    "ada_base_subp_spec_p_primitive_subp_tagged_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_subp_spec_p_return_type = _import_func(
    "ada_base_subp_spec_p_return_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_entry_spec_f_entry_name = _import_func(
    "ada_entry_spec_f_entry_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_spec_f_family_type = _import_func(
    "ada_entry_spec_f_family_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_spec_f_entry_params = _import_func(
    "ada_entry_spec_f_entry_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_spec_f_subp_kind = _import_func(
    "ada_subp_spec_f_subp_kind",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_spec_f_subp_name = _import_func(
    "ada_subp_spec_f_subp_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_spec_f_subp_params = _import_func(
    "ada_subp_spec_f_subp_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_spec_f_subp_returns = _import_func(
    "ada_subp_spec_f_subp_returns",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_list_f_components = _import_func(
    "ada_component_list_f_components",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_list_f_variant_part = _import_func(
    "ada_component_list_f_variant_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_known_discriminant_part_f_discr_specs = _import_func(
    "ada_known_discriminant_part_f_discr_specs",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_completion_formal_params_f_params = _import_func(
    "ada_entry_completion_formal_params_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_formal_part_f_decls = _import_func(
    "ada_generic_formal_part_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_record_def_f_components = _import_func(
    "ada_base_record_def_f_components",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_assoc_p_get_params = _import_func(
    "ada_basic_assoc_p_get_params",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_aggregate_assoc_f_designators = _import_func(
    "ada_aggregate_assoc_f_designators",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_aggregate_assoc_f_r_expr = _import_func(
    "ada_aggregate_assoc_f_r_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_assoc_f_ids = _import_func(
    "ada_discriminant_assoc_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_assoc_f_discr_expr = _import_func(
    "ada_discriminant_assoc_f_discr_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_iterated_assoc_f_spec = _import_func(
    "ada_iterated_assoc_f_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_iterated_assoc_f_r_expr = _import_func(
    "ada_iterated_assoc_f_r_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_assoc_f_designator = _import_func(
    "ada_param_assoc_f_designator",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_assoc_f_r_expr = _import_func(
    "ada_param_assoc_f_r_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_is_formal = _import_func(
    "ada_basic_decl_p_is_formal",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_basic_decl_p_doc_annotations = _import_func(
    "ada_basic_decl_p_doc_annotations",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_DocAnnotationArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_doc = _import_func(
    "ada_basic_decl_p_doc",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_basic_decl_p_previous_part_for_decl = _import_func(
    "ada_basic_decl_p_previous_part_for_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_canonical_part = _import_func(
    "ada_basic_decl_p_canonical_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_is_static_decl = _import_func(
    "ada_basic_decl_p_is_static_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_basic_decl_p_is_imported = _import_func(
    "ada_basic_decl_p_is_imported",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_basic_decl_f_aspects = _import_func(
    "ada_basic_decl_f_aspects",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_get_aspect_assoc = _import_func(
    "ada_basic_decl_p_get_aspect_assoc",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_get_aspect_spec_expr = _import_func(
    "ada_basic_decl_p_get_aspect_spec_expr",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_get_aspect = _import_func(
    "ada_basic_decl_p_get_aspect",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.c_uint8,
        ctypes.POINTER(Aspect._c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_has_aspect = _import_func(
    "ada_basic_decl_p_has_aspect",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.c_uint8,
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_basic_decl_p_get_pragma = _import_func(
    "ada_basic_decl_p_get_pragma",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_get_representation_clause = _import_func(
    "ada_basic_decl_p_get_representation_clause",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.c_uint8,
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_is_compilation_unit_root = _import_func(
    "ada_basic_decl_p_is_compilation_unit_root",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_basic_decl_p_is_visible = _import_func(
    "ada_basic_decl_p_is_visible",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_basic_decl_p_base_subp_declarations = _import_func(
    "ada_basic_decl_p_base_subp_declarations",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_basic_decl_p_root_subp_declarations = _import_func(
    "ada_basic_decl_p_root_subp_declarations",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_find_all_overrides = _import_func(
    "ada_basic_decl_p_find_all_overrides",
    [
        ctypes.POINTER(_Entity_c_type),
        _AnalysisUnitArrayConverter.c_type,
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_defining_names = _import_func(
    "ada_basic_decl_p_defining_names",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_basic_decl_p_defining_name = _import_func(
    "ada_basic_decl_p_defining_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_type_expression = _import_func(
    "ada_basic_decl_p_type_expression",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_subp_spec_or_null = _import_func(
    "ada_basic_decl_p_subp_spec_or_null",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_is_subprogram = _import_func(
    "ada_basic_decl_p_is_subprogram",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_basic_decl_p_relative_name = _import_func(
    "ada_basic_decl_p_relative_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_relative_name_text = _import_func(
    "ada_basic_decl_p_relative_name_text",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_symbol_type)],
    ctypes.c_int,
)
_basic_decl_p_next_part_for_decl = _import_func(
    "ada_basic_decl_p_next_part_for_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_body_part_for_decl = _import_func(
    "ada_basic_decl_p_body_part_for_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_basic_decl_p_fully_qualified_name_array = _import_func(
    "ada_basic_decl_p_fully_qualified_name_array",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_UnboundedTextTypeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_basic_decl_p_fully_qualified_name = _import_func(
    "ada_basic_decl_p_fully_qualified_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_basic_decl_p_canonical_fully_qualified_name = _import_func(
    "ada_basic_decl_p_canonical_fully_qualified_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_basic_decl_p_unique_identifying_name = _import_func(
    "ada_basic_decl_p_unique_identifying_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_base_formal_param_decl_p_formal_type = _import_func(
    "ada_base_formal_param_decl_p_formal_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_component_decl_f_ids = _import_func(
    "ada_component_decl_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_decl_f_component_def = _import_func(
    "ada_component_decl_f_component_def",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_decl_f_default_expr = _import_func(
    "ada_component_decl_f_default_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_spec_f_ids = _import_func(
    "ada_discriminant_spec_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_spec_f_type_expr = _import_func(
    "ada_discriminant_spec_f_type_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_spec_f_default_expr = _import_func(
    "ada_discriminant_spec_f_default_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_formal_f_decl = _import_func(
    "ada_generic_formal_f_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_spec_f_ids = _import_func(
    "ada_param_spec_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_spec_f_has_aliased = _import_func(
    "ada_param_spec_f_has_aliased",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_spec_f_mode = _import_func(
    "ada_param_spec_f_mode",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_spec_f_type_expr = _import_func(
    "ada_param_spec_f_type_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_param_spec_f_default_expr = _import_func(
    "ada_param_spec_f_default_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_package_decl_f_package_name = _import_func(
    "ada_base_package_decl_f_package_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_package_decl_f_public_part = _import_func(
    "ada_base_package_decl_f_public_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_package_decl_f_private_part = _import_func(
    "ada_base_package_decl_f_private_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_package_decl_f_end_name = _import_func(
    "ada_base_package_decl_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_package_decl_p_body_part = _import_func(
    "ada_base_package_decl_p_body_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_f_name = _import_func(
    "ada_base_type_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_base_subtype = _import_func(
    "ada_base_type_decl_p_base_subtype",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_private_completion = _import_func(
    "ada_base_type_decl_p_private_completion",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_is_inherited_primitive = _import_func(
    "ada_base_type_decl_p_is_inherited_primitive",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_get_record_representation_clause = _import_func(
    "ada_base_type_decl_p_get_record_representation_clause",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_get_enum_representation_clause = _import_func(
    "ada_base_type_decl_p_get_enum_representation_clause",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_is_record_type = _import_func(
    "ada_base_type_decl_p_is_record_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_array_type = _import_func(
    "ada_base_type_decl_p_is_array_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_find_derived_types = _import_func(
    "ada_base_type_decl_p_find_derived_types",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_real_type = _import_func(
    "ada_base_type_decl_p_is_real_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_float_type = _import_func(
    "ada_base_type_decl_p_is_float_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_fixed_point = _import_func(
    "ada_base_type_decl_p_is_fixed_point",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_enum_type = _import_func(
    "ada_base_type_decl_p_is_enum_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_access_type = _import_func(
    "ada_base_type_decl_p_is_access_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_char_type = _import_func(
    "ada_base_type_decl_p_is_char_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_discrete_range = _import_func(
    "ada_base_type_decl_p_discrete_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(DiscreteRange._c_type)],
    ctypes.c_int,
)
_base_type_decl_p_is_discrete_type = _import_func(
    "ada_base_type_decl_p_is_discrete_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_int_type = _import_func(
    "ada_base_type_decl_p_is_int_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_accessed_type = _import_func(
    "ada_base_type_decl_p_accessed_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_tagged_type = _import_func(
    "ada_base_type_decl_p_is_tagged_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_base_type = _import_func(
    "ada_base_type_decl_p_base_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_base_types = _import_func(
    "ada_base_type_decl_p_base_types",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_find_all_derived_types = _import_func(
    "ada_base_type_decl_p_find_all_derived_types",
    [
        ctypes.POINTER(_Entity_c_type),
        _AnalysisUnitArrayConverter.c_type,
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_comp_type = _import_func(
    "ada_base_type_decl_p_comp_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_index_type = _import_func(
    "ada_base_type_decl_p_index_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_int,
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_derived_type = _import_func(
    "ada_base_type_decl_p_is_derived_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_interface_type = _import_func(
    "ada_base_type_decl_p_is_interface_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_matching_type = _import_func(
    "ada_base_type_decl_p_matching_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_canonical_type = _import_func(
    "ada_base_type_decl_p_canonical_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_previous_part = _import_func(
    "ada_base_type_decl_p_previous_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_next_part = _import_func(
    "ada_base_type_decl_p_next_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_full_view = _import_func(
    "ada_base_type_decl_p_full_view",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_type_decl_p_is_definite_subtype = _import_func(
    "ada_base_type_decl_p_is_definite_subtype",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_base_type_decl_p_is_private = _import_func(
    "ada_base_type_decl_p_is_private",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_base_type_decl_p_discriminants_list = _import_func(
    "ada_base_type_decl_p_discriminants_list",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_base_type_decl_p_root_type = _import_func(
    "ada_base_type_decl_p_root_type",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_base_type_decl_p_shapes = _import_func(
    "ada_base_type_decl_p_shapes",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_ShapeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_subtype_decl_f_subtype = _import_func(
    "ada_subtype_decl_f_subtype",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_incomplete_type_decl_f_discriminants = _import_func(
    "ada_incomplete_type_decl_f_discriminants",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_incomplete_tagged_type_decl_f_has_abstract = _import_func(
    "ada_incomplete_tagged_type_decl_f_has_abstract",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_type_decl_f_discriminants = _import_func(
    "ada_protected_type_decl_f_discriminants",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_type_decl_f_interfaces = _import_func(
    "ada_protected_type_decl_f_interfaces",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_type_decl_f_definition = _import_func(
    "ada_protected_type_decl_f_definition",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_type_decl_f_discriminants = _import_func(
    "ada_task_type_decl_f_discriminants",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_type_decl_f_definition = _import_func(
    "ada_task_type_decl_f_definition",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_decl_f_discriminants = _import_func(
    "ada_type_decl_f_discriminants",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_decl_f_type_def = _import_func(
    "ada_type_decl_f_type_def",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_decl_p_get_primitives = _import_func(
    "ada_type_decl_p_get_primitives",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_basic_subp_decl_p_subp_decl_spec = _import_func(
    "ada_basic_subp_decl_p_subp_decl_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_classic_subp_decl_f_overriding = _import_func(
    "ada_classic_subp_decl_f_overriding",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_classic_subp_decl_f_subp_spec = _import_func(
    "ada_classic_subp_decl_f_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_classic_subp_decl_p_body_part = _import_func(
    "ada_classic_subp_decl_p_body_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_formal_subp_decl_f_default_expr = _import_func(
    "ada_formal_subp_decl_f_default_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_decl_f_overriding = _import_func(
    "ada_entry_decl_f_overriding",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_decl_f_spec = _import_func(
    "ada_entry_decl_f_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_decl_p_body_part = _import_func(
    "ada_entry_decl_p_body_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_enum_literal_decl_f_name = _import_func(
    "ada_enum_literal_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_enum_literal_decl_p_enum_type = _import_func(
    "ada_enum_literal_decl_p_enum_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_internal_f_subp_spec = _import_func(
    "ada_generic_subp_internal_f_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_body_node_p_previous_part = _import_func(
    "ada_body_node_p_previous_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_body_node_p_decl_part = _import_func(
    "ada_body_node_p_decl_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_body_node_p_subunit_root = _import_func(
    "ada_body_node_p_subunit_root",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_subp_body_f_overriding = _import_func(
    "ada_base_subp_body_f_overriding",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_subp_body_f_subp_spec = _import_func(
    "ada_base_subp_body_f_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_expr_function_f_expr = _import_func(
    "ada_expr_function_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_body_f_decls = _import_func(
    "ada_subp_body_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_body_f_stmts = _import_func(
    "ada_subp_body_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_body_f_end_name = _import_func(
    "ada_subp_body_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_renaming_decl_f_renames = _import_func(
    "ada_subp_renaming_decl_f_renames",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_body_stub_p_syntactic_fully_qualified_name = _import_func(
    "ada_body_stub_p_syntactic_fully_qualified_name",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_UnboundedTextTypeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_package_body_stub_f_name = _import_func(
    "ada_package_body_stub_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_body_stub_f_name = _import_func(
    "ada_protected_body_stub_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_body_stub_f_overriding = _import_func(
    "ada_subp_body_stub_f_overriding",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subp_body_stub_f_subp_spec = _import_func(
    "ada_subp_body_stub_f_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_body_stub_f_name = _import_func(
    "ada_task_body_stub_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_entry_name = _import_func(
    "ada_entry_body_f_entry_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_index_spec = _import_func(
    "ada_entry_body_f_index_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_params = _import_func(
    "ada_entry_body_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_barrier = _import_func(
    "ada_entry_body_f_barrier",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_decls = _import_func(
    "ada_entry_body_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_stmts = _import_func(
    "ada_entry_body_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_body_f_end_name = _import_func(
    "ada_entry_body_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_body_f_package_name = _import_func(
    "ada_package_body_f_package_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_body_f_decls = _import_func(
    "ada_package_body_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_body_f_stmts = _import_func(
    "ada_package_body_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_body_f_end_name = _import_func(
    "ada_package_body_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_body_f_name = _import_func(
    "ada_protected_body_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_body_f_decls = _import_func(
    "ada_protected_body_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_body_f_end_name = _import_func(
    "ada_protected_body_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_body_f_name = _import_func(
    "ada_task_body_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_body_f_decls = _import_func(
    "ada_task_body_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_body_f_stmts = _import_func(
    "ada_task_body_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_body_f_end_name = _import_func(
    "ada_task_body_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_index_spec_f_id = _import_func(
    "ada_entry_index_spec_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_entry_index_spec_f_subtype = _import_func(
    "ada_entry_index_spec_f_subtype",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exception_decl_f_ids = _import_func(
    "ada_exception_decl_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exception_decl_f_renames = _import_func(
    "ada_exception_decl_f_renames",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exception_handler_f_exception_name = _import_func(
    "ada_exception_handler_f_exception_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exception_handler_f_handled_exceptions = _import_func(
    "ada_exception_handler_f_handled_exceptions",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exception_handler_f_stmts = _import_func(
    "ada_exception_handler_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_for_loop_var_decl_f_id = _import_func(
    "ada_for_loop_var_decl_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_for_loop_var_decl_f_id_type = _import_func(
    "ada_for_loop_var_decl_f_id_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_decl_f_formal_part = _import_func(
    "ada_generic_decl_f_formal_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_decl_f_package_decl = _import_func(
    "ada_generic_package_decl_f_package_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_decl_p_body_part = _import_func(
    "ada_generic_package_decl_p_body_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_decl_f_subp_decl = _import_func(
    "ada_generic_subp_decl_f_subp_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_decl_p_body_part = _import_func(
    "ada_generic_subp_decl_p_body_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_instantiation_p_designated_generic_decl = _import_func(
    "ada_generic_instantiation_p_designated_generic_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_instantiation_f_name = _import_func(
    "ada_generic_package_instantiation_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_instantiation_f_generic_pkg_name = _import_func(
    "ada_generic_package_instantiation_f_generic_pkg_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_instantiation_f_params = _import_func(
    "ada_generic_package_instantiation_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_f_overriding = _import_func(
    "ada_generic_subp_instantiation_f_overriding",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_f_kind = _import_func(
    "ada_generic_subp_instantiation_f_kind",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_f_subp_name = _import_func(
    "ada_generic_subp_instantiation_f_subp_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_f_generic_subp_name = _import_func(
    "ada_generic_subp_instantiation_f_generic_subp_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_f_params = _import_func(
    "ada_generic_subp_instantiation_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_instantiation_p_designated_subp = _import_func(
    "ada_generic_subp_instantiation_p_designated_subp",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_renaming_decl_f_name = _import_func(
    "ada_generic_package_renaming_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_package_renaming_decl_f_renames = _import_func(
    "ada_generic_package_renaming_decl_f_renames",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_renaming_decl_f_kind = _import_func(
    "ada_generic_subp_renaming_decl_f_kind",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_renaming_decl_f_name = _import_func(
    "ada_generic_subp_renaming_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_generic_subp_renaming_decl_f_renames = _import_func(
    "ada_generic_subp_renaming_decl_f_renames",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_label_decl_f_name = _import_func(
    "ada_label_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_named_stmt_decl_f_name = _import_func(
    "ada_named_stmt_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_number_decl_f_ids = _import_func(
    "ada_number_decl_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_number_decl_f_expr = _import_func(
    "ada_number_decl_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_ids = _import_func(
    "ada_object_decl_f_ids",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_has_aliased = _import_func(
    "ada_object_decl_f_has_aliased",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_has_constant = _import_func(
    "ada_object_decl_f_has_constant",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_mode = _import_func(
    "ada_object_decl_f_mode",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_type_expr = _import_func(
    "ada_object_decl_f_type_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_default_expr = _import_func(
    "ada_object_decl_f_default_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_f_renaming_clause = _import_func(
    "ada_object_decl_f_renaming_clause",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_p_public_part_decl = _import_func(
    "ada_object_decl_p_public_part_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_object_decl_p_private_part_decl = _import_func(
    "ada_object_decl_p_private_part_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_renaming_decl_f_name = _import_func(
    "ada_package_renaming_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_renaming_decl_f_renames = _import_func(
    "ada_package_renaming_decl_f_renames",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_renaming_decl_p_renamed_package = _import_func(
    "ada_package_renaming_decl_p_renamed_package",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_package_renaming_decl_p_final_renamed_package = _import_func(
    "ada_package_renaming_decl_p_final_renamed_package",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_single_protected_decl_f_name = _import_func(
    "ada_single_protected_decl_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_single_protected_decl_f_interfaces = _import_func(
    "ada_single_protected_decl_f_interfaces",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_single_protected_decl_f_definition = _import_func(
    "ada_single_protected_decl_f_definition",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_single_task_decl_f_task_type = _import_func(
    "ada_single_task_decl_f_task_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_stmt_alternative_f_choices = _import_func(
    "ada_case_stmt_alternative_f_choices",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_stmt_alternative_f_stmts = _import_func(
    "ada_case_stmt_alternative_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_compilation_unit_f_prelude = _import_func(
    "ada_compilation_unit_f_prelude",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_compilation_unit_f_body = _import_func(
    "ada_compilation_unit_f_body",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_compilation_unit_f_pragmas = _import_func(
    "ada_compilation_unit_f_pragmas",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_compilation_unit_p_syntactic_fully_qualified_name = _import_func(
    "ada_compilation_unit_p_syntactic_fully_qualified_name",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_UnboundedTextTypeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_compilation_unit_p_unit_kind = _import_func(
    "ada_compilation_unit_p_unit_kind",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_int)],
    ctypes.c_int,
)
_compilation_unit_p_withed_units = _import_func(
    "ada_compilation_unit_p_withed_units",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_compilation_unit_p_imported_units = _import_func(
    "ada_compilation_unit_p_imported_units",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_compilation_unit_p_unit_dependencies = _import_func(
    "ada_compilation_unit_p_unit_dependencies",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_compilation_unit_p_decl = _import_func(
    "ada_compilation_unit_p_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_compilation_unit_p_is_preelaborable = _import_func(
    "ada_compilation_unit_p_is_preelaborable",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_component_clause_f_id = _import_func(
    "ada_component_clause_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_clause_f_position = _import_func(
    "ada_component_clause_f_position",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_clause_f_range = _import_func(
    "ada_component_clause_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_def_f_has_aliased = _import_func(
    "ada_component_def_f_has_aliased",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_def_f_has_constant = _import_func(
    "ada_component_def_f_has_constant",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_component_def_f_type_expr = _import_func(
    "ada_component_def_f_type_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_constant_node_p_as_bool = _import_func(
    "ada_constant_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_delta_constraint_f_digits = _import_func(
    "ada_delta_constraint_f_digits",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_delta_constraint_f_range = _import_func(
    "ada_delta_constraint_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_digits_constraint_f_digits = _import_func(
    "ada_digits_constraint_f_digits",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_digits_constraint_f_range = _import_func(
    "ada_digits_constraint_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discriminant_constraint_f_constraints = _import_func(
    "ada_discriminant_constraint_f_constraints",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_index_constraint_f_constraints = _import_func(
    "ada_index_constraint_f_constraints",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_range_constraint_f_range = _import_func(
    "ada_range_constraint_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_declarative_part_f_decls = _import_func(
    "ada_declarative_part_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_elsif_expr_part_f_cond_expr = _import_func(
    "ada_elsif_expr_part_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_elsif_expr_part_f_then_expr = _import_func(
    "ada_elsif_expr_part_f_then_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_elsif_stmt_part_f_cond_expr = _import_func(
    "ada_elsif_stmt_part_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_elsif_stmt_part_f_stmts = _import_func(
    "ada_elsif_stmt_part_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_expr_p_expression_type = _import_func(
    "ada_expr_p_expression_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_expr_p_is_dynamically_tagged = _import_func(
    "ada_expr_p_is_dynamically_tagged",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_expr_p_is_static_expr = _import_func(
    "ada_expr_p_is_static_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_expr_p_first_corresponding_decl = _import_func(
    "ada_expr_p_first_corresponding_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_expr_p_eval_as_int = _import_func(
    "ada_expr_p_eval_as_int",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_big_integer.c_type)],
    ctypes.c_int,
)
_expr_p_eval_as_int_in_env = _import_func(
    "ada_expr_p_eval_as_int_in_env",
    [
        ctypes.POINTER(_Entity_c_type),
        _SubstitutionArrayConverter.c_type,
        ctypes.POINTER(_big_integer.c_type),
    ],
    ctypes.c_int,
)
_expr_p_eval_as_string = _import_func(
    "ada_expr_p_eval_as_string",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_expr_p_eval_as_string_in_env = _import_func(
    "ada_expr_p_eval_as_string_in_env",
    [
        ctypes.POINTER(_Entity_c_type),
        _SubstitutionArrayConverter.c_type,
        ctypes.POINTER(_TextTypeConverter.c_type),
    ],
    ctypes.c_int,
)
_expr_p_matching_nodes = _import_func(
    "ada_expr_p_matching_nodes",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_allocator_f_subpool = _import_func(
    "ada_allocator_f_subpool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_allocator_f_type_or_expr = _import_func(
    "ada_allocator_f_type_or_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_allocator_p_get_allocated_type = _import_func(
    "ada_allocator_p_get_allocated_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_aggregate_f_ancestor_expr = _import_func(
    "ada_base_aggregate_f_ancestor_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_aggregate_f_assocs = _import_func(
    "ada_base_aggregate_f_assocs",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_bin_op_f_left = _import_func(
    "ada_bin_op_f_left",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_bin_op_f_op = _import_func(
    "ada_bin_op_f_op",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_bin_op_f_right = _import_func(
    "ada_bin_op_f_right",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_expr_alternative_f_choices = _import_func(
    "ada_case_expr_alternative_f_choices",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_expr_alternative_f_expr = _import_func(
    "ada_case_expr_alternative_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_cond_expr_p_dependent_exprs = _import_func(
    "ada_cond_expr_p_dependent_exprs",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_case_expr_f_expr = _import_func(
    "ada_case_expr_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_expr_f_cases = _import_func(
    "ada_case_expr_f_cases",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_expr_f_cond_expr = _import_func(
    "ada_if_expr_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_expr_f_then_expr = _import_func(
    "ada_if_expr_f_then_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_expr_f_alternatives = _import_func(
    "ada_if_expr_f_alternatives",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_expr_f_else_expr = _import_func(
    "ada_if_expr_f_else_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_contract_cases_f_contract_cases = _import_func(
    "ada_contract_cases_f_contract_cases",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decl_expr_f_decls = _import_func(
    "ada_decl_expr_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decl_expr_f_expr = _import_func(
    "ada_decl_expr_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_membership_expr_f_expr = _import_func(
    "ada_membership_expr_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_membership_expr_f_op = _import_func(
    "ada_membership_expr_f_op",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_membership_expr_f_membership_exprs = _import_func(
    "ada_membership_expr_f_membership_exprs",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_enclosing_defining_name = _import_func(
    "ada_name_p_enclosing_defining_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_is_defining = _import_func(
    "ada_name_p_is_defining",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_name_is = _import_func(
    "ada_name_p_name_is",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_symbol_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_name_p_is_direct_call = _import_func(
    "ada_name_p_is_direct_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_access_call = _import_func(
    "ada_name_p_is_access_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_call = _import_func(
    "ada_name_p_is_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_dot_call = _import_func(
    "ada_name_p_is_dot_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_failsafe_referenced_def_name = _import_func(
    "ada_name_p_failsafe_referenced_def_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(RefdDef._c_type)],
    ctypes.c_int,
)
_name_p_referenced_defining_name = _import_func(
    "ada_name_p_referenced_defining_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_all_env_elements = _import_func(
    "ada_name_p_all_env_elements",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_AdaNodeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_name_p_called_subp_spec = _import_func(
    "ada_name_p_called_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_referenced_decl = _import_func(
    "ada_name_p_referenced_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_failsafe_referenced_decl = _import_func(
    "ada_name_p_failsafe_referenced_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(RefdDecl._c_type)],
    ctypes.c_int,
)
_name_p_referenced_decl_internal = _import_func(
    "ada_name_p_referenced_decl_internal",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(RefdDecl._c_type)],
    ctypes.c_int,
)
_name_p_name_designated_type = _import_func(
    "ada_name_p_name_designated_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_is_static_subtype = _import_func(
    "ada_name_p_is_static_subtype",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_name_matches = _import_func(
    "ada_name_p_name_matches",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(ctypes.c_uint8),
    ],
    ctypes.c_int,
)
_name_p_relative_name = _import_func(
    "ada_name_p_relative_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_name_p_is_operator_name = _import_func(
    "ada_name_p_is_operator_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_write_reference = _import_func(
    "ada_name_p_is_write_reference",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_dispatching_call = _import_func(
    "ada_name_p_is_dispatching_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_is_static_call = _import_func(
    "ada_name_p_is_static_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_name_p_as_symbol_array = _import_func(
    "ada_name_p_as_symbol_array",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_UnboundedTextTypeArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_attribute_ref_f_prefix = _import_func(
    "ada_attribute_ref_f_prefix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_attribute_ref_f_attribute = _import_func(
    "ada_attribute_ref_f_attribute",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_attribute_ref_f_args = _import_func(
    "ada_attribute_ref_f_args",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_call_expr_f_name = _import_func(
    "ada_call_expr_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_call_expr_f_suffix = _import_func(
    "ada_call_expr_f_suffix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_call_expr_p_is_array_slice = _import_func(
    "ada_call_expr_p_is_array_slice",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_defining_name_f_name = _import_func(
    "ada_defining_name_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_defining_name_p_basic_decl = _import_func(
    "ada_defining_name_p_basic_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_defining_name_p_find_refs = _import_func(
    "ada_defining_name_p_find_refs",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.c_uint8,
        ctypes.POINTER(_RefResultArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_defining_name_p_find_all_references = _import_func(
    "ada_defining_name_p_find_all_references",
    [
        ctypes.POINTER(_Entity_c_type),
        _AnalysisUnitArrayConverter.c_type,
        ctypes.c_uint8,
        ctypes.POINTER(_RefResultArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_defining_name_p_find_all_calls = _import_func(
    "ada_defining_name_p_find_all_calls",
    [
        ctypes.POINTER(_Entity_c_type),
        _AnalysisUnitArrayConverter.c_type,
        ctypes.c_uint8,
        ctypes.POINTER(_RefResultArrayConverter.c_type),
    ],
    ctypes.c_int,
)
_defining_name_p_next_part = _import_func(
    "ada_defining_name_p_next_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_defining_name_p_previous_part = _import_func(
    "ada_defining_name_p_previous_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_defining_name_p_canonical_part = _import_func(
    "ada_defining_name_p_canonical_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_discrete_subtype_name_f_subtype = _import_func(
    "ada_discrete_subtype_name_f_subtype",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_dotted_name_f_prefix = _import_func(
    "ada_dotted_name_f_prefix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_dotted_name_f_suffix = _import_func(
    "ada_dotted_name_f_suffix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_end_name_f_name = _import_func(
    "ada_end_name_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_end_name_p_basic_decl = _import_func(
    "ada_end_name_p_basic_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_explicit_deref_f_prefix = _import_func(
    "ada_explicit_deref_f_prefix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_qual_expr_f_prefix = _import_func(
    "ada_qual_expr_f_prefix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_qual_expr_f_suffix = _import_func(
    "ada_qual_expr_f_suffix",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_single_tok_node_p_canonical_text = _import_func(
    "ada_single_tok_node_p_canonical_text",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_symbol_type)],
    ctypes.c_int,
)
_char_literal_p_denoted_value = _import_func(
    "ada_char_literal_p_denoted_value",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint32)],
    ctypes.c_int,
)
_string_literal_p_denoted_value = _import_func(
    "ada_string_literal_p_denoted_value",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_TextTypeConverter.c_type)],
    ctypes.c_int,
)
_int_literal_p_denoted_value = _import_func(
    "ada_int_literal_p_denoted_value",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_big_integer.c_type)],
    ctypes.c_int,
)
_paren_expr_f_expr = _import_func(
    "ada_paren_expr_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_quantified_expr_f_quantifier = _import_func(
    "ada_quantified_expr_f_quantifier",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_quantified_expr_f_loop_spec = _import_func(
    "ada_quantified_expr_f_loop_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_quantified_expr_f_expr = _import_func(
    "ada_quantified_expr_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_raise_expr_f_exception_name = _import_func(
    "ada_raise_expr_f_exception_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_raise_expr_f_error_message = _import_func(
    "ada_raise_expr_f_error_message",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_un_op_f_op = _import_func(
    "ada_un_op_f_op",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_un_op_f_expr = _import_func(
    "ada_un_op_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_handled_stmts_f_stmts = _import_func(
    "ada_handled_stmts_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_handled_stmts_f_exceptions = _import_func(
    "ada_handled_stmts_f_exceptions",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_library_item_f_has_private = _import_func(
    "ada_library_item_f_has_private",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_library_item_f_item = _import_func(
    "ada_library_item_f_item",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_limited_node_p_as_bool = _import_func(
    "ada_limited_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_for_loop_spec_f_var_decl = _import_func(
    "ada_for_loop_spec_f_var_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_for_loop_spec_f_loop_type = _import_func(
    "ada_for_loop_spec_f_loop_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_for_loop_spec_f_has_reverse = _import_func(
    "ada_for_loop_spec_f_has_reverse",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_for_loop_spec_f_iter_expr = _import_func(
    "ada_for_loop_spec_f_iter_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_while_loop_spec_f_expr = _import_func(
    "ada_while_loop_spec_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_not_null_p_as_bool = _import_func(
    "ada_not_null_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_params_f_params = _import_func(
    "ada_params_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_pragma_node_f_id = _import_func(
    "ada_pragma_node_f_id",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_pragma_node_f_args = _import_func(
    "ada_pragma_node_f_args",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_pragma_node_p_associated_decls = _import_func(
    "ada_pragma_node_p_associated_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_AdaNodeArrayConverter.c_type)],
    ctypes.c_int,
)
_private_node_p_as_bool = _import_func(
    "ada_private_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_protected_def_f_public_part = _import_func(
    "ada_protected_def_f_public_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_def_f_private_part = _import_func(
    "ada_protected_def_f_private_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_def_f_end_name = _import_func(
    "ada_protected_def_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_protected_node_p_as_bool = _import_func(
    "ada_protected_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_range_spec_f_range = _import_func(
    "ada_range_spec_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_renaming_clause_f_renamed_object = _import_func(
    "ada_renaming_clause_f_renamed_object",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_reverse_node_p_as_bool = _import_func(
    "ada_reverse_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_select_when_part_f_cond_expr = _import_func(
    "ada_select_when_part_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_select_when_part_f_stmts = _import_func(
    "ada_select_when_part_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_accept_stmt_f_name = _import_func(
    "ada_accept_stmt_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_accept_stmt_f_entry_index_expr = _import_func(
    "ada_accept_stmt_f_entry_index_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_accept_stmt_f_params = _import_func(
    "ada_accept_stmt_f_params",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_accept_stmt_with_stmts_f_stmts = _import_func(
    "ada_accept_stmt_with_stmts_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_accept_stmt_with_stmts_f_end_name = _import_func(
    "ada_accept_stmt_with_stmts_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_loop_stmt_f_spec = _import_func(
    "ada_base_loop_stmt_f_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_loop_stmt_f_stmts = _import_func(
    "ada_base_loop_stmt_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_base_loop_stmt_f_end_name = _import_func(
    "ada_base_loop_stmt_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_begin_block_f_stmts = _import_func(
    "ada_begin_block_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_begin_block_f_end_name = _import_func(
    "ada_begin_block_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decl_block_f_decls = _import_func(
    "ada_decl_block_f_decls",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decl_block_f_stmts = _import_func(
    "ada_decl_block_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decl_block_f_end_name = _import_func(
    "ada_decl_block_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_stmt_f_expr = _import_func(
    "ada_case_stmt_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_case_stmt_f_alternatives = _import_func(
    "ada_case_stmt_f_alternatives",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_extended_return_stmt_f_decl = _import_func(
    "ada_extended_return_stmt_f_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_extended_return_stmt_f_stmts = _import_func(
    "ada_extended_return_stmt_f_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_stmt_f_cond_expr = _import_func(
    "ada_if_stmt_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_stmt_f_then_stmts = _import_func(
    "ada_if_stmt_f_then_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_stmt_f_alternatives = _import_func(
    "ada_if_stmt_f_alternatives",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_if_stmt_f_else_stmts = _import_func(
    "ada_if_stmt_f_else_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_named_stmt_f_decl = _import_func(
    "ada_named_stmt_f_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_named_stmt_f_stmt = _import_func(
    "ada_named_stmt_f_stmt",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_select_stmt_f_guards = _import_func(
    "ada_select_stmt_f_guards",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_select_stmt_f_else_stmts = _import_func(
    "ada_select_stmt_f_else_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_select_stmt_f_abort_stmts = _import_func(
    "ada_select_stmt_f_abort_stmts",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_abort_stmt_f_names = _import_func(
    "ada_abort_stmt_f_names",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_assign_stmt_f_dest = _import_func(
    "ada_assign_stmt_f_dest",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_assign_stmt_f_expr = _import_func(
    "ada_assign_stmt_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_call_stmt_f_call = _import_func(
    "ada_call_stmt_f_call",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_delay_stmt_f_has_until = _import_func(
    "ada_delay_stmt_f_has_until",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_delay_stmt_f_expr = _import_func(
    "ada_delay_stmt_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exit_stmt_f_loop_name = _import_func(
    "ada_exit_stmt_f_loop_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_exit_stmt_f_cond_expr = _import_func(
    "ada_exit_stmt_f_cond_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_goto_stmt_f_label_name = _import_func(
    "ada_goto_stmt_f_label_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_label_f_decl = _import_func(
    "ada_label_f_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_raise_stmt_f_exception_name = _import_func(
    "ada_raise_stmt_f_exception_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_raise_stmt_f_error_message = _import_func(
    "ada_raise_stmt_f_error_message",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_requeue_stmt_f_call_name = _import_func(
    "ada_requeue_stmt_f_call_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_requeue_stmt_f_has_abort = _import_func(
    "ada_requeue_stmt_f_has_abort",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_return_stmt_f_return_expr = _import_func(
    "ada_return_stmt_f_return_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subunit_f_name = _import_func(
    "ada_subunit_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subunit_f_body = _import_func(
    "ada_subunit_f_body",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subunit_p_body_root = _import_func(
    "ada_subunit_p_body_root",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_synchronized_node_p_as_bool = _import_func(
    "ada_synchronized_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_tagged_node_p_as_bool = _import_func(
    "ada_tagged_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_task_def_f_interfaces = _import_func(
    "ada_task_def_f_interfaces",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_def_f_public_part = _import_func(
    "ada_task_def_f_public_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_def_f_private_part = _import_func(
    "ada_task_def_f_private_part",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_task_def_f_end_name = _import_func(
    "ada_task_def_f_end_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_access_def_f_has_not_null = _import_func(
    "ada_access_def_f_has_not_null",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_access_to_subp_def_f_has_protected = _import_func(
    "ada_access_to_subp_def_f_has_protected",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_access_to_subp_def_f_subp_spec = _import_func(
    "ada_access_to_subp_def_f_subp_spec",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_anonymous_type_access_def_f_type_decl = _import_func(
    "ada_anonymous_type_access_def_f_type_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_access_def_f_has_all = _import_func(
    "ada_type_access_def_f_has_all",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_access_def_f_has_constant = _import_func(
    "ada_type_access_def_f_has_constant",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_access_def_f_subtype_indication = _import_func(
    "ada_type_access_def_f_subtype_indication",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_array_type_def_f_indices = _import_func(
    "ada_array_type_def_f_indices",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_array_type_def_f_component_type = _import_func(
    "ada_array_type_def_f_component_type",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_has_abstract = _import_func(
    "ada_derived_type_def_f_has_abstract",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_has_limited = _import_func(
    "ada_derived_type_def_f_has_limited",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_has_synchronized = _import_func(
    "ada_derived_type_def_f_has_synchronized",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_subtype_indication = _import_func(
    "ada_derived_type_def_f_subtype_indication",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_interfaces = _import_func(
    "ada_derived_type_def_f_interfaces",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_record_extension = _import_func(
    "ada_derived_type_def_f_record_extension",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_derived_type_def_f_has_with_private = _import_func(
    "ada_derived_type_def_f_has_with_private",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_enum_type_def_f_enum_literals = _import_func(
    "ada_enum_type_def_f_enum_literals",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_interface_type_def_f_interface_kind = _import_func(
    "ada_interface_type_def_f_interface_kind",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_interface_type_def_f_interfaces = _import_func(
    "ada_interface_type_def_f_interfaces",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_mod_int_type_def_f_expr = _import_func(
    "ada_mod_int_type_def_f_expr",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_private_type_def_f_has_abstract = _import_func(
    "ada_private_type_def_f_has_abstract",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_private_type_def_f_has_tagged = _import_func(
    "ada_private_type_def_f_has_tagged",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_private_type_def_f_has_limited = _import_func(
    "ada_private_type_def_f_has_limited",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decimal_fixed_point_def_f_delta = _import_func(
    "ada_decimal_fixed_point_def_f_delta",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decimal_fixed_point_def_f_digits = _import_func(
    "ada_decimal_fixed_point_def_f_digits",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_decimal_fixed_point_def_f_range = _import_func(
    "ada_decimal_fixed_point_def_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_floating_point_def_f_num_digits = _import_func(
    "ada_floating_point_def_f_num_digits",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_floating_point_def_f_range = _import_func(
    "ada_floating_point_def_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ordinary_fixed_point_def_f_delta = _import_func(
    "ada_ordinary_fixed_point_def_f_delta",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_ordinary_fixed_point_def_f_range = _import_func(
    "ada_ordinary_fixed_point_def_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_type_def_f_has_abstract = _import_func(
    "ada_record_type_def_f_has_abstract",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_type_def_f_has_tagged = _import_func(
    "ada_record_type_def_f_has_tagged",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_type_def_f_has_limited = _import_func(
    "ada_record_type_def_f_has_limited",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_record_type_def_f_record_def = _import_func(
    "ada_record_type_def_f_record_def",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_signed_int_type_def_f_range = _import_func(
    "ada_signed_int_type_def_f_range",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_expr_p_type_name = _import_func(
    "ada_type_expr_p_type_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_expr_p_designated_type_decl = _import_func(
    "ada_type_expr_p_designated_type_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_type_expr_p_designated_type_decl_from = _import_func(
    "ada_type_expr_p_designated_type_decl_from",
    [
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
        ctypes.POINTER(_Entity_c_type),
    ],
    ctypes.c_int,
)
_anonymous_type_f_type_decl = _import_func(
    "ada_anonymous_type_f_type_decl",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subtype_indication_f_has_not_null = _import_func(
    "ada_subtype_indication_f_has_not_null",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subtype_indication_f_name = _import_func(
    "ada_subtype_indication_f_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subtype_indication_f_constraint = _import_func(
    "ada_subtype_indication_f_constraint",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_subtype_indication_p_is_static_subtype = _import_func(
    "ada_subtype_indication_p_is_static_subtype",
    [ctypes.POINTER(_Entity_c_type), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_unconstrained_array_index_f_subtype_indication = _import_func(
    "ada_unconstrained_array_index_f_subtype_indication",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_until_node_p_as_bool = _import_func(
    "ada_until_node_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)
_use_package_clause_f_packages = _import_func(
    "ada_use_package_clause_f_packages",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_use_type_clause_f_has_all = _import_func(
    "ada_use_type_clause_f_has_all",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_use_type_clause_f_types = _import_func(
    "ada_use_type_clause_f_types",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_variant_f_choices = _import_func(
    "ada_variant_f_choices",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_variant_f_components = _import_func(
    "ada_variant_f_components",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_variant_part_f_discr_name = _import_func(
    "ada_variant_part_f_discr_name",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_variant_part_f_variant = _import_func(
    "ada_variant_part_f_variant",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_with_clause_f_has_limited = _import_func(
    "ada_with_clause_f_has_limited",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_with_clause_f_has_private = _import_func(
    "ada_with_clause_f_has_private",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_with_clause_f_packages = _import_func(
    "ada_with_clause_f_packages",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_Entity_c_type)],
    ctypes.c_int,
)
_with_private_p_as_bool = _import_func(
    "ada_with_private_p_as_bool",
    [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(ctypes.c_uint8)],
    ctypes.c_int,
)

# Unit providers
_dec_ref_unit_provider = _import_func(
    "ada_dec_ref_unit_provider", [_unit_provider], None
)


class _project_scenario_variable(ctypes.Structure):
    _fields_ = [("name", ctypes.c_char_p), ("value", ctypes.c_char_p)]


_create_project_unit_provider = _import_func(
    "ada_create_project_unit_provider",
    [
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.POINTER(_project_scenario_variable),
        ctypes.c_char_p,
        ctypes.c_char_p,
    ],
    _unit_provider,
)

_create_auto_provider = _import_func(
    "ada_create_auto_provider",
    [ctypes.POINTER(ctypes.c_char_p), ctypes.c_char_p],
    _unit_provider,
)


# Misc
_token_kind_name = _import_func(
    "ada_token_kind_name", [ctypes.c_int], ctypes.POINTER(ctypes.c_char)
)
_token_next = _import_func(
    "ada_token_next", [ctypes.POINTER(Token), ctypes.POINTER(Token)], None
)
_token_is_equivalent = _import_func(
    "ada_token_is_equivalent",
    [ctypes.POINTER(Token), ctypes.POINTER(Token)],
    ctypes.c_int,
)
_token_previous = _import_func(
    "ada_token_previous", [ctypes.POINTER(Token), ctypes.POINTER(Token)], None
)
_token_range_text = _import_func(
    "ada_token_range_text",
    [ctypes.POINTER(Token), ctypes.POINTER(Token), ctypes.POINTER(_text)],
    ctypes.c_int,
)
_entity_image = _import_func(
    "ada_entity_image", [ctypes.POINTER(_Entity_c_type), ctypes.POINTER(_text)], None
)


#
# Layering helpers
#


def _unwrap_str(c_char_p_value):
    """
    Assuming c_char_p_value is a valid char*, convert it to a native Python
    string and free the C pointer.
    """
    result = ctypes.c_char_p(ctypes.addressof(c_char_p_value.contents)).value
    _free(c_char_p_value)
    return _py2to3.bytes_to_text(result)


_kind_to_astnode_cls = {
    1: AbortAbsent,
    2: AbortPresent,
    3: AbstractAbsent,
    4: AbstractPresent,
    5: AdaNodeList,
    6: AlternativesList,
    7: ConstraintList,
    8: DeclList,
    9: StmtList,
    10: AspectAssocList,
    11: BaseAssocList,
    12: AssocList,
    13: BasicDeclList,
    14: CaseExprAlternativeList,
    15: CaseStmtAlternativeList,
    16: CompilationUnitList,
    17: ContractCaseAssocList,
    18: DefiningNameList,
    19: DiscriminantSpecList,
    20: ElsifExprPartList,
    21: ElsifStmtPartList,
    22: EnumLiteralDeclList,
    23: ExprAlternativesList,
    24: DiscriminantChoiceList,
    25: NameList,
    26: ParentList,
    27: ParamSpecList,
    28: PragmaNodeList,
    29: SelectWhenPartList,
    30: UnconstrainedArrayIndexList,
    31: VariantList,
    32: AliasedAbsent,
    33: AliasedPresent,
    34: AllAbsent,
    35: AllPresent,
    36: ConstrainedArrayIndices,
    37: UnconstrainedArrayIndices,
    38: AspectAssoc,
    39: AtClause,
    40: AttributeDefClause,
    41: EnumRepClause,
    42: RecordRepClause,
    43: AspectSpec,
    44: ContractCaseAssoc,
    45: PragmaArgumentAssoc,
    46: EntrySpec,
    47: EnumSubpSpec,
    48: SubpSpec,
    49: ComponentList,
    50: KnownDiscriminantPart,
    51: UnknownDiscriminantPart,
    52: EntryCompletionFormalParams,
    53: GenericFormalPart,
    54: NullRecordDef,
    55: RecordDef,
    56: AggregateAssoc,
    57: MultiDimArrayAssoc,
    58: DiscriminantAssoc,
    59: IteratedAssoc,
    60: ParamAssoc,
    61: ComponentDecl,
    62: DiscriminantSpec,
    63: GenericFormalObjDecl,
    64: GenericFormalPackage,
    65: GenericFormalSubpDecl,
    66: GenericFormalTypeDecl,
    67: ParamSpec,
    68: GenericPackageInternal,
    69: PackageDecl,
    70: DiscreteBaseSubtypeDecl,
    71: SubtypeDecl,
    72: ClasswideTypeDecl,
    73: IncompleteTypeDecl,
    74: IncompleteTaggedTypeDecl,
    75: ProtectedTypeDecl,
    76: TaskTypeDecl,
    77: SingleTaskTypeDecl,
    78: TypeDecl,
    79: AnonymousTypeDecl,
    80: SynthAnonymousTypeDecl,
    81: AbstractSubpDecl,
    82: AbstractFormalSubpDecl,
    83: ConcreteFormalSubpDecl,
    84: SubpDecl,
    85: EntryDecl,
    86: EnumLiteralDecl,
    87: GenericSubpInternal,
    88: ExprFunction,
    89: NullSubpDecl,
    90: SubpBody,
    91: SubpRenamingDecl,
    92: PackageBodyStub,
    93: ProtectedBodyStub,
    94: SubpBodyStub,
    95: TaskBodyStub,
    96: EntryBody,
    97: PackageBody,
    98: ProtectedBody,
    99: TaskBody,
    100: EntryIndexSpec,
    101: ErrorDecl,
    102: ExceptionDecl,
    103: ExceptionHandler,
    104: ForLoopVarDecl,
    105: GenericPackageDecl,
    106: GenericSubpDecl,
    107: GenericPackageInstantiation,
    108: GenericSubpInstantiation,
    109: GenericPackageRenamingDecl,
    110: GenericSubpRenamingDecl,
    111: LabelDecl,
    112: NamedStmtDecl,
    113: NumberDecl,
    114: ObjectDecl,
    115: AnonymousObjectDecl,
    116: ExtendedReturnStmtObjectDecl,
    117: PackageRenamingDecl,
    118: SingleProtectedDecl,
    119: SingleTaskDecl,
    120: CaseStmtAlternative,
    121: CompilationUnit,
    122: ComponentClause,
    123: ComponentDef,
    124: ConstantAbsent,
    125: ConstantPresent,
    126: DeltaConstraint,
    127: DigitsConstraint,
    128: DiscriminantConstraint,
    129: IndexConstraint,
    130: RangeConstraint,
    131: DeclarativePart,
    132: PrivatePart,
    133: PublicPart,
    134: ElsifExprPart,
    135: ElsifStmtPart,
    136: Allocator,
    137: Aggregate,
    138: BracketAggregate,
    139: DeltaAggregate,
    140: BracketDeltaAggregate,
    141: NullRecordAggregate,
    142: BinOp,
    143: RelationOp,
    144: BoxExpr,
    145: CaseExprAlternative,
    146: CaseExpr,
    147: IfExpr,
    148: ContractCases,
    149: DeclExpr,
    150: MembershipExpr,
    151: AttributeRef,
    152: UpdateAttributeRef,
    153: CallExpr,
    154: DefiningName,
    155: DiscreteSubtypeName,
    156: DottedName,
    157: EndName,
    158: ExplicitDeref,
    159: QualExpr,
    160: CharLiteral,
    161: Identifier,
    162: OpAbs,
    163: OpAnd,
    164: OpAndThen,
    165: OpConcat,
    166: OpDiv,
    167: OpDoubleDot,
    168: OpEq,
    169: OpGt,
    170: OpGte,
    171: OpIn,
    172: OpLt,
    173: OpLte,
    174: OpMinus,
    175: OpMod,
    176: OpMult,
    177: OpNeq,
    178: OpNot,
    179: OpNotIn,
    180: OpOr,
    181: OpOrElse,
    182: OpPlus,
    183: OpPow,
    184: OpRem,
    185: OpXor,
    186: StringLiteral,
    187: NullLiteral,
    188: IntLiteral,
    189: RealLiteral,
    190: TargetName,
    191: ParenExpr,
    192: QuantifiedExpr,
    193: RaiseExpr,
    194: UnOp,
    195: HandledStmts,
    196: InterfaceKindLimited,
    197: InterfaceKindProtected,
    198: InterfaceKindSynchronized,
    199: InterfaceKindTask,
    200: IterTypeIn,
    201: IterTypeOf,
    202: LibraryItem,
    203: LimitedAbsent,
    204: LimitedPresent,
    205: ForLoopSpec,
    206: WhileLoopSpec,
    207: ModeDefault,
    208: ModeIn,
    209: ModeInOut,
    210: ModeOut,
    211: NotNullAbsent,
    212: NotNullPresent,
    213: NullComponentDecl,
    214: OthersDesignator,
    215: OverridingNotOverriding,
    216: OverridingOverriding,
    217: OverridingUnspecified,
    218: Params,
    219: PragmaNode,
    220: PrimTypeAccessor,
    221: PrivateAbsent,
    222: PrivatePresent,
    223: ProtectedDef,
    224: ProtectedAbsent,
    225: ProtectedPresent,
    226: QuantifierAll,
    227: QuantifierSome,
    228: RangeSpec,
    229: RenamingClause,
    230: SyntheticRenamingClause,
    231: ReverseAbsent,
    232: ReversePresent,
    233: SelectWhenPart,
    234: AcceptStmt,
    235: AcceptStmtWithStmts,
    236: ForLoopStmt,
    237: LoopStmt,
    238: WhileLoopStmt,
    239: BeginBlock,
    240: DeclBlock,
    241: CaseStmt,
    242: ExtendedReturnStmt,
    243: IfStmt,
    244: NamedStmt,
    245: SelectStmt,
    246: ErrorStmt,
    247: AbortStmt,
    248: AssignStmt,
    249: CallStmt,
    250: DelayStmt,
    251: ExitStmt,
    252: GotoStmt,
    253: Label,
    254: NullStmt,
    255: RaiseStmt,
    256: RequeueStmt,
    257: ReturnStmt,
    258: TerminateAlternative,
    259: SubpKindFunction,
    260: SubpKindProcedure,
    261: Subunit,
    262: SynchronizedAbsent,
    263: SynchronizedPresent,
    264: TaggedAbsent,
    265: TaggedPresent,
    266: TaskDef,
    267: AccessToSubpDef,
    268: AnonymousTypeAccessDef,
    269: TypeAccessDef,
    270: ArrayTypeDef,
    271: DerivedTypeDef,
    272: EnumTypeDef,
    273: FormalDiscreteTypeDef,
    274: InterfaceTypeDef,
    275: ModIntTypeDef,
    276: PrivateTypeDef,
    277: DecimalFixedPointDef,
    278: FloatingPointDef,
    279: OrdinaryFixedPointDef,
    280: RecordTypeDef,
    281: SignedIntTypeDef,
    282: AnonymousType,
    283: EnumLitSynthTypeExpr,
    284: SubtypeIndication,
    285: ConstrainedSubtypeIndication,
    286: DiscreteSubtypeIndication,
    287: UnconstrainedArrayIndex,
    288: UntilAbsent,
    289: UntilPresent,
    290: UsePackageClause,
    291: UseTypeClause,
    292: Variant,
    293: VariantPart,
    294: WithClause,
    295: WithPrivateAbsent,
    296: WithPrivatePresent,
}


def _field_address(struct, field_name):
    """
    Get the address of a structure field from a structure value.

    For instance::

        class Foo(ctypes.Structure):
            _fields_ = [('i', ctypes.c_int)]

        f = Foo()
        i_addr =_field_address(f, 'i')
    """
    struct_type = type(struct)
    struct_addr = ctypes.addressof(struct)
    field = getattr(struct_type, field_name)
    field_type = None
    for f_name, f_type in struct_type._fields_:
        if f_name == field_name:
            field_type = f_type
            break
    assert field_type is not None
    return struct_addr + field.offset


#
# Language specific extensions #
#


def token_match(self, other):
    """
    Helper for the finditer/find/findall methods, so that a token matches
    another token even if they are not strictly equivalent.
    """
    return self == other or self.text == other


@property
def doc_name(n):
    """
    Format this name to be a readable qualified name for the entity designated
    by it. Meant to be used in documentation context.

    If the entity is local, it will return the relative name. If it is
    non-local, return the shortest qualified name not taking use clauses into
    account.

    .. WARNING:: This is an EXPERIMENTAL feature. This is a python specific
        method, because for the moment this is not conveniently implementable
        directly as a libadalang property.  Consider it an experimental API
        endpoint, and use it at your own risk.
    """
    if n.p_is_defining and not n.is_a(DefiningName):
        n = n.p_enclosing_defining_name

    ref_decl = n.p_basic_decl if n.p_is_defining else n.p_referenced_decl()
    ref_decl_fqn = ref_decl.p_fully_qualified_name

    enclosing_package = next((p for p in n.parents if p.is_a(BasePackageDecl)), None)

    if enclosing_package is None or enclosing_package == ref_decl:
        return ref_decl_fqn

    enclosing_decl_fqn = enclosing_package.p_fully_qualified_name

    if ref_decl_fqn.lower().startswith(enclosing_decl_fqn.lower()):
        return ref_decl_fqn[len(enclosing_decl_fqn) :].strip(".")
    else:
        return ref_decl_fqn


Token.match = token_match
Name.doc_name = doc_name


#
# App base class
#


class App(object):
    """
    Base class to regroup logic for an app. We use a class so that
    specific languages implementations can add specific arguments and
    processing by overriding specific methods:

    - `main`, which will be the main method of the app.

    - `add_arguments` to add arguments to the argparse.Parser instance

    - `create_unit_provider` to return a custom unit provider to be used by the
      AnalysisContext.

    - `description` to change the description of the app.

    Inside of `main`, the user can access app specific state:

    - `self.units` is a map of filenames to analysis units.
    - `self.ctx` is the analysis context.
    - `self.u` is the last parsed unit.

    The user can then run the app by calling `App.run()`.

    Here is a small example of an app subclassing `App`, that will simply print
    the tree of every unit passed as argument:

    .. code-block:: python

        from libadalang import App


        class ExampleApp(App):
            def main(self):
                for u in self.units.values():
                    print u.filename
                    print u.root.dump()

        ExampleApp.run()
    """

    @property
    def description(self):
        """
        Description for this app. Empty by default.
        """
        return ""

    def __init__(self, args=None):
        self.parser = argparse.ArgumentParser(description=self.description)
        self.parser.add_argument("files", nargs="+", help="Files")
        self.add_arguments()

        # Parse command line arguments
        self.args = self.parser.parse_args(args)

        self.ctx = AnalysisContext(
            "utf-8", with_trivia=True, unit_provider=self.create_unit_provider()
        )

        # Parse files
        self.units = {}
        for file_name in self.args.files:
            self.u = self.ctx.get_from_file(file_name)
            self.units[file_name] = self.u

    def add_arguments(self):
        """
        Hook for subclasses to add arguments to self.parser. Default
        implementation does nothing.
        """
        pass

    def create_unit_provider(self):
        """
        Hook for subclasses to return a custom unit provider.
        Default implementation returns None.
        """
        return None

    def main(self):
        """
        Default implementation for App.main: just iterates on every units and
        call ``process_unit`` on it.
        """
        for u in sorted(self.units.values(), key=lambda u: u.filename):
            self.process_unit(u)

    def process_unit(self, unit):
        """
        Abstract method that processes one unit. Needs to be subclassed by
        implementors.
        """
        raise NotImplementedError()

    @classmethod
    def run(cls, args=None):
        """
        Instantiate and run this application.
        """
        cls(args).main()

    def add_arguments(self):
        self.parser.add_argument(
            "-X", action="append", help="Scenario variables to pass along to GPR"
        )
        self.parser.add_argument(
            "-P", "--project", type=str, default="", help="GPR project file"
        )

    def create_unit_provider(self):

        from tempfile import mkdtemp

        if not self.args.project:
            return None

        self.scenario_vars = {}
        if self.args.X:
            for var in self.args.X:
                k, v = var.split("=")
                self.scenario_vars[k] = v

        return UnitProvider.for_project(
            self.args.project, scenario_vars=self.scenario_vars
        )
